suppressMessages(require(data.table))

getObs <- function(edges, file = "LS_WOD.rds", file_zoo = "LS_WOD_plankton.rds", # nolint: cyclocomp_linter.
                   file_zoo_int = "LSzooInt.rds", oppla_inp = "oppla.inp",
                   datecols = "month", prefix, transform = "log",
                   georange = list(lat = units::set_units(c(min = 40, max = 60), "degrees_N"),
                                   lon = units::set_units(c(min = -70, max = -50), "degrees_E"))) {
  require(oppla) # needed to make readRDS assign units
  fse <- function(x, se_na = NA_real_, ...) {
    ## standard error
    s <- sum(complete.cases(x))
    return(if (s < 2) se_na else sqrt(var(x, ...) / s))
  }
  fsd <- function(x, sd_na = NA_real_, ...) {
    ## standard deviation
    s <- sum(complete.cases(x))
    return(if (s < 2) sd_na else sd(x, ...))
  }
  if (missing(edges)) {
    inml <- oppla::nml_read(file = oppla_inp)
    ## extract depth edges from initial conditions
    if (substr(inml$files[[1]]$ifn, 1, 1) %in% c("/", "~")) ifn <- inml$files[[1]]$ifn
    else ifn <- file.path(if (missing(prefix)) dirname(oppla_inp) else prefix, inml$files[[1]]$ifn)
    if (file.exists(ifn))
      edges <- attr(oppla::oppla_slab(data = oppla::oppla_read(file = ifn, forcing = FALSE),
                                      depthrange = c(0, inml$costfun[[1]]$max_depth)), "depth_edges")
    else
      stop("file ", ifn, " does not exist.")
  } else {
    if (length(edges) == 0) stop("Argument edges must not be empty!")
  }
  dnames <- c("Nitrate", "Phosphate", "Chlorophyll")
  rho0 <- units::set_units(1025, "kg m-3") # default for missing density data
  ## nutrient and Chl data
  lsd <- data.table::as.data.table(readRDS(file = file))[
                       lat %between% georange$lat & lon %between% georange$lon,
                       j = .SD, .SDcols = c("lon", "lat", "date", "depth", dnames, "density")][
                       j = c("Nitrate", "Phosphate") := .(Nitrate * fcoalesce(density, rho0),
                                                          Phosphate * fcoalesce(density, rho0))]
  thresholds <- list(Nitrate = units::set_units(0.05, "mmol m-3"),
                     Phosphate = units::set_units(0.01, "mmol m-3"),
                     Chlorophyll = units::set_units(0.05, "mg m-3"),
                     measure_val_bio = units::set_units(0.1, "mmol m-3"),
                     zooI = units::set_units(5, "mmol m-2"))
  if (identical(transform, "log")) {
    ## normalise and log-transform nutrients and Chl
    one <- units::set_units(1, 1)
    lsd[j = Nitrate := log(units::drop_units(pmax(units::set_units(Nitrate / thresholds$Nitrate, 1), one)))][
      j = Phosphate := log(units::drop_units(pmax(units::set_units(Phosphate / thresholds$Phosphate, 1), one)))][
      j = Chlorophyll := log(units::drop_units(pmax(units::set_units(Chlorophyll / thresholds$Chlorophyll, 1), one)))]
  }
  ## zooplankton data
  lsz <- data.table::as.data.table(readRDS(file = file_zoo))[
                       lat %between% georange$lat & lon %between% georange$lon &
                       ! measure_units_bio %in% c("code/haul", "#/ul", "#/ml", "#/haul", "ml/m2"),
                       j = .(lon, lat, date, lower_z_bio, upper_z_bio, measure_val_bio, measure_units_bio)][
                       j = depth := units::set_units(rowMeans(.SD), "m"), .SDcols = c("lower_z_bio", "upper_z_bio")]
  imlm3 <- which(lsz$measure_units_bio == "ml/m3") # local concentrations
  iugm3 <- which(lsz$measure_units_bio == "ug/m3")
  ## convert: ml (*1000) -> mm3/m3 (/0.95, Sato et al. 2015) -> wet weight (*0.064) -> mg (/12) -> mmol
  ## factor 0.064 = 0.4 * 0.16 (dry weight = 16% of wet weight, copepods contain 84% water, Postel et al., 2000, p.91)
  lsz$measure_val_bio[imlm3] <- lsz$measure_val_bio[imlm3] * 1000 / 0.95 * 0.064 / 12
  lsz$measure_val_bio[iugm3] <- lsz$measure_val_bio[iugm3] / 1000 * 0.4 / 12 # from ug dry weight/m3 to mmol C/m3
  lszc <- lsz[c(imlm3, iugm3)][j = measure_val_bio := units::set_units(measure_val_bio, "mmol m-3")]
  lszi <- data.table::as.data.table(readRDS(file = file_zoo_int)) # 100m depth-integrated zooplankton data in mmol C/m2
  if (identical(transform, "log")) {
    ## log-transform zooplankton
    lszc[j = measure_val_bio := log(units::drop_units(pmax(units::set_units(measure_val_bio /
                                                                            thresholds$measure_val_bio, 1), one)))]
    llsz <- log(units::drop_units(pmax(units::set_units(lszi$total / thresholds$zooI, 1), one)))
    llszi <- aggregate(llsz, by = oppla_split_date(date = lszi$date, columns = datecols),
                       FUN = function(x, ...) c(mean = mean(x, ...), se = fse(x, ...), sd = fsd(x, ...)))
    zoo <- data.table::data.table(llszi[datecols], llszi$x)
  } else {
    lsza <- aggregate(lszi$total, by = oppla_split_date(date = lszi$date, columns = datecols),
                      FUN = function(x, ...) c(mean = units::drop_units(mean(x, ...)), se = fse(x, ...), sd = fsd(x, ...)))
    zoo <- data.table::data.table(lsza[datecols], units::set_units(lsza$x, units(lszi$total)))
  }
  lsall <- merge(lsd, lszc, all = TRUE)
  avg <- oppla::op_aggregate(data = lsall, edges = edges, datecols = datecols,
                             columns = c(dnames, "measure_val_bio"), fun = mean)
  if ("units" %in% class(avg$Nitrate)) {
    se <- oppla::op_aggregate(data = lsall, edges = edges, datecols = datecols,
                              columns = c(dnames, "measure_val_bio"), drop_na = FALSE,
                              fun = function(x, ...) units::keep_units(fse, x, ...))
    ## apply the thresholds also to the standard errors
    for (name in dnames) se[j = c(name) := pmax(thresholds[[..name]], se[[..name]])]
    sd <- oppla::op_aggregate(data = lsall, edges = edges, datecols = datecols,
                              columns = c(dnames, "measure_val_bio"), drop_na = FALSE,
                              fun = function(x, ...) units::keep_units(fsd, x, ...))
  } else {
    se <- oppla::op_aggregate(data = lsall, edges = edges, datecols = datecols, drop_na = FALSE,
                              columns = c(dnames, "measure_val_bio"), fun = fse)
    ## eliminate unreasonably small standard error for Chl
    for (name in dnames) se[j = c(name) := ifelse(se[[..name]] > 0.01, se[[..name]], NA)]
    sd <- oppla::op_aggregate(data = lsall, edges = edges, datecols = datecols, drop_na = FALSE,
                              columns = c(dnames, "measure_val_bio"), fun = fsd)
  }
  data.table::setkeyv(avg, c(datecols, "depth"))
  data.table::setkeyv(se, c(datecols, "depth"))
  data.table::setkeyv(sd, c(datecols, "depth"))
  data.table::setkeyv(zoo, datecols)
  obs <- list(avg = avg, se = se, sd = sd, zoo = zoo, datecols = datecols)
  if (identical(transform, "log")) obs$thresholds <- thresholds
  return(obs)
}

readTwin <- function(file = "twin.nc", obs) {
  fse <- function(x, ...) {
    ## standard error
    s <- sum(complete.cases(x))
    return(if (s == 1) 0 else sqrt(var(x, ...) / s))
  }
  fsd <- function(x, ...) {
    ## standard deviation
    s <- sum(complete.cases(x))
    return(if (s == 1) 0 else sd(x, ...))
  }
  th_names <- c(DIN = "Nitrate", DIP = "Phosphate", Chl = "Chlorophyll", zooTC = "measure_val_bio")
  twin_all <- oppla::oppla_read(file = file, forcing = FALSE)
  cfn <- file.path(sub(pattern = "[^/]+$", replacement = ".", x = attr(twin_all, "file")), attr(twin_all, "control_file"))
  cf <- oppla::nml_read(file = cfn)$costfun[[1]]
  if (missing(obs)) obs <- getObs(oppla_inp = cfn)
  datecols <- obs$datecols
  pfn <- file.path(sub(pattern = "[^/]+$", replacement = ".", x = attr(twin_all, "file")), attr(twin_all, "parameter_file"))
  chl_vars <- paste(oppla::nml_read(file = pfn)$coco[[1]]$cmo, "Chl", sep = "_")
  zoo_vars <- paste(oppla::nml_read(file = pfn)$coco[[1]]$cfo, "C", sep = "_")
  twin_slab <- oppla::oppla_slab(data = twin_all,
                                 daterange = c(as.POSIXct(cf$cut_date, tz = "UTC"), tail(twin_all$date, 1)),
                                 depthrange = c(0, cf$max_depth))[
                        j = Chl := units::set_units(rowSums(.SD), units(.SD[[1, 1]])), .SDcols = chl_vars][
                        j = zooTC := units::set_units(rowSums(.SD), units(.SD[[1, 1]])), .SDcols = zoo_vars]
  names(twin_slab)[match(x = names(th_names), table = names(twin_slab))] <- th_names
  zooI <- oppla::oppla_int(data = twin_slab, columns = "measure_val_bio", avc = "month",
                           depth = units::set_units(100, "m", mode = "standard"))[month %in% obs$zoo$month]
  if (transform == "log") {
    one <- units::set_units(1, "")
    for (thn in th_names) # log transformation
      twin_slab[j = c(thn) := lapply(X = .SD,
                                     FUN = function(x) {
                                       log(units::drop_units(pmax(units::set_units(x / obs$thresholds[[thn]], 1), one)))}),
                .SDcols = thn]
    zooI[j = measure_val_bio := log(units::drop_units(pmax(units::set_units(measure_val_bio / obs$thresholds$zooI, 1), one)))]
  }
  twin_cmt <- oppla::oppla_ave(data = twin_slab, avc = datecols, columns = th_names)
  twin_se <- oppla::oppla_ave(data = twin_slab, avc = datecols, columns = th_names, fun = fse)
  twin_sd <- oppla::oppla_ave(data = twin_slab, avc = datecols, columns = th_names, fun = fsd)
  twin_zoo <- data.frame(mon = obs$zoo$mon, mean = zooI$measure_val_bio, se = obs$zoo[, "se"], sd = obs$zoo[, "sd"])
  return(list(avg = twin_cmt, se = twin_se, sd = twin_sd, zoo = twin_zoo, thresholds = obs$thresholds,
              datecols = datecols))
}

## cost function across all depth-time combinations
## model is an oppla result as obtained with oppla_read()
cost_dt <- function(obs, model, cut_date, max_depth, coco, datecols = obs$datecols, norm = "se") {
  se_standardise_dt <- function(mod, obs, err, assoc) {
    em <- sapply(X = names(assoc), FUN = function(x) {
      ise <- 1 / err[[assoc[[x]]]]
      ise[is.infinite(ise)] <- NA
      return(units::set_units((obs[[assoc[[x]]]] - mod[[x]]) * ise, 1) ^ 2)
    })
    return(cbind(data.table::as.data.table(em), mod[j = .(month, depth, height, dt)]))
  }
  se_standardize <- function(mod, obs, se) {
    ise <- 1 / se
    ise[is.infinite(ise)] <- NA
    return(mean(na.omit((obs - mod) * ise) ^ 2))
  }
  chl_vars <- paste(coco$cmo, "Chl", sep = "_")
  zoo_vars <- paste(coco$cfo, "C", sep = "_")
  model_slab <- oppla::oppla_slab(data = model,
                                  daterange = c(as.POSIXct(cut_date, tz = "UTC"), tail(model$date, 1)),
                                  depthrange = c(0, max_depth))[
                         j = Chl := units::set_units(rowSums(.SD), units(.SD[[1, 1]])), .SDcols = chl_vars][
                         j = zoo_TC := units::set_units(rowSums(.SD), units(.SD[[1, 1]])), .SDcols = zoo_vars]
  zooI <- oppla::oppla_int(data = model_slab, columns = c("zoo_TC", zoo_vars), avc = "month",
                           depth = units::set_units(100, "m", mode = "standard"))[month %in% obs$zoo$month]
  th_names <- c(DIN = "Nitrate", DIP = "Phosphate", Chl = "Chlorophyll", zoo_TC = "measure_val_bio", zoo_C = "measure_val_bio")
  if ("thresholds" %in% names(obs)) {
    ## in this case the data are log-transformed
    one <- units::set_units(1, 1)
    for (thn in names(th_names))
      model_slab[j = (thn) := log(units::drop_units(pmax(units::set_units(get(thn) /
                                                                          obs$thresholds[[th_names[[thn]]]], 1), one)))]
    zooI[j = zoo_TC := log(units::drop_units(pmax(units::set_units(zoo_TC / obs$thresholds$zooI, 1), one)))][
      j = zoo_C := log(units::drop_units(pmax(units::set_units(zoo_C / obs$thresholds$zooI, 1), one)))][
      j = zoo_noSVM_C := log(units::drop_units(pmax(units::set_units(zoo_noSVM_C / obs$thresholds$zooI, 1), one)))]
  }
  cmt <- oppla::oppla_ave(data = model_slab, columns = names(th_names), avc = datecols)
  data.table::setkeyv(cmt, c(datecols, "depth"))
  if ("year" %in% datecols) {
    mdd <- paste(cmt$date, cmt$depth) # match model and observations
    odd <- paste(obs$avg$date, obs$avg$depth)
    cmt <- cmt[which(mdd %in% odd)]
    iobs <- which(odd %in% mdd)
    obs$avg <- obs$avg[iobs]
    obs[[norm]] <- obs[[norm]][iobs]
  }
  ZooInt <- se_standardize(mod = zooI$zoo_TC, obs = obs$zoo$mean, se = obs$zoo[[norm]]) # all zooplankton
  ZooIntSVM <- se_standardize(mod = zooI$zoo_C, obs = obs$zoo$mean, se = obs$zoo[[norm]]) # only migrators
  ZooIntNoSVM <- se_standardize(mod = zooI$zoo_noSVM_C, obs = obs$zoo$mean, se = obs$zoo[[norm]]) # only non-migrators
  ZooIntT <- sqrt((ZooInt / ZooIntSVM) ^ 2 + (ZooIntSVM / 209) ^ 2) * 209 # transformed
  return(list(mod = cmt, obs = obs$avg, err = obs[[norm]],
              cost = se_standardise_dt(mod = cmt, obs = obs$avg, err = obs[[norm]],
                                       assoc = c(DIN = "Nitrate", DIP = "Phosphate",
                                                 Chl = "Chlorophyll", zoo_TC = "measure_val_bio")),
              zoo = c(zoo_I = ZooInt, zoo_SVM_I = ZooIntSVM, zoo_noSVM_I = ZooIntNoSVM)))
}

cost_pattern <- function(obs, model, cut_date, max_depth, coco) {
  control <- oppla::nml_read(attr(model, "control_file"))
  param <- oppla::nml_read(attr(model, "parameter_file"))
  if (missing(coco)) {
    cut_date <- control$costfun[[1]]$cut_date
    max_depth <- control$costfun[[1]]$max_depth
    coco <- if ("coco" %in% names(control)) control$coco[[1]] else param$coco[[1]]
  }
  daterange <- c(as.POSIXct(cut_date, tz = "UTC"), max(model$date))
  depthrange <- c(min(model$depth), max_depth)
  chl_vars <- paste(coco$cmo, "Chl", sep = "_")
  zoo_vars <- paste(coco$cfo, "C", sep = "_")
  model_slab <- oppla::oppla_slab(data = model, daterange = daterange, depthrange = depthrange)[
                         j = Chl := units::set_units(rowSums(.SD), units(.SD[[1, 1]])), .SDcols = chl_vars][
                         j = zoo_TC := units::set_units(rowSums(.SD), units(.SD[[1, 1]])), .SDcols = zoo_vars]
  zooI <- oppla::oppla_int(data = model_slab, columns = c("zoo_TC", "zoo_C", "zoo_noSVM_C"), avc = "month",
                           depth = units::set_units(100, "m", mode = "standard"))[month %in% obs$zoo$month]
  cost_names <- c(DIN = "Nitrate", DIP = "Phosphate", Chl = "Chlorophyll", zoo_TC = "measure_val_bio",
                  zoo_C = "measure_val_bio", zoo_noSVM_C = "measure_val_bio")
  if ("thresholds" %in% names(obs)) {
    one <- units::set_units(1, 1)
    for (name in names(cost_names))
      model_slab[j = (name) := log(units::drop_units(pmax(units::set_units(get(name) /
                                                                           obs$thresholds[[cost_names[[name]]]], 1), one)))]
    zooI[j = zoo_TC := log(units::drop_units(pmax(units::set_units(zoo_TC / obs$thresholds$zooI, 1), one)))][
      j = zoo_C := log(units::drop_units(pmax(units::set_units(zoo_C / obs$thresholds$zooI, 1), one)))][
      j = zoo_noSVM_C := log(units::drop_units(pmax(units::set_units(zoo_noSVM_C / obs$thresholds$zooI, 1), one)))]
  }
  cmt <- oppla::oppla_cmt(data = model_slab, columns = names(cost_names))
  cost <- list()
  for (name in names(cost_names)) {
    ind_obs <- c(which.min(obs$avg[[cost_names[[name]]]]), which.max(obs$avg[[cost_names[[name]]]]))
    ind_model <- c(which.min(cmt[[name]]), which.max(cmt[[name]]))
    ranges_obs <- obs$avg[ind_obs, .(value = get(cost_names[[name]]), month, depth = as.numeric(as.factor(depth)))]
    ranges_model <- cmt[ind_model, .(value = get(name), month, depth = as.numeric(as.factor(depth)))]
    cost[[name]] <- sum((ranges_obs$value - ranges_model$value) ^ 2 +
                        (ranges_obs$month - ranges_model$month) ^ 2 +
                        (ranges_obs$depth - ranges_model$depth) ^ 2)
  }
  for (name in c("zoo_TC", "zoo_C", "zoo_noSVM_C")) {
    ind_obs <- c(which.min(obs$zoo$mean), which.max(obs$zoo$mean))
    ind_model <- c(which.min(zooI[[name]]), which.max(zooI[[name]]))
    ranges_obs <- obs$zoo[ind_obs, .(mean, month)]
    ranges_model <- zooI[ind_model, .(mean = get(name), month)]
    cost[[paste0(name, "I")]] <- sum((ranges_obs$mean - ranges_model$mean) ^ 2 +
                                     (ranges_obs$month - ranges_model$month) ^ 2)
  }
  return(unlist(cost))
}

cost <- function(...) {
  cdt <- cost_dt(...)
  return(c(cdt$cost[j = colMeans(.SD, na.rm = TRUE), .SDcols = c("DIN", "DIP", "Chl", "zoo_TC")], cdt$zoo))
}

plot_cost <- function(cost, ylim = c(100, 0)) {
  require(ggplot2)
  require(units)
  p <- ggplot(data = data.table::melt(cost, id.vars = c("month", "depth", "height", "dt"), value.name = "error"),
              mapping = aes(x = month, y = depth, fill = error)) +
    geom_tile(mapping = aes(width = dt, height = units::drop_units(height))) + facet_wrap(vars(variable)) +
    coord_cartesian(ylim = ylim)
  return(p)
}

costfun <- function(obs, model, datecols = obs$datecols, norm = "se") {
  control <- oppla::nml_read(file = attr(model, "control_file"))
  param <- oppla::nml_read(file = attr(model, "parameter_file"))
  cf <- control$costfun[[1]]
  cdt <- cost_dt(obs = obs, model = model, cut_date = cf$cut_date, max_depth = cf$max_depth,
                 coco = if ("coco" %in% names(control)) control$coco[[1]] else param$coco[[1]],
                 datecols = datecols, norm = norm)
  return(c(cdt$cost[j = colMeans(.SD, na.rm = TRUE), .SDcols = c("DIN", "DIP", "Chl", "zoo_TC")], cdt$zoo))
}

costfunGA <- function(pop_size, cost_file = "oppla_cost.csv", error_file = "errors.csv", model_file,
                      cost_columns = c("DIN", "DIP", "Chl", "zoo_TC", "zoo_SVM_I", "zoo_noSVM_I"), cost = cost, ...) {
  oi <- oppla::onc_info(filename = file.path("fitness1", model_file), close = TRUE)
  control <- oppla::nml_read(file = file.path("fitness1", oi$att$control_file))
  param <- oppla::nml_read(file = file.path("fitness1", oi$att$parameter_file))
  cf <- control$costfun[[1]]
  if (! all(c("cut_date", "max_depth") %in% names(cf)))
    stop(paste("namelist costfun, containing cut_date and max_depth, missing or incomplete in", oi$att$control_file))
  coco <- if ("coco" %in% names(control)) control$coco[[1]] else param$coco[[1]]
  error <- sapply(X = seq_len(pop_size), FUN = function(n) {
    model <- oppla::oppla_read(file = file.path(paste0("fitness", n), model_file), forcing = FALSE)
    return(cost(model = model, cut_date = cf$cut_date, max_depth = cf$max_depth,
                coco = coco, ...))
  })
  write.table(x = error, file = error_file, sep = ",", col.names = FALSE, append = TRUE)
  write(x = "", file = error_file, append = TRUE)
  ## write averages of the cost_columns
  write(x = -colMeans(error[cost_columns, ]), file = cost_file, sep = ",", ncolumns = pop_size)
}
