! This program is the Fortran 90 version of gafort.f
!*************************************************************************
MODULE gafort_globals
  IMPLICIT NONE
  INTEGER, PARAMETER :: DP=KIND(1.0D0), K4=KIND(1), K8=8
  END MODULE gafort_globals
!*************************************************************************
MODULE ran_state
  USE gafort_globals
  IMPLICIT NONE
  INTEGER(K8), PARAMETER :: madim=55
  TYPE ranstate
     REAL(DP) :: ma(madim),mj
     INTEGER(K8) :: iff,inext,inextp,iduma
     INTEGER(K8) :: padding(128)
  END TYPE ranstate
  TYPE(ranstate) :: procstates
  INTEGER(K8) :: lenran
  !  According to Knuth, any large mbig, and any smaller (but still large)
  !  mseed can be substituted for the above values.                       
  REAL(DP), PARAMETER :: mbig=4000000.,mseed=1618033.,mz=0.,fac=1./mbig
END MODULE ran_state
!
MODULE rosenbrock_function
  IMPLICIT NONE
CONTAINS
  ! Involved multidimensional generalisation of the Rosenbrock function.  Reference:
  ! Rosenbrock function, https://en.wikipedia.org/w/index.php?title=Rosenbrock_function&oldid=993612771
  !   (last visited Mar. 9, 2021).
  FUNCTION rosenbrock (x) RESULT (r)
    USE gafort_globals
    IMPLICIT NONE
    REAL(DP), INTENT(IN) :: x(:)
    REAL(DP) :: r
    INTEGER(K4) :: n
    n=SIZE(x)
    r = SUM(100.*(x(2:n) - x(1:n-1)**2)**2 + (1 - x(1:n-1))**2)
  END FUNCTION rosenbrock
END MODULE rosenbrock_function
!*************************************************************************
!
! To download the original version 1.6.4, go to website:
!    <http://cuaerospace.com/carroll/ga.html>
!
! This update on 4/14/99.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! David L. Carroll
! CU Aerospace
! 2004 South Wright Street Extended
! Urbana, IL  61802
!
! e-mail: carroll@cuaerospace.com
! Phone:  217-333-8274
! fax:    217-244-7757
!
! This genetic algorithm (GA) driver is free for public use.  My only
! request is that the user reference and/or acknowledge the use of this
! driver in any papers/reports/articles which have results obtained
! from the use of this driver.  I would also appreciate a copy of such
! papers/articles/reports, or at least an e-mail message with the
! reference so I can get a copy.  Thanks.
!
! This program is a FORTRAN version of a genetic algorithm driver.
! This code initializes a random sample of individuals with different
! parameters to be optimized using the genetic algorithm approach, i.e.
! evolution via survival of the fittest.  The selection scheme used is
! tournament selection with a shuffling technique for choosing random
! pairs for mating.  The routine includes binary coding for the
! individuals, jump mutation, creep mutation, and the option for
! single-point or uniform crossover.  Niching (sharing) and an option
! for the number of children per pair of parents has been added.
! An option to use a micro-GA was recently added.
!
! For companies wishing to link this GA driver with an existing code,
! I am available for some consulting work.  Regardless, I suggest
! altering this code as little as possible to make future updates
! easier to incorporate.
!
! Any users new to the GA world are encouraged to read David Goldberg's
! "Genetic Algorithms in Search, Optimization and Machine Learning,"
! Addison-Wesley, 1989.
!
! Other associated files are:  gafort.in
!                              gafort.out
!                              gafort.restart (not used in SPEC version)
!                              params.f (not used in SPEC version)
!                              ReadMe
!
! I have provided a sample subroutine "func", but ultimately
! the user must supply this subroutine "func" which should be your
! cost function.  You should be able to run the code with the
! sample subroutine "func" and the provided ga.inp file and obtain
! the optimal function value of 1.0 at generation 97 with the
! uniform crossover micro-GA enabled (this is 485 function evaluations).
!
! The code is presently set for a maximum population size of 200,
! 30 chromosomes (binary bits) and 8 parameters.  These values can be
! changed in params.f as appropriate for your problem.  Correspondingly
! you will have to change a few 'write' and 'format' statements if you
! change nchrome and/or nparam.  In particular, if you change nchrome
! and/or nparam, then you should change the 'format' statement numbers
! 1075, 1275, and 1500 (see ReadMe file).
!
! Please feel free to contact me with questions, comments, or errors
! (hopefully none of latter).
!
! Disclaimer:  this program is not guaranteed to be free of error
! (although it is believed to be free of error), therefore it should
! not be relied on for solving problems where an error could result in
! injury or loss.  If this code is used for such solutions, it is
! entirely at the user's risk and the author disclaims all liability.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Input variable definitions:
!
! icreep   = 0 for no creep mutations
!          = 1 for creep mutations; creep mutations are recommended.
! idum     The initial random number seed for the GA run.  Must equal
!          a negative integer, e.g. idum=-1000.
! ielite   = 0 for no elitism (best individual not necessarily
!              replicated from one generation to the next).
!          = 1 for elitism to be invoked (best individual replicated
!              into next generation); elitism is recommended.
! iend     = 0 for normal GA run (this is standard).
!          = number of last population member to be looked at in a set
!            of individuals.  Setting iend>0 is only used for debugging
!            purposes and is commonly used in conjunction with iskip.
! iniche   = 0 for no niching
!          = 1 for niching; niching is recommended.
! irestrt  = 0 for a new GA run, or for a single function evaluation
!          = 1 for a restart continuation of a GA run.
! iskip    = 0 for normal GA run (this is standard).
!          = number in population to look at a specific individual or
!            set of individuals.  Setting iskip-0 is only used for
!            debugging purposes.
! iunifrm  = 0 for single-point crossover
!          = 1 for uniform crossover; uniform crossover is recommended.
! kountmx  = the maximum value of kount before a new restart file is
!            written; presently set to write every fifth generation.
!            Increasing this value will reduce I/O time requirements
!            and reduce wear and tear on your storage device
! maxgen   The maximum number of generations to run by the GA.
!          For a single function evaluation, set equal to 1.
! microga  = 0 for normal conventional GA operation
!          = 1 for micro-GA operation (this will automatically reset
!            some of the other input flags).  I recommend using
!            npopsiz=5 when microga=1.
! nchild   = 1 for one child per pair of parents (this is what I
!              typically use). This code is written to use 1 child.
!          = 2 for two children per pair of parents (2 is more common
!              in GA work).
! nichflg  = array of 1/0 flags for whether or not niching occurs on
!            a particular parameter.  Set to 0 for no niching on
!            a parameter, set to 1 for niching to operate on parameter.
!            The default value is 1, but the implementation of niching
!            is still controlled by the flag iniche.
! nowrite  = 0 to write detailed mutation and parameter adjustments
!          = 1 to not write detailed mutation and parameter adjustments
! nparam   Number of parameters (groups of bits) of each individual.
!          Make sure that nparam matches the number of values in the
!          parmin, parmax and nposibl input arrays.
! npopsiz  The population size of a GA run (typically 100 works well).
!          For a single calculation, set equal to 1.
! nposibl  = array of integer number of possibilities per parameter.
!            For optimal code efficiency set nposibl=2**n, i.e. 2, 4,
!            8, 16, 32, 64, etc.
! parmax   = array of the maximum allowed values of the parameters
! parmin   = array of the minimum allowed values of the parameters
! pcreep   The creep mutation probability.  Typically set this
!          = (nchrome/nparam)/npopsiz.
! pcross   The crossover probability.  For single-point crossover, a
!          value of 0.6 or 0.7 is recommended.  For uniform crossover,
!          a value of 0.5 is suggested.
! pmutate  The jump mutation probability.  Typically set = 1/npopsiz.
!
!
! For single function evaluations, set npopsiz=1, maxgen=1, & irestrt=0.
!
! My favorite initial choices of GA parameters are:
!    microga=1, npopsiz=5, inunifrm=1, maxgen=100
!    microga=1, npopsiz=5, inunifrm=0, maxgen=100
! I generally get good performance with both the uniform and single-
! point crossover micro-GA.
!
! For those wishing to use the more conventional GA techniques,
! my old favorite choice of GA parameters was:
!    iunifrm=1, iniche=1, ielite=1, nchild=1
! For most problems I have dealt with, I get good performance using
!    npopsiz=100, pcross=0.5, pmutate=0.01, pcreep=0.02, maxgen=26
! or
!    npopsiz= 50, pcross=0.5, pmutate=0.02, pcreep=0.04, maxgen=51
!
! Any negative integer for idum should work.  I typically arbitrarily
! choose idum=-10000 or -20000.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Code variable definitions (those not defined above):
!
! best     = the best fitness of the generation
! child    = the floating point parameter array of the children
! cpu      = cpu time of the calculation
! creep    = +1 or -1, indicates which direction parameter creeps
! delta    = del/nparam
! diffrac  = fraction of total number of bits which are different
!            between the best and the rest of the micro-GA population.
!            Population convergence arbitrarily set as diffrac<0.05.
! fbar     = average fitness of population
! fitness  = array of fitnesses of the parents
! fitsum   = sum of the fitnesses of the parents
! g0       = lower bound values of the parameter array to be optimized.
!            The number of parameters in the array should match the
!            dimension set in the above parameter statement.
! g1       = the increment by which the parameter array is increased
!            from the lower bound values in the g0 array.  The minimum
!            parameter value is g0 and the maximum parameter value
!            equals g0+g1*(2**g2-1), i.e. g1 is the incremental value
!            between min and max.
! ig2      = array of the number of bits per parameter, i.e. the number
!            of possible values per parameter.  For example, ig2=2 is
!            equivalent to 4 (=2**2) possibilities, ig2=4 is equivalent
!            to 16 (=2**4) possibilities.
! ig2sum   = sum of the number of possibilities of ig2 array
! ibest    = binary array of chromosomes of the best individual
! ichild   = binary array of chromosomes of the children
! icount   = counter of number of different bits between best
!            individual and other members of micro-GA population
! icross   = the crossover point in single-point crossover
! indmax   = maximum # of individuals allowed, i.e. max population size
! iparent  = binary array of chromosomes of the parents
! istart   = the generation to be started from
! jelite   = a counter which tracks the number of bits of an individual
!             which match those of the best individual
! jend     = used in conjunction with iend for debugging
! jstart   = used in conjunction with iskip for debugging
! kount    = a counter which controls how frequently the restart
!            file is written
! kelite   = kelite set to unity when jelite=nchrome, indicates that
!            the best parent was replicated amongst the children
! mate1    = the number of the population member chosen as mate1
! mate2    = the number of the population member chosen as mate2
! nchrmax  = maximum # of chromosomes (binary bits) per individual
! nchrome  = number of chromosomes (binary bits) of each individual
! ncreep   = # of creep mutations which occurred during reproduction
! nmutate  = # of jump mutations which occurred during reproduction
! nparmax  = maximum # of parameters which the chromosomes make up
! paramav  = the average of each parameter in the population
! paramsm  = the sum of each parameter in the population
! parent   = the floating point parameter array of the parents
! pardel   = array of the difference between parmax and parmin
! rand     = the value of the current random number
! npossum  = sum of the number of possible values of all parameters
! time0    = clock time at start of run
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! Subroutines:
! ____________
!
! code     = Codes floating point value to binary string.
! crosovr  = Performs crossover (single-point or uniform). This is inlined.
!	     It is no longer called from main generation loop (compiler issue!).
! decode   = Decodes binary string to floating point value.
! evalout  = Evaluates the fitness of each individual and outputs
!            generational information to the 'ga.out' file.
! func     = The function which is being evaluated.
! gamicro  = Implements the micro-GA technique.
! input    = Inputs information from the 'ga.inp' file.
! initial  = Program initialization and inputs information from the
!            'gafort.restart' file.
! mutate   = Performs mutation (jump and/or creep).
! newgen   = Writes child array back into parent array for new
!            generation; also checks to see if best individual was
!            replicated (elitism).
! niche    = Performs niching (sharing) on population.
! possibl  = Checks to see if decoded binary string falls within
!            specified range of parmin and parmax.
! ran3     = The random number generator.
! restart  = Writes the 'gafort.restart' file.
! selectn  = Performs selection; tournament selection is the only
!            option in this version of the code. This is inlined. No
!	     longer called from the main generation loop (compiler issue!).
! shuffle  = Shuffles the population randomly for selection.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
PROGRAM gafortran
  USE gafort_globals
  USE ran_state
  IMPLICIT NONE
  INTEGER(K4), PARAMETER :: indmax=400000, nchrmax=450, nparmax=30
  REAL(DP), DIMENSION(nparmax,indmax) :: parent, child
  REAL(DP), DIMENSION(indmax) :: fitness
  REAL(DP), DIMENSION(nparmax) :: g0, g1, parmax, parmin, pardel
  REAL(DP) :: pcross,pmutate,pcreep
  INTEGER(K8), DIMENSION(nparmax) :: ig2, nposibl, nichflg
  INTEGER(K8), DIMENSION(nchrmax,indmax) :: iparent, ichild
  INTEGER(K8) :: maxgen,idum,irestrt,ielite,icreep,iunifrm, &
       iniche,iskip,iend,nchild,microga,kountmx,npopsiz, &
       nowrite,nparam,nchrome
  INTEGER(K8) :: igen, maincnt,evaloutcnt,childcnt,mutatecnt,newgencnt, &
       crosovrcnt,selectcnt,shufflecnt
  CHARACTER(LEN=200) :: fmt(4), gafort_path, oppla_path='', param_file, cost_file, &
       param_name(nparmax), obs='obs', default_param_file='', oppla_file='oppla.inp',&
       costfun_script='costfun.R', costfun='cost'
  CHARACTER(LEN=1000) :: command_line
  INTEGER(K8) :: kount, istart, ig2sum, npossum, ncross, ipick, ifirst, &
       isecond, mate1, mate2, n, j, &
       icross,k,offset,ithird,iforth,lun_out,lun_vld,lun_restart,lun_param
  REAL(DP) :: fbar, best, rand, evals
  REAL(DP), DIMENSION(1000000) :: genavg,genmax
  INTEGER(K8), DIMENSION(1000000) :: geni
  REAL :: cpu0,cpu1,cpu
  INTEGER(K8), DIMENSION(nchrmax) :: ibest
  !
  CALL CPU_TIME (cpu0)
  !
  ! Call the input subroutine.
  ! TIME0=SECNDS(0.0)
  CALL input
  !
  ! Perform necessary initialization and read the gafort.restart file.
  CALL initial(istart,npossum,ig2sum)
  !
  WRITE (fmt(1),'("(I3,1X,",I0,"(1X,ES11.4))")') nparam + 1
  WRITE (fmt(2),'("(/""Average Values:"",1X,",I0,"(1X,ES11.4)/)")') nparam + 1
  WRITE (fmt(4),'("(""&PARAM""/""PARAM_VALUE = "",",I0,"(ES11.4,"",""),ES11.4/""/"")")') nparam - 1
!!!!!!! Main generational processing loop. !!!!!!!!!!!
  kount=0
  DO maincnt=1,maxgen
     igen = maincnt + istart - 1
     WRITE (6,1111) igen
     WRITE (lun_out,'("Generation:",I12/"Restart info:"/2I12)') igen, igen, npopsiz
     DO j=1,npopsiz
        WRITE(lun_out,TRIM(fmt(3))) j, iparent(1:nchrome,j)
     END DO
     WRITE (lun_out,'("Parameters and fitness:")')
     !
     !  Evaluate the population, assign fitness, establish the best
     !  individual, and write output information.
     CALL evalout(iskip,iend,ibest,fbar,best,igen)
     geni(maincnt) = igen
     genavg(maincnt)=fbar
     genmax(maincnt)=best
     IF(npopsiz.EQ.1 .OR. iskip.NE.0) THEN
        CLOSE(lun_out)
        STOP
     END IF
     !
     !  Implement "niching".
     IF (iniche.NE.0) CALL niche
     !
     !  Enter selection, crossover and mutation loop.
     ncross=0
     ipick = 1
     offset = npopsiz/4  ! npopsiz should be a multiple of 4
     ! following loop is correct for nchild=1.
     DO k = 1,4
        CALL shuffle
        ! generate one-forth of the children population now
        childcnt = childcnt + offset
        crosovrcnt = crosovrcnt + offset*nchrome
        selectcnt = selectcnt + offset*nchrome
        DO j=ipick,ipick+offset-1,nchild
           ! Perform selection.
           CALL selectn(k,j,mate1,mate2)
           ! implement "tournament" selection for selection of new population.
           ifirst=(4*(j-1)+1)-(npopsiz*(k-1))
           isecond=ifirst+1
           ithird=ifirst+2
           iforth=ifirst+3
           IF (fitness(ifirst).GT.fitness(isecond)) THEN
              mate1=ifirst
           ELSE
              mate1=isecond
           END IF
           IF (fitness(ithird).GT.fitness(iforth)) THEN
              mate2=ithird
           ELSE
              mate2=iforth
           END IF
           DO n=1,nchrome
              ichild(n,j)=iparent(n,mate1)
              IF(nchild.EQ.2) ichild(j+1,n)=iparent(n,mate2)
           END DO
           ! Now perform crossover between the randomly selected pair.
           CALL crosovr(ncross,j,mate1,mate2)
           IF (iunifrm.EQ.0) THEN
              !  Single-point crossover at a random chromosome point.
              CALL ran3(INT(1, K8),rand)
              IF (rand.GT.pcross) CYCLE
              ncross=ncross+1
              CALL ran3(INT(1, K8),rand)
              icross = 2 + INT(DBLE(nchrome-1)*rand)
              DO n=icross,nchrome
                 ichild(n,j)=iparent(n,mate2)
                 IF(nchild.EQ.2) ichild(j+1,n)=iparent(n,mate1)
              END DO
           ELSE
              !  Perform uniform crossover between the randomly selected pair.
              DO n=1,nchrome
                 CALL ran3(INT(1, K8),rand)
                 IF (rand.LE.pcross) THEN
                    ichild(n,j)=iparent(n,mate2)
                    IF(nchild.EQ.2) ichild(j+1,n)=iparent(n,mate1)
                 END IF
              END DO
           END IF
        END DO
        ipick = ipick + offset
     END DO
     write(6,1225) ncross
     write(lun_out,1225) ncross
     !
     ! Now perform random mutations.  If running micro-GA, skip mutation.
     !
     IF (microga.EQ.0) CALL mutate
     !
     ! Write child array back into parent array for new generation.  Check
     ! to see IF the best parent was replicated.
     !
     CALL newgen(ielite,npossum,ig2sum,ibest)
     !
     !  Implement micro-GA IF enabled.
     !
     IF (microga.NE.0) CALL gamicro(igen,npossum,ig2sum,ibest)
     !
     !  Write to restart file.
     CALL restart(igen,istart,kount)
  END DO
!!!!!!! End of main generational processing loop. !!!!!!
  ! 999 CONTINUE 
  WRITE(lun_out,3000)
  DO maincnt=1,maxgen
     evals=npopsiz*geni(maincnt)
     WRITE(lun_out,3100) geni(maincnt),evals,genavg(maincnt),genmax(maincnt)
  END DO
  ! write validation counters
  WRITE(lun_vld,5100) 
  WRITE(lun_vld,5200) maincnt
  WRITE(lun_vld,5200) evaloutcnt
  WRITE(lun_vld,5200) childcnt
  WRITE(lun_vld,5200) crosovrcnt
  WRITE(lun_vld,5200) selectcnt
  WRITE(lun_vld,5200) shufflecnt
  WRITE(lun_vld,5200) newgencnt
  WRITE(lun_vld,5200) mutatecnt
  WRITE(lun_vld,5300) genmax(maxgen)
  CALL CPU_TIME (cpu1)
  cpu=(cpu1-cpu0)
  WRITE(6,1400) cpu,cpu/60.0
  WRITE(lun_out,1400) cpu,cpu/60.0
  CLOSE (lun_out)
1111 FORMAT(//'#################  Generation',I5,'  #################')
1225 FORMAT(/'  Number of Crossovers      =',I5)
1400 FORMAT(2X,'CPU time for all generations=',E13.6,' sec'/  &
       2X,'                             ',E13.6,' min')
3000 FORMAT(2X,'Generation   Evaluations   Avg.Fitness   Best Fitness')
3100 FORMAT(2X,I0,3(E11.4,3X))
  !
5100 FORMAT(2X,'Gafort Validation Counters and Best Fitness')
5200 FORMAT(2X,I20)
5300 FORMAT(2X,F10.9)
  !
  IF (lenran > 0) THEN
     lenran = 0
  END IF
  CLOSE(lun_vld)
  CLOSE(lun_param)
  !
CONTAINS
  !***********************************************************************
  SUBROUTINE input
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    ! This subroutine inputs information from the ga.inp (gafort.in) file. !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    INTEGER(K8) :: i,itourny,lun_in
    LOGICAL :: have_restart
    NAMELIST /ga/ irestrt,npopsiz,pmutate,maxgen,idum,pcross,      &
         itourny,ielite,icreep,pcreep,iunifrm,iniche,              &
         iskip,iend,nchild,nparam,parmin,parmax,nposibl,           &
         nowrite,nichflg,microga,kountmx,param_name,obs,           &
         default_param_file,oppla_file,costfun_script,costfun
    !
    kountmx=1
    irestrt=0
    ielite=0
    iunifrm=0
    iniche=0
    iskip=0
    iend=0
    nchild=1
    nichflg=1
    microga=0
    nposibl=0
    param_name=''
    !
    OPEN (NEWUNIT=lun_out, FILE='gafort.out', STATUS='UNKNOWN')
    REWIND lun_out
    OPEN (NEWUNIT=lun_in, FILE='gafort.in', STATUS='OLD')
    READ (lun_in, NML=ga)
    CLOSE (lun_in)
    IF (COUNT((/(LEN_TRIM(param_name(i)) > 0, i=INT(1, K8),nparam)/)).NE.nparam) &
         STOP 'Number of parameter names must be nparam!'
    OPEN (NEWUNIT=lun_vld, FILE='gafort.vld', STATUS='UNKNOWN')
    REWIND lun_vld
    IF (irestrt.EQ.1) THEN
       INQUIRE (FILE='gafort.restart', EXIST=have_restart)
       IF (.NOT.have_restart) irestrt = 0
    END IF
    IF (irestrt.EQ.1) THEN
       OPEN (NEWUNIT=lun_param, FILE='gafort_param.csv', STATUS='OLD', POSITION='APPEND')
    ELSE
       OPEN (NEWUNIT=lun_param, FILE='gafort_param.csv', STATUS='UNKNOWN')
       WRITE (UNIT=lun_param, FMT='(100(A,:,","))') (TRIM(param_name(i)), i=1,nparam), "fitness"
    END IF
    ! Check for array sizing errors.
    IF ((npopsiz.GT.1).AND.INT(npopsiz/4)*4.NE.npopsiz) THEN
       WRITE(6,1800)
       WRITE(lun_out,1800)
       CLOSE(lun_out)
       STOP
    END IF
    IF (npopsiz.GT.indmax) THEN
       WRITE(6,1600) npopsiz
       WRITE(lun_out,1600) npopsiz
       CLOSE(lun_out)
       STOP
    END IF
    IF (nparam.GT.nparmax) THEN
       WRITE(6,1700) nparam
       WRITE(lun_out,1700) nparam
       CLOSE(lun_out)
       STOP
    END IF
    IF (COUNT(nposibl.GT.0).EQ.1) nposibl(1:nparam) = nposibl(1)
    IF (ANY(nposibl(1:nparam).EQ.0)) THEN
       WRITE(6,1900) nparam
       WRITE(lun_out,1900) nparam
       CLOSE(lun_out)
       STOP
    END IF
    ! If using the microga option, reset some input variables
    IF (microga.NE.0) THEN
       pmutate=0.0d0
       pcreep=0.0d0
       ielite=1
       iniche=0
       nchild=1
       IF (iunifrm.EQ.0) THEN
          pcross=1.0d0
       ELSE
          pcross=0.5d0
       END IF
    END IF
    !
1600 FORMAT(1X,'ERROR: npopsiz > indmax.  Set indmax = ',I6)
1700 FORMAT(1X,'ERROR: nparam > nparmax.  Set nparmax = ',I6)
1800 FORMAT(1X,'ERROR: npopsiz should be a multiple of 4')
1900 FORMAT(1X,'ERROR: nposibl should contain 1 or ',I0,' entries')
  END SUBROUTINE input
!***********************************************************************
  SUBROUTINE initial(istart,npossum,ig2sum)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!! This subroutine sets up the program by generating the g0, g1 and    !
!!! ig2 arrays, and counting the number of chromosomes required for the !
!!! specified input.  The subroutine also initializes the random number !
!!! generator, parent and iparent arrays (reads gafort.restart).        !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    INTEGER(K8), INTENT(OUT) :: istart, npossum, ig2sum
    INTEGER(K8) :: i, j, k, l, lun
    REAL(DP) :: rand
    INTEGER(K8), DIMENSION(nparam) :: ig2sum1
    LOGICAL :: oppla_found
    CALL GET_COMMAND_ARGUMENT (NUMBER=0, VALUE=gafort_path)
    gafort_path = gafort_path(1:SCAN(TRIM(gafort_path), '/', BACK=.TRUE.))
    IF (LEN_TRIM(gafort_path).GT.0) &
         oppla_path = gafort_path(1:SCAN(gafort_path(1:(LEN_TRIM(gafort_path)-1)), '/', BACK=.TRUE.))
    INQUIRE (FILE=TRIM(oppla_path)//'oppla', EXIST=oppla_found)
    IF (.NOT.oppla_found) THEN
       WRITE (*,'("Cannot find oppla, expected in ",A)') TRIM(oppla_path)
       STOP
    END IF
    ! generate obs.rds for use in costfun.R
    SELECT CASE (TRIM(obs))
    CASE ('obs')
       WRITE (command_line,'("Rscript --slave -e ''source(""",A,"""); obs <- getObs(oppla_inp = """,A,"""); &
            &saveRDS(obs, ""obs.rds"")''")') TRIM(costfun_script), TRIM(oppla_file)
    CASE ('twin')
       WRITE (command_line,'("Rscript --slave -e ''source(""",A,"""); &
            &twin <- readTwin(obs = getObs()); saveRDS(twin, ""obs.rds"")''")') TRIM(costfun_script)
    CASE DEFAULT
       STOP 'invalid setting for obs in gafort.in'
    END SELECT
    CALL EXECUTE_COMMAND_LINE (TRIM(command_line))
    g0(1:nparam)=parmin(1:nparam)
    pardel(1:nparam)=parmax(1:nparam)-parmin(1:nparam)
    g1(1:nparam)=pardel(1:nparam)/DBLE(nposibl(1:nparam)-1)
    ig2(1:nparam)=CEILING(LOG(REAL(nposibl(1:nparam), DP))/LOG(2._DP))
    IF (ANY(ig2(1:nparam).GT.30)) THEN
       WRITE(6,2000)
       WRITE(lun_out,2000)
       CLOSE(lun_out)
       STOP
    END IF
    ! Count the total number of chromosomes (bits) required
    nchrome=SUM(ig2(1:nparam))
    npossum=SUM(nposibl(1:nparam))
    ig2sum1(1:nparam)=2**ig2(1:nparam)
    ig2sum=SUM(ig2sum1(1:nparam))
    IF (nchrome.GT.nchrmax) THEN
       WRITE(6,1800) nchrome
       WRITE(lun_out,1800) nchrome
       CLOSE(lun_out)
       STOP
    END IF
    IF (npossum.LT.ig2sum .AND. microga.NE.0) THEN
       WRITE(6,2100)
       WRITE(lun_out,2100)
    END IF
    !
    DO i=1,npopsiz
       fitness(i)=0
       DO j=1,nchrome
          iparent(j,i)=0
          ichild(j,i)=0
       ENDDO
       DO j=1,nparam
          parent(j,i)=0
          child(j,i)=0
       END DO
    END DO
    !
    WRITE (fmt(3),'("(I5,3X,",I0,"I1)")') nchrome
    !
    ! Initialize random number generator
    CALL ran3(idum,rand)
    !
    IF (LEN_TRIM(default_param_file).EQ.0) STOP 'default_param_file must be specified in namelist ga!'
    IF (irestrt.EQ.0) THEN
       !  Initialize the random distribution of parameters in the individual
       !  parents when irestrt=0.
       istart=1
       !  Use the default parameter set for the first individual if it is not a
       !  restart.  The other individuals are assigned random values.
       CALL EXECUTE_COMMAND_LINE ('Rscript --slave -e ''source("' // TRIM(gafort_path) &
            // 'get_default_parameters.R"); get_default_parameters(default_parameter_file = "' &
            // TRIM(default_param_file) // '", csv_file = "default_parameters.csv")''')
       OPEN (NEWUNIT=lun, FILE='default_parameters.csv', STATUS='OLD')
       READ (lun,*) parent(1:nparam,1)
       CLOSE (lun)
       IF (ANY((parent(1:nparam,1).LT.parmin(1:nparam)).AND.(parent(1:nparam,1).GT.parmax(1:nparam)))) THEN
          DO k=1,nparam
             WRITE (*,'("default for parameter ",A," is not between ",G0," and ",G0,".")') param_name(k), parmin(k), parmax(k)
          END DO
       END IF
       DO k=1,nparam
          CALL code (INT(1, K8),k,parent,iparent)
       END DO
       DO i=2,npopsiz
          DO j=1,nchrome
             CALL ran3(INT(1, K8),rand)
             iparent(j,i)=NINT(rand)
          END DO
       END DO
       IF (npossum.LT.ig2sum) CALL possibl(parent,iparent)
    ELSE
       OPEN (NEWUNIT=lun_restart, FILE='gafort.restart', STATUS='OLD')
       READ(lun_restart,*) istart,npopsiz
       DO j=1,npopsiz
          READ(lun_restart,TRIM(fmt(3))) k,(iparent(l,j),l=1,nchrome)
       END DO
       CLOSE (lun_restart)
    END IF
    IF(irestrt.NE.0) CALL ran3(idum-istart,rand)
1800 FORMAT(1X,'ERROR: nchrome > nchrmax.  Set nchrmax = ',I6)
2000 FORMAT(1X,'ERROR: You have a parameter with a number of '/           &
         1X,'   possibilities > 2**30!  If you really desire this,'/   &
         1X,'   change the code and recompile.'//       &
         1X,'   You may also need to alter the code to work with'/     &
         1X,'   REAL numbers rather than INTEGER numbers; Fortran'/    &
         1X,'   does not like to compute 2**j when j>30.')             
2100 FORMAT(1X,'WARNING: for some cases, a considerable performance'/     &
         1X,'   reduction has been observed when running a non-'/      &
         1X,'   optimal number of bits with the micro-GA.'/            &
         1X,'   If possible, use values for nposibl of 2**n,'/         &
         1X,'   e.g. 2, 4, 8, 16, 32, 64, etc.  See ReadMe file.')
  END SUBROUTINE initial
  !***********************************************************************
  SUBROUTINE evalout(iskip,iend,ibest,fbar,best,igen)
    USE rosenbrock_function
    SAVE
    INTEGER(K8), INTENT(IN) :: iskip,iend,igen
    INTEGER(K8), INTENT(OUT), DIMENSION(nchrmax) :: ibest
    REAL(DP), INTENT(OUT) :: fbar,best
    REAL(DP), DIMENSION(nparmax) :: paramav,paramsm
    INTEGER(K8) :: j, k, kk, jstart, jend
    REAL(DP) :: fitsum
    REAL(DP), DIMENSION(nparmax) :: paramsm1
    INTEGER(K8) :: jmax(1), lun_pf, lun_cost
    !
    paramsm(1:nparam)=0.0D0
    jstart=1
    jend=npopsiz
    IF(iskip.NE.0) jstart=iskip
    IF(iend.NE.0) jend=iend
    paramsm1(1:nparmax) = 0.0
    WRITE (param_file, '("oppla_gar_",I0,".nml")') igen
    OPEN (NEWUNIT=lun_pf, FILE=param_file)     ! rbtest
    DO j=jstart,jend
       CALL decode(j,parent,iparent)
       IF(iskip.NE.0 .AND. iend.NE.0 .AND. iskip.EQ.iend) &
            WRITE(6,TRIM(fmt(1))) j, (parent(kk,j),kk=1,nparam),0.0
       !
       !  Call function evaluator, write out individual and fitness, and add
       !  to the summation for later averaging.
       paramsm1(1:nparam) = paramsm1(1:nparam) + parent(1:nparam,j)
       ! write a namelist with parameter values to be read by R
       WRITE(UNIT=lun_pf, FMT=TRIM(fmt(4))) parent(1:nparam,j)   ! rbtest
    END DO
    CLOSE (UNIT=lun_pf)
    WRITE (cost_file,'("oppla_cost_GA",I0,".csv")') igen
    WRITE (command_line,'(A,"costfun.sh",1X,I0,1X,A,1X,A,1X,I0,1X,A,1X,A,1X,A,1X,A,1X,A,1X,A,1X,A)') &
         TRIM(gafort_path), npopsiz, 'gafort.in', TRIM(cost_file), igen, TRIM(param_file),&
         TRIM(default_param_file), TRIM(oppla_file), TRIM(gafort_path), TRIM(oppla_path),&
         TRIM(costfun_script), TRIM(costfun)
    CALL EXECUTE_COMMAND_LINE (TRIM(command_line))
    OPEN(NEWUNIT=lun_cost, FILE=cost_file)
    READ(UNIT=lun_cost,FMT=*) fitness(jstart:jend)
    CLOSE(UNIT=lun_cost)
    !
    DO j=jstart,jend
       WRITE(lun_out,TRIM(fmt(1))) j,parent(1:nparam,j),fitness(j)
       WRITE(UNIT=lun_param, FMT='(100(ES14.6,:,","))') parent(1:nparam,j),fitness(j)
    END DO
    ! Identify individual jmax with the best fitness.
    jmax = MAXLOC(fitness(jstart:jend))
    best = fitness(jmax(1))
    ibest(1:nchrome)=iparent(1:nchrome,jmax(1))
    paramsm(1:nparam)=paramsm(1:nparam)+paramsm1(1:nparam)
    evaloutcnt = evaloutcnt + (jend-jstart)
    fitsum = SUM(fitness(jstart:jend))
    fbar=fitsum/DBLE(npopsiz)
    paramav(1:nparam)=paramsm(1:nparam)/DBLE(npopsiz)
    !
    !  Write output information
    IF (npopsiz.EQ.1) THEN 
       WRITE(lun_out,TRIM(fmt(1))) 1, (parent(k,1),k=1,nparam),fitness(1)
       WRITE(lun_out,fmt(2)) (parent(k,1),k=1,nparam),fbar
    ELSE 
       WRITE(lun_out,fmt(2)) (paramav(k),k=1,nparam),fbar
    END IF
    WRITE(6,1100) fbar
    WRITE(lun_out,1100) fbar
    WRITE(6,1200) best
    WRITE(lun_out,1200) best

1100 FORMAT(1X,'Average Function Value of Generation = ',ES11.4)
1200 FORMAT(1X,'Maximum Function Value               = ',ES11.4/)
    
  END SUBROUTINE evalout
  !*************************************************************************
  SUBROUTINE niche
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !  Implement "niching" through Goldberg's multidimensional phenotypic    !
    !  sharing scheme with a triangular sharing function.  To find the       !
    !  multidimensional distance from the best individual, normalize all     !
    !  parameter differences.                                                !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !   Variable definitions:                                                !
    !                                                                        !
    !  alpha   = power law exponent for sharing function; typically = 1.0    !
    !  del     = normalized multidimensional distance between ii and all     !
    !            other members of the population                             !
    !            (equals the square root of del2)                            !
    !  del2    = sum of the squares of the normalized multidimensional       !
    !            distance between member ii and all other members of         !
    !            the population                                              !
    !  nniche  = number of niched parameters                                 !
    !  sigshar = normalized distance to be compared with del; in some sense, !
    !            1/sigshar can be viewed as the number of regions over which !
    !            the sharing function should focus, e.g. with sigshar=0.1,   !
    !            the sharing function will try to clump in ten distinct      !
    !            regions of the phase space.  A value of sigshar on the      !
    !            order of 0.1 seems to work best.                            !
    !  share   = sharing function between individual ii and j                !
    !  sumshar = sum of the sharing functions for individual ii              !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    SAVE
    INTEGER(K8) :: nniche, ii, j, k
    ! REAL(DP) :: alpha
    REAL(DP) :: sigshar, sumshar, del2, del, share
    ! alpha=1.0
    sigshar=0.1d0
    nniche=SUM(nichflg(1:nparam))
    IF (nniche.EQ.0) THEN
       WRITE(6,1900)
       WRITE(lun_out,1900)
       CLOSE(lun_out)
       STOP
    END IF
    DO ii=1,npopsiz
       sumshar=0.0d0
       DO j=1,npopsiz
          del2=0.0d0
          DO k=1,nparam
             IF (nichflg(k).NE.0) THEN
                del2=del2+((parent(k,j)-parent(k,ii))/pardel(k))**2
             END IF
          END DO
          del=(DSQRT(del2))/DBLE(nniche)
          IF (del.LT.sigshar) THEN
             !        share=1.0-((del/sigshar)**alpha)
             share=1.0d0-(del/sigshar)
          ELSE
             share=0.0d0
          END IF
          sumshar=sumshar+share/DBLE(npopsiz)
       END DO
       IF (ABS(sumshar).GT.0.0d0) fitness(ii)=fitness(ii)/sumshar
    END DO
1900 FORMAT(1X,'ERROR: iniche=1 and all values in nichflg array = 0'/   &
          1X,'       Do you want to niche or not?')
   END SUBROUTINE niche
   !*************************************************************************
   SUBROUTINE selectn(k,j,mate1,mate2)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     ! Subroutine for selection operator.  Presently, tournament selection !
     ! is the only option available.                                       !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     INTEGER(K8), INTENT(IN) :: j,k
     INTEGER(K8), INTENT(OUT) :: mate1,mate2
     INTEGER(K8) :: n,ifirst,isecond,ithird,iforth
     ! implement "tournament" selection for selection of new population.
     ifirst=(4*(j-1)+1)-(npopsiz*(k-1))
     isecond=ifirst+1
     ithird=ifirst+2
     iforth=ifirst+3
     ! select first mate
     IF (fitness(ifirst).GT.fitness(isecond)) THEN
        mate1=ifirst
     ELSE
        mate1=isecond
     END IF
     ! select second mate
     IF (fitness(ithird).GT.fitness(iforth)) THEN
        mate2=ithird
     ELSE
        mate2=iforth
     END IF
     !  WRITE(3,*) mate1,mate2,fitness(mate1),fitness(mate2)
     DO n=1,nchrome
        ichild(n,j)=iparent(n,mate1)
        IF(nchild.EQ.2) ichild(j+1,n)=iparent(n,mate2)
     END DO
   END SUBROUTINE selectn
   !*************************************************************************
   SUBROUTINE crosovr(ncross,j,mate1,mate2)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     ! Subroutine for crossover between the randomly selected pair. !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     INTEGER(K8), INTENT(IN) :: j, mate1, mate2
     INTEGER(K8), INTENT(INOUT) :: ncross
     INTEGER(K8) :: icross, n
     REAL(DP) :: rand
     !
     IF (iunifrm.EQ.0) THEN
        !  Single-point crossover at a random chromosome point.
        CALL ran3(INT(1, K8),rand)
        IF (rand.GT.pcross) GOTO 69
        ncross=ncross+1
        CALL ran3(INT(1, K8),rand)
        icross = 2 + INT(DBLE(nchrome-1)*rand)
        DO n=icross,nchrome
           ichild(n,j)=iparent(n,mate2)
           IF(nchild.EQ.2) ichild(j+1,n)=iparent(n,mate1)
        END DO
     ELSE
        !  Perform uniform crossover between the randomly selected pair.
        DO n=1,nchrome
           CALL ran3(INT(1, K8),rand)
           IF (rand.LE.pcross) THEN
              ncross=ncross+1
              ichild(n,j)=iparent(n,mate2)
              IF(nchild.EQ.2) ichild(j+1,n)=iparent(n,mate1)
           END IF
        END DO
     END IF
69   CONTINUE
   END SUBROUTINE crosovr
   !*************************************************************************
   SUBROUTINE mutate
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     ! This subroutine performs mutations on the children generation.        !
     ! Perform random jump mutation if a random number is less than pmutate. !
     ! Perform random creep mutation if a different random number is less    !
     ! than pcreep.                                                          !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     INTEGER(K8) :: nmutate, ncreep
     INTEGER(K8) :: j, k
     REAL(DP) :: rand, creep
     nmutate=0
     ncreep=0
     DO j=1,npopsiz
        DO k=1,nchrome
           !     Jump mutation
           CALL ran3(INT(1, K8),rand)
           IF (rand.LE.pmutate) THEN
              nmutate=nmutate+1
              ichild(k,j) = 1 - ichild(k,j)
              IF (nowrite.EQ.0) WRITE(6,1300) j,k
              IF (nowrite.EQ.0) WRITE(lun_out,1300) j,k
           END IF
        END DO
        !  Creep mutation (one discrete position away).
        IF (icreep.NE.0) THEN
           DO k=1,nparam
              CALL ran3(INT(1, K8),rand)
              IF (rand.LE.pcreep) THEN
                 CALL decode(j,child,ichild)
                 ncreep=ncreep+1
                 creep=1.0d0
                 CALL ran3(INT(1, K8),rand)
                 IF (rand.LT.0.5d0) creep=-1.0d0
                 child(k,j)=child(k,j)+g1(k)*creep
                 IF (child(k,j).GT.parmax(k)) THEN
                    child(k,j)=parmax(k)-1.0d0*g1(k)
                 ELSE IF (child(k,j).LT.parmin(k)) THEN
                    child(k,j)=parmin(k)+1.0d0*g1(k)
                 END IF
                 CALL code(j,k,child,ichild)
                 IF (nowrite.EQ.0) WRITE(6,1350) j,k
                 IF (nowrite.EQ.0) WRITE(lun_out,1350) j,k
              END IF
           END DO
        END IF
     END DO
     mutatecnt = mutatecnt + npopsiz
     !WRITE(6,1250) nmutate,ncreep
     !WRITE(lun_vld,1250) nmutate,ncreep
     !
     !1250 FORMAT(/'  Number of Jump Mutations  =',I5/  &
     !           '  Number of Creep Mutations =',I5)
1300 FORMAT('*** Jump mutation performed on individual  ',I4,  &
          ', chromosome ',I3,' ***')
1350 FORMAT('*** Creep mutation performed on individual ',I4,  &
          ', parameter  ',I3,' ***')
     !
   END SUBROUTINE mutate
   !*************************************************************************
   SUBROUTINE newgen (ielite,npossum,ig2sum,ibest)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     ! Write child array back into parent array for new generation.  Check !
     ! to see if the best parent was replicated; if not, and if ielite=1,  !
     ! then reproduce the best parent into a random slot.                  !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     SAVE
     INTEGER(K8), INTENT(IN) :: ielite, npossum, ig2sum
     INTEGER(K8), DIMENSION(nchrmax) :: ibest
     INTEGER(K8) :: j, n, kelite, jelite, irand
     INTEGER(K8), DIMENSION(1) :: kelite1
     REAL(DP) :: rand
     !
     IF (npossum.LT.ig2sum) CALL possibl(child,ichild)
     kelite1(1) = 0
     DO j=1,npopsiz
        jelite=0
        DO n=1,nchrome
           iparent(n,j)=ichild(n,j)
           IF (iparent(n,j).EQ.ibest(n)) jelite=jelite+1
           IF (jelite.EQ.nchrome) kelite1(1)=1
        END DO
     END DO
     kelite=SUM(kelite1) ! If any elements are 1, kelite>0, else kelite=0.
     newgencnt = newgencnt + npopsiz*nchrome
     IF (ielite.NE.0 .AND. kelite.EQ.0) THEN
        CALL ran3(INT(1, K8),rand)
        irand = 1 + INT(DBLE(npopsiz)*rand)
        iparent(1:nchrome,irand)=ibest(1:nchrome)
        WRITE(lun_out,1260) irand
     END IF
     !
1260 FORMAT('  Elitist Reproduction on Individual ',I4)
     !
   END SUBROUTINE newgen
   !*************************************************************************
   SUBROUTINE gamicro(i,npossum,ig2sum,ibest)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     ! Micro-GA implementation subroutine !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     SAVE
     INTEGER(K8), INTENT(IN) :: i, npossum, ig2sum
     INTEGER(K8), DIMENSION(nchrmax) :: ibest
     INTEGER(K8) :: icount, j, n
     REAL(DP) :: diffrac, rand
     ! First, check for convergence of micro population.
     ! If converged, start a new generation with best individual and fill
     ! the remainder of the population with new randomly generated parents.
     !
     ! Count number of different bits from best member in micro-population
     icount=0
     DO j=1,npopsiz
        DO n=1,nchrome
           IF(iparent(n,j).NE.ibest(n)) icount=icount+1
        END DO
     END DO
     !  If icount less than 5% of number of bits, then consider population
     !  to be converged.  Restart with best individual and random others.
     diffrac=DBLE(icount)/DBLE((npopsiz-1)*nchrome)
     IF (diffrac.LT.0.05d0) THEN
        iparent(1:nchrome,1)=ibest(1:nchrome)
        DO j=2,npopsiz
           DO n=1,nchrome
              CALL ran3(INT(1, K8),rand)
              iparent(n,j)=1
              IF(rand.LT.0.5d0) iparent(n,j)=0
           END DO
        END DO
        IF (npossum.LT.ig2sum) CALL possibl(parent,iparent)
        WRITE(6,1375) i
        WRITE(lun_out,1375) i
     END IF
     !
1375 FORMAT(//'%%%%%%%  Restart micro-population at generation',    &
          I5,'  %%%%%%%')
   END SUBROUTINE gamicro
   !*************************************************************************
   SUBROUTINE shuffle
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     ! This routine shuffles the parent array and its corresponding fitness !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     INTEGER(K8) :: j, itemp(nchrome), iother
     REAL(DP) :: rand, temp
     DO j=1,npopsiz-1
        CALL ran3(INT(1, K8),rand)
        iother = j + 1 + INT(DBLE(npopsiz-j)*rand)
        itemp(1:nchrome)=iparent(1:nchrome,iother)
        iparent(1:nchrome,iother)=iparent(1:nchrome,j)
        iparent(1:nchrome,j)=itemp(1:nchrome)
        temp=fitness(iother)
        fitness(iother)=fitness(j)
        fitness(j)=temp
     END DO
     shufflecnt = shufflecnt + npopsiz - 1
   END SUBROUTINE shuffle
   !*************************************************************************
   SUBROUTINE decode(i,array,iarray)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     ! This routine decodes a binary string to a real number.!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     INTEGER(K8), INTENT(IN) :: i
     INTEGER(K8), DIMENSION(nchrmax,indmax) :: iarray
     REAL(DP), INTENT(OUT), DIMENSION(nparmax,indmax) :: array
     INTEGER(k8) :: l, k, iparam, j, m
     l=1
     DO k=1,nparam
        iparam=0
        m=l
        DO j=m,m+ig2(k)-1
           l=l+1
           iparam=iparam+iarray(j,i)*(2**(m+ig2(k)-1-j))
        END DO
        array(k,i)=g0(k)+g1(k)*DBLE(iparam)
     END DO
   END SUBROUTINE decode
   !*************************************************************************
   SUBROUTINE code(j,k,array,iarray)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This routine codes a parameter into a binary string.!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     INTEGER(K8), INTENT(IN) :: j, k
     INTEGER(K8), INTENT(INOUT), DIMENSION(nchrmax,indmax) :: iarray
     REAL(DP), INTENT(IN), DIMENSION(nparmax,indmax) :: array
     INTEGER(K8) :: i, iparam, istart, m
     ! First, establish the beginning location of the parameter string of
     ! interest.
     istart=1+SUM(ig2(1:k-1))
     !  Find the equivalent coded parameter value, and back out the binary
     !  string by factors of two.
     m=ig2(k)-1
     IF (g1(k).EQ.0.0d0) RETURN
     iparam=nint((array(k,j)-g0(k))/g1(k))
     DO i=istart,istart+ig2(k)-1
        iarray(i,j)=0
        IF ((iparam+1).GT.(2**m)) THEN
           iarray(i,j)=1
           iparam=iparam-2**m
        END IF
        m=m-1
     END DO
     !  WRITE(3,*)array(k,j),iparam,(iarray(i,j),i=istart,istart+ig2(k)-1)
   END SUBROUTINE code
   !*************************************************************************
   SUBROUTINE possibl(array,iarray)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!  This subroutine determines whether or not all parameters are within !
!!!  the specified range of possibility.  If not, the parameter is       !
!!!  randomly reassigned within the range.  This subroutine is only      !
!!!  necessary when the number of possibilities per parameter is not     !
!!!  optimized to be 2**n, i.e. if npossum < ig2sum.                     !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     REAL(DP), INTENT(OUT), DIMENSION(nparmax,indmax) :: array
     INTEGER(K8), INTENT(INOUT), DIMENSION(nchrmax,indmax) :: iarray
     INTEGER(K8) :: i, j, n2ig2j, irand
     REAL(DP) :: rand
     DO i=1,npopsiz
        CALL decode(i,array,iarray)
        DO j=1,nparam
           n2ig2j=2**ig2(j)
           IF (nposibl(j).NE.n2ig2j .AND. array(j,i).GT.parmax(j)) THEN
              CALL ran3(INT(1, K8),rand)
              irand = INT(DBLE(nposibl(j))*rand)
              array(j,i)=g0(j)+DBLE(irand)*g1(j)
              CALL code(i,j,array,iarray)
              IF (nowrite.EQ.0) WRITE(6,1000) i,j
              IF (nowrite.EQ.0) WRITE(lun_out,1000) i,j
           END IF
        END DO
     END DO
1000 FORMAT('*** Parameter adjustment to individual     ',I4,   &
          ', parameter  ',I3,' ***')
   END SUBROUTINE possibl
   !*************************************************************************
   SUBROUTINE restart(i,istart,kount)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     ! This subroutine writes restart information to the gafort.restart file.!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     SAVE
     INTEGER(K8), INTENT(IN) :: i, istart
     INTEGER(K8), INTENT(INOUT) :: kount
     INTEGER(K8) :: j, l
     kount=kount+1
     IF (i.EQ.maxgen+istart-1 .OR. kount.EQ.kountmx) THEN
        OPEN (NEWUNIT=lun_restart, FILE='gafort.restart', STATUS='UNKNOWN')
        REWIND lun_restart
        WRITE(lun_restart,*) i+1,npopsiz
        DO j=1,npopsiz
           WRITE(lun_restart,TRIM(fmt(3))) j,(iparent(l,j),l=1,nchrome)
        END DO
        CLOSE (lun_restart)
        kount=0
     END IF

   END SUBROUTINE restart
   !*************************************************************************
   SUBROUTINE ran_init(idum)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     ! Initialize the state of the random number generator per processor.   !
     ! To avoid false sharing and increase spatial locality, we allocate    !
     ! one structure ranstate per processor and maintain pointers to it.    !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     USE ran_state
     INTEGER(K8), INTENT(IN) :: idum
     INTEGER(K8) :: i,ii,k,j
     REAL(DP) :: mk
     !
     ! See if we need to deallocate space.
     !
     IF (lenran > 0) THEN
        lenran=0
     END IF
     !
     ! Allocate memory for the state variables.
     !
     procstates%iff=0
     procstates%iduma=idum*j  ! Generate seed value for each processor.
     IF (procstates%iff.EQ.0) THEN
        procstates%iff=1
        procstates%mj=mseed-DBLE(ABS(procstates%iduma))
        procstates%mj=DMOD(procstates%mj,mbig)
        procstates%ma(madim)=procstates%mj
        mk=1
        DO i=1,madim-1
           ii=MOD(INT(21, K8)*i,madim)
           procstates%ma(ii)=mk
           mk=procstates%mj-mk
           IF(mk.LT.mz) mk=mk+mbig
           procstates%mj=procstates%ma(ii)
        END DO
        DO k=1,4
           DO i=1,madim
              procstates%ma(i)=procstates%ma(i)-procstates%ma(1+MOD(i+30,madim))
              IF(procstates%ma(i).LT.mz) procstates%ma(i)=procstates%ma(i)+mbig
           END DO
        END DO
        procstates%inext=0
        procstates%inextp=31
        procstates%iduma=1
     END IF
   END SUBROUTINE ran_init
   !*************************************************************************
   SUBROUTINE ran3(idum,rand)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     !  Returns a uniform random deviate between 0.0 and 1.0.  Set idum to  !
     !  any negative value to initialize or reinitialize the sequence.      !
     !  This function is taken from W.H. Press', "Numerical Recipes" p. 199.!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     USE ran_state
     INTEGER(K8), INTENT(IN) :: idum
     REAL(DP), INTENT(OUT) :: rand
     !
     IF (idum.LT.0) CALL ran_init(idum)
     procstates%inext=procstates%inext+1
     IF(procstates%inext.EQ.56) procstates%inext=1
     procstates%inextp=procstates%inextp+1
     IF(procstates%inextp.EQ.56) procstates%inextp=1
     procstates%mj=procstates%ma(procstates%inext) & 
          - procstates%ma(procstates%inextp)
     IF(procstates%mj.LT.mz) procstates%mj=procstates%mj+mbig
     procstates%ma(procstates%inext)=procstates%mj
     rand=procstates%mj*fac
   END SUBROUTINE ran3
   !*************************************************************************
   SUBROUTINE func(j,funcval)
     INTEGER(K4), INTENT(IN) :: j
     REAL(DP), INTENT(OUT) :: funcval
     INTEGER(K4) :: nvalley
     REAL(DP) :: pi
     REAL(DP), DIMENSION(nparam) :: f1,f2,funcval1
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     !  This is an N-dimensional version of the multimodal function with  !
     !  decreasing peaks used by Goldberg and Richardson (1987, see ReadMe! 
     !  file for complete reference).  In N dimensions, this function has !
     !  (nvalley-1)^nparam peaks, but only one global maximum.  It is a   !
     !  reasonably tough problem for the GA.                              !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     nvalley=4
     pi=4.0d0*datan(1.d0)
     f1(1:nparam)=(SIN(5.1d0*pi*parent(1:nparam,j) + 0.5d0))**nvalley
     f2(1:nparam)=EXP(-4.0d0*log(2.0d0)*((parent(1:nparam,j)-0.0667d0)**2)/0.64d0)
     funcval1(1:nparam)=f1(1:nparam)*f2(1:nparam)
     funcval=PRODUCT(funcval1(1:nparam))
   END SUBROUTINE func
!*************************************************************************
END PROGRAM gafortran
