function data = ads (file, varargin)
%ODB.ADS add new dataset to OPPLA database file
% DATA = ODB.ADS(FILE, NEWDAT1, ...)
% DATA = ODB.ADS(OLDDATA, NEWDAT1, ...)
%   FILE    contains OLDDATA; will be overwritten with DATA
%   OLDDATA is the original dataset
%   NEWDAT1 etc. must have the structure described in and returned by odb.rec
%   DATA    is the new dataset
%
% See also ODB.REC
  pfns = {'stn' 'file' 'crs' 'lat' 'lon' 'date'}; % profile fields
  cfns = {'cruise' 'station' 'file' 'rec'};
  npfs = length(pfns);
  data = [];
  if isstruct(file)
    data = file;
  elseif ischar(file)
    if (exist(file, 'file') || exist([file '.mat'], 'file'))
      data = load(file);
    end
  end
  for newdata = varargin
    nd = newdata{1};
    if ~all(isfield(nd, cfns))  % check for missing fields
      for nf=1:length(cfns)
        if ~isfield(nd, cfns{nf})
          error('OPPLA:odb', 'Missing field %s in newdata.', cfns{nf})
        end
      end
    end
    if isempty(data)
      nr0 = 0;
      np0 = 0;
      iprev = [];
      data.prf = repmat(cell2struct({[] [] [] [] [] []}, pfns, 2), 1, 0);
      data.cruise = nd.cruise(:)';
      data.station = nd.station(:)';
      data.file = nd.file(:)';
      data.rec.prf = [];
      data.rec.depth = [];
    else
      if ~isfield(data, 'prf')
        data = odb.ads([], data);
      end
      nr0 = length(data.rec.prf);           % previous number of data records
      np0 = length(data.prf);               % previous number of profiles
      idfns = {'cruise' 'station' 'file'};
      idpfns = {'crs' 'stn' 'file'};
      for nf=1:length(idfns)
        idfn = idfns{nf};
        idpfn = idpfns{nf};
        nid0 = length(data.(idfn));         % previous number of entries
        newids = setdiff(nd.(idfn)(:)', data.(idfn)); % new entries
        ni = nd.(idfn)([nd.prf.(idpfn)]);
        [iidn, jidn] = ismember(ni, newids);
        [iido, jido] = ismember(ni, data.(idfn));
        data.(idfn) = [data.(idfn)(:)' newids];
        jidn = num2cell(jidn(iidn) + nid0);
        jido = num2cell(jido(iido));
        [nd.prf(iidn).(idpfn)] = jidn{:};
        [nd.prf(iido).(idpfn)] = jido{:};
      end
      datprf = squeeze(struct2cell(rmfield(data.prf, {'stn' 'file'})))';
      newdatprf = squeeze(struct2cell(rmfield(nd.prf, {'stn' 'file'})))';
      stn = [data.prf.stn nd.prf.stn]';
      clear datloc newdatloc
      [~, datloc(:,1)] = ismember([data.prf.stn], stn);
      [~, newdatloc(:,1)] = ismember([nd.prf.stn], stn);
      datloc(:,3:npfs) = cell2mat(datprf);
      newdatloc(:,3:npfs) = cell2mat(newdatprf);
      [isprev, iold] = ismember(newdatloc, datloc, 'rows');
      iprev = find(isprev);	% profiles in nd already present in data
      iold = iold(iprev);	% indices of these in data
    end
    if isempty(iprev)
      data.prf = [data.prf nd.prf];
    else
      for np=1:length(iprev)
        nd.rec.prf(nd.rec.prf == iprev(np)) = iold(np) - np0;
      end
      inew = setdiff(1:length(nd.prf), iprev); % new profiles
      for np=1:length(inew)
        nd.rec.prf(nd.rec.prf == inew(np)) = np;
      end
      data.prf = [data.prf nd.prf(inew)];
    end
    [prf, ~, jprf] = unique([nd.rec.prf(:) nd.rec.depth(:)], 'rows');
    nd.rec.prf = prf(:,1)';	% remove duplicate records
    nd.rec.depth = prf(:,2)';
    data.rec.prf = [data.rec.prf nd.rec.prf + np0];
    data.rec.depth = [data.rec.depth nd.rec.depth];
    dfns = setdiff(fieldnames(nd), ['prf' cfns]);
    for ndf=1:length(dfns)
      dfn = dfns{ndf};              % data field name
      nd.(dfn).rec = jprf(nd.(dfn).rec)';	% correct record indices
      fns = fieldnames(nd.(dfn));
      if (isfield(data, dfn))
        if ~isempty(setdiff(fns, fieldnames(nd.(dfn))))
          error('OPPLA:odb', ['Cannot mix datasets with different contents\n' ...
                              'for field %s.'], dfn)
        end
        if (~strcmp(data.(dfn).units, nd.(dfn).units))
          error('OPPLA:odb', ...
                'Inconsistent units in field %s.\nPrevious: %s, new: %s.', dfn, ...
                data.(dfn).units, nd.(dfn).units)
        end
        data.(dfn).rec = [data.(dfn).rec nd.(dfn).rec + nr0];
        fns = setdiff(fns, {'rec' 'units'});
        for nf=1:length(fns)
          data.(dfn).(fns{nf}) = [data.(dfn).(fns{nf}) nd.(dfn).(fns{nf})];
        end
      else
        data.(dfn).rec = nd.(dfn).rec + nr0;
        fns = setdiff(fns, 'rec');
        for nf=1:length(fns)
          data.(dfn).(fns{nf}) = nd.(dfn).(fns{nf});
        end
      end
    end
  end
  if ischar(file)
    save(file, '-struct', 'data');
  end
end
