function cmt = cmt(data, depth, varargin)
%ODB.CMT calculate monthly climatologies
% CMT = ODB.CMT(DATA, DEPTH, FIELDS, OPTIONS)
%   DATA        must be a structure as defined in odbrec
%   DEPTH       [cell] array of mean box depths [for each data field]
%   FIELDS      are the data fields for which climatologies are calculated
%   OPTIONS:
%     'location', LOC     location ({[lon lat] rad} or ID) of region for climatologies
%     'timerange', TR     time limits as date vectors ([y0 m0 d1; y1 m1 d1]) or years ([y0; y1])
%     'latrange', [l0 l1] limit latitudes to between l0 and l1
%     'lonrange', [l0 l1] limit longitudes to between l0 and l1
%   CMT         structure with climatologies in fields mean, sd, civ (95% confidence interval), N
%
% See also ODB.REC
  [~, fields, prfarg] = oppla.pop({'location' 'timerange' 'latrange' 'lonrange'}, ...
                                  {}, [], varargin{:});
  nf = length(fields);
  if iscell(depth)
    if (length(depth) ~= nf)
      error('ODB:cmt', 'Number of fields must match number of depth structures')
    end
  else
    cmt.depth = depth;
  end
  for n=1:nf
    f = fields{n};
    [fdat, rec, units, fdb] = odb.var(data, f, prfarg{:});
    if iscell(depth)
      cmt.(f) = cmtdat(fdat, [fdb.prf(fdb.rec.prf(rec)).date], ...
                       fdb.rec.depth(rec), depth{n});
      cmt.depth.(f) = depth{n};
    else
      cmt.(f) = cmtdat(fdat, [fdb.prf(fdb.rec.prf(rec)).date], ...
                       fdb.rec.depth(rec), cmt.depth);
    end
    cmt.units.(f) = units;
  end
end


function cmt = cmtdat(data, dates, depdat, depcmt)
% monthly climatology and 95% confidence intervals from raw data
% CMT = ODB.CMTDAT(DATA, DATES, DEPDAT, DEPCMT)
%   DATA    raw data
%   DATES   dates for raw data
%   DEPDAT  depths for raw data
%   DEPCMT  depths for climatology
%   CMT     climatology structure (length(depcmt) by 12 array)
%           with fields mean, sd, N
  ndc = length(depcmt);                % number of layers
  dcbb = odb.depth(depcmt);            % layer boundaries
  dd = datetime(datevec(dates));
  [ym, iym, jym] = unique([dd.Year dd.Month], 'rows');
  nym = length(ym);
  ymdat = zeros(ndc, nym);
  for n=1:nym     % for each year-month combination
    kym = find(jym == n);
    for nd=1:ndc  % for each box
      indat = find((depdat(kym) >= dcbb(nd)) & (depdat(kym) < dcbb(nd+1)));
      if isempty(indat)
        ymdat(nd,n) = NaN;
      else
        ymdat(nd,n) = mean(data(kym(indat)));
      end
    end
  end
  % average values from all years    
  cmt = repmat(struct('mean', NaN, 'sd', NaN, 'civ', [NaN NaN], 'N', 0), [ndc 12]);
  for nm=1:12	% months
    im = find(dd.Month(iym) == nm);
    for nd=1:ndc	% depth levels
      indat = find(~isnan(ymdat(nd,im)));
      if ~isempty(indat)
        cmt(nd,nm).mean = mean(ymdat(nd,im(indat)));
        cmt(nd,nm).N = length(indat);
        if length(indat) > 1
          cmt(nd,nm).sd = std(ymdat(nd,im(indat)));
          cmt(nd,nm).civ = quantile(ymdat(nd,im(indat)), [0.05 0.95]);
        end
      end
    end
  end
end
