function varargout = contour(db, field, varargin)
%ODB.CNT contour plot for data from an ODB data base
% [HAX, HCB] = ODB.CNT(DB, FIELD, OPTIONS)
%   DB        is the data base
%   FIELD       is the name of the data field in the data base
%   OPTIONS     optional arguments:
%     'location', {pos rad} selects only data from region within rad degrees
%                           around pos; pos is a geographical position in the
%                           form [lon lat]
%     'location', loc       select named region loc (BATS, LS, etc.)
%     'scale', factor       multiply data by factor (also supply units)
%     'units', us           units string
%     'depth', depths       vector of mean layer depths
%     'dtmin', dt           minimum temporal distance (days) for plotting
%     'dzmin', dz           minimum vertical distance (m) for plotting
%     'timestep', ndays     interpolate to regular time steps every ndays days

  opt.dtmin = 1;
  opt.dzmin = 5;
  opt.row = 1;
  opt.fontsize = 14;
  opt.nticks = 6;
  opt.MarkerSize = 2.5;
  [opt, varargin, prfarg] = oppla.pop({'location' 'station' 'timerange' 'latrange' 'lonrange'}, ...
                                      {}, opt, varargin{:});
  opt = oppla.pop({'depth' 'date' 'dzmin' 'fontsize' 'nticks' 'row' 'scale' 'units' 'MarkerSize'}, ...
                  {}, opt, varargin{:});
  if isfield(db, 'prf')
    if ~isempty(prfarg)
      db = odb.subset(db, prfarg{:});
    end
    data = odb.get(db, field);
    if ~isfield(opt, 'date')
      opt.date = unique(round(datenum(unique([data.prf.date]))./opt.dtmin)).*opt.dtmin;
    end
    if ~isfield(opt, 'depth')
      opt.depth = unique(round(unique(data.depth)./opt.dzmin)).*opt.dzmin;
    end
    if opt.depth(1) == 0
      opt.depth = odb.depth(opt.depth);
    end
    % set up time-depth grid
    if (size(data.data, 1) > 1)
      data.data = data.data(opt.row,:);
    end
    gdat = griddata(datenum([data.prf.date]), data.depth, data.data, opt.date(:)', opt.depth(:));
  else
    data = db;
    data.units = data.units.(field);
    opt.date = 1:12;
    if ~isfield(opt, 'depth')
      opt.depth = unique(round(unique(data.depth)./opt.dzmin)).*opt.dzmin;
    end
    gdat = reshape([data.(field).mean], size(data.(field)));
  end
  if isfield(opt, 'scale')
    gdat = gdat.*opt.scale;
    if ~isfield(opt, 'units')
      warning('ODB:cnt: You should supply units when applying a scale factor.')
    end
  end
  if ~isfield(opt, 'units')
    opt.units = data.units;
  end
  axopts = {'YDir' 'reverse' 'FontSize' opt.fontsize};
  [~, h] = contourf(opt.date(:)', opt.depth(:), gdat);
  set(h, 'LineStyle', 'none')
  ylabel('depth / m')
  title(field)
  hax = gca;
  [b0, b1] = bounds(gdat(:));
  hcb = odb.colorbar([b0, b1], opt.units);
  if isfield(data, 'prf')
    odb.tic(hax, opt.date([1, end]), opt.nticks, axopts{:});
    xlabel('time')
    hold on
    scatter(hax, datenum([data.prf.date]), data.depth, opt.MarkerSize, data.data)
    hold off
  else                                  % climatology
    odb.tic(hax, opt.date([1, end]), opt.nticks, axopts{:}, 'cmt');
    xlabel('month')
  end
  if nargout > 0
    varargout = {hax, hcb};
  end
end
