function [data, header] = csv(fid, varargin)
%ODB.CSV read data from csv file or text into record data structure for odb.rec
% DATA = ODB.CSV(FID, OPTIONS, FN1, FN2, ...)
%   FID 	file ID or data string
%   FN1, ...    data field names
% OPTIONS:
%   'fns', {}           field names for numeric data
%   'delimiter', 'd'    data delimiter in file or string (default: ,)
%   'crsnam', 's'	cruise header
%   'stnnam', 's'	station header
%   'datnam', 's'	date header
%   'depnam', 's'	depth header
%   'lines', n          number of data lines to read from file or string
%   'low', x            vector of lower limits of valid data ranges
%   'missing', 's'      missing value indicator in file or string
%   'dateform', 's'     date format string
%   'mindate', n        earliest allowed date (Matlab date number)
%   'maxdate', n        latest allowed date (Matlab date number)
%   'header', n		(line) number of header (lines to skip)
%   'skip', n           number of lines to skip after header
%   'units', n (or {})	(line number of) units
%   'data', n           first line number of data
%   'fmt', 's'		format string corresponding to fns
% Returns:
%   DATA 	data record structure for odb.rec
%
% ODB.CSV can be called in two ways.  If the data file contains a header line
% whose content can be used as field names, header and optionally data and
% units options indicate the lines where this information is stored in the
% file.  In this case the option fmt must not be used and option low
% corresponds to the numeric fields in option fns.  If option fmt is
% provided, option header indicates the number of header lines and options
% units and low correspond to the field names FN1, FN2, etc.
%
% See also ODB.REC
  opt.delimiter = ',';
  opt.lines = -1;
  opt.header = 0;
  opt.skip = 0;
  opt.datnam = 'date';
  opt.dateform = 'yyyy-mm-dd';
  opt.mindate = datetime(0, 0, 0);
  opt.maxdate = datetime(datevec(now));
  opt.latnam = {'lat' 'lat.' 'latitude'};
  opt.lonnam = {'lon' 'long' 'long.' 'longitude'};
  opt.timnam = 'time';
  opt.depnam = 'depth';
  opt.crsnam = 'cruise';
  opt.stnnam = {'station' 'stnnbr' 'stn'};
  opt.filnam = 'file';
  opt.fns = {};
  opt.low = 0;
  [opt, xfns] = oppla.pop({'delimiter' 'low' 'missing' 'lines' 'dateform' 'header' 'skip' ...
                      'units' 'data' 'datnam' 'latnam' 'lonnam' 'timnam' 'crsnam' 'stnnam' ...
                      'filnam' 'depnam' 'fmt' 'fns' 'mindate' 'maxdate'}, {}, opt, varargin{:});
  if isfield(opt, 'missing')
    missing = {'treatAsEmpty' opt.missing};
  else
    missing = {};
  end
  if ischar(opt.delimiter)
    if strcmp(opt.delimiter, ' ')
      delimiter = {'delimiter', ' ', 'MultipleDelimsAsOne', 1};
    else
      delimiter = {'delimiter' opt.delimiter};
    end
  else
    delimiter = {};
  end
  lh = max(opt.header, 1);
  hdat = textscan(fid, '%s', 1, 'HeaderLines', lh-1, 'delimiter', '');
  hc = textscan(hdat{1}{1}, '%q', delimiter{:});
  header = hc{1};
  if isfield(opt, 'fmt')
    fmt = opt.fmt;
    fns = xfns;
    if isfield(opt, 'units')
      units = opt.units;
    else
      units = odb.units(fns);
    end
  else
    nhs = length(header);
    if isfield(opt, 'units')
      udat = textscan(fid, '%s', 1, 'HeaderLines', opt.units-lh-1, 'delimiter', '');
      ru = textscan(udat{1}{1}, '%q', delimiter{:});
    else 
      ru = {repmat({[]}, 1, nhs)};
    end
    fm = cell(1, nhs);
    fn = cell(1, nhs);
    unit = cell(1, nhs);
    for nh=1:nhs
      [fm{nh}, fn{nh}] = fmthdr(header{nh}, opt.datnam, opt.latnam, opt.lonnam, ...
                                opt.timnam, opt.crsnam, opt.stnnam, opt.depnam, ...
                                opt.filnam, opt.fns, xfns);
      if ~isempty(fn{nh})
        if ismember(fn{nh}, {'depth' 'lat' 'lon' 'cruise' 'station' 'file'})
          unit(nh) = odb.units(fn{nh});
        elseif ismember(fn{nh}, xfns(:)) || ~isempty(strfind(fn{nh}, '_flag'))
          unit(nh) = {[]};
        elseif isempty(ru{1}{nh}) 
          unit(nh) = odb.units(fn{nh});
        else
          unit(nh) = {fmtunt(ru{1}{nh})};
        end
      end
    end
    ifn = ~cellfun(@isempty, fn);
    fns = fn(ifn);
    fmt = [fm{:} '%*[^\n]'];
    units = unit(ifn);
    if isfield(opt, 'data')
      if isfield(opt, 'units')
        opt.skip = opt.data - opt.units - 1;
      else
        opt.skip = opt.data - lh - 1;
      end
    end
  end
  headerlines = {'HeaderLines' opt.skip};
  nfs = length(fns);
  cdat = textscan(fid, fmt, opt.lines, delimiter{:}, ...
                  'EmptyValue', NaN, missing{:}, headerlines{:});
  if length(opt.low) == 1
    opt.low = repmat(opt.low, nfs);
    opt.low(ismember(fns, 'Temperature')) = -2;
  end
  if (length(cdat) ~= nfs) || (nfs ~= length(units))
    error('ODB:csv', ['Number of unskipped format columns must equal ' ...
                      'number of fields and units.'])
  end
  [idep, jdep] = ismember('depth', fns);
  if idep
    ival = ~isnan(cdat{jdep});
    for n=1:nfs
      cdat{n} = cdat{n}(ival);
    end
  end
  if nfs == 0
    data = struct([]);
  end
  for nf=1:nfs
    if ~ischar(units{nf})                 % profile fields have no units
      data.(fns{nf}) = cdat{nf};
    else    			        % data fields
      if iscell(cdat{nf})
        cdat{nf} = cellfun(@str2double, cdat{nf});
      end
      if ~isnan(opt.low(nf))
        cdat{nf}(cdat{nf} < opt.low(nf)) = NaN;
      end
      if any(~isnan(cdat{nf}))
        data.(fns{nf}) = cdat{nf}';
        data.units.(fns{nf}) = units{nf};
      end
    end
  end
  if isfield(data, 'date')
    if isfield(data, 'time')
      if isnumeric(data.time)
        data.time(isnan(data.time)) = 0;
        data.time = num2str(data.time, '%04d');
      end
      data.date = strcat(data.date, data.time);
      data = rmfield(data, 'time');
    end
    data.date = datetime(datestr(data.date, opt.dateform));
    if any((data.date < opt.mindate) | (data.date > opt.maxdate))
      warning('ODB:csv', 'Date(s) outside allowed range encountered.')
    end
  end
  if isfield(data, 'depth')               % ensure correct orientation of depth
    data.depth = abs(data.depth(:))';     % vector
  end
  if ~isfield(data, 'file')
    data.file = repmat({fopen(fid)}, size(data.cruise));
  end
end


function [fm, fn] = fmthdr(header, datnam, latnam, lonnam, timnam, crsnam, ...
                           stnnam, depnam, filnam, fns, xfns)
  fm = '%f';
  ff = regexp(lower(header), '_flag(_w)?$');
  if ~isempty(ff)
    header = header(1:ff-1);
  end
  switch lower(header)
    case crsnam
      fm = '%q';
      fn = 'cruise';
    case stnnam
      fm = '%q';
      fn = 'station';
    case datnam
      fm = '%q';
      fn = 'date';
    case timnam
      fm = '%q';
      fn = 'time';
    case filnam
      fm = '%q';
      fn = 'file';
    case latnam
      fn = 'lat';
    case lonnam
      fn = 'lon';
    case depnam
      fn = 'depth';
    case {'salinity' 'salnity'}
      fn = 'Salinity';
    case 'ctdsal'
      fn = 'Salinity_CTD';
    case {'ctdtmp' 'tmp' 'temp' 'temperature'}
      fn = 'Temperature';
    case 'pressure'
      fn = 'Pressure';
    case 'theta'
      fn = 'theta';
    case 'sigmatheta'
      fn = 'sigma_theta';
    case 'density'
      fn = 'Density';
    case 'ph'
      fn = 'pH';
    case {'bac' 'bact' 'bacteria' 'hbact' 'h.bact'}
      fn = 'Bacteria';
    case {'pbact' 'p.bact'}
      fn = 'Prochlorococcus';
    case {'sbact' 's.bact'}
      fn = 'Synechococcus';
    case 'fluor'
      fn = 'Fluorescence';
    case {'chl' 'chla' 'chlorophyll'}
      fn = 'Chlorophyll';
    case 'phaeo'
      fn = 'Phaeophytin';
    case 'bacteria_prod'
      fn = 'Bacteria_prod';
    case 'primary_prod'
      fn = 'Primary_Prod';
    case 'lightc14'
      fn = 'LightC14';
    case 'darkc14'
      fn = 'DarkC14';
    case {'oxygen' 'o2'}
      fn = 'O2';
    case {'ctdoxy' 'ctd_o2'}
      fn = 'O2_CTD';
    case 'alkalinity'
      fn = 'Alkalinity';
    case {'silcat' 'silicate' 'sio3' 'sio4'}
      fn = 'SiO3';
    case {'po4' 'phspht' 'phosphate'}
      fn = 'PO4';
    case {'no3' 'nitrat' 'nitrate'}
      fn = 'NO3';
    case {'no2' 'nitrit' 'nitrite'}
      fn = 'NO2';
    case {'nh4' 'ammonia' 'ammonium'}
      fn = 'NH4';
    case {'tcarbn' 'dic' 'co2' 'tco2'}
      fn = 'DIC';
    case 'pco2'
      fn = 'pCO2';
    case {'nit' 'no2no3' 'no3no2' 'no2+no3'}
      fn = 'NO2NO3';
    case {'tn' 'total_nitrogen'}
      fn = 'TN';
    case {'tp' 'total_phosphorus'}
      fn = 'TP';
    case {'din' 'poc' 'pon' 'doc' 'don' 'p13c' 'p15n' 'tdn' 'tdp' 'toc' 'ton' 'par' 'ccl4'}
      fn = upper(header);
    case 'psi'
      fn = 'PSi';
    case {'pc' 'tpc'}
      fn = 'PC';
    case {'pn' 'tpn'}
      fn = 'PN';
    case {'pp' 'tpp'}
      fn = 'PP';
    case {'zoo_200' 'zoo_500' 'zoo_1000' 'zoo_2000' 'zoo_5000'}
      fn = lower(header);
    case 'toc_kg'
      fn = 'TOC_kg';
    case {'cfc-11' 'cfc11'}
      fn = 'CFC11';
    case {'cfc-12' 'cfc12'}
      fn = 'CFC12';
    case {'cfc-113' 'cfc113'}
      fn = 'CFC113';
    case fns
      fn = header;
    case xfns
      fm = '%q';
      fn = regexprep(lower(header), '[()/ -]', '_');
    otherwise
      fm = '%*q';
      fn = [];
  end
  if ~isempty(ff) && ~isempty(fn)
    fn = [fn '_flag'];
  end
end


function unit = fmtunt (unit)
  switch regexprep(lower(unit), ' +', ' ')
    case {'pss-78' 'psu'}
      unit = 'psu';
    case {'umol/kg' 'umol/l' 'e umol/k'}
      unit = '{\mu}mol kg^{-1}';
    case {'nmol/kg' 'nmol/l'}
      unit = 'nmol kg^{-1}';
    case 'mg C/m3'
      unit = 'mg C m^{-3}';
    case {'ug/l' 'µg/l'}
      unit = '{\mu}g L^{-1}';
    case {'um' 'µm'}
      if strcmp(unit, {'uM' 'µM'})
        unit = '{\mu}mol kg^{-1}';
      else
        unit = '{\mu}m';
      end
    case {'pmol/kg' 'pmol/l' 'pm/k' 'pm/kg' 'g pm/k'}
      unit = 'pmol kg^{-1}';
    case {'uatm' 'uatm@t' 'uatm@20'}
      unit = '{\mu}atm';
    case {'deg c' 'deg-c' 'deg_c' 'ipts-68' 'ipts-90' 'its-90'}
      unit = '^{\circ}C';
    case 'ml/l'
      unit = 'mL L^{-1}';
    case {'dbar' 'dbars'}
      unit = 'dbar';
    case {'' 't_scale' 'sw' 'sw@25' '@25_deg' '25_deg'}
      unit = '';
    case {'cells/l' 'number/l'}
      unit = 'cells L^{-1}';
    otherwise
      warning('ODB:csv', ['Unknown units: ' unit])
  end
end
