function idat = int(data, depth, varargin)
%ODB.INT depth integrals with annual and seasonal means for OPPLA database
% IDAT = ODB.INT(DATA, DEPTH, FIELDS, OPTIONS)
% DATA      OPPLA database
% DEPTH     depth for integration
%           depth may be a structure with fields date and depth
% FIELDS    field names in DATA
% OPTIONS:
%   'mindep', d     profiles ending shallower than d are excluded
%   'maxdep', d     profiles starting deeper than d are excluded
%   'dephigh', c    cell array of field names whose deep values are more
%                   likely close to the deepest measurement than close to 0
%   options for odb.prf ('location' and 'timerange')
% Returns:
% IDAT      structure with depth integrals, annual and seasonal means
% Odb.int assumes that surface values are equal to the first measurement and
% data below the deepest measurements are zero.
% NaNs are removed from sorted depth integrals and field prf is added for
% reference to profiles in DATA.
%
% See also ODB.PRF
  dfnsl = setdiff(fieldnames(data), {'cruise' 'prf' 'rec'});
  opt.mindep = 0;
  if isstruct(depth)
    opt.maxdep = max(depth.depth(:));
  else
    opt.maxdep = depth;
  end
  opt.dephigh = {};
  [opt, varargin] = oppla.pop({'mindep' 'maxdep' 'dephigh'}, dfnsl, opt, varargin{:});
  dfns = intersect(dfnsl, fieldnames(opt));
  if isstruct(depth)
    if (opt.maxdep > max(depth.depth(:)))
      opt.maxdep = max(depth.depth(:));
    end
  else
    if (opt.maxdep > depth)
      opt.maxdep = depth;
    end
    pdep = depth;
  end
  nfs = length(dfns);
  idat = [];
  for nf=1:nfs
    dfn = dfns{nf};
    datprf = odb.prf(data, dfn, varargin{:});
    if isempty(datprf)
      warning('ODB:int', 'No data for %s in specified time and region.', dfn)
      continue
    end
    nps = length(datprf);
    idat.(dfn).date = [data.prf([datprf.prf]).date]';
    idat.(dfn).val = NaN(1,nps);
    idv = ismember(dfn, opt.dephigh);
    for np=1:nps
      if (min(datprf(np).depth) <= opt.maxdep) ...
            && (max(datprf(np).depth) >= opt.mindep)
        if isstruct(depth)
          pdep = depth.depth(find(depth.date >= idat.(dfn).date(np), 1, 'first'));
        end
        id = find(datprf(np).depth <= pdep);
        dd = datprf(np).depth(id);
        [datdep, datval] = unidat(datprf(np).(dfn).val(id), dd, pdep, idv);
        idat.(dfn).val(np) = trapz(datdep, datval);
      end
    end
    ival = find(~isnan(idat.(dfn).val)); % remove NaNs and sort wrt. time
    [idat.(dfn).date, idate] = sort(idat.(dfn).date(ival));
    idat.(dfn).val = idat.(dfn).val(ival(idate));
    idat.(dfn).prf = [datprf(ival(idate)).prf];
    [iysd, iymd, iyd, idat.(dfn).dates, idat.(dfn).datem, idat.(dfn).datey, ...
     nysd, idat.(dfn).year] = odb.season(idat.(dfn).date);
    idat.(dfn).mmn = NaN(nysd, 12);
    idat.(dfn).smn = NaN(nysd, 4);
    idat.(dfn).amn = NaN(nysd, 1);
    for ny=1:nysd                         % annual means
      idat.(dfn).amn(ny) = mean(idat.(dfn).val(iyd{ny}));
      for ns=1:4                          % seasonal means
        idat.(dfn).smn(ny,ns) = mean(idat.(dfn).val(iysd{ns,ny}));
      end
      for nm=1:12                         % monthly means
        idat.(dfn).mmn(ny,nm) = mean(idat.(dfn).val(iymd{nm,ny}));
      end
    end
  end
end


function [udep, udat] = unidat (dat, dep, maxdep, idv)
% calculate unique profile by averaging values at the same depth
  udep = unique([0 dep maxdep]);
  imb = ismember(udep, dep);
  udat = zeros(size(udep));
  for n=1:length(udep)
    if imb(n)
      udat(n) = mean(dat(dep == udep(n)));
    end
  end
  if ~imb(1)
    udat(1) = udat(2);
  end
  if ~(imb(end))
    if idv
      udat(end) = udat(end-1);
    else
      udat(end) = 0;
    end
  end
end
