function db = pangaea(file)
%ODB.PANGAEA: read Pangaea tab-delimited data file
% DATA = ODB.PANGAEA(FILE)
%  FILE     file name
%  DATA     ODB data structure
% recognised parameter names
  names = {'Date/Time' 'Depth water [m]' 'Chl a [pmol/l]' 'Chl a [µg/l]' ...
           'Chl tot [µg/l]' 'PON [µg/l]' 'POC [µg/l]' 'NO3+NO2 [µmol/l]' ...
           'NO2 [µmol/l]' 'NO2 [µmol/l]' ...
           'NH4 [µmol/l]' 'PO4 [µmol/l]' 'Si(OH)4 [µmol/l]' 'O2 [µmol/l]' 'Sal'};
  % corresponding ODB field names
  fields = {'date' 'depth' 'Chlorophyll' 'Chlorophyll' 'Chlorophyll' 'PON' 'POC' ...
            'NO2NO3' 'NO2' 'NO3' 'NH4' 'PO4' 'SiO3' 'O2' 'Salinity'};
  strdat = fileread(file);
  stndat = regexp(strdat, ['Event\(s\):[[:space:]]*(?<station>[^*]+)' ...
                      '\* LATITUDE: +(?<lat>[0-9.---]+)' ...
                      ' \* LONGITUDE: +(?<lon>[0-9.---]+)' ...
                      '(?<date>.* \* DATE/TIME: +[0-9:T---]+)?' ...
                      '(.* \* CAMPAIGN: (?<cruise>[A-Za-z0-9_---/():. ]+)[ \n])?' ...
                      '.*\*/(\n.*)?\n(?<params>(Date/Time|Depth water)[^\n]+)' ...
                      '\n(?<data>.+)'], 'names');
  if isempty(stndat)
    error('ODB:pangaea:header', ['Could not extract header in ' file])
  end
  if ~isfield(stndat, 'cruise')
    stndat.cruise = regexp(stndat.station, '^[^_]+', 'match');
  end
  params = textscan(stndat.params, '%s', 'delimiter', '\t');
  [ihs, jhs] = ismember(params{1}, names);
  iuh = find(jhs);
  % take only the first occurrence of repeated parameter names
  for ni=length(iuh):-1:2
    if ismember(jhs(iuh(ni)), jhs(1:iuh(ni)-1))
      ihs(iuh(ni)) = 0;                   % remove duplicate parameter names
    end
  end
  [fmt{ihs}] = deal('%f');
  [fmt{~ihs}] = deal('%*s');              % skip unregognised or duplicate parameters
  jhs = jhs(ihs);                         % indices of recognised parameters
  fns = fields(jhs);                      % field names for data
  nfs = length(fns);
  cf = ones(1, nfs);                      % conversion factors
  optcsv = {};
  for nf=1:nfs
    switch jhs(nf)
      case 1                              % date
        fmt{iuh(nf)} = '%s';
        optcsv = [optcsv 'dateform' 'yyyy-mm-dd']; %#ok
      case 3                              % Chl a in pmol/l
        cf(nf) = 1.121e-3;                % convert pmol/l -> µg/kg
    end
  end
  data = odb.csv(stndat.data, 'fmt', [fmt{:} '%*[^\n]'], 'delimiter', '\t', ...
                 optcsv{:}, fns{:});
  if isfield(data, 'date')                % use date info from date parameter
    nd = length(data.depth);
    data.lat = repmat(str2double(stndat.lat), 1, nd);
    data.lon = repmat(str2double(stndat.lon), 1, nd);
    data.file = repmat({file}, 1, nd);
    data.station = repmat({strtrim(stndat.station)}, 1, nd);
    data.cruise = repmat({strtrim(stndat.cruise)}, 1, nd);
  else                                    % use date info from header
    if isempty(stndat.date)
      error('ODB:pangaea:date', ['No date found in file ' file])
    else
      stndate = regexp(stndat.date, 'DATE/TIME: +(?<date>[0-9:T---]+)', ...
                       'names');
    end
    data.lat = str2double(stndat.lat);
    data.lon = str2double(stndat.lon);
    data.date = datenum(stndate.date, 'yyyy-mm-ddTHH:MM:SS');
    data.file = file;
    data.station = strtrim(stndat.station);
    data.cruise = strtrim(stndat.cruise);
  end
  for nf=1:nfs
    data.(fns{nf}) = data.(fns{nf}).*cf(nf); % apply unit conversion factors
  end
  db = odb.rec(data);
end
