function prf = prf(db, varargin)
%ODB.PRF get profile data for specific fields
% PRF = ODB.PRF(DB, FIELD1, ..., OPTIONS)
%   DB        	OPPLA database
%   FIELD1 ...  field names of data to retrieve
%   OPTIONS:
%     'station',   'STN', see ODB.IPSLAB
%     'location',  {[lon lat] rad}
%     'depthrange, [d0 d1]
%     'timerange,  [y0 m0 d1; y1 m1 d1] or [y0; y1]
%     'intersect'  retrieve only records containing all fields
%   PRF         structure with four or more fields:
%                 prf         profile index in DB
%                 FIELD1, ... arrays of data values
%                 depth       array of corresponding depths
%                 units       units from DB.FIELD1
%                 other optional fields from DB.FIELD1, ... if present
% See also ODB.GET, ODB.SUBSET
  [opt, varargin] = oppla.pop({'depthrange'}, {'intersect'}, [], varargin{:});
  [~, vnms, args] = oppla.pop({'location' 'station' 'timerange' 'latrange' 'lonrange'}, ...
                              {}, [], varargin{:});
  if any(~isfield(db, vnms))
    error('ODB:prf', '%s is not a field in DB.', [vnms{~isfield(db, vnms)}])
  end
  nvar = length(vnms);
  [db.prf.station] = db.station{[db.prf.stn]};
  iloc = odb.ipslab(db.prf, args{:});
  rec = cell(1, nvar);                    % collect all record indices
  for nv=1:nvar
    rec{nv} = db.(vnms{nv}).rec(:)';
  end
  if isfield(opt, 'intersect')
    recs = rec{1};
    for nv=2:nvar
      recs = intersect(recs, rec{nv});
    end
  else
    recs = union([rec{:}], []);
  end
  if isfield(opt, 'depthrange')
    recs = recs((db.rec.depth(recs) >= opt.depthrange(1)) & (db.rec.depth(recs) <= opt.depthrange(2)));
  end
  prfs = intersect(db.rec.prf(recs), iloc); % all profile indices
  nps = length(prfs);
  prf = repmat(struct([]), 1);
  for np=1:nps
    prf(np).prf = prfs(np);
    precs = recs(db.rec.prf(recs) == prfs(np)); % records for this profile
    [prf(np).depth, idep] = sort(db.rec.depth(precs));
    nrs = length(precs);
    for nv=1:nvar
      varnam = vnms{nv};
      [idat, ldat] = ismember(precs(idep), db.(varnam).rec);
      fns = setdiff(fieldnames(db.(varnam)), {'rec'});
      for nf=1:length(fns)
        if ismember(fns{nf}, {'units' 'description'})
          prf(np).(varnam).(fns{nf}) = db.(varnam).(fns{nf});
        else
          prf(np).(varnam).(fns{nf}) = NaN(1, nrs);
          prf(np).(varnam).(fns{nf})(idat) = db.(varnam).(fns{nf})(ldat(idat));
        end
      end
    end
  end
end
