function recdat = rec(rawdat)
%[DATA] = ODB.REC(RAWDAT) organize raw data in records
%   RAWDAT must be a scalar structure with the following fields:
%     cruise: cruise ID (cell column vector)
%     station:station ID (cell column vector)
%     file:   file name (cell column vector)
%     date:   date of sampling as a datetime vector
%     lat:    latitude in degree N
%     lon:    longitude in degree E
%     depth:  depth of sampling in m or dbar
%     units:  structure with one field (TYPE) for each datafield giving that
%             data's units
%     TYPE:   data vector or structure
%     one datafield for each type of data contained in the dataset
%     cruise, station, file, date, lat, lon can be scalars or vectors of size(depth)
%     type can be a row vector or a structure with field val (and others)
%     the depth and all data fields must be row vectors of the same size
%     optional fields within a type structure must have their second
%       dimension match the size of the val field if they are numeric
%   DATA will have the following fields:
%     cruise          cell array of unique cruise IDs
%     station         cell array of unique station IDs
%     file            cell array of unique file names
%     prf             struct array identifying casts with sub-fields:
%             crs     index into cruise field
%             stn     index into station field
%             file    index into file field
%             station, date, lat, lon as described above
%     rec.prf         is the index of the cast in prf field
%     rec.depth       is the depth of sampling
%     datafields with the following sub-fields:
%             rec     is the record index vector
%             val     is the data vector
%             units   is the units string
%             other (optional) fields present in RAWDAT.type
  ndat = length(rawdat.date);
  if ndat == 1
    if ~iscell(rawdat.cruise)
      rawdat.cruise = {rawdat.cruise};
    end
    if ~iscell(rawdat.station)
      rawdat.station = {rawdat.station};
    end
    if ~iscell(rawdat.file)
      rawdat.file = {rawdat.file};
    end
  elseif ~iscell(rawdat.cruise) || ~iscell(rawdat.station) || ~iscell(rawdat.file)
    error('odb.rec: Fields cruise, station, and file must be cell arrays.')
  end
  [recdat.cruise, ~, rawdat.crs] = unique(rawdat.cruise);
  [recdat.station, ~, rawdat.stn] = unique(rawdat.station);
  [recdat.file, ~, rawdat.file] = unique(rawdat.file);
  pfns = {'station' 'crs' 'stn' 'file' 'date' 'lat' 'lon'}; % profile field names
  rawdat.date = datenum(rawdat.date);
  if ndat == 1
    rdfns = setdiff(fieldnames(rawdat), [pfns 'cruise' 'units']);
    inan = isnan(rawdat.depth(:));
  else
    rdfns = setdiff(fieldnames(rawdat), 'units');
    inan = any(isnan([rawdat.date(:) rawdat.lon(:) rawdat.lat(:) ...
                      rawdat.depth(:)]), 2);
    for pfn=pfns
      rawdat.(pfn{1})(inan) = [];
    end
  end
  nrdat = numel(rawdat.depth);
  rawdat.depth(inan) = [];
  ndat = length(rawdat.date);
  npfs = length(pfns);
  pfd = zeros(ndat,npfs);
  [~, ip, jp] = unique(rawdat.station);
  pfd(:,1) = ip(jp);
  for npf=2:npfs
    pfd(:,npf) = rawdat.(pfns{npf});          % profile field data
  end
  [pd, ipd, iprf] = unique(pfd, 'rows');      % unique (sorted) profiles
  nps = length(ipd);
  for npf=2:npfs
    pdc = num2cell(pd(:,npf));
    [recdat.prf(1:nps).(pfns{npf})] = pdc{:}; % assign unique profile data
  end
  if ndat == 1
    ndat = numel(rawdat.depth);
    iprf = repmat(iprf, ndat, 1);
  end
  recdat.rec.depth = rawdat.depth(:)';
  recdat.rec.prf = iprf';                     % associate records with profiles
  dfns = setdiff(rdfns, [pfns 'cruise' 'station' 'file' 'depth']); % data field names
  dfns = dfns(~contains(dfns, '_flag'));
  for ndf=1:length(dfns)
    dfn = dfns{ndf};
    if isstruct(rawdat.(dfn))
      df = rawdat.(dfn);
    else
      clear df
      df.val = rawdat.(dfn)(:)';
      df.units = rawdat.units.(dfn);
      if isfield(rawdat, [dfn '_flag'])
        df.flag = rawdat.([dfn '_flag'])(:)';
      end
    end
    if (length(df.val) ~= nrdat)
      error('ODB:rec', 'All data fields must have the same size.')
    end
    if iscell(df.val)
      error('ODB:rec', 'Field %s must be a numeric array.', dfn)
    end
    rec = find(~any(isnan(df.val(:,~inan)), 1));
    if ~isempty(rec)                      % skip empty datasets
      recdat.(dfn).rec = rec;             % associate data with records
      for fn=fieldnames(df)'
        if isnumeric(df.(fn{1}))
          df.(fn{1})(inan) = [];
          if (size(df.(fn{1}), 2) ~= ndat)
            if isstruct(rawdat.(dfn))
              dfn = [dfn '.' fn{1}]; %#ok
            end
            error('ODB:rec', ['All data fields must have the same second ' ...
                              'dimension (data field %s).'], dfn)
          end
          recdat.(dfn).(fn{1}) = df.(fn{1})(:,rec);
        else
          recdat.(dfn).(fn{1}) = df.(fn{1});
        end
      end
    end
  end
end
