function [iys, iym, iy, ts, tm, ty, nys, years] = season(ndate, varargin)
%ODB.SEASON calculate season indices
% [IYS, NY, TS] = ODB.SEASON(NDATE, 'daterange', dr, 'solar')
%   NDATE       dates as matlab date numbers
%   options:
%      'solar'          shifts start of years and seasons by 19 days
%      'daterange', dr  daterange for which to calculate indices
% Returns:
%   YEARS       all years with any data
%   NYS         number of YEARS
%   IYS(n,y)    time indices for all data in year y and season n (cell array)
%   IYM(n,y)    time indices for all data in year y and month n (cell array)
%   IY(y)       time indices for all data in year y
%   TS(n,:)     corresponding mid-season times as matlab date numbers
%   TM(n,:)     corresponding mid-month times as matlab date numbers
%   TY(n,:)     corresponding mid-year times as matlab date numbers
  opt = oppla.pop({'daterange'}, {'solar'}, struct, varargin{:});
  if isfield(opt, 'daterange')
    dr = datenum(opt.daterange);
    idate = find(ndate >= dr(1) & ndate < dr(2));
    ndate = ndate(idate);
  else
    idate = 1:length(ndate);
  end
  if isfield(opt, 'solar')
    ndate = ndate - 19;
  end
  dates = datevec(ndate);
  years = unique(dates(:,1));
  nys = length(years);
  imon = cell(1,12);
  for m=1:12
    imon{m} = find(ismember(dates(:,2), m));
  end
  isns{1} = find(ismember(dates(:,2), [1 2 12]));% winter
  isns{2} = find(ismember(dates(:,2), 3:5));	% spring
  isns{3} = find(ismember(dates(:,2), 6:8));	% summer
  isns{4} = find(ismember(dates(:,2), 9:11));	% autumn
  i12 = find(ismember(dates(:,2), 12));	% December

  iys = cell( 4, nys);
  iym = cell(12, nys);
  iy =  cell( 1, nys);
  ts = zeros( 4, nys);
  tm = zeros(12, nys);
  ty = zeros( 1, nys);
  for ny=1:nys
    iny = find(dates(:,1) == years(ny));
    if ~isempty(iny)
      iy{1,ny} = idate(iny);
      ty(1,ny) = mean(ndate(iny));
      for nm=1:12
        inym = intersect(imon{nm}, iny);  % month
        iym{nm,ny} = idate(inym);
        if isempty(inym)
          tm(nm,ny) = NaN;
        else
          tm(nm,ny) = mean(ndate(inym));
        end
      end
    end
  end
  dates(i12,1) = dates(i12,1) + 1;	% add December to next year
  for ny=1:nys
    iny = find(dates(:,1) == years(ny));
    if ~isempty(iny)
      for ns=1:4
        inys = intersect(isns{ns}, iny);  % season
        iys{ns,ny} = idate(inys);
        if isempty(inys)
          ts(ns,ny) = NaN;
        else
          ts(ns,ny) = mean(ndate(inys));
        end
      end
    end
  end
  iemp = find(cellfun(@isempty, iy));
  iy(iemp) = [];
  ty(iemp) = [];
  years(iemp) = [];
  nys = length(years);
  [iym{cellfun(@isempty, iym)}] = deal([]);
  [iys{cellfun(@isempty, iys)}] = deal([]);
end
