function secdat = sec(db, dfn, varargin)
%ODB.SEC 2D section from ODB dataset
% SD = ODB.SEC(DB, DFN, OPTIONS)
%   DB      OPPLA database
%   DFN     data field name in DB
%   OPTIONS:
%     options for odb.prf (see there)
%     'date', [d0; d1]  date range of section
%     'depth', [d0 d1]  depth range of section
%     'lat', [l0 l1]    latitude range of section
%     'lon', [l0 l1]    longitude range of section
%     'depthbin', d     vector of depth boundaries for depth binning
%     'latbin', l, 'datebin', d, 'lonbin', l: similar to depthbin
%     'ncfile', 'f'     filename for netcdf output
%     two of date, depth, lat, lon must be specified to define the section;
%     data within these ranges are averaged for each point along the other
%     two dimensions
%
%   Example:
%     extract a lat-depth section at 32°W for May 2000 for chl:
%     sd = odb.sec(db, 'chl', 'lat', [-33 -31], ...
%                  'date', datenum([2000 4 30; 2000 5 31]));
%
% See also ODB.PRF, ODB.GRD
  dimens = {'date' 'lon' 'lat' 'depth'};
  bins = strcat(dimens, 'bin');
  [opt, varargin] = oppla.pop([dimens bins 'ncfile'], {}, [], varargin{:});
  prfdat = odb.prf(db, dfn, varargin{:});
  if isfield(opt, 'ncfile')
    ncfile = opt.ncfile;
    opt = rmfield(opt, 'ncfile');
  end
  nodims = setdiff(fieldnames(opt), [bins 'ncfile']);
  dims = setdiff(dimens, nodims);
  idims = ismember(dimens, dims);
  if sum(idims) ~= 2
    error('ODB:dimen', 'Exactly two dimensions must be specified.')
  end
  idep = 0;
  for nd=1:2
    switch (nodims{nd})
      case 'depth'
        idep = nd;
        if length(opt.(nodims{nd})) == 1
          if opt.depth < 5
            opt.depth = [0 5];
          else
            opt.depth = [opt.depth-2.5 opt.depth+2.5];
          end
        end
        if isfield(opt, 'depthbin')
          error('ODB:options', 'Cannot specify both depth and depthbin.')
        end
      otherwise
        if length(opt.(nodims{nd})) == 1
          opt.(nodims{nd}) = [opt.(nodims{nd})-0.5 opt.(nodims{nd})+0.5];
        end
        dimind = [db.prf([prfdat.prf]).(nodims{nd})] >= opt.(nodims{nd})(1) ...
                 & [db.prf([prfdat.prf]).(nodims{nd})] < opt.(nodims{nd})(2);
        prfdat = prfdat(dimind);          % profiles within section
        if isfield(opt, [nodims{nd} 'bin'])
          error('ODB:options', 'Cannot specify both %s and %sbin.', ...
                nodims{nd}, nodims{nd})
        end
    end
  end
  if idep == 0
    jdep = find(strcmp(dims, 'depth'));
  end
  for nd=1:2
    if isfield(opt, [dims{nd} 'bin'])
      secdat.(dims{nd}) = opt.([dims{nd} 'bin']);
      if secdat.(dims{nd})(1) == 0
        secdat.(dims{nd}) = secdat.(dims{nd})(2:end);
      end
    else
      switch (dims{nd})
        case 'depth'
          secdat.depth = unique([prfdat.depth]);
        otherwise
          secdat.(dims{nd}) = unique([db.prf([prfdat.prf]).(dims{nd})]);
      end
    end
  end
  secdat.(dfn) = zeros(length(secdat.(dims{1})), length(secdat.(dims{2})));
  ndata = zeros(size(secdat.(dfn)));
  idim = cell(1, 2);
  for np=1:length(prfdat)
    if idep > 0                           % depth is not a secdat dimension
      for nd=1:2
        idim{nd} = find(secdat.(dims{nd}) >= [db.prf(prfdat(np).prf).(dims{nd})], ...
                        1, 'first');
      end
      secdat.(dfn)(idim{:}) = secdat.(dfn)(idim{:}) ...
          + mean(prfdat(np).(dfn)(prfdat(np).depth >= opt.depth(1) ...
                                  & prfdat(np).depth < opt.depth(2)));
    else
      npdep = length(prfdat(np).depth);
      ndat = zeros(1, length(secdat.depth));
      data = zeros(1, length(secdat.depth));
      for ndep=1:npdep
        indep = find(secdat.depth >= prfdat(np).depth(ndep), 1, 'first');
        ndat(indep) = ndat(indep) + 1;
        data(indep) = data(indep) + prfdat(np).(dfn)(ndep);
      end
      indat = find(ndat > 0);
      idim{jdep} = indat;
      nd = 3 - jdep;
      idim{nd} = find(secdat.(dims{nd}) >= [db.prf(prfdat(np).prf).(dims{nd})], ...
                      1, 'first');
      secdat.(dfn)(idim{:}) = secdat.(dfn)(idim{:}) + data(indat)./ndat(indat);
    end
    ndata(idim{:}) = ndata(idim{:}) + 1;
  end
  for nd=1:2
    if isfield(opt, [dims{nd} 'bin'])
      secdat.(dims{nd}) = odb.depth([0 secdat.(dims{nd})]);
    end
  end
  secdat.(dims{1}) = secdat.(dims{1})';
  idat = ndata > 0;
  secdat.(dfn)(~idat) = NaN;
  secdat.(dfn)(idat) = secdat.(dfn)(idat)./ndata(idat);
  if exist('ncfile', 'var')
    nc_create_empty(ncfile);
    ncvar.Nctype = 'double';
    for nd=1:2
      nc_add_dimension(ncfile, dims{nd}, length(secdat.(dims{nd})));
      ncvar.Name = dims{nd};
      ncvar.Dimension = dims(nd);
      nc_addvar(ncfile, ncvar);
      nc_varput(ncfile, dims{nd}, secdat.(dims{nd}));
    end
    ncvar.Name = dfn;
    ncvar.Dimension = dims;
    ncvar.Attribute(1).Name = 'units';
    ncvar.Attribute(1).Value = db.(dfn).units;
    nc_addvar(ncfile, ncvar);
    nc_varput(ncfile, dfn, secdat.(dfn));
  end
end
