function [ticks, labels] = tic(hax, time, niv, varargin)
%ODB.TIC generate ticks and labels showing dates and apply them to the (x-)axis
% [TICKS, LABELS] = ODB.TIC(HAX, YR0, TIME, NIV, OPTIONS)
%   HAX  is the axis handle
%   TIME is the time as matlab date numbers
%   NIV  is the number of intervals desired
%   OPTIONS:
%     'axis', 'A'     apply ticks to axis (X, Y or Z)
%     'format', 'fmt' is the date format
%     'nolabels'      suppresses tick labels
%     any options which apply to matlab axes
%   TICKS  vector of tick mark positions on specified axis
%   LABELS cell array of tick labels
  opt.axis = 'X';
  [opt, varargin] = oppla.pop({'format' 'axis'}, {'nolabels' 'cmt'}, opt, varargin{:});
  dtivs = [1/24 1/12 0.125 1/6 0.25 0.5 1 2 3 7 14];
  deld0 = [1/48 1/24 1/24 1/24 0.125 0.125 0.25 0.5 0.5 1 1];
  dfmt = {15 15 15 'mmmdd,HH:MM' 'mmmdd,HH' 'mmmdd,HH' ...
          'mmmdd,HH' 'dd mmm' 'ddmmmyy' 'ddmmmyy' 'ddmmmyy'};
  mtivs = [0.5 1 2 3 4 6 12 24 36 48 60];
  delm0 = [0.5 0.5 0.5 0.5 0.5 1 1 3 3 6 6];
  mfmt = {'ddmmmyy' 'ddmmmyy' 'ddmmmyy' 'ddmmmyy' 'ddmmmyy' ...
          'ddmmmyy' 'mmmyyyy' 'mmmyyyy' 'mmmyyyy' 'mmmyyyy' 'mmmyyyy'};
  if isfield(opt, 'cmt')    	% climatology
    months = ['Jan'; 'Feb'; 'Mar'; 'Apr'; 'May'; 'Jun'; 'Jul'; 'Aug'; 'Sep'; ...
              'Oct'; 'Nov'; 'Dec'];
    switch (niv)
      case 1
        ticks = [1 12];
      case 2
        ticks = [1 6 12];
      case 3
        ticks = [1 4 8 12];
      case 4
        ticks = 1:3:10;
      case {5 6 7 8}
        ticks = 1:2:11;
      case {9 10 11 12 13}
        ticks = 1:12;
    end
    labels = months(ticks,:);
    if (min(time(:)) <= 0)
      ticks = ticks - 0.5;
    end
  else
    t0 = time(1); t1 = time(end);
    ddt = (t1 - t0)/niv;            % approximate time interval
    if (ddt <= dtivs(end))          % regular spacing in terms of days
      idtiv = find(ddt <= dtivs, 1, 'first');
      ddt0 = t0 - floor(t0);
      if (ddt0 > 0)
        ddt0 = (floor(ddt0/deld0(idtiv)) + 1)*deld0(idtiv);
      end
      ticks = floor(t0) + ddt0:dtivs(idtiv):t1;
      opt.format = dfmt{idtiv};
    else
      mt0 = datevec(t0);
      mt1 = datevec(t1);
      m0 = mt0(1)*12 + mt0(2);
      m1 = mt1(1)*12 + mt1(2);
      dmt = (m1 - m0)/niv;
      if (dmt <= mtivs(end))        % tick spacing in terms of months
        imtiv = find(dmt <= mtivs, 1, 'first');
        dm0 = datenum([mt0(1:2) 1]);        % first day of first month
        dm1 = datenum([mt0(1:2) + [0 1] 1]);% first day of second month
        dmt0 = (t0 - dm0)/(dm1 - dm0);
        if (dmt0 > 0)
          dmt0 = (floor(dmt0/delm0(imtiv)) + 1)*delm0(imtiv);
        end
        mdates = m0 + dmt0:mtivs(imtiv):m1 + 1;
        ntics = length(mdates);
        ydates = floor((mdates - 1)./12);
        ddates = ones(1, ntics);
        idates = find(mdates > floor(mdates));
        if ~isempty(idates)
          ndm = datenum(ydates(idates), mdates(idates) + 1, ddates(idates)) ...
                - datenum(ydates(idates), mdates(idates), ddates(idates));
          ddates(idates) = floor(ndm./2);
        end
        mdates = mod(mdates - 1, 12) + 1;
        ticks = datenum(ydates, mdates, ddates);
        if (ticks(end) > t1)
          ticks = ticks(1:end-1);
        end
        opt.format = mfmt{imtiv};
      else                          % tick spacing in terms of years
        dely = floor(mt1(1) - mt0(1))/niv;
        if (t0 > 0)
          mt0(1) = mt0(1) + 1;
        end
        ydates = mt0(1):dely:mt1(1);
        ntics = length(ydates);
        ticks = datenum(ydates, ones(1, ntics), ones(1, ntics));
        opt.format = 10;
      end
    end
    labels = datestr(ticks, opt.format);
  end
  if isfield(opt, 'nolabels')
    labels = {''};
  end
  set(hax, [opt.axis 'TickMode'], 'manual', ...
           [opt.axis 'Tick'], ticks, ...
           [opt.axis 'TickLabelMode'], 'manual', ...
           [opt.axis 'TickLabel'], labels, ...
           varargin{:});
end
