## zooplankton net tows from the central Labrador Sea
ls_zoo <- data.frame(biomass = c(6558.2, 6925.6, 4194.6, 8185.5, 5425.5, 6649.3),
                     start = as.POSIXct(c("1995-06-07", "1996-05-12", "1997-07-07",
                                          "1998-06-22", "1999-06-27", "2000-05-20")),
                     end = as.POSIXct(c("1995-07-05", "1996-06-01", "1997-08-07",
                                        "1998-07-09", "1999-07-13", "2000-06-08")))
attr(ls_zoo, "description") <- "These data are from the central area of the Labrador Sea.
 The values are average biomass of mesozooplankton: 200μm - 1000μm for 0 - 100m, measured in mg/m^2.
 A single value is given for each cruise - the start and end dates of the cruises are given."
attr(ls_zoo$biomass, "units") <- "mg m-2"

## Labrador Sear biogeochemical data
read_ls <- function(folder = "~/data/LS/biology") {
  require("oppla")
  read_ls_csv <- function(file, columns, units, exclude = "time", datecols = "date", datefmt = "%Y-%m-%d", ...) {
    data <- read.csv(file = file, check.names = FALSE, ...)
    valid <- names(units)[columns %in% names(data)]
    names(data)[match(columns[columns %in% names(data)], names(data))] <- valid
    data <- within(data, {
      if (("time" %in% datecols) & ("time" %in% valid))
        time <- as.character(ifelse(test = is.na(time), yes = "12:00:00", no = time))
      date <- as.POSIXct(do.call(paste, lapply(datecols, FUN = function(x) eval(parse(text = x)))),
                         format = datefmt, tz = "GMT")
      depth <- abs(depth)
    })
    data <- subset(data[setdiff(valid, exclude)], subset = !(is.na(date) | is.na(depth)))
    for (v in valid) if (!is.na(units[[v]])) units(data[[v]]) <- units[[v]]
    return(data)
  }

  ## nutrient data
  units_nuts <- c(latitude = "degrees_N", longitude = "degrees_E", date = NA, depth = "m", Chl = "mg m-3",
                  cruise = NA, DIC = "mmol m-3", temperature = "°C", salinity = "psu",
                  phaeo = "mg m-3", NO3 = "mmol m-3", PO4 = "mmol m-3", SIO3 = "mmol m-3",
                  POC = "mg m-3",PON = "mg m-3", O2 = "mmol m-3", sigma_theta = "permil")
  nuts <- read_ls_csv(file = file.path(folder, "nutrient_data.csv"),
                      columns = list("LAT", "LONG", "CORRECTED DATETIME", "integrated depth", "CHL",
                                     "CRUISE_NAME", "co2", "TEMP", "SALINITY", "phaeo", "NO3", "PO4", "SIO3",
                                     "POC","PON", "O2", "sigma_theta"),
                      units = units_nuts,
                      datefmt = "%d/%m/%Y %H:%M:%S")
  io2 <- nuts$O2 > units::set_units(1e3, units_nuts["O2"], mode = "standard") & !is.na(nuts$O2)
  nuts$O2[io2] <- nuts$O2[io2] / 22.4 # convert l -> mol
  io2 <- nuts$O2 < units::set_units(15, units_nuts["O2"], mode = "standard") & !is.na(nuts$O2)
  nuts$O2[io2] <- nuts$O2[io2] / 22.4e-3 # convert l -> mmol
  nuts$O2[nuts$O2 < units::set_units(100, units_nuts["O2"], mode = "standard")] <- NA # suspiciously low
  nuts <- subset(nuts, subset = cruise != "ICEALGAE84")

  ## bacteria and chlorophyll
  bac <- read_ls_csv(file = file.path(folder, "bacteria&chlorophyll.csv"),
                     columns = list("Date", "LATITUDE", "LONGITUDE", "DEPTH",
                                    "Time (added by malloch from CTD odf files or nutrient file or interpolation)",
                                    "CHL (ug/liter)", "BACTERIA (cells/liter)", "Cruise"),
                     units = c(date = NA, latitude = "degrees_N", longitude = "degrees_E", depth = "m", time = NA,
                               Chl = "mg m-3", bact_number = "cells L-1", cruise = NA),
                     datecols = list("date", "time"), datefmt = "%d %b %Y %H:%M:%S", na.strings = "")

  ## BRAVO
  bravo <- read_ls_csv(file = file.path(folder, "BIOCHEM_data_BRAVO.csv"),
                       columns = list("LATITUDE", "LONGITUDE", "PARM_DATE", "DEPTH", "NITRATE", "CHL_A"),
                       units = c(latitude = "degrees_N", longitude = "degrees_E", date = NA, depth = "m",
                                 NO3 = "mmol m-3", Chl = "mg m-3"))

  ## bacterial production
  bp <- read_ls_csv(file = file.path(folder, "Bact_prod_Lab_Sea.csv"),
                    columns = list("LAT", "LON", "Date", "DEPTH", "Cruise Number",
                                   "Bacterial Production (mgC m-3 h-1)"),
                    units = c(latitude = "degrees_N", longitude = "degrees_E", date = NA, depth = "m", cruise = NA,
                              bact_prod = "mgC m-3 h-1"),
                    colClasses = c(DEPTH = "numeric"))

  ## TOC
  toc <- read_ls_csv(file = file.path(folder, "TOC_data.csv"),
                     columns = list("Lat.", "Long.", "Date", "time (from CTD)", "data Depth (m)", "TOC"),
                     units = c(latitude = "degrees_N", longitude = "degrees_E", date = NA, time = NA, depth = "m",
                               TOC = "mmol m-3"),
                     datecols = c("date", "time * 24"), datefmt = "%Y-%m-%d %H")

  ## primary production
  pp <- read_ls_csv(file = file.path(folder, "PP10m-LSea-99&00.csv"),
                    columns = list("Date", "STN", "LON", "LAT", "Depth (m)", "POC (mg/m3)", "PP (mgC/m3/h)"),
                    units = c(date = NA, cruise = NA, longitude = "degrees_E", latitude = "degrees_N", depth = "m",
                              POC = "mg m-3", PP = "mg m-3 h-1"),
                    colClasses = c("Depth (m)" = "numeric"))

  ## BRAVO biological data
  bravo_bio <- read_ls_csv(file = file.path(folder, "BRAVO", "BRAVO_biological_data2.csv"),
                           columns = list("lat", "long", "date", "depth", "chlorophyll", "carbon", "nitrate", "ammonium",
                                           "bacteria", "bacteria_mass"),
                           units = c(latitude = "degrees_N", longitude = "degrees_E", date = NA, depth = "m",
                                     Chl = "mg m-3", POC = "mg m-3", NO3 = "mmol m-3", NH4 = "mmol m-3",
                                     bact_number = "cells L-1", bact_mass = "mg m-3"),
                           colClasses = c(depth = "numeric"))

  ## BIOCHEM bottle data
  bc <- read_ls_csv(file = file.path(folder, "BIOCHEM", "bc_data_2795_new.csv"),
                    columns = list("START LAT", "START LON", "START_DATE", "START TIME", "AREA NAME", "START DEPTH",
                                    "NITRATE", "AMMONIA", "CHLOROPHYLL_A"),
                    units = c(latitude = "degrees_N", longitude = "degrees_E", date = NA, time = NA, cruise = NA,
                              depth = "m", NO3 = "mmol m-3", NH4 = "mmol m-3", Chl = "mg m-3"),
                    datecols = list("date", "time"), datefmt = "%Y-%m-%d %H:%M")

  ## PI and chl data
  pipp <- read_ls_csv(file = file.path(folder, "PIPPLSea-10m-96-00.csv"),
                      columns = list("lat", "lon", "year", "z", "temp", "chl", "cruise"),
                      units = c(latitude = "degrees_N", longitude = "degrees_E", date = NA, depth = "m",
                                temperature = "°C", Chl = "mg m-3", cruise = NA),
                      datecols = list("date", "mon", "day"), datefmt = "%Y %m %d")

  files <- paste0(c("LS_1994_ar07we_bot", "LS_1994_a01ew_bot", "LS_1993_ar07wd_bot", "LS_1992_ar07wc_bot",
                    "LS_1990_ar7wa_bot", "LS_1996_ar07wf_bot", "LS_1997_ar07wh_bot", "LS_1998_ar07wj_bot",
                    "LS_1995_a01w_bot"), ".csv")
  lsl <- list()
  for (n in seq_along(files)) {
    lsl[[n]] <- read_ls_csv(file = file.path(folder, files[n]),
                     columns = list("DATE", "TIME", "LATITUDE", "LONGITUDE", "DEPTH", "CTDTMP", "SALNTY", "CCL4",
                                    "OXYGEN", "SILCAT", "NO2+NO3", "PHSPHT", "CFC-11", "CFC113", "CFC-12", "TCARBN"),
                     units = c(date = NA, time = NA, latitude = "degrees_N", longitude = "degrees_E", depth = "m",
                               temperature = "°C", salinity = "psu", CCL4 = "pmol kg-1",
                               O2 = "mmol m-3", SIO3 = "mmol m-3", NO23 = "mmol m-3", PO4 = "mmol m-3",
                               CFC11 = "pmol kg-1", CFC113 = "pmol kg-1", CFC12 = "pmol kg-1", CO2 = "mmol m-3"),
                     datecols = list("date", "sprintf(\"%04d\", time)"), datefmt = "%d/%m/%Y %H%M",
                     skip = 5, comment.char = "#", na.strings = "-999", colClasses = c(DEPTH = "numeric"))
  }

  lsd <- Reduce(f = function(x, y) merge(x, y, all = TRUE),
                x = c(list(nuts, bac, bc, bp, bravo, bravo_bio, toc, bc, pp, pipp), lsl))
  lsd$cruise <- as.factor(lsd$cruise)
  return(structure(lsd, file = "~/data/LS/biology"))
}

read_ls_wod <- function(file = "~/data/LS/biology/WOD/ocldb_OSD.nc") {
  require("oppla")
  require("units")
  nc <- oppla::onc_info(filename = file)
  zrs <- RNetCDF::var.get.nc(ncfile = nc$self, variable = "z_row_size")
  zrs[is.na(zrs)] <- 0
  fieldnames <- matrix(c("Chlorophyll", "Nitrate", "Phosphate", "Silicate", "Oxygen", "tCO2",
                         "Chl",         "NO3",     "PO4",       "SIO3",     "O2",     "DIC"),
                       ncol = 2)
  data <- data.frame(latitude = rep(set_units(RNetCDF::var.get.nc(ncfile = nc$self, variable = "lat"),
                                              nc$var$lat$units, mode = "standard"), times = zrs),
                     longitude = rep(set_units(RNetCDF::var.get.nc(ncfile = nc$self, variable = "lon"),
                                               nc$var$lon$units, mode = "standard"), times = zrs),
                     date = rep(as.POSIXct(RNetCDF::var.get.nc(ncfile = nc$self, variable = "time") * 86400,
                                           origin = sub(pattern = "days since ", replacement = "",
                                                        x = nc$var$time$units), tz = "UTC"),
                                times = zrs))
  z <- RNetCDF::var.get.nc(ncfile = nc$self, variable = "z")
  data$depth <- units::set_units(z[z < 1e10], nc$var$z$units, mode = "standard")
  crs <- RNetCDF::var.get.nc(ncfile = nc$self, variable = "WOD_cruise_identifier")
  crs[nchar(crs) == 0] <- "WOD"
  data$cruise <- rep(crs, times = zrs)
  rs_names <- intersect(paste(fieldnames[, 1], "row_size", sep = "_"), names(nc$var))
  for (rs_name in rs_names) {
    rs <- RNetCDF::var.get.nc(ncfile = nc$self, variable = rs_name)
    name <- sub(pattern = "_row_size", replacement = "", x = rs_name)
    raw <- RNetCDF::var.get.nc(ncfile = nc$self, variable = name)
    nmiss <- sum(rs, na.rm = TRUE) - nc$var[[name]]$size
    data[[name]] <- NA
    data[[name]][rep(!is.na(rs), times = zrs)] <- c(raw, rep(NA, times = nmiss))
    if (name %in% fieldnames) data[[name]][data[[name]] < 0] <- NA
    if (nchar(nc$var[[name]]$units) > 0)
      units(data[[name]]) <- sub(pattern = "1e-1percent", replacement = "permil", x = nc$var[[name]]$units)
  }
  RNetCDF::close.nc(nc$self)
  ifns <- fieldnames[, 1] %in% names(data)
  names(data)[match(fieldnames[ifns, 1], names(data))] <- fieldnames[ifns, 2]
  attr(data, "file") <- file
  return(data[apply(X = data[fieldnames[ifns, 2]], MARGIN = 1, FUN = function(x) any(!is.na(x))), ])
}
