suppressMessages(require(data.table))
require(oppla) # needed to make readRDS assign units

getObsKDE <- function(edges, file = "../data/LS_WOD.rds", file_zoo = "../data/LS_WOD_plankton.rds",
                      file_zoo_int = "../data/LSzooInt.rds", n_kde = 10000L,
                      georange = list(lat = units::set_units(c(min = 40, max = 60), "degrees_N"),
                                      lon = units::set_units(c(min = -70, max = -50), "degrees_E"))) {
  ## kernel density estimates for (monthly climatologies of) observations
  f_kde <- function(x, n_run = 10000L) {
    run_x <- seq(from = 0 * x[1], to = max(x) * 2, length = n_run)
    cdd <- oppla_cdd(x = x, run = run_x, quiet = TRUE)
    return(c(cdd, run = list(run_x)))
  }
  thresholds <- list(Nitrate = units::set_units(0.05, "mmol m-3"),
                     Phosphate = units::set_units(0.01, "mmol m-3"),
                     Chlorophyll = units::set_units(0.05, "mg m-3"),
                     measure_val_bio = units::set_units(0.1, "mmol m-3"),
                     zooI = units::set_units(5, "mmol m-2"))
  if (length(edges) == 0) stop("Argument edges must not be empty!")
  dnames <- c("Nitrate", "Phosphate", "Chlorophyll")
  rho0 <- units::set_units(1025, "kg m-3") # default for missing density data
  ## nutrient and Chl data
  lsd <- data.table::as.data.table(readRDS(file = file))[
    lat %between% georange$lat & lon %between% georange$lon,
    j = .SD, .SDcols = c("lon", "lat", "date", "depth", dnames, "density")][
      j = c("Nitrate", "Phosphate") := .(Nitrate * fcoalesce(density, rho0),
                                         Phosphate * fcoalesce(density, rho0))]
  lsz <- data.table::as.data.table(readRDS(file = file_zoo))[
    lat %between% georange$lat & lon %between% georange$lon
    & ! measure_units_bio %in% c("code/haul", "#/ul", "#/ml", "#/haul", "ml/m2"),
    j = .(lon, lat, date, lower_z_bio, upper_z_bio, measure_val_bio, measure_units_bio)][
      j = depth := units::set_units(rowMeans(.SD), "m"), .SDcols = c("lower_z_bio", "upper_z_bio")]
  imlm3 <- which(lsz$measure_units_bio == "ml/m3") # local concentrations
  iugm3 <- which(lsz$measure_units_bio == "ug/m3")
  ## convert: ml (*1000) -> mm3/m3 (/0.95, Sato et al. 2015) -> wet weight (*0.064) -> mg (/12) -> mmol
  ## factor 0.064 = 0.4 * 0.16 (dry weight = 16% of wet weight, copepods contain 84% water, Postel et al., 2000, p.91)
  lsz$measure_val_bio[imlm3] <- lsz$measure_val_bio[imlm3] * 1000 / 0.95 * 0.064 / 12
  lsz$measure_val_bio[iugm3] <- lsz$measure_val_bio[iugm3] / 1000 * 0.4 / 12 # from ug dry weight/m3 to mmol C/m3
  lszc <- lsz[c(imlm3, iugm3)][j = measure_val_bio := units::set_units(measure_val_bio, "mmol m-3")]
  lszi <- data.table::as.data.table(readRDS(file = file_zoo_int))[ # 100m depth-integrated zooplankton data in mmol C/m2
    j = c("year", "month") := oppla_split_date(date, c("year", "month"))][
      j = .(total = mean(total)), by = c("year", "month")] # date columns
  scale_zoo <- median(lszi$total, na.rm = TRUE)
  cdd_zoo <- lapply(X = unique(lszi[j = month]),
                    FUN = function(x) {
                      c(month = x, f_kde(x = lszi[month == x, total], n_run = n_kde),
                        list(threshold = thresholds$zooI, scale = scale_zoo))
                    })
  lsall <- merge(lsd, lszc, all = TRUE)
  level_means <- oppla::op_aggregate(data = lsall, edges = edges, datecols = c("year", "month"),
                                     columns = c(dnames, "measure_val_bio"), fun = mean)
  ml <- unique(level_means[j = .(month, level)])
  cdd_ls <- sapply(X = c(dnames, "measure_val_bio"),
                   FUN = function(column) {
                     scale <- median(level_means[[column]], na.rm = TRUE)
                     cdd <- lapply(X = seq_len(nrow(ml)),
                                   FUN = function(x) {
                                     data <- na.omit(level_means[(month == ml[x, month]) & (level == ml[x, level]), ..column])
                                     if (length(data[[column]]) > 3)
                                       c(ml[x, ], f_kde(x = pmax(data[[column]], thresholds[[column]]), n_run = n_kde),
                                         list(threshold = thresholds[[column]], scale = scale))
                                   })
                     return(cdd[! sapply(cdd, is.null)])
                   })
  return(c(cdd_ls[sapply(cdd_ls, function(x) length(x) > 0)], zooI = list(cdd_zoo)))
}

costIQD <- function(obs, model, cut_date, max_depth, coco) {
  ## trapezoidal integral scaled (divided) by obs$scale (set to the median in getObsKDE)
  IQD <- function(obs, mod) {
    cdd_mod <- oppla_cdd(x = pmax(units::set_units(mod, units(obs$run)), obs$threshold), run = obs$run, quiet = TRUE)
    delta <- if (is.null(cdd_mod$cdd)) obs$cdd ^ 2 else (obs$cdd - cdd_mod$cdd) ^ 2
    return(sum((head(delta, -1L) + delta[-1L]) / 2 * diff(obs$run)) / obs$scale)
  }
  chl_vars <- paste(coco$cmo, "Chl", sep = "_")
  zoo_vars <- paste(coco$cfo, "C", sep = "_")
  monthly <- oppla_ave(data = oppla_slab(data = model,
                                         daterange = c(as.POSIXct(cut_date, tz = "UTC"), tail(model$date, 1)),
                                         depthrange = c(0, max_depth)),
                       columns = c("DIN", "DIP", chl_vars, zoo_vars), avc = c("year", "month"))[
                         j = Chl := units::set_units(rowSums(.SD), units(.SD[[1, 1]])), .SDcols = chl_vars][
                           j = zoo_TC := units::set_units(rowSums(.SD), units(.SD[[1, 1]])), .SDcols = zoo_vars][
                             j = month := month(date)]
  model_depth <- sort(unique(monthly$depth))
  zooI <- oppla::oppla_int(data = monthly, columns = c("zoo_TC", zoo_vars),
                           depth = units::set_units(100, "m", mode = "standard"))[j = month := month(date)]
  cost_names <- c(DIN = "Nitrate", DIP = "Phosphate", Chl = "Chlorophyll",
                  zoo_TC = "measure_val_bio", zoo_C = "measure_val_bio")
  IQDmod <- sapply(X = c("DIN", "DIP", "Chl"), # , "zoo_TC", "zoo_C"
                   FUN = function(column) {
                     sapply(X = obs[[cost_names[[column]]]],
                            FUN = function(x) {
                              IQD(obs = x,
                                  mod = monthly[(depth == model_depth[x$level]) & (month == x$month),
                                                ..column][[1]])
                            })
                   })
  IQDzooI <- sapply(X = c("zoo_C", "zoo_noSVM_C", "zoo_TC"),
                    FUN = function(column) {
                      sapply(X = obs$zooI, FUN = function(x) {
                        IQD(obs = x, mod = zooI[month == x$month, ..column][[1]])
                      })
                    }, simplify = FALSE)
  ## exlude zeros because they indicate missing observations
  return(sapply(X = c(IQDmod, IQDzooI), FUN = function(x) mean(x[x > 0])))
}

costfunIQD <- function(obs, model) {
  control <- oppla::nml_read(file = attr(model, "control_file"))
  param  <- oppla::nml_read(file = attr(model, "parameter_file"))
  cf     <- control$costfun[[1]]
  coco   <- if ("coco" %in% names(control)) control$coco[[1]] else param$coco[[1]]
  return(costIQD(obs = obs, model = model, cut_date = cf$cut_date, max_depth = cf$max_depth, coco = coco))
}
