plot_cmt_ls <- function(data, model, depthrange = units::set_units(c(0, 100), "m"),
                        daterange = as.POSIXct(c("1988-01-01", "1995-01-01"), tz = "UTC"), ...) {
  require(oppla)
  mc <- oppla::oppla_cmt(oppla_slab(model, depthrange = depthrange, daterange = daterange), ...)
  if ("month" %in% names(data)) {
    dc <- data
  } else {
    dc <- oppla::op_aggregate(data = data, datecols = "month", edges = attr(mc, "depth_edges"))
  }
  mc[j = Chlorophyll := phyt_Chl]
  dc[
    j = DIN := units::set_units(rowSums(.SD, na.rm = TRUE), units(Nitrate)),
    .SDcols = c("Nitrate", "Ammonia")
  ]
  dc[j = DIP := Phosphate]
  print(oppla_tile(Chlorophyll, data = list(data = dc, model = mc)))
  print(oppla_tile(DIN, data = list(data = dc, model = mc)))
  print(oppla_tile(DIP, data = list(data = dc, model = mc)))
  invisible(dc)
}

## use fun = median in ...
## export this figure as 7.5 x 5 inches:
#' dev.print(device = cairo_pdf, width = 7.5, height = 5, file = "LS_15_7_0529.pdf")
plot_surface_ls <- function(obs, model, daterange = as.POSIXct(c("1988-01-01", "1995-01-01"), tz = "UTC"), label = NULL, ...) {
  require(oppla)
  require(ggplot2)
  require(egg)
  require(ggnewscale)
  suppressMessages(require(data.table))
  plot_one <- function(obs, model, model_name, ylabel = model_name, colours = c("model"),
                       xticklabels = TRUE) {
    if ("level" %in% names(obs[[1]])) {
      data <- rbindlist(lapply(X = obs, FUN = function(x) x[c("month", "level", "25%", "50%", "75%")]))[level == 1]
    } else {
      data <- rbindlist(lapply(X = obs, FUN = function(x) x[c("month", "25%", "50%", "75%")]))
    }
    units <- units(model[[model_name]]) # convert all data to model units
    data[j = c("25%", "50%", "75%") := lapply(.SD, function(x) units::set_units(x, units)), .SDcols = c("25%", "50%", "75%")]
    return(ggplot(data = units::drop_units(data), mapping = aes(x = month, y = `50%`)) +
           geom_pointrange(mapping = aes(ymin = `25%`, ymax = `75%`, linetype = "data")) +
           geom_line(data = units::drop_units(model),
                     mapping = aes(x = month, y = .data[[model_name]], colour = colours[[1]])) +
           scale_x_continuous(breaks = 1:12, labels = if (xticklabels) month.abb else NULL, name = NULL) +
           scale_y_continuous(limits = c(0, NA), name = units::make_unit_label(ylabel, units)) +
           scale_colour_viridis_d(end = 0.9, limits = colours, labels = parse(text = colours)) +
           guides(colour = guide_legend(override.aes = list(linetype = c(1, 1, 2)[seq_along(colours)]), order = 1),
                  linetype = guide_legend(order = 2)) +
           theme_bw() +
           theme(legend.text.align = 0, legend.position = c(0.95, 0.98), legend.justification = c(1, 1),
                 legend.background = element_rect(fill = "transparent"), legend.spacing = unit(0, "mm"),
                 legend.margin = margin(), axis.ticks.length.x = if (xticklabels) unit(2.75, "pt") else unit(0, "pt"),
                 legend.title = element_blank()))
  }
  ms <- oppla_slab(model, daterange = daterange)
  mc <- oppla::oppla_cmt(ms, columns = c("phyt_Chl", "DIN", "DIP"), ...)[depth == min(depth)] # extract surface layer
  pc <- plot_one(obs = obs$Chlorophyll, model = mc, model_name = "phyt_Chl",
                 ylabel = "Chlorophyll", xticklabels = FALSE)
  pn <- plot_one(obs = obs$Nitrate, model = mc, model_name = "DIN")
  pp <- plot_one(obs = obs$Phosphate, model = mc, model_name = "DIP")
  mci <- oppla::oppla_int(ms, avc = "month", columns = c("zoo_C", "zoo_noSVM_C"), ...)
  pz <- plot_one(obs = obs$zooI, model = mci, model_name = "zoo_noSVM_C", ylabel = "Zooplankton",
                 xticklabels = FALSE, colours = c("mu * Z", "SVM", "total")) +
    geom_line(data = units::drop_units(mci), mapping = aes(x = month, y = zoo_C, colour = "SVM")) +
    geom_line(data = units::drop_units(mci), mapping = aes(x = month, y = zoo_C + zoo_noSVM_C, colour = "total"),
              linetype = 2)
  return(ggarrange(pz, pc, pn, pp, ncol = 2, nrow = 2, top = label, labels = c("A", "B", "C", "D"),
                   label.args = list(hjust = 0, gp = grid::gpar(cex = 1.2))))
}

## use fun = median in ...
## export button in RStudio does not allow this aspect ratio:
## use dev.print(device = cairo_pdf, width = 7.5, height = 2.5, file = "LS_chl_locations.pdf")
plot_surface_chl <- function(obs, models, daterange = as.POSIXct(c("1988-01-01", "1995-01-01"), tz = "UTC"),
                             label = "", tags = names(models), legend_label = "location", nrow = 1,
                             labels = c("A", "B"), months = 1:12, ...) {
  require(oppla)
  require(ggplot2)
  require(egg)
  require(ggnewscale)
  suppressMessages(require(data.table))
  cmts <- list()
  units <- units(models[[1]]$phyt_Chl)
  chl <- rbindlist(lapply(X = obs$Chlorophyll, FUN = function(x) x[c("month", "level", "25%", "50%", "75%")]))[level == 1]
  chl[j = c("25%", "50%", "75%") := lapply(.SD, function(x) units::set_units(x, units)), .SDcols = c("25%", "50%", "75%")]
  for (n in seq_along(models)) {
    models[[n]] <- oppla::oppla_slab(models[[n]], daterange = daterange)[depth == min(depth)]
    cmts[[n]] <- oppla::oppla_cmt(models[[n]], columns = c("phyt_Chl"), ...)
    models[[n]]$tag <- tags[n]
    cmts[[n]]$tag <- tags[n]
  }
  all_models <- data.table::rbindlist(models)
  all_cmts <- data.table::rbindlist(cmts)
  ylabel <- units::make_unit_label("Chl", units(all_models[["phyt_Chl"]]))
  ## time-series plot
  tsp <- ggplot(data = units::drop_units(all_models), mapping = aes(x = date, y = phyt_Chl, colour = tag)) +
    scale_x_datetime(name = NULL) +
    scale_y_continuous(limits = c(0, NA), name = ylabel) +
    geom_line() + scale_colour_viridis_d(end = 0.9, name = legend_label, guide = guide_legend(nrow = 2)) + theme_bw() +
    theme(legend.position = c(0.03, 0.97), legend.background = element_rect(fill = "transparent"),
          legend.spacing.y = unit(0, "mm"), legend.margin = margin(), legend.justification = c(0, 1),
          legend.title.align = 0.5)
  cmtp <- ggplot() +
    geom_pointrange(data = units::drop_units(chl),
                    mapping = aes(x = month, y = `50%`, ymin = `25%`, ymax = `75%`, linetype = "data")) +
    scale_linetype(name = NULL, guide = guide_legend(order = 0)) +
    geom_line(data = units::drop_units(all_cmts), mapping = aes(x = month, y = phyt_Chl, colour = tag), show.legend = FALSE) +
    scale_x_continuous(breaks = months, labels = month.abb[months], name = NULL) +
    scale_y_continuous(limits = c(0, NA), name = ylabel) +
    scale_colour_viridis_d(end = 0.9) +    ## scale_colour_discrete(name = legend_label, guide = guide_legend(order = 1)) +
    theme_bw() +
    theme(legend.position = c(0.03, 0.97), legend.background = element_rect(fill = "transparent"),
          legend.margin = margin(), legend.justification = c(0, 1))
  return(ggarrange(tsp, cmtp, nrow = nrow, labels = labels,
                   label.args = list(hjust = 0, gp = grid::gpar(cex = 1.2))))
}

## export to size 7.5 x 6 inches (5 panels) or 7.5 x 4 inches (3 panels)
plot_surface_zoo <- function(models, daterange = as.POSIXct(c("1988-01-01", "1995-01-01"), tz = "UTC"),
                             nrow = 3, with_tags = TRUE) {
  require(oppla)
  require(units)
  require(ggplot2)
  suppressMessages(require(data.table))
  izoo <- list()
  for (config in names(models)) {
    columns <- intersect(names(models[[config]]), c("zoo_C", "zoo_noSVM_C", "zoo_id_C"))
    izoo[[config]] <- oppla::oppla_int(oppla_slab(models[[config]], daterange = daterange), columns = columns)[
                               j = configuration := ..config] # add configuration column
  }
  all_zoo <- data.table::rbindlist(izoo, fill = TRUE)[ # add column with total zooplankton
                           j = total := rowSums(x = cbind(zoo_noSVM_C, zoo_C, zoo_id_C), na.rm = TRUE)]
  ylabel <- units::make_unit_label("Integrated zooplankton (100 m)", units(all_zoo$zoo_C))
  tags <- data.frame(configuration = names(models), tag = LETTERS[seq_along(models)],
                     x = as.POSIXct(-Inf), y = Inf)
  zplot <- ggplot(data = units::drop_units(all_zoo), mapping = aes(x = date)) +
    geom_line(mapping = aes(y = zoo_noSVM_C, colour = "mu * Z")) +
    geom_line(mapping = aes(y = zoo_C, colour = "SVM")) +
    geom_line(mapping = aes(y = zoo_id_C, colour = "Stayers")) +
    scale_x_datetime(name = NULL) + scale_y_continuous(name = ylabel) +
    scale_colour_viridis_d(end = 0.9, name = NULL, limits = c("mu * Z", "SVM", "Stayers"),
                           labels = expression(mu * Z, SVM, Stayers)) +
    coord_cartesian(clip = "off") +
    theme_bw() +
    theme(legend.text.align = 0,
          legend.position = if (with_tags) c(0.75, 0.05) else c(0.5, 0.1),
          legend.justification = c(0.5, 0),
          legend.direction = if (with_tags) "vertical" else "horizontal",
          panel.border = element_rect(linewidth = 0.2), legend.background = element_blank(),
          strip.background = element_rect(fill = "transparent", linetype = "blank")) +
    facet_wrap(facets = vars(configuration), nrow = nrow, labeller = label_parsed)
  if (with_tags) zplot <- zplot +
                   geom_text(data = tags, mapping = aes(x = x, y = y, label = tag), hjust = 0, vjust = -0.5, size = 4)
  return(zplot)
}

## export to size 7.5 x 5 inches:
#' dev.print(device = cairo_pdf, width = 7.5, height = 5, file = "LS_cmt_svm_0529.pdf")
plot_cmt_svm <- function(obs, models, daterange = as.POSIXct(c("1988-01-01", "1995-01-01"), tz = "UTC"),
                         label = NULL, labels = c("A", "B", "C", "D"), months = 1:12, ...) {
  require(oppla)
  require(ggplot2)
  require(egg)
  require(ggnewscale)
  suppressMessages(require(data.table))
  plot_one <- function(obs, model, model_name, ylabel = model_name, colours = c("data", "model"),
                       legend = FALSE, xticklabels = TRUE, months = 1:12) {
    if ("level" %in% names(obs[[1]])) {
      data <- rbindlist(lapply(X = obs, FUN = function(x) x[c("month", "level", "25%", "50%", "75%")]))[level == 1]
    } else {
      data <- rbindlist(lapply(X = obs, FUN = function(x) x[c("month", "25%", "50%", "75%")]))
    }
    units <- units(model[[model_name]]) # convert all data to model units
    data[j = c("25%", "50%", "75%") := lapply(.SD, function(x) units::set_units(x, units)),
         .SDcols = c("25%", "50%", "75%")]
    cmtplot <- ggplot(data = units::drop_units(data), mapping = aes(x = month, y = `50%`)) +
      geom_pointrange(mapping = aes(ymin = `25%`, ymax = `75%`, colour = "data"),
                      show.legend = if (legend) NA else FALSE) +
      scale_colour_manual(name = NULL, values = "black", limits = "data") +
      new_scale_colour() +
      geom_line(data = units::drop_units(model), mapping = aes(x = month, y = .data[[model_name]], colour = configuration),
                show.legend = if (legend) NA else FALSE) +
      scale_x_continuous(breaks = months, labels = if (xticklabels) month.abb[months] else NULL, name = NULL) +
      scale_y_continuous(limits = c(0, NA), name = units::make_unit_label(ylabel, units)) +
      scale_colour_viridis_d(begin = 0.3, end = 0.7, direction = -1, name = NULL, guide = guide_legend(reverse = TRUE)) +
      theme_bw() +
      theme(axis.ticks.length.x = if (xticklabels) unit(2.75, "pt") else unit(0, "pt"))
    if (legend)
      cmtplot <- cmtplot + theme(legend.position = c(0.95, 0.98), legend.background = element_rect(fill = "transparent"),
                                 legend.spacing = unit(0, "mm"), legend.margin = margin(), legend.justification = c(1, 1))
    return(cmtplot)
  }
  cmts <- list()
  for (config in names(models)) {
    ms <- oppla_slab(models[[config]], daterange = daterange)
    columns <- intersect(names(ms), c("zoo_C", "zoo_noSVM_C", "zoo_id_C"))
    mci <- oppla::oppla_int(ms, avc = "month", columns = columns)
    cmts[[config]] <- oppla_cmt(ms, columns = c("phyt_Chl", "DIN", "DIP"), ...)[depth == min(depth)][
      j = zoo := units::set_units(x = rowSums(mci[j = .SD, .SDcols = columns]),
                                  value = units(mci[[columns[1]]]))][j = configuration := ..config]
  }
  all_cmts <- data.table::rbindlist(cmts)
  pz <- plot_one(obs = obs$zooI, model = all_cmts, model_name = "zoo", ylabel = "Zooplankton",
                 xticklabels = FALSE, months = months)
  pc <- plot_one(obs = obs$Chlorophyll, model = all_cmts, model_name = "phyt_Chl",
                 ylabel = "Chlorophyll", legend = TRUE, xticklabels = FALSE, months = months)
  pn <- plot_one(obs = obs$Nitrate, model = all_cmts, model_name = "DIN", months = months)
  pp <- plot_one(obs = obs$Phosphate, model = all_cmts, model_name = "DIP", months = months)
  return(ggarrange(pz, pc, pn, pp, ncol = 2, nrow = 2, top = label,
                   labels = labels, label.args = list(hjust = 0, gp = grid::gpar(cex = 1.2))))
}

plot_cmt_svm_2 <- function(obs, models, daterange = as.POSIXct(c("1988-01-01", "1995-01-01"), tz = "UTC"),
                           label = NULL, labels = NULL, months = seq(from = 1, to = 12, by = 2), ...) {
  require(oppla)
  require(ggplot2)
  require(egg)
  require(ggnewscale)
  suppressMessages(require(data.table))
  plot_one <- function(obs, model, model_name, ylabel = model_name, colours = c("data", "model"),
                       legend = FALSE, xticklabels = TRUE, months = 1:12) {
    if ("level" %in% names(obs[[1]])) {
      data <- rbindlist(lapply(X = obs, FUN = function(x) x[c("month", "level", "25%", "50%", "75%")]))[level == 1]
    } else {
      data <- rbindlist(lapply(X = obs, FUN = function(x) x[c("month", "25%", "50%", "75%")]))
    }
    units <- units(model[[model_name]]) # convert all data to model units
    data[j = c("25%", "50%", "75%") := lapply(.SD, function(x) units::set_units(x, units)),
         .SDcols = c("25%", "50%", "75%")]
    cmtplot <- ggplot(data = units::drop_units(data), mapping = aes(x = month, y = `50%`)) +
      geom_pointrange(mapping = aes(ymin = `25%`, ymax = `75%`, colour = "data"),
                      show.legend = if (legend) NA else FALSE) +
      scale_colour_manual(name = NULL, values = "black", limits = "data") +
      ggnewscale::new_scale_colour() +
      geom_line(data = units::drop_units(model), mapping = aes(x = month, y = .data[[model_name]], colour = configuration),
                show.legend = if (legend) NA else FALSE) +
      scale_x_continuous(breaks = months, labels = if (xticklabels) month.abb[months] else NULL, name = NULL) +
      scale_y_continuous(limits = c(0, NA), name = units::make_unit_label(ylabel, units)) +
      scale_colour_viridis_d(begin = 0.3, end = 0.7, direction = -1, name = NULL, guide = guide_legend(reverse = TRUE)) +
      theme_bw() +
      theme(axis.ticks.length.x = if (xticklabels) unit(2.75, "pt") else unit(0, "pt"))
    if (legend)
      cmtplot <- cmtplot + theme(legend.position = c(0.95, 0.98), legend.background = element_rect(fill = "transparent"),
                                 legend.spacing = unit(0, "mm"), legend.margin = margin(), legend.justification = c(1, 1))
    return(cmtplot)
  }
  cmts <- list()
  for (config in names(models)) {
    ms <- oppla_slab(models[[config]], daterange = daterange)
    columns <- intersect(names(ms), c("zoo_C", "zoo_noSVM_C", "zoo_id_C"))
    mci <- oppla::oppla_int(ms, avc = "month", columns = columns)
    cmts[[config]] <- oppla_cmt(ms, columns = c("phyt_Chl", "DIN", "DIP"), ...)[depth == min(depth)][
      j = zoo := units::set_units(x = rowSums(mci[j = .SD, .SDcols = columns]),
                                  value = units(mci[[columns[1]]]))][j = configuration := ..config]
  }
  all_cmts <- data.table::rbindlist(cmts)
  pz <- plot_one(obs = obs$zooI, model = all_cmts, model_name = "zoo", ylabel = "Zooplankton",
                 months = months, xticklabels = FALSE)
  pc <- plot_one(obs = obs$Chlorophyll, model = all_cmts, model_name = "phyt_Chl",
                 ylabel = "Chlorophyll", legend = TRUE, months = months)
  return(ggarrange(pz, pc, ncol = 1, nrow = 2, top = label,
                   labels = labels, label.args = list(hjust = 0, gp = grid::gpar(cex = 1.2))))
}
