#' Read Fortran namelists into a named list of lists.
#'
#' \code{read.nml} returns a list of lists whose names are those of the
#'   namelists
#'
#' @param file name of a file with fortran namelists
#'
#' @return A list of namelists.
#'
#' Each element is a list of namelist entries with the same name,
#'   whose elements are the individual variables in each namelist.
#'   Variable contents are stored as matrices with units if provided.
#'   Row names, if present, represent tags.
#' @export
nml_read <- function(file) {
  c <- paste(gsub(pattern = "[ \t]*!.*", replacement = "", # remove comments
                  x = scan(file = file, what = character(), sep = "\n", quiet = TRUE)), collapse = " ")
  ## read namelists into a matrix with columns nlname and content
  nls <- namedCapture::str_match_all_named(c, pattern = "&(?<nlname>\\w+)[ \t]*(?<content>.*?/)[ \t]*(?=&|$)")[[1]]
  nlnames <- nls[, "nlname"] # names of all namelists
  nl <- list()               # list to collect output in
  for (nnl in seq_along(nlnames)) {
    n <- length(nl[[nlnames[nnl]]]) + 1 # size of this nml so far
    if (n == 1) nl[[nlnames[nnl]]] <- list() # add new list for nmls of this name if not present already
    nl[[c(nlnames[nnl], n)]] <- list()       # list to hold content of this namelist
    vars <- namedCapture::str_match_all_named(nls[nnl, "content"],
                          pattern = "(?<name>\\w+)[ \t]*=[ \t]*(?<content>[^=]+?)(?=,[ \t]*\\w+[ \t]*=|/$)")[[1]]
    for (nv in seq_along(vars)) {
      if (nzchar(trimws(vars[nv, "content"]))) {
        vl <- namedCapture::str_match_all_named(
                              strsplit(trimws(vars[nv, "content"]), split = ",")[[1]],
                              pattern = "\\s*(?<tag>'[^']*')?\\s*(?<value>'[^']*'|[[:graph:]]+)\\s*(?<units>'[^']+')?")
        cm <- gsub("'", "", do.call(what = rbind, args = vl)) # remove quotes, combine into character matrix
        vm <- tryCatch(as.numeric(cm[, "value"]), warning = function(x) as.vector(cm[, "value"]))
        if (nzchar(cm[1, "tag"])) names(vm) <- cm[, "tag"]
        if (nzchar(cm[1, "units"]))
          vm <- units::as_units(units::mixed_units(vm, cm[, "units"])) # convert to uniform units
        if (all(toupper(vm) %in% c("F", "T", ".FALSE.", ".TRUE.")))
          vm <- as.logical(gsub("\\.", "", toupper(vm)))
        if (all(is.na(vm))) vm[] <- "" # must be an empty string
        nl[[c(nlnames[nnl], n, rownames(vars)[nv])]] <- vm
      }
    }
  }
  return(nl)
}

#' Write a list of namelists to a fortran namelist file.
#'
#' \code{nml_write} writes a namelist as obtained with \code{nml_read} and
#'   possibly modified with \code{nml_set}
#'
#' @param nml list of namelists obtained with \code{nml_read} and possibly
#'   modified with \code{nml_set}
#' @param file name of namelist file to be written
#' @export
nml_write <- function(nml, file) {
  fc <- file(description = file, open = "wt")
  for (nm in names(nml)) {
    for (n in seq_along(nml[[nm]])) {
      cat("&", nm, " ", file = fc, sep = "", append = TRUE) # write start of namelist
      nms <- names(nml[[nm]][[n]])                          # variable names
      nvs <- length(nms)
      for (nv in seq_len(nvs)) {
        v <- nml[[nm]][[n]][[nv]]
        vn <- names(v)
        v <- switch(class(v),
                    logical = paste0(".", v, "."),
                    character = paste0("'", v, "'"),
                    units = cbind(v, paste0("'", units::deparse_unit(v), "'")),
                    numeric = v)
        if (! is.null(vn)) v <- cbind(paste0("'", vn, "'"), v)
        vm <- as.matrix(v)
        cat(nms[nv], "=", do.call(what = paste, args = c(split(vm, col(vm)), collapse = ", ")),
            if (nv < nvs) ", ",
            file = fc, sep = "", append = TRUE) # write entry for variable
      }
      cat("/\n", file = fc, sep = "", append = TRUE) # write end of namelist
    }
  }
  close(fc)
}

#' Modify namelist entries.
#'
#' \code{nml_set} sets one or more entries in a list of namelists obtained with
#'   \code{nml_read}
#'
#' @param nml list of namelists obtained with \code{nml_read}
#' @param vars list of lists, each of the form
#'   \code{list("namelist name", index, "variable name"[, indices])}
#'   or vector or list of variable specifications of the form
#'   \code{"namelist name-index-variable name[-indices]"}
#' @param values named list of values corresponding to indices in vars
#' @return Modified list of namelists.
#' @export
nml_set <- function(nml, vars, values) {
  for (nv in seq_along(vars)) {
    vi <- vars[[nv]]
    if (length(vi) == 1) {
      vi <- as.list(strsplit(vi, split = "-", fixed = TRUE)[[1]])
      if (length(vi) == 4) vi[[4]] <- eval(str2lang(vi[[4]]))
    }
    ne <- nml[[unlist(vi[1:3])]]
    if (is.null(ne)) ne <- 0 # in case the variable is not present yet
    if (length(vi) < 4) vi[[4]] <- seq_along(ne)
    nml[[unlist(vi[1:3])]][vi[[4]]] <- values[[nv]]
  }
  return(nml)
}

#' Extract values from a namelist.
#'
#' \code{nml_get} returns a list of values corresponding to vars
#'
#' @param nml list of namelists as obtained with \code{nml_read}
#' @param vars a named list of lists, each of the form
#'   \code{name = list("namelist name", index, "variable name"[, indices])}
#'   or a named vector or list of variable specifications of the form
#'   \code{name = "namelist name-index-variable name[-indices]"}
#' @return Named list of values.  The names are those of \code{vars}.
#' @export
nml_get <- function(nml, vars) {
  values <- list()
  for (nv in names(vars)) {
    vi <- vars[[nv]]
    if (length(vi) == 1) {
      vi <- as.list(strsplit(vi, split = "-", fixed = TRUE)[[1]])
      if (length(vi) == 4) vi[[4]] <- eval(str2lang(vi[[4]]))
    }
    tryCatch(ne <- nml[[unlist(vi[1:3])]], error = function(e) {
      if (! vi[1] %in% names(nml))
        stop(simpleError(paste0("No namelist ", vi[[1]], " in ",
                                deparse1(substitute(nml, env = parent.env(environment()))), "."),
                         call = sys.call(which = -4)))
      else if (vi[2] > length(nml[[vi[[1]]]]))
        stop(simpleError(paste0(deparse(substitute(nml)), " has only ", length(nml[[vi[[1]]]]),
                                " namelists ", vi[[1]], "."), call = sys.call(which = -4)))
      else stop(e)
    })
    if (length(vi) < 4) vi[[4]] <- seq_along(ne)
    if (length(vi[[4]]) == 0)
      stop("No Variable ", vi[[3]], " in namelist ", vi[[1]], "(", vi[[2]], ")")
    values[[nv]] <- ne[vi[[4]], drop = FALSE]
  }
  return(values)
}
