#' Read \code{oppla} results.
#'
#' Read \code{oppla} netcdf results into an R data table.
#'
#' @return [`Data table`][data.table::data.table] with a single-column vector date
#' of class [`POSIXct`][base::DateTimeClasses].  Numeric data are stored as [`units`][units::units] objects.
#' The depth edges of the layers are stored as attribute "depth_edges".
#'
#' @details Note that reading forcing data data with \code{forcing = TRUE} may not work if the
#' \code{file} was moved or copied if the \code{control_file} and \code{data_file} were specified
#' as relative paths.
#'
#' @param file name of oppla output file in netcdf format
#' @param forcing flag indicating whether to read forcing data
#' @param datelim 2-element vector of first and last times to extract
#' @seealso See [`units`][units::units], [`set_units`][units::set_units], and
#' [`mixed_units`][units::mixed_units] for details about units objects.
#' @export
oppla_read <- function(file, forcing = TRUE, datelim = NULL) {
  absolute_path <- function(path, file) {
    return(if (substr(file, 1, 1) %in% c("/", "~")) file else file.path(path, file))
  }
  nc <- oppla::onc_info(filename = file)
  depth <- units::set_units(nc$dim$depth$vals, nc$var$depth$units, mode = "standard")
  depth_edges <- units::set_units(x = RNetCDF::var.get.nc(ncfile = nc$self, variable = "depth_edges"),
                                  value = nc$var$depth_edges$units, mode = "standard")
  if ((length(dim(depth_edges)) < 2) && (length(depth_edges) == (length(depth) + 1)))
    depth_edges <- units::keep_units(FUN = rbind, head(depth_edges, -1), depth_edges[-1])
  time <- as.POSIXct(x = nc$dim$time$vals, tz = "UTC",
                     origin = as.POSIXct(x = nc$var$time$units,
                                         format = "seconds since %Y-%m-%d", tz = "UTC"))
  dl <- oppla_datelim(time, datelim)
  oppla <- data.table::data.table(date = rep(time[dl$start:dl$end], each = length(depth)),
                                  depth = rep(depth, times = dl$count),
                                  height = rep(diff(depth_edges), times = dl$count))
  if (dl$count > 1) {
    dt <- units::set_units(diff(as.numeric(time[dl$start:dl$end])), "s")
    oppla[j = dt := rep(c(dt, tail(dt, n = 1L)), each = length(..depth))]
  } else {
    forcing <- FALSE
  }
  start <- c(1, dl$start, 1, 1)
  count <- c(NA, dl$count, NA, NA)
  data.table::setattr(oppla, name = "file", value = normalizePath(file))
  for (name in names(nc$var)) {
    ## read all regular state variables from the netcdf file
    if (setequal(nc$var[[name]]$dimensions, nc$var$DIC$dimensions)) {
      oppla[j = c(name) := as.vector(RNetCDF::var.get.nc(ncfile = nc$self, variable = ..name,
                                                         start = start, count = count))]
      if ("units" %in% names(nc$var[[name]]))
        oppla[j = c(name) := units::set_units(.SD[[1]], nc$var[[name]]$units), .SDcols = c(name)]
    }
  }
  uo <- units::units_options(auto_convert_names_to_symbols = FALSE)
  lat <- units::set_units(x = nc$dim$latitude$vals, value = nc$var$latitude$units, mode = "standard")
  lon <- units::set_units(x = nc$dim$longitude$vals, value = nc$var$longitude$units, mode = "standard")
  units::units_options(uo)
  gatts <- c(nc$att, list(latitude = lat, longitude = lon, depth = depth, depth_edges = depth_edges))
  for (att in names(gatts)) data.table::setattr(oppla, name = att, value = gatts[[att]]) # global attributes
  ## replace file names with their canonical paths
  path <- sub(pattern = "[^/]+$", replacement = ".", x = file)
  for (file_name in intersect(names(gatts),
                              c("control_file", "data_file", "parameter_file", "initial_conditions"))) {
    file_path <- absolute_path(path, gatts[[file_name]])
    if (file.exists(file_path)) file_path <- normalizePath(file_path)
    data.table::setattr(oppla, name = file_name, value = file_path)
  }
  if (!("soma" %in% names(nc$att))) {
    if ("soma" %in% names(nc$var)) data.table::setattr(oppla, name = "soma", value = "soma")
    if ("flux" %in% names(nc$var)) data.table::setattr(oppla, name = "flux", value = "flux")
  }
  RNetCDF::close.nc(con = nc$self)
  if (forcing) oppla_forcing(oppla)
  return(oppla)
}

#' Add forcing data.
#'
#' Add the forcing data columns to \code{oppla} output.
#'
#' @param oppla oppla output as obtained with [`oppla_read`]
#'
#' @return \code{oppla_forcing} adds forcing columns to its \code{oppla} argument, so it is not
#' necessary to assign the result
#' @export
oppla_forcing <- function(oppla) {
  cfn <- attr(oppla, "control_file")
  dfn <- attr(oppla, "data_file")
  if (! all(file.exists(cfn, dfn))) {
    warning("Could not find ", paste(c(cfn, dfn)[! file.exists(cfn, dfn)], collapse = " and "),
            ".  Forcing data were omitted.")
    invisible(oppla)
  }
  fdf <- oppla::nml_read(file = cfn)$physics[[1]]$forcing
  fds <- oppla::nml_read(file = cfn)$physics[[1]]$surface
  fdx <- oppla::nml_read(file = cfn)$physics[[1]]$surflx
  ncf <- oppla::onc_info(filename = dfn)
  ft0 <- as.POSIXct(x = ncf$var$time$units,
                    format = "seconds since %Y-%m-%d", tz = "UTC") # origin of time data
  fdat <- as.POSIXct(x = RNetCDF::var.get.nc(ncf$self, variable = "time"), origin = ft0, tz = "UTC")
  time <- unique(oppla$date)
  for (name in names(fdf)) # interpolate forcing to dates of model output
    oppla[j = c(name) :=
            units::set_units(x = as.vector(t(apply(X = RNetCDF::var.get.nc(ncf$self, variable = fdf[[..name]][1]),
                                                   MARGIN = 1,
                                                   FUN = function(y) approx(x = fdat, y = y, rule = 2, xout = time)$y))),
                             value = ncf$var[[fdf[[..name]][1]]]$units, mode = "standard")]
  for (name in names(fds))
    data.table::setattr(oppla, name = name,
                        value = units::set_units(x = approx(x = fdat, y = RNetCDF::var.get.nc(ncf$self,
                                                                                              variable = fds[[name]][1]),
                                                            rule = 2, xout = time)$y,
                                                 value = ncf$var[[fds[[name]][1]]]$units, mode = "standard"))
  for (name in names(fdx))
    data.table::setattr(oppla, name = name,
                        value = units::set_units(x = approx(x = fdat, y = RNetCDF::var.get.nc(ncf$self,
                                                                                              variable = fdx[[name]][1]),
                                                            rule = 2, xout = time)$y,
                                                 value = ncf$var[[fdx[[name]][1]]]$units, mode = "standard"))
  RNetCDF::close.nc(con = ncf$self)
  invisible(oppla)
}

## Calculate netcdf indices for date limits.
##
## time     time vector of oppla output
## datelim  2-element vector of first and last times to extract
##
## This function is only used internally.
oppla_datelim <- function(time, datelim) {
  if (is.null(datelim)) datelim <- c(time[1], utils::tail(time, n = 1))
  start <- which(time >= datelim[1])[1]
  end <- utils::tail(which(time <= datelim[2]), n = 1)
  count <- end - start + 1
  return(list(start = start, end = end, count = count))
}

#' Read flux array.
#'
#' Read the vertical fluxes (sinking + advection + mixing) for all state variables.
#'
#' @param data oppla output as obtained with [`oppla_read`]
#' @param datelim 2-element vector of first and last times to extract
#'
#' @return [`Data table`][data.table::data.table] with vertical fluxes for all state variables,
#' including surface fluxes.
#' @export
oppla_flux <- function(data, datelim = range(data$date)) {
  if (!("flux" %in% names(attributes(data))))
    stop("Cannot find flux array in file ", attr(data, "file"), ".")
  nc <- oppla::onc_info(filename = attr(data, "file"))
  time <- as.POSIXct(x = nc$dim$time$vals,
                     origin = as.POSIXct(x = nc$var$time$units, format = "seconds since %Y-%m-%d"))
  depths <- RNetCDF::var.get.nc(ncfile = nc$self, variable = "depth_flux")
  dl <- oppla_datelim(time, datelim)
  flux <- RNetCDF::var.get.nc(ncfile = nc$self, variable = attr(data, "flux"),
                              start = c(1, 1, dl$start, 1, 1), count = c(NA, NA, dl$count, NA, NA))
  RNetCDF::close.nc(nc$self)
  dimnames(flux) <- list(state = nc$dim$States$vals, NULL, NULL)
  flux_list <- sapply(X = nc$dim$States$vals, simplify = FALSE,
                      FUN = function(state) {
                        units::set_units(as.vector(flux[state, , ]), units(data[[state]])) *
                          units::set_units(86400, "m d-1")
                      })
  return(data.table::as.data.table(c(expand.grid(depth = units::set_units(depths, "m"),
                                                 date = time[dl$start:dl$end]),
                                     flux_list)))
}

#' Read the soma array.
#'
#' The soma (sources-minus-sinks matrix) is a 3D (\code{nsms}, \code{depth}, \code{time}) array,
#' where \code{nsms} is the number of sources-minus-sinks terms.
#'
#' @param data oppla output as obtained with [`oppla_read`]
#' @param datelim 2-element vector of first and last times to extract
#'
#' @return \code{oppla_soma} adds the soma columns to \code{data}, so it is not necessary to assign the result
#' @export
oppla_soma <- function(data, datelim = range(data$date)) {
  if ("soma" %in% names(attributes(data))) {
    nc <- oppla::onc_info(filename = attr(data, "file"))
    time <- as.POSIXct(x = nc$dim$time$vals,
                       origin = as.POSIXct(x = nc$var$time$units, format = "seconds since %Y-%m-%d", tz = "UTC"),
                       tz = "UTC")
    dl <- oppla_datelim(time, datelim = datelim)
    soma <- RNetCDF::var.get.nc(ncfile = nc$self, variable = attr(data, "soma"), start = c(1, 1, dl$start, 1, 1),
                                count = c(NA, NA, dl$count, NA, NA))
    ## first row in soma is a phantom row for aggregation
    colnames <- trimws(nc$dim$sms$vals)
    stvnames <- sub(pattern = "[[:alnum:]_]+->([[:alnum:]_]+)", replacement = "\\1", x = colnames)
    if (! identical(datelim, range(data$date))) data <- data[date %between% datelim]
    for (n in seq_along(colnames))
      data[j = c(colnames[n]) := units::set_units(x = as.vector(soma[n, , ]) * 86400,
                                                  value = units(data[[stvnames[n]]])) * units::as_units("d-1")]
    RNetCDF::close.nc(nc$self)
  } else {
    warning("Cannot find soma in file ", attr(data, "file"), ".")
  }
  invisible(data)
}

#' Concatenate \code{oppla} simulations.
#'
#' Concatenate results of several \code{oppla} simulations as if they were parts
#' of one long simulation.
#'
#' @param ... oppla results as obtained with[`oppla_read`]
#'
#' @return [`Data table`][data.table::data.table] with all simulations in \code{...} concatenated.
#' If all \code{date} columns are identical, a dummy date column is created to
#' allow plotting the results on a continuous date axis.
#' @export
oppla_cat <- function(...) {
  if (data.table::is.data.table(..1)) models <- list(...)
  else models <- ..1
  ## check whether all simulations in ... have the same date column
  if (Reduce(f = function(x, y) x & identical(y$date, models[[1]]$date), x = models[-1], init = TRUE))
    for (n in 2:length(models)) # if so, create a dummy time axis
      models[[n]][j = date := date + (tail(models[[n - 1]]$date, n = 1) - models[[1]]$date[1])]
  return(data.table::rbindlist(models))
}

#' Time-slice from oppla results.
#'
#' Extract one time-slice from oppla results.
#'
#' @param oppla oppla output as obtained with [`oppla_read`]
#' @param date target date: \code{oppla_slice} will extract
#' the time slice closest to \code{date} in \code{oppla}
#' @export
oppla_slice <- function(oppla, date = oppla[1, date]) {
  dts <- oppla[j = date[which.min(abs(date - as.POSIXct(..date, tz = "UTC")))]]
  return(oppla[date == dts])
}

#' Means (or other functions) of oppla results.
#'
#' Means (or other functions) (of temporal subsets) of oppla results.
#'
#' @return [`Data table`][data.table::data.table] with aggregated temporal subsets.
#'
#' @param data oppla data as obtained with [`oppla_read`]
#' @param columns character vector of data columns to include
#' @param avc character vector of subsetting columns in data.  \code{avc = "month"}
#'   produces a monthly climatology.  \code{avc = c("mday", "month", "year")}
#'   gives daily means (if \code{fun = mean})
#' @param fun function to apply to the temporal subsets
#'
#' @details
#' The order of the averaging columns (avc) determines how the resulting
#' data table is sorted.  If the \code{columns} argument is left out, all data
#' columns are included.
#' @export
oppla_ave <- function(data, columns = setdiff(names(data), c(avc, "depth", "height")), avc, fun = mean) {
  avc[avc == "mon"] <- "month"
  if ("year" %in% avc) {
    columns <- union(columns, "date")
  } else {
    columns <- setdiff(columns, "date")
    if (length(avc) > 1)
      stop("avc must be scalar or contain year.")
  }
  data[j = c(avc) := oppla_split_date(date = date, columns = avc)]
  if ("depth" %in% names(data)) avc <- c(avc, "depth", "height")
  oma <- data[j = lapply(.SD, FUN = fun), by = avc, .SDcols = columns]
  if ("year" %in% avc) {
    ud <- unique(oma$date)
    if (length(ud) > 1) oma[j = dt := min(diff(as.numeric(ud)))] else oma[j = dt := 1]
  } else {
    ut <- sort(unique(oma[[avc[1]]]))
    if (length(ut) > 1) oma[j = dt := max(diff(as.numeric(ut)))] else oma[j = dt := 1]
    data.table::setattr(oma$dt, "class", NULL) # remove units from dt
    data.table::setattr(oma$dt, "units", NULL)
  }
  data.table::setattr(oma, name = "depth_edges", value = attr(data, "depth_edges"))
  data.table::setattr(oma, name = "file", value = attr(data, "file"))
  return(oma)
}

#' Split a date vector into year, month, etc.
#'
#' @param date vetor of class [`Date`][Dates] or one of the [`Date-Time Classes`][DateTimeClasses]
#' @param columns character vector (or list) of date components; must be names of the
#' single-argument functions (year, month, etc.) listed under [`IDateTime`][data.table::IDateTime]
#' @return named list of integer-vectors for the requested date components
#' @export
oppla_split_date <- function(date, columns = c("year", "month", "mday")) {
  columns[columns == "mon"] <- "month"
  return(sapply(X = unlist(columns), FUN = function(x) do.call(what = x, args = list(date)), simplify = FALSE))
}

#' Monthly climatology.
#'
#' Monthly climatology of (temporally averaged) oppla results.
#'
#' @param data oppla results from [`oppla_read`]
#' @param avc columns used for time averaging
#' @param ... other options passed to [`oppla_ave`]
#' @export
oppla_cmt <- function(data, avc = "month", ...) {
  return(oppla_ave(data = data, avc = avc, ...))
}

#' Vertical integrals.
#'
#' Vertical integrals of (temporally averaged) oppla results.
#'
#' @param data oppla data as obtained with [`oppla_read`]
#' @param avc character vector of averaging columns in data.  \code{avc = "month"}
#'   produces a monthly climatology.  \code{avc = c("mday", "month", "year")} gives daily
#'   averages
#' @param columns character vector of columns in data to be integrated
#' @param depth units object specifying the depth for integration
#' @param ... parameters passed to [`oppla_ave`]
#' @export
oppla_int <- function(data, avc, columns = setdiff(names(data), c("depth", "height")),
                      depth = units::set_units(100, "m", mode = "standard"), ...) {
  hts <- diff(attr(data, "depth_edges")) # layer heights
  dps <- attr(data, "depth_edges")[2, ]
  if (missing("avc")) {
    avc <- "date"
    data <- data[j = .SD, .SDcols = intersect(c(columns, "depth", "height", "date", "dt"), names(data))]
  } else {
    data <- oppla_ave(data = data, avc = avc, columns = columns, ...)
  }
  oi <- unique(data[j = .SD, .SDcols = c(avc, "dt", if ("year" %in% avc) "date")])
  if ("year" %in% avc) oi[j = date := round(x = data.table::as.IDate(date), digits = "months") + 14]
  units(depth) <- "m"            # ensure that depth has units
  depth <- max(pmin(depth, dps)) # limit depth to model domain
  nh <- which.max(dps >= depth)
  ht <- hts[1:nh]
  ht[nh] <- ht[nh] - (dps[nh] - depth) # correct height of deepest layer
  uht <- units::as_units(units(ht))
  for (column in setdiff(names(data), c("depth", "height", "date", "dt", avc)))
    oi[j = c(column) := units::set_units(as.vector(t(matrix(data[[column]], nrow = length(hts))[1:nh, ]) %*% ht),
                                         units(data[[column]])) * uht]
  data.table::setattr(oi, name = "depth", value = depth)
  return(oi)
}

#' Date-depth slab from oppla results.
#'
#' Extract a date-depth slab from an oppla results data table.
#'
#' @param data oppla data as obtained with [`oppla_read`]
#' @param daterange length-two vector of class [`POSIXt`]
#' @param depthrange length-two vector of class [`units`][units::units]
#' or numeric in which case units of meters are assumed
#' @export
oppla_slab <- function(data, daterange = range(data$date), depthrange = range(attr(data, "depth_edges"))) {
  units(depthrange) <- "m" # ensure that depthrange has units
  os <- data[(date %between% daterange) & (depth %between% depthrange)]
  dros <- range(os$depth)
  de <- attr(data, "depth_edges")
  deos <- de[, (de[2, ] >= dros[1]) & (de[1, ] <= dros[2])] # limit depth edges to depthrange
  anms <- setdiff(names(attributes(data)), c("names", "row.names", "class", "depth_edges"))
  for (anm in anms) data.table::setattr(os, name = anm, value = attr(data, anm)) # copy attributes
  data.table::setattr(os, name = "depth_edges", value = deos)
  return(data.table::copy(os))
}

#' @inherit oppla_int
#' @export
oppla_icmt <- oppla_int

#' Monthly averages.
#'
#' Calculate monthly averages of \code{oppla} results.
#'
#' @param data oppla results from [`oppla_read`]
#' @param avc columns used for time averaging
#' @param ... other options passed to [`oppla_ave`]
#' @export
oppla_monthly <- function(data, avc = c("month", "year"), ...) {
  return(oppla_ave(data = data, avc = avc, ...))
}

#' Daily averages.
#'
#' Calculate daily averages of \code{oppla} results.
#'
#' @param data oppla results from [`oppla_read`]
#' @param avc columns used for time averaging
#' @param ... other options passed to [`oppla_ave`]
#' @export
oppla_daily <- function(data, avc = c("mday", "month", "year"), ...) {
  return(oppla_ave(data = data, avc = avc, ...))
}

#' Potential density.
#'
#' Potential density in kg/m^3 (Millero et al. 1980).
#'
#' @return Potential density as a [`units`][units::units] object with units set to kg/m^3.
#' @param S  salinity in psu
#' @param TC temperature in °C
#' @export
oppla_density <- function(TC, S) {
  if ("units" %in% class(TC)) TC <- units::drop_units(units::set_units(TC, "Celsius"))
  if ("units" %in% class(S)) S <- units::drop_units(units::set_units(S, "psu"))
  return(units::set_units(
                  999.842594 + TC * (6.793952e-2 + TC * (-9.09529e-3 + TC * (1.001685e-4 + TC * (-1.120083e-6 + TC * 6.536332e-9)))) +
                  S * ((8.24493e-1 + TC * (-4.0899e-3 + TC * (7.6438e-5 + TC * (-8.2467e-7 + TC * 5.3875e-9)))) +
                       sqrt(S) * (-5.72466e-3 + TC * (1.0227e-4 + TC * (-1.6546e-6))) + S * 4.8314e-4),
                  "kg m-3"))
}
