#' @useDynLib oppla
NULL

#' Add columns with DIC components to \code{oppla} output.
#'
#' DIC components \code{CO2}, \code{HCO3}, \code{CO3} and \code{pH} are calculated
#' and added to \code{oppla}, together with density (\code{rho}), aragonite
#' saturation (\code{Omega}) and calcite saturation (\code{OmegaCalcite}).
#'
#' @return [`Data table`][data.table::data.table] \code{oppla} with columns \code{CO2},
#' \code{HCO3}, \code{CO3}, \code{pH}, \code{rho}, \code{Omega} and \code{OmegaCalcite} added.
#' @param oppla oppla output as obtained with [`oppla_read`]
#' @details This function calls the Fortran subroutine \code{dic:dicsys:co2},
#' which is also called by the \code{oppla} program.  It will work only if
#' the forcing data (temperature, salinity) have been read into \code{oppla}.
#' @export
oppla_dic <- function(oppla) {
  if (! "alkalinity" %in% names(oppla)) {
    ## calculate alkalinity if missing in oppla
    pA <- oppla::nml_read(file = attr(oppla, "control_file"))$envi[[1]]$potalk
    oppla[j = alkalinity := pA * units::drop_units(Salinity) / 35 - DIN]
  }
  n0 <- 0L
  nr <- 1L
  res <- list()
  pfn <- attr(oppla, "parameter_file")
  lpfn <- nchar(pfn)
  repeat {
    n1 <- min(nrow(oppla), n0 + 10000L)
    oppla_subset <- oppla[(n0 + 1):n1]
    if ("SiO3" %in% names(oppla_subset)) SiO3 <- oppla_subset$SiO3 else SiO3 <- oppla_subset$DIN
    ndat <- nrow(oppla_subset)
    rho <- rep_len(0, ndat)
    CO2 <- rep_len(0, ndat)
    HCO3 <- rep_len(0, ndat)
    CO3 <- rep_len(0, ndat)
    hc <- rep_len(0, ndat)
    Omega <- rep_len(0, ndat)
    OmegaCalcite <- rep_len(0, ndat)
    r <- suppressWarnings(.Fortran("rdic", pfn = pfn, lpfn = lpfn,
                                   depth = oppla_subset$depth,
                                   temp = oppla_subset$Temperature, salt = oppla_subset$Salinity,
                                   alk  = oppla_subset$alkalinity,  tco2 = oppla_subset$DIC,
                                   dip  = oppla_subset$DIP,         silc = SiO3,
                                   ndat = ndat,
                                   rho  = rho, CO2 = CO2, HCO3 = HCO3, CO3 = CO3, hc = hc,
                                   Omega = Omega, OmegaCalcite = OmegaCalcite))
    res[[nr]] <- data.table::as.data.table(r[c("rho", "CO2", "HCO3", "CO3", "Omega", "OmegaCalcite")])[j = pH := -log10(r$hc)]
    if (n1 == nrow(oppla)) break
    nr <- nr + 1L
    n0 <- n1
  }
  rdt <- data.table::rbindlist(res)
  units <- c(rho = "g m-3", CO2 = "mol kg-1", HCO3 = "mol kg-1", CO3 = "mol kg-1")
  for (name in names(units)) rdt[j = (name) := units::set_units(get(name), units[[name]])]
  rdt[j = "rho" := units::set_units(rho, "kg m-3")]
  return(cbind(oppla, rdt))
}
