#' Generate a set of parameter files with Latin Hypercube Sampling.
#'
#' @param default name of default parameter file.
#' @param npfs number of parameter files to generate.
#' @param ranges data.frame with one row for each parameter and three columns named low, high and name.
#' The row names specify the namellist entries and the three columns specify the ranges and names of the parameters.
#' @param format template passed as \code{fmt} to [`sprintf`][base::sprintf] to generate the names
#' of the parameter-files to be written; must contain exactly one conversion specification.
#' @param LHSfun one of the Latin Hypercube Sample functions in the [`lhs`][lhs::lhs] package.
#' @param seed initial random-number seed, useful for reproducible results.
#' @param ... arguments passed on to \code{LHSfun}.
#' @return matrix with named columns.  Each row is one parameter set.
#' @import lhs
#' @export
opplaLHS <- function(default, npfs, ranges, format = "oppla_param_%04d.inp", LHSfun = lhs::maximinLHS, seed, ...) {
  if (! missing(seed)) set.seed(seed)
  lhs <- LHSfun(n = npfs, k = nrow(ranges), ...)
  values <- matrix(data = rep(ranges$low, each = npfs), nrow = npfs) +
    lhs * matrix(data = rep(ranges$high - ranges$low, each = npfs), nrow = npfs)
  icfos <- na.omit(strcapture(pattern = "cfo-([0-9]+)-phi-0", x = rownames(ranges), proto = data.frame(cfo = numeric()))$cfo)
  for (icfo in icfos) {
    ## treat phi-0 as prey-capture coefficient and phi-1, phi-2, etc. as food preferences
    iphi0 <- grep(pattern = paste("cfo", icfo, "phi", "0", sep = "-"), x = rownames(ranges))
    iphis <- grep(pattern = paste("cfo", icfo, "phi", "[1-9][0-9]*", sep = "-"), x = rownames(ranges))
    values[, iphi0] <- values[, iphi0] / rowSums(values[, iphis]) # ensure that food preferences add up to 1
    for (iphi in iphis) values[, iphi] <- values[, iphi] * values[, iphi0]
    values <- values[, -iphi0]
    ranges <- ranges[-iphi0, ]
  }
  defaultNL <- oppla::nml_read(file = default)
  vars <- rownames(ranges)
  ## parameters which may be applied to all zooplankton types simultaneously
  for (all_zoo in c("Emax", "mort")) {
    if (all_zoo %in% rownames(ranges)) {
      ## apply Emax to all cfo groups
      iaz <- which(rownames(ranges) == all_zoo)
      vars <- c(vars[-iaz], paste("cfo", seq_along(icfos), all_zoo, sep = "-"))
      values <- cbind(values[, -iaz], values[, rep(iaz, length(icfos))])
      ranges <- rbind(ranges[-iaz, ], ranges[rep(iaz, length(icfos)), ])
    }
  }
  pfns <- seq_len(npfs)
  files <- sprintf(fmt = format, pfns)
  for (npf in pfns) {
    pnl <- oppla::nml_set(nml = defaultNL, vars = vars, values = values[npf, ])
    oppla::nml_write(nml = pnl, file = files[npf])
  }
  colnames(values) <- ranges$name
  return(values)
}

#' Generate a set of parameter files for a sensitivity analysis
#'
#' @param default name of default parameter file.
#' @param parameters (named) vector of parameter specifications of the form
#' code{name = "namelist name-index-variable name[-indices]"}.  If the vector is named, the names
#' are used for the generated file names, otherwise the parameter specifications will be used as names.
#' @param delta fraction by with each parameter is varied.
#' @param folder where to save the generated parameter files.
#' @param base_name base name to construct the parameter file names from
#' @details
#' The names of the generated parameter files are constructed from the \code{base_name}
#' and the names of the \code{parameters}.  A \code{+} or \code{-} before the extension \code{.nml}
#' indicates whether the parameter was increased or decreased relative to its default.
#' @export
opplaSensitivity <- function(default, parameters, delta = 0.1, folder = ".", base_name = "oppla_param") {
  defaultNL <- oppla::nml_read(file = default)
  tags <- names(parameters)
  if (is.null(tags)) tags <- parameters
  else tags <- ifelse(test = nzchar(tags), yes = tags, no = parameters)
  if (! dir.exists(folder)) dir.create(path = folder)
  for (np in seq_along(parameters)) {
    value <- oppla::nml_get(nml = defaultNL, vars = parameters[np])[[1]]
    nl1 <- oppla::nml_set(nml = defaultNL, vars = parameters[np], values = value * (1 - delta))
    nl2 <- oppla::nml_set(nml = defaultNL, vars = parameters[np], values = value * (1 + delta))
    oppla::nml_write(nml = nl1, file = file.path(folder, paste0(base_name, "_", tags[[np]], "-", ".nml")))
    oppla::nml_write(nml = nl2, file = file.path(folder, paste0(base_name, "_", tags[[np]], "+", ".nml")))
  }
}
