#' Aggregate data onto a regular time-depth grid.
#'
#' Spatial and/or temporal averages of \code{oppla} results.
#'
#' @return [`Data table`][data.table::data.table] of spatially and temporally averaged observations.
#'
#' @param data     bottle data from \code{read_ls} or \code{read_bats} or one of the .rds files
#' @param edges    number of layers or vector or matrix of edges
#' @param datecols date columns (subset of names of the single-argument functions (year, month, etc.)
#'                 listed under [`IDateTime`][data.table::IDateTime]) or \code{"date"}
#' @param layers   number of layers or vector or matrix of edges
#' @param columns  names of columns to include
#' @param drop_na  whether to drop empty rows
#' @param fun      function to apply to subsets
#' @param ...      optional arguments to \code{fun}
#'
#' @details
#' In order to obtain layers matching oppla model output, use
#' ```
#' oa <- op_aggregate(data, edges = attr(model, "depth_edges"), ...)
#' ```
#'
#' If \code{datecols} does not contain "year", data are first averaged within years and then
#' \code{fun} is applied across years, in order to obtain better climatological values.
#' If \code{datecols} is \code{"date"}, aggregate only by depth level
#' @export
op_aggregate <- function(data, edges = layers, datecols = c("year", "month"), layers = 20, drop_na = TRUE,
                         columns = names(data)[sapply(X = data, FUN = is.numeric)], fun = mean, ...) {
  if (length(dim(edges)) == 2) edges <- c(edges[1, 1], edges[2, ])
  if (length(edges) == 0) stop("Argument egdes must not be empty!")
  if (length(edges) == 1) {
    edges <- seq(from = 0, to = max(data$depth), length.out = edges + 1)
    height <- diff(edges)
  } else {
    height <- diff(edges)
  }
  levels <- height / 2 + head(edges, n = -1)
  datecols_year <- union("year", setdiff(datecols, "date"))
  ad <- data.table::as.data.table(data)[j = c(datecols_year, "level") :=
                                          c(oppla::oppla_split_date(date = date, columns = datecols_year),
                                            list(findInterval(x = depth, vec = edges)))][level < length(edges)]
  ## aggregate by date and level
  if (! any(c("year", "date") %in% datecols)) # average first within years, then apply fun across years
    ad <- ad[j = lapply(.SD, FUN = mean, na.rm = TRUE), by = c(datecols_year, "level"),
             .SDcols = c(columns, "date")]
  ad <- ad[j = lapply(.SD, FUN = fun, na.rm = TRUE, ...), by = c(datecols, "level"),
           .SDcols = c(columns, if ("year" %in% datecols) "date")]
  ad[j = c("depth", "height") := .(units::set_units(levels[level], "m"),
                                   units::set_units(height[level], "m"))] # nominal depths and heights
  ## remove empty rows
  if (drop_na) ad <- ad[j = .SD[apply(X = !is.na(ad[j = columns, with = FALSE]), MARGIN = 1, FUN = any)]]
  if ("year" %in% datecols) {
    if (any(c("wday", "mday", "yday") %in% datecols))
      ad[j = c("date", "dt") := list(as.POSIXct(data.table::as.IDate(date) + 0.5), units::set_units(1, "day"))]
    else if ("week" %in% datecols)
      ad[j = c("date", "dt") := list(as.POSIXct(round(x = data.table::as.IDate(date), digits = "weeks") + 2.5),
                                     units::set_units(1, "week"))]
    else if ("month" %in% datecols)
      ad[j = c("date", "dt") := list(as.POSIXct(round(x = data.table::as.IDate(date), digits = "months") + 14),
                                     units::set_units(1, "month"))]
    else if ("quarter" %in% datecols)
      ad[j = c("date", "dt") := list(as.POSIXct(round(x = data.table::as.IDate(date), digits = "quarters") + 45),
                                     units::set_units(3, "months"))]
    else
      ad[j = c("date", "dt") := list(as.POSIXct(round(x = data.table::as.IDate(date), digits = "years") + 181.5),
                                     units::set_units(1, "year"))]
  } else {
    if (length(datecols) > 1) stop("datecols must be scalar or contain year.")
    else ad[j = dt := 1]
  }
  data.table::setattr(ad, name = "file", value = attr(data, "file"))
  return(ad)
}

#' Convert mass-specific to volume-specific units.
#'
#' Many data are provides in mass-specific units whereas \code{oppla} results are in volume-specific
#' units.  This function converts columns given in mass-specific units to make them comparable with
#' \code{oppla} results.
#'
#' @return Modified [`Data table`][data.table::data.table].  All columns containing kg-1 in their
#' units string are multiplied with the corresponding potential density, which is taken from the
#' \code{Density} column, if it exists, or calculated from the \code{Temperature} and \code{Salinity}
#' columns.  Missing density data are set to 1024 kg/m^3.
#' @param data dataframe or data table as obtained with \code{read_ls} or \code{read_bats} or one of the .rds files
#' @export
op_volume_units <- function(data) {
  data <- data.table::as.data.table(data)
  rho <- if ("Density" %in% names(data)) data$Density else oppla::oppla_density(TC = data$Temperature, S = data$Salinity)
  rho[is.na(rho)] <- units::set_units(1024, "kg m-3")
  for (name in names(data))
    if ("units" %in% class(data[[name]])) {
      u <- units::deparse_unit(data[[name]])
      if (grepl(pattern = "kg-1", x = u, fixed = TRUE))
        data[j = c(name) := .SD[[1]] * rho, .SDcols = name]
    }
  return(data)
}
