#' Read surface data for a specific location from NCEP NetCDF files.
#'
#' NCEP NetCDF files contain global datasets for specific dates or years.  This
#' function extracts the data for a specific location.
#'
#' @return Dataframe with a time (class [`POSIXt`][base::DateTimeClasses])
#' and a data column (class [`units`][units::units]) whose name is \code{varid}
#' with "_NCEP" appended.
#' The data are those closest to the specified position.
#'
#' @param path path name of the folder where to find the NCEP NetCDF files
#' @param varid name of the variable to be retrieved from the NetCDF files
#' @param lon,lat longitude and latitude of the location
#' @details
#' The \code{path} must not contain any other NetCDF files.
#' @export
oppla_read_ncep <- function(path, varid, lon, lat) {
  lon[lon < 0] <- lon[lon < 0] + 360 # ensure 0 <= lon <= 360
  vdata <- list()
  dates <- list()
  files <- dir(path = path, pattern = paste0("^", varid, "\\..+\\.nc$"))
  nc <- oppla::onc_info(filename = file.path(path, files[1]))
  ## find the grid point closest to lon/lat
  lola <- expand.grid(lon = nc$dim$lon$vals, lat = nc$dim$lat$vals)
  ill <- which.min((lola$lon - lon) ^ 2 + (lola$lat - lat) ^ 2)
  ilon <- which(nc$dim$lon$vals == lola$lon[ill])
  ilat <- which(nc$dim$lat$vals == lola$lat[ill])
  ndims <- length(nc$var[[varid]]$dimids) # can be 3 (dswrf) or 4 (wind)
  RNetCDF::close.nc(nc$self)
  start <- c(ilon, ilat, 1, 1)[1:ndims]
  count <- c(rep(1, ndims - 1), NA) # last dimension is time
  for (nf in seq_along(files)) {
    nc <- oppla::onc_info(filename = file.path(path, files[nf]))
    ## extract units and origin of time axis
    times <- strcapture(pattern = "(\\w+) *since *([0-9 :.-]+)", x = nc$var$time$units,
                        proto = data.frame(units = character(), origin = character()))
    dates[[nf]] <- as.POSIXct(times$origin, format = "%Y-%m-%d %H:%M:%S", tz = "UTC") +
      as.difftime(tim = nc$dim$time$vals, units = times$units)
    vdata[[nf]] <- units::set_units(RNetCDF::var.get.nc(ncfile = nc$self, variable = varid, start = start, count = count),
                                    nc$var[[varid]]$units, mode = "standard")
    RNetCDF::close.nc(nc$self)
  }
  data <- data.frame(time = do.call(what = c, args = dates))
  data[[paste(varid, "NCEP", sep = "_")]] <- do.call(what = c, args = vdata)
  if (varid == "icec") {
    data$icec_NCEP <- data$icec_NCEP * 100 # units are percent in the attribute but numbers are fractions
    ## NA is 0 in these files
    data$icec_NCEP[is.na(data$icec_NCEP)] <- units::set_units(0, nc$var[[varid]]$units, mode = "standard")
  }
  for (att in names(nc$att)) attr(data, att) <- nc$att[[att]]
  return(structure(data[complete.cases(data),], lon = lon, lat = lat))
}

#' Convert forcing file to a new depth structure.
#'
#' This can be used to compare the effects of different vertical resolutions separately
#' from thos of the actual forcing data.
#'
#' @param old name of original forcing file
#' @param depth new depth structure (vector of mean depths)
#' @param new name of new forcing file
#' @export
oppla_convert_forcing <- function(old, depth, new) {
  nc_type <- function(x)
    switch(class(x),
           integer = "NC_INT",
           character = "NC_CHAR")
  nco <- oppla::onc_info(filename = old)
  ncn <- RNetCDF::create.nc(filename = new, format = "netcdf4")
  dims <- nco$dim
  dims$depth$vals <- depth
  for (dim in dims) RNetCDF::dim.def.nc(ncfile = ncn, dimname = dim$name, dimlength = dim$length, unlim = dim$unlim)
  vals_new <- list()
  de <- Reduce(f = function(x, y) 2 * y - x, x = depth, init = 0, accumulate = TRUE)
  for (var in nco$var) {
    RNetCDF::var.def.nc(ncfile = ncn, varname = var$name, vartype = var$type, dimensions = names(var$dimensions))
    for (name in setdiff(names(var), c("id", "name", "type", "ndims", "dimids", "natts", "dimensions")))
      RNetCDF::att.put.nc(ncfile = ncn, variable = var$name, name = name, type = nc_type(var[[name]]), value = var[[name]])
    vals_new[[var$name]] <-
      if (nco$dim$depth$id %in% var$dimids)
        if (var$name %in% c("depth_bounds", "depth_edges"))
          matrix(c(head(de, n = -1), de[-1]), ncol = 2)
        else
          apply(X = RNetCDF::var.get.nc(nco$self, variable = var$name, collapse = FALSE),
                MARGIN = setdiff(var$dimids, nco$dim$depth$id) + 1,
                FUN = function(x) approx(x = nco$dim$depth$vals, xout = depth, y = x, rule = 2)$y)
      else
        RNetCDF::var.get.nc(nco$self, variable = var$name)
  }
  RNetCDF::close.nc(nco$self)
  for (varid in names(vals_new)) RNetCDF::var.put.nc(ncn, variable = varid, data = vals_new[[varid]])
  RNetCDF::att.put.nc(ncfile = ncn, variable = "NC_GLOBAL", name = "description", type = "NC_CHAR",
                      value = paste0("Oppla forcing file generated by oppla_convert_forcing on ", date(),
                                     " from ", old, "."))
  RNetCDF::close.nc(ncn)
}

#' Create oppla forcing file from FOCI output.
#'
#' Detect forcing data for \code{oppla} in a FOCI output file and create an \code{oppla} forcing file.
#'
#' @param foci_file FOCI output file (NetCDF) with datasets for vertical eddy diffusivity,
#' temperature, salinity, surface wind speed, etc. (see Details).
#' Each dataset is expected to contain 9 vertical columns corresponding to a \eqn{3 \times 3}
#' subset of the horizontal FOCI grid.
#' @param ilon,ilat longitude and latitude indices to include in the horizontal averages
#' @param filename  name of the oppla forcing file to be created
#' @param vdc_cap   upper limit of vertical diffusivity (\eqn{\mathrm{m^2 s^{-1}}})
#' @param surface   named list of paths to NCEP dataset collections;
#' The names are passed as \code{varid}s to [`oppla_read_ncep`]
#' @param max_depth if provided, exclude only layers below this depth (m), otherwise include only
#' layers above the sea floor at all positions in \code{foci_file}
#' @details
#' \code{oppla_forcing_foci} detects forcing data according to this scheme:
#' \tabular{ll}{
#' **FOCI** \tab **oppla** \cr
#' votkeavt \tab D (vertical eddy diffusivity) \cr
#' votemper \tab Temperature  \cr
#' vosaline \tab Salinity     \cr
#' sowindsp \tab wind         \cr
#' ileadfra \tab ice_fraction
#' }
#' Additional time-series (1D) datasets can be added with [`oppla_forcing_add_ts`].  Note that the
#' first level of vertical diffusivities is for the sea surface in FOCI, whereas OPPLA requires them
#' at the bottom of each layer.  Thus, \code{oppla_forcing_foci} omits the surface diffusivities.
#' @export
oppla_forcing_foci <- function(foci_file, ilon = 1:3, ilat = 1:3, filename, vdc_cap = Inf, surface = list(), max_depth) {
  ncf <- oppla::onc_info(filename = foci_file)
  lon <- mean(RNetCDF::var.get.nc(ncfile = ncf$self, variable = "nav_lon")[ilon, ilat])
  lat <- mean(RNetCDF::var.get.nc(ncfile = ncf$self, variable = "nav_lat")[ilon, ilat])
  time_origin <- sub(pattern = "seconds since *", x = ncf$var$time_counter$units, replacement = "")
  dates <- as.POSIXct(ncf$dim$time_counter$vals, origin = time_origin, tz = "UTC")
  vdc <- RNetCDF::var.get.nc(ncfile = ncf$self, variable = "votkeavt",  # vertical eddy diffusivity
                             start = c(1, 1, 2, 1), count = c(NA, NA, NA, NA))
  ## find the index of the bottom-most positive diffusivity, add 1 to find the bottom
  if (missing(max_depth))
    nbox <- max(which(apply(X = vdc[ilon, ilat, , 1, drop = FALSE], MARGIN = 3, FUN = prod) > 0)) + 1
  else
    nbox <- which.max(ncf$dim$deptht$vals > max_depth) - 1
  ## create a mask for invalid data (below bottom), considering that vdc data are shifted upwards by
  ## one level because it is provided on a different grid (and surface diffusivity is ignored)
  invalid <- which(vdc[, , c(1, 1:(nbox - 1)), ] == 0)
  setNA <- function(x, invalid = invalid) {
    x[invalid] <- NA
    return(x)
  }
  count_in <- c(NA, NA, nbox, NA)
  count_out <- c(1, 1, nbox, dim(vdc)[4])
  nc_out <- RNetCDF::create.nc(filename = filename, format = "netcdf4")
  dims <- list(longitude = list(units = ncf$var$nav_lon$units, axis = "X", vals = lon, unlim = FALSE),
               latitude = list(units = ncf$var$nav_lat$units, axis = "Y", vals = lat, unlim = FALSE),
               depth = list(units = ncf$var$deptht$units, axis = "Z", positive = "down", bounds = "depth_bounds",
                            vals = ncf$dim$deptht$vals[1:nbox], unlim = FALSE),
               time = list(units = "seconds since 1970-01-01 00:00:00", axis = "T", vals = as.numeric(dates), unlim = TRUE))
  dimnames <- onc_make_axes(ncfile = nc_out, dims = dims)
  RNetCDF::dim.def.nc(ncfile = nc_out, dimname = "bounds", dimlength = 2)
  datasets <- list(D = list(foci_name = "votkeavt", units = ncf$var$votkeavt$units,
                            foci_start = c(1, 1, 2, 1), # omit surface diffusivity
                            fun = function(x) pmin(setNA(x, invalid = invalid), vdc_cap)),
                   Temperature = list(foci_name = "votemper", units = ncf$var$votemper$units,
                                      foci_start = c(1, 1, 1, 1), fun = function(x) setNA(x, invalid = invalid)),
                   Salinity = list(foci_name = "vosaline", units = "psu", foci_start = c(1, 1, 1, 1),
                                   fun = function(x) setNA(x, invalid = invalid)))
  for (name in names(datasets)) {
    RNetCDF::var.def.nc(ncfile = nc_out, varname = name, vartype = "NC_DOUBLE", dimensions = dimnames)
    RNetCDF::att.put.nc(ncfile = nc_out, variable = name, name = "units", type = "NC_CHAR", value = datasets[[name]]$units)
    data <- colMeans(datasets[[name]]$fun(RNetCDF::var.get.nc(ncfile = ncf$self, variable = datasets[[name]]$foci_name,
                                                              count = count_in, start = datasets[[name]]$foci_start)[
                                                     ilon, ilat, , , drop = FALSE]), dims = 2, na.rm = TRUE)
    RNetCDF::var.put.nc(ncfile = nc_out, variable = name, data = data, count = count_out)
  }
  RNetCDF::var.def.nc(ncfile = nc_out, varname = "depth_bounds", vartype = "NC_DOUBLE", dimensions = c("bounds", "depth"))
  RNetCDF::att.put.nc(ncfile = nc_out, variable = "depth_bounds", name = "units", type = "NC_CHAR", value = dims$depth$units)
  RNetCDF::var.put.nc(ncfile = nc_out, variable = "depth_bounds", count = c(NA, nbox),
                      data = RNetCDF::var.get.nc(ncfile = ncf$self, variable = "deptht_bnds", count = c(NA, nbox)))
  surfoci <- c(sowindsp = "wind", ileadfra = "ice_fraction")
  for (surname in intersect(names(surfoci), names(ncf$var))) {
    RNetCDF::var.def.nc(ncfile = nc_out, varname = surfoci[[surname]], vartype = "NC_DOUBLE", dimensions = c("time"))
    RNetCDF::att.put.nc(ncfile = nc_out, variable = surfoci[[surname]], name = "units", type = "NC_CHAR",
                        value = ncf$var[[surname]]$units)
    RNetCDF::var.put.nc(ncfile = nc_out, variable = surfoci[[surname]],
                        data = colMeans(RNetCDF::var.get.nc(ncfile = ncf$self, variable = surname), dims = 2))
  }
  RNetCDF::close.nc(ncf$self)
  for (nsf in names(surface)) {
    sfd <- oppla_read_ncep(path = surface[[nsf]], varid = nsf, lon = lon, lat = lat)
    varid <- setdiff(names(sfd), "time")
    RNetCDF::var.def.nc(ncfile = nc_out, varname = varid, vartype = "NC_DOUBLE", dimensions = c("time"))
    RNetCDF::att.put.nc(ncfile = nc_out, variable = varid, name = "units", type = "NC_CHAR",
                        value = as.character(units(sfd[[varid]])))
    RNetCDF::var.put.nc(ncfile = nc_out, variable = varid,
                        data = approx(x = sfd$time, y = sfd[[varid]], xout = dates, rule = 2)$y)
  }
  RNetCDF::att.put.nc(ncfile = nc_out, variable = "NC_GLOBAL", name = "description", type = "NC_CHAR",
                      value = paste0("Oppla forcing file generated by oppla_forcing_foci on ", date(),
                                     " from ", foci_file, "."))
  RNetCDF::close.nc(nc_out)
}

#' Add 1D datasets to \code{oppla} forcing file.
#'
#' Add 1D (time-series) datasets to oppla forcing file.  These can be surface or bottom datasets
#' as specified by their associations in namelist physics.
#'
#' @param forcing_file name of the forcing file
#' @param new_data dataframe with a \code{time} column (class [`POSIXt`][base::DateTimeClasses])
#' and further columns containing the time-series data (as [`units`][units::units] objects) to be added;
#' the new \code{varid}s will be the names of the columns
#' @details
#' Further information about the added datasets can be passed as attributes of the columns of
#' \code{new_data}, which will be copied into the \code{forcing_file}.
#' @export
oppla_forcing_add_ts <- function(forcing_file, new_data) {
  ncf <- oppla::onc_info(filename = forcing_file, write = TRUE)
  time_origin <- sub(pattern = "seconds since *", x = ncf$var$time$units, replacement = "")
  dates <- as.POSIXct(ncf$dim$time$vals, origin = time_origin, tz = "UTC")
  varids <- setdiff(names(new_data), "time")
  for (varid in varids) {
    RNetCDF::var.def.nc(ncfile = ncf$self, varname = varid, vartype = "NC_DOUBLE", dimensions = "time")
    for (att in setdiff(names(attributes(new_data[[varid]])), c("class", "dim")))
      RNetCDF::att.put.nc(ncfile = ncf$self, variable = varid, name = att, type = "NC_CHAR",
                          value = as.character(attr(new_data[[varid]], att)))
    RNetCDF::var.put.nc(ncfile = ncf$self, variable = varid,
                        data = approx(x = new_data$time, y = new_data[[varid]], xout = dates, rule = 2)$y)
  }
  RNetCDF::close.nc(ncf$self)
}

#' Create oppla forcing file from Heiner's mom model.
#'
#' @param filename   name of the forcing file to be created
#' @param region     region identifyer string, one of \code{"aloh"}, \code{"bats"},
#'            \code{"esto"}, \code{"irmi"}, \code{"k2pa"}, \code{"labr"}, \code{"ligu"}, \code{"s1pa"}
#' @param path       path to the folder where the model output is stored
#' @param years      vector of years for which to create the forcing file, must be a
#'            contiguous subset of \code{1949:2009}
#' @param ndep       maximum number of levels to include
#' @param ilon,ilat  longitude, latitude indices in the mom output files, which contain 3x3 columns
#' @export
oppla_forcing_mom <- function(filename, region, path = "~/data/mom_sis_ciaf_repeat", years = 1949:2009,
                              ndep, ilon = 2, ilat = 2) {
  count_r <- c(NA, NA, NA, 1)
  base <- file.path(path, paste0("rregionocean_phys_", region, "_"))
  nc0 <- oppla::onc_info(filename = paste0(base, years[1], ".nc"))
  vdc <- RNetCDF::var.get.nc(ncfile = nc0$self, variable = "diff_cbt_kpp_s",
                             count = count_r, collapse = FALSE)[ilon, ilat, , , drop = FALSE]
  RNetCDF::close.nc(nc0$self)
  getMeans <- prod(dim(vdc)[1:2]) > 1
  if (getMeans) vdc <- colMeans(vdc, dims = 2)
  nc_out <- RNetCDF::create.nc(filename = filename, format = "netcdf4")
  if (missing(ndep)) ndep <- length(which(!is.na(vdc))) # cut off NAs below sea floor
  ## extract units and origin of time axis (origin is useless here because of missing leap-years)
  times <- strcapture(pattern = "(\\w+) *since *([\\d :.-]+)", x = nc0$var$time$units,
                      proto = data.frame(units = character(), origin = character()))
  lon <- mean(nc0$dim$xt_ocean_sub1$vals[ilon])
  lat <- mean(nc0$dim$yt_ocean_sub1$vals[ilat])
  dims <- list(longitude = list(units = "degrees_E", axis = "X", vals = lon, unlim = FALSE),
               latitude = list(units = "degrees_N", axis = "Y", vals = lat, unlim = FALSE),
               depth = list(units = "m", axis = "Z", positive = "down", bounds = "depth_bounds",
                            vals = nc0$dim$st_ocean$vals[1:ndep], unlim = FALSE),
               time = list(units = "seconds since 1970-01-01 00:00:00", axis = "T", vals = 0, unlim = TRUE))
  dimnames <- onc_make_axes(ncfile = nc_out, dims = dims)
  RNetCDF::dim.def.nc(ncfile = nc_out, dimname = "bounds", dimlength = 2)
  vars <- list(D = list(name = "diff_cbt_kpp_s", units = nc0$var$diff_cbt_kpp_s$units, dim = dimnames),
               Temperature = list(name = "temp", units = "degrees_Celsius", dim = dimnames),
               wt = list(name = "wt", units = nc0$var$wt$units, dim = dimnames))
  for (varname in names(vars)) {
    RNetCDF::var.def.nc(ncfile = nc_out, varname = varname, vartype = "NC_DOUBLE", dimensions = vars[[varname]]$dim)
    RNetCDF::att.put.nc(ncfile = nc_out, variable = varname, name = "units", type = "NC_CHAR", value = vars[[varname]]$units)
  }
  RNetCDF::var.def.nc(ncfile = nc_out, varname = "depth_bounds", vartype = "NC_DOUBLE", dimensions = c("bounds", "depth"))
  RNetCDF::att.put.nc(ncfile = nc_out, variable = "depth_bounds", name = "units", type = "NC_CHAR", value = dims$depth$units)
  RNetCDF::var.put.nc(ncfile = nc_out, variable = "depth_bounds", count = c(NA, ndep),
                      data = t(matrix(nc0$dim$st_edges_ocean$vals[c(1:ndep, 2:(ndep + 1))], ncol = 2)))
  count_r[3:4] <- c(ndep, NA)
  start_w <- c(1, 1, 1, 1)
  count_w <- c(1, 1, NA, NA)
  for (year in years) {
    nc_in <- oppla::onc_info(filename = paste0(base, year, ".nc"))
    count_w[4] <- nc_in$dim$time$len   # No. of time slices in file
    h <- as.difftime(tim = nc_in$dim$time$vals, units = times$units)
    for (name in names(vars)) {
      data <- RNetCDF::var.get.nc(ncfile = nc_in$self, variable = vars[[name]]$name,
                                  count = count_r, collapse = FALSE)[ilon, ilat, , , drop = FALSE]
      if (getMeans) data <- colMeans(data, dims = 2)
      RNetCDF::var.put.nc(ncfile = nc_out, variable = name, start = start_w, count = count_w, data = data)
    }
    RNetCDF::close.nc(nc_in$self)
    dates <- as.POSIXct(paste0(year, "-01-01"), format = "%Y-%m-%d", tz = "UTC") + h - h[1]
    RNetCDF::var.put.nc(ncfile = nc_out, variable = "time", start = start_w[4], count = count_w[4], data = as.numeric(dates))
    start_w[4] <- start_w[4] + count_w[4]
  }
  RNetCDF::att.put.nc(ncfile = nc_out, variable = "NC_GLOBAL", name = "description", type = "NC_CHAR",
                      value = paste0("Oppla forcing file generated by oppla_forcing_mom on ", date(),
                                     " from files in ", base))
  RNetCDF::close.nc(nc_out)
  pardata <- oppla_read_ncep(path = "~/data/NCEP/dswrf", varid = "dswrf", lon = lon, lat = lat)
  oppla_forcing_add_ts(forcing_file = filename, new_data = pardata)
}

#' Create oppla forcing file from GOTM output.
#'
#' @param GOTMfile   name of GOTM output file
#' @param filename   name of the forcing file to be created
#' @export
oppla_forcing_gotm <- function(GOTMfile, filename) {
  gnc <- oppla::onc_info(filename = GOTMfile)
  depth <- rev(abs(gnc$dim$z$vals))
  dz <- rev(gnc$dim$z1$vals - gnc$dim$z$vals) # half-heights
  idep <- rev(seq_along(depth))
  dims <- list(longitude = list(units = "degrees_E", axis = "X", vals = gnc$dim$lon$vals, unlim = FALSE),
               latitude = list(units = "degrees_N", axis = "Y", vals = gnc$dim$lat$vals, unlim = FALSE),
               depth = list(units = "m", axis = "Z", positive = "down", bounds = "depth_bounds",
                            vals = depth, unlim = FALSE),
               time = list(units = gnc$var$time$units, axis = "T", vals = gnc$dim$time$vals, unlim = TRUE))
  nc_out <- RNetCDF::create.nc(filename = filename, format = "netcdf4")
  dimnames <- onc_make_axes(ncfile = nc_out, dims = dims)
  RNetCDF::dim.def.nc(ncfile = nc_out, dimname = "bounds", dimlength = 2)
  RNetCDF::var.def.nc(ncfile = nc_out, varname = "depth_bounds", vartype = "NC_DOUBLE", dimensions = c("bounds", "depth"))
  RNetCDF::att.put.nc(ncfile = nc_out, variable = "depth_bounds", name = "units", type = "NC_CHAR", value = dims$depth$units)
  RNetCDF::var.put.nc(ncfile = nc_out, variable = "depth_bounds", data = rbind(depth - dz, depth + dz))
  vars <- list(D = list(name = "nus", units = gnc$var$nus$units, dim = dimnames),
               Temperature = list(name = "temp", units = "degrees_Celsius", dim = dimnames),
               Salinity = list(name = "salt", units = "psu", dim = dimnames),
               SWR = list(name = "swr", units = gnc$var$swr$units, dim = dimnames[c(1, 2, 4)]))
  for (varname in names(vars)) {
    RNetCDF::var.def.nc(ncfile = nc_out, varname = varname, vartype = "NC_DOUBLE", dimensions = vars[[varname]]$dim)
    RNetCDF::att.put.nc(ncfile = nc_out, variable = varname, name = "units", type = "NC_CHAR", value = vars[[varname]]$units)
    data <- RNetCDF::var.get.nc(ncfile = gnc$self, variable = vars[[varname]]$name, collapse = FALSE)
    if (length(dim(data)) == 4) data <- data[, , idep, , drop = FALSE]
    RNetCDF::var.put.nc(ncfile = nc_out, variable = varname, data = data)
  }
  RNetCDF::close.nc(gnc$self)
  RNetCDF::att.put.nc(ncfile = nc_out, variable = "NC_GLOBAL", name = "description", type = "NC_CHAR",
                      value = paste0("Oppla forcing file generated by oppla_forcing_gotm on ", date(),
                                     " from ", GOTMfile))
  RNetCDF::close.nc(nc_out)
}

#' Create NetCDF forcing file from pop model output.
#'
#' Create NetCDF forcing file from pop model output in HDF5 format.
#'
#' @param pop_file  pop-model forcing file in HDF5 format
#' @param filename  name of the forcing file to be created
#' @export
#' @import hdf5r
oppla_forcing_pop <- function(pop_file, filename = sub(pattern = "\\.h5$", replacement = ".nc", x = pop_file)) {
  h5 <- H5File$new(filename = pop_file)
  tn <- h5$attr_open(attr_name = "Time")$read()  # name of time dataset
  ln <- h5$attr_open(attr_name = "Light")$read() # name of light dataset
  lon <- h5$attr_open(attr_name = "Longitude")$read()
  lat <- h5$attr_open(attr_name = "Latitude")$read()
  time <- h5$open(name = tn)$read()
  depth <- h5$open(name = "Depth")$read()
  lig <- h5$open(name = ln)$read()
  vdc <- h5$open(name = "D")$read()
  tmp <- h5$open(name = "Temperature")$read()
  h5$close_all()
  pardata <- oppla_read_ncep(path = "~/data/NCEP/dswrf", varid = "dswrf", lon = lon, lat = lat)
  nc_out <- RNetCDF::create.nc(filename = filename, format = "netcdf4")
  dims <- list(longitude = list(units = "degrees_E", axis = "X", vals = lon, unlim = FALSE),
               latitude = list(units = "degrees_N", axis = "Y", vals = lat, unlim = FALSE),
               depth = list(units = "m", axis = "Z", positive = "down", vals = depth, unlim = FALSE),
               time = list(units = sprintf("seconds since %s-01-01 00:00:00", substr(x = tn, start = 2, stop = 5)),
                           axis = "T", vals = time, unlim = TRUE))
  dimnames <- onc_make_axes(ncfile = nc_out, dims = dims)
  date <- as.POSIXct(time, tz = "UTC", origin = sub(pattern = "seconds since ", replacement = "", x = dims$time$units))
  vars <- list(D = list(units = "m2 s-1", dim = dimnames, vals = vdc),
               Temperature = list(units = "degrees_Celsius", dim = dimnames, vals = tmp),
               avgSWR_NCEP = list(units = "W m-2", dim = dimnames[c(1, 2, 4)],
                                  vals = approx(x = pardata$time, y = pardata$dswrf, xout = date, rule = 2)$y))
  vars[[ln]] <- list(units = "", dim = dimnames[c(1, 2, 4)], vals = lig)
  varnames <- names(vars)
  for (varname in varnames) {
    RNetCDF::var.def.nc(ncfile = nc_out, varname = varname, vartype = "NC_DOUBLE", dimensions = dimnames)
    RNetCDF::att.put.nc(ncfile = nc_out, variable = varname, name = "units", type = "NC_CHAR", value = vars[[varname]]$units)
    RNetCDF::var.put.nc(ncfile = nc_out, variable = varname, data = vars[[varname]]$vals)
  }
  RNetCDF::att.put.nc(ncfile = nc_out, variable = "NC_GLOBAL", name = "description", type = "NC_CHAR",
                      value = sprintf("Oppla forcing file generated by oppla_forcing_pop on %s from %s", date(), pop_file))
  RNetCDF::close.nc(nc_out)
}
