## Helper function for oppla_level and oppla_tile, creating a data table from a list
## of oppla results or data tables with the relevant columns. The resulting data table contains
## a column \code{value} containing the result of evaluating the column expression inside the individual
## data tables, the columns indicated by datecol and a tag column.
oppla_eval_tag <- function(column, data, datecol) {
  if (! "names" %in% names(attributes(data))) {
    ans <- names(attributes(data[[1]]))
    if ("tag" %in% ans) names(data) <- sapply(data, FUN = function(x) attr(x, "tag"))
    else if ("file" %in% ans) names(data) <- sapply(data, FUN = function(x) {
      sub(pattern = ".*/(.+)\\..+$", replacement = "\\1", x = attr(x, "file"))
    })
  }
  for (n in seq_along(data)) {
    if (!data.table::is.data.table(data[[n]])) data[[n]] <- data.table::as.data.table(data[[n]])
    edc <- eval(column, envir = data[[n]])
    data[[n]] <- data[[n]][j = c(.SD, .(value = ..edc)), .SDcols = c(datecol, "depth", "height", "dt")]
  }
  return(data.table::rbindlist(data, idcol = "tag"))
}

#' Levelplot(s) of a column-expression of oppla results.
#'
#' \code{oppla_level} produces (a) levelplot(s) of oppla results
#'
#' @param column expression involving columns of data
#' @param data (list of) [`Data table(s)`][data.table::data.table] with oppla results or temporal averages
#' @param log use a pseudo-log scale for the colour key
#' @param xlab x-axis label
#' @param ylim vertical (depth) limits for plot panels
#' @param main plot title; default is \code{column} with units
#' @param palette colour palette for levelplots
#' @param ... other parameters to be passed to [`levelplot`][lattice::levelplot]
#' @export
oppla_level <- function(column, data, ylim, palette = "viridis", xlab = "date", log = FALSE, main, ...) {
  if (is.data.frame(data)) data <- list(data)
  datecols <- intersect(c("mday", "yday", "mon", "year", "date"), names(data[[1]]))
  dt <- oppla_eval_tag(substitute(column), data = data, datecol = datecols)
  ucs <- units::deparse_unit(dt$value)                    # save units for labels
  ud <- units::as_units(units::deparse_unit(dt$depth))
  dt <- units::drop_units(dt) # remove units, as they confuse levelplot
  if (log) {
    rng <- pretty(log10(range(dt[value > 0, j = value])))
    r0 <- rng[1] - diff(rng[1:2]) # r0 represents 0 in the pseudo-log scale
    dt[j = value := ifelse(value == 0, r0, log10(value))]
    colorkey <- list(at = seq(from = r0, to = tail(rng, n = 1), length.out = 100),
                     labels = list(at = c(r0, rng), labels = c(0, 10 ^ rng)))
  } else { colorkey <- list() }
  if (missing(main)) main <- if (nzchar(ucs)) units::make_unit_label(as.expression(substitute(column)), units::as_units(ucs))
                             else as.expression(substitute(column))
  at <- as.numeric(seq(from = min(dt[j = value]), to = max(dt[j = value]), length.out = 100))
  lattice::levelplot(stats::formula(paste("value", "~", paste0(datecols, collapse = "+"), "+", "depth | tag")),
                     data = dt, at = at, xlab = xlab, ylab = units::make_unit_label("depth", ud),
                     col.regions = grDevices::hcl.colors(n = length(at), palette = palette),
                     ylim = ylim, main = main, colorkey = colorkey, ...)
}

## log transform for ggplot2-based plotting functions
op_log_trans <- function(x, breaks = 4) {
  rng <- log10(range(x[x > 0], na.rm = TRUE))
  r0 <- rng[1] - 1
  return(scales::trans_new(name = "op_log",
                           transform = function(x) ifelse(x == 0, r0, log10(x)),
                           inverse = function(x) ifelse(x == r0, 0, 10 ^ x),
                           domain = c(0, Inf),
                           breaks = function(x) c(if (min(x) == 0) 0, 10 ^ pretty(x = rng, n = breaks))
                           ))
}

#' Tile plot(s) of a column-expression of oppla results.
#'
#' \code{oppla_tile} produces (a) tile plot(s) of oppla results (and/or data).
#'
#' @param column expression involving columns of \code{data}
#' @param data (named list of) [`Data table(s)`][data.table::data.table] with oppla results or temporal averages.
#' If it is a named list, the names will be used as facet labels
#' @param log use a pseudo-log scale for the colour key
#' @param order character vector containing all \code{tag} or \code{file} attributes of \code{data},
#' setting the order of the facets, which are arranged alphabetically by default
#' @param palette colour palette type for [`scale_fill_continuous`][ggplot2::scale_fill_continuous]
#' @param scales passed on to [`facet_wrap`][ggplot2::facet_wrap]
#' @param type either "tile" for a tile plot or "contour" for filled contours
#' @param xlab x-axis label
#' @param breaks (number of) contour levels for \code{type = "contour"}
#' @return A [`ggplot`][ggplot2::ggplot] object.
#' @param ylim vertical (depth) limits for plot panels
#' @details If \code{data} is a list including model results and data, you must ensure that the \code{column}
#' expression is valid and has the same units in all elements of \code{data}.  Use [`op_volume_units`]
#' to convert mass-specific to volume-specific units.
#'
#' The units for the colour bar label are taken from (the first element in) \code{data}.
#' @export
oppla_tile <- function(column, data, ylim, palette = "viridis", xlab = "date", log = FALSE, type = "tile", order,
                       scales = "fixed", breaks = 10) {
  if (is.data.frame(data)) data <- list(data)
  datecol <- intersect(c("yday", "week", "month", "quarter", "year"), names(data[[1]]))
  if ("date" %in% names(data[[1]])) {
    datecol <- "date"
    scale_x <- ggplot2::scale_x_datetime()
    dtfun <- function(x) units::drop_units(units::set_units(x, "s"))
  } else {
    scale_x <- ggplot2::scale_x_continuous(breaks = switch(datecol, yday = (0:5) * 61 + 15, quarter = 1:4,
                                                           week = (0:6) * 8 + 2, month = (1:6) * 2),
                                           name = datecol)
    dtfun <- identity
  }
  df <- oppla_eval_tag(substitute(column), data = data, datecol = datecol)
  ucs <- units::deparse_unit(df$value)     # save units
  units(df$value) <- NULL
  if (! missing(order)) df$tag <- factor(df$tag, levels = order)
  if (! missing(ylim))
    df <- subset(x = df, subset = findInterval(x = depth, vec = range(ylim), rightmost.closed = TRUE) == 1)
  p <- ggplot2::ggplot(data = df, mapping = ggplot2::aes(x = date + 0.5 * units::as_difftime(dt - dt[1]), y = depth)) +
    scale_x + ggplot2::xlab(label = xlab) + units::scale_y_units(trans = scales::reverse_trans())
  p <- switch(
    EXPR = type,
    tile = p + ggplot2::geom_tile(mapping = ggplot2::aes(fill = value, width = dtfun(dt),
                                                         height = units::drop_units(height))) +
      ggplot2::scale_fill_continuous(
                 name = if (nzchar(ucs)) units::make_unit_label(as.expression(substitute(column)), units::as_units(ucs))
                        else as.expression(substitute(column)),
                 type = palette,
                 trans = if (log) op_log_trans(df$value) else scales::identity_trans(),
                 guide = ggplot2::guide_colourbar(title.position = "right", title.hjust = 0.5,
                                                  title.theme = ggplot2::element_text(angle = 90),
                                                  barheight = grid::unit(x = 0.5, units = "npc"))),
    contour = {
      if (length(breaks) == 1)
        breaks <- if (log) op_log_trans(df$value, breaks = breaks)$breaks(df$value) else pretty(df$value, n = breaks)
      p + ggplot2::geom_contour_filled(mapping = ggplot2::aes(z = value), breaks = breaks) +
        ggplot2::scale_fill_discrete(
                   name = if (nzchar(ucs)) units::make_unit_label(as.expression(substitute(column)), units::as_units(ucs))
                          else as.expression(substitute(column)),
                   type = palette,
                   guide = ggplot2::guide_coloursteps(title.position = "right", title.hjust = 0.5, show.limits = TRUE,
                                                      title.theme = ggplot2::element_text(angle = 90),
                                                      barheight = grid::unit(x = 0.5, units = "npc")))
    }
  ) + ggplot2::facet_wrap(facets = "tag", scales = scales)
  return(p)
}

#' Tile plot(s) of column-expression strings of oppla results.
#'
#' \code{oppla_tile_vars} produces (a) tile plot(s) of oppla results.
#'
#' @param columns named vector of expression-strings involving columns of \code{data}
#' @param data [`Data table`][data.table::data.table] with oppla results or temporal averages.
#' @param log use a pseudo-log scale for the colour key
#' @param palette colour palette type for [`scale_fill_continuous`][ggplot2::scale_fill_continuous]
#' @param scales passed on to [`facet_wrap`][ggplot2::facet_wrap]
#' @param type either "tile" for a tile plot or "contour" for filled contours
#' @param xlab x-axis label
#' @param breaks (number of) contour levels for \code{type = "contour"}
#' @return A [`ggplot`][ggplot2::ggplot] object.
#' @param ylim vertical (depth) limits for plot panels
#' @details You must ensure that the \code{columns}
#' expression strings are all valid and have the same units.  Use [`op_volume_units`]
#' to convert mass-specific to volume-specific units.
#' @export
oppla_tile_vars <- function(columns, data, ylim = range(data$depth), palette = "viridis", xlab = "date",
                            log = FALSE, type = "tile", scales = "fixed", breaks = 10) {
  datecol <- intersect(c("yday", "week", "month", "quarter", "year"), names(data))
  if ("date" %in% names(data)) {
    datecol <- "date"
    scale_x <- ggplot2::scale_x_datetime()
    dtfun <- function(x) units::drop_units(units::set_units(x, "s"))
  } else {
    scale_x <- ggplot2::scale_x_continuous(breaks = switch(datecol, yday = (0:5) * 61 + 15, quarter = 1:4,
                                                           week = (0:6) * 8 + 2, month = (1:6) * 2),
                                           name = datecol)
    dtfun <- identity
  }
  cns <- names(columns)
  dt <- data[depth %between% sort(ylim)]
  for (name in cns) {
    col <- eval(parse(text = columns[[name]]), envir = dt)
    dt[j = (name) := col]
  }
  dtm <- data.table::melt(dt[j = .SD, .SDcols = c(cns, datecol, "dt", "depth", "height")],
                          id.vars = c(datecol, "dt", "depth", "height"))
  ucs <- units::deparse_unit(dtm$value)     # save units
  units(dtm$value) <- NULL
  p <- ggplot2::ggplot(data = dtm, mapping = ggplot2::aes(x = date + 0.5 * units::as_difftime(dt - dt[1]), y = depth)) +
    scale_x + ggplot2::xlab(label = xlab) + units::scale_y_units(trans = scales::reverse_trans())
  p <- switch(
    EXPR = type,
    tile = p + ggplot2::geom_tile(mapping = ggplot2::aes(fill = value, width = dtfun(dt),
                                                         height = units::drop_units(height))) +
      ggplot2::scale_fill_continuous(
                 name = if (nzchar(ucs)) units::make_unit_label(NULL, units::as_units(ucs), group = c("", ""))
                        else NULL,
                 type = palette,
                 trans = if (log) op_log_trans(dtm$value) else scales::identity_trans(),
                 guide = ggplot2::guide_colourbar(title.position = "right", title.hjust = 0.5,
                                                  title.theme = ggplot2::element_text(angle = 90),
                                                  barheight = grid::unit(x = 0.5, units = "npc"))),
    contour = {
      if (length(breaks) == 1)
        breaks <- if (log) op_log_trans(dtm$value, breaks = breaks)$breaks(dtm$value) else pretty(dtm$value, n = breaks)
      p + ggplot2::geom_contour_filled(mapping = ggplot2::aes(z = value), breaks = breaks) +
        ggplot2::scale_fill_discrete(
                   name = if (nzchar(ucs)) units::make_unit_label(NULL, units::as_units(ucs), group = c("", ""))
                          else NULL,
                   type = palette,
                   guide = ggplot2::guide_coloursteps(title.position = "right", title.hjust = 0.5, show.limits = TRUE,
                                                      title.theme = ggplot2::element_text(angle = 90),
                                                      barheight = grid::unit(x = 0.5, units = "npc")))
    }
  ) + ggplot2::facet_wrap(facets = "variable", scales = scales)
  return(p)
}

#' Plot data as time series
#'
#' \code{op_plot_date} can either plot a column of \code{data} vs. date, coloured by depth
#' or create a scatter plot depth vs. date coloured by the column data.
#'
#' @param data    dataframe as returned by \code{read_ls} or \code{read_bats}
#' @param y       name of the column in data to be used for the y axis
#' @param col     name of the column in data to be used for the colour of the points
#' @param palette palette use fot the colour scale: "gradient", "viridis" or a colour scale
#' function (see [`scale_colour_continuous`][ggplot2::scale_colour_continuous])
#' @param log     logical: whether to apply a pseudo-log transformation to the colour scale
#' @export
op_plot_date <- function(data, y, col = depth, palette = "viridis", log = FALSE) {
  data <- eval(substitute(within(data, expr = {
    ydat <- y
    coldat <- col })))
  if ("date" %in% names(data)) datecol <- "date"
  else datecol <- intersect(c("mday", "yday", "week", "month", "quarter"), names(data))
  ucs <- units::deparse_unit(data$coldat)
  units(data$coldat) <- NULL
  y <- deparse(substitute(y))
  col <- deparse(substitute(col))
  p <- ggplot2::ggplot(data = ggplot2::remove_missing(df = data, na.rm = TRUE, vars = c("ydat", "coldat")),
                       mapping = ggplot2::aes(x = .data[[datecol]], y = ydat, col = coldat)) +
    ggplot2::geom_point() +
    ggplot2::scale_colour_continuous(
               name = if (nzchar(ucs)) units::make_unit_label(lab = col, u = units::as_units(ucs)) else col,
               type = palette, trans = if (log) op_log_trans(data[[col]]) else scales::identity_trans(),
               guide = ggplot2::guide_colourbar(title.position = "right", title.hjust = 0.5,
                                                title.theme = ggplot2::element_text(angle = 90),
                                                barheight = grid::unit(x = 0.5, units = "npc"))) +
    if (y == "depth")
      units::scale_y_units(name = y, trans = scales::reverse_trans())
    else
      units::scale_y_units(name = y)
  return(p)
}
