#' Create an initial-conditions file from an oppla time slice.
#'
#' \code{oppla_icf} can be used for switching from one depth structure to another
#' or for adding or modifying states.  You can generate an initial-conditions file
#' from a time slice of oppla results, but oppla can also directly restart from one
#' of its output files.
#'
#' @param oppla oppla output as obtained with [`oppla_read`]
#' @param date date of the time slice in \code{oppla}
#' @param filename name of the initial-conditions file
#' @param icdate date for the initial conditions
#' @param depth optional depth vector ([`units`][units::units] object) for the initial-conditions
#'
#' @details
#' The \code{depth} argument can be used to specify a different depth structure
#' for the initial-conditions file.  If the vertical extent of the new depth structure
#' is within that of the original, total mass within this depth range will be preserved.
#' If the new depth structure extends below the original, the lower levels are filled
#' with the bottom-most values of the original profiles.
#' @export
oppla_icf <- function(oppla, date = oppla$date[1], filename, icdate = date, depth = unique(oppla$depth)) {
  svns <- setdiff(names(oppla), c("date", "depth", "dt", "height", "Temperature", "Salinity", "VDC", "tag"))
  dts <- oppla$date[which.min(abs(oppla$date - as.POSIXct(date, tz = "UTC")))]
  ts <- oppla[date == dts, j = .SD, .SDcols = c("date", svns)]
  lon <- attr(oppla, "longitude")
  lat <- attr(oppla, "latitude")
  nc_out <- RNetCDF::create.nc(filename = filename, format = "netcdf4")
  dims <- list(longitude = list(units = units::deparse_unit(lon), axis = "X", vals = lon, unlim = FALSE),
               latitude = list(units = units::deparse_unit(lat), axis = "Y", vals = lat, unlim = FALSE),
               depth = list(units = units::deparse_unit(depth), axis = "Z", positive = "down", bounds = "depth_edges",
                            vals = depth, unlim = FALSE),
               time = list(units = "seconds since 1970-01-01 00:00:00", axis = "T",
                           vals = as.numeric(as.POSIXct(icdate, tz = "UTC")), unlim = TRUE))
  dimnames <- onc_make_axes(ncfile = nc_out, dims = dims)
  RNetCDF::dim.def.nc(ncfile = nc_out, dimname = "bounds", dimlength = 2)
  if (!identical(depth, unique(oppla$depth))) ts <- oppla_convert(ts, depth)
  for (svn in svns) {
    RNetCDF::var.def.nc(ncfile = nc_out, varname = svn, vartype = "NC_DOUBLE", dimensions = dimnames)
    RNetCDF::var.put.nc(ncfile = nc_out, variable = svn, data = as.numeric(ts[[svn]]), count = rep(NA, length(dimnames)))
    RNetCDF::att.put.nc(ncfile = nc_out, variable = svn, name = "units", type = "NC_CHAR",
                        value = units::deparse_unit(ts[[svn]]))
  }
  depth_edges <- attr(ts, "depth_edges")
  RNetCDF::var.def.nc(ncfile = nc_out, varname = "depth_edges", vartype = "NC_DOUBLE", dimensions = c("bounds", "depth"))
  RNetCDF::att.put.nc(ncfile = nc_out, variable = "depth_edges", name = "units", type = "NC_CHAR",
                      value = units::deparse_unit(depth_edges))
  RNetCDF::att.put.nc(ncfile = nc_out, variable = "NC_GLOBAL", name = "description", type = "NC_CHAR",
                      value = paste("Oppla initial conditions created on", Sys.time(),
                                    "from timeslice", as.POSIXct(dts, tz = "UTC"), "in", attr(oppla, "file")))
  RNetCDF::var.put.nc(ncfile = nc_out, variable = "depth_edges", data = depth_edges)
  RNetCDF::close.nc(nc_out)
}

#' Convert oppla results to a new depth structure.
#'
#' The conversion tries to conserve mass as much and as locally as possible.
#'
#' @param oppla  oppla results [`Data table`][data.table::data.table]
#' @param depth  new depth structure ([`units`][units::units] object)
#' @details Mass is conserved only if the new depth structure is within the limits of the old.
#' Otherwise the layers extending below the previous lower limit are filled with the bottom-most
#' values of the original profiles.
#' @export
oppla_convert <- function(oppla, depth) {
  depth_edges_orig <- attr(oppla, "depth_edges")
  depth_bounds_orig <- c(depth_edges_orig[1, 1], depth_edges_orig[2, ])
  depth_bounds <- units::set_units(0, "m")
  for (n in seq_along(depth)) depth_bounds[n + 1] <- 2 * depth[n] - depth_bounds[n]
  height <- diff(depth_bounds)
  depth_bounds_union <- sort(union(depth_bounds, depth_bounds_orig))
  height_union <- diff(depth_bounds_union) # virtual level heights
  depth_union <- (depth_bounds_union[-1] + head(depth_bounds_union, n = -1)) / 2
  ilev0 <- as.numeric(cut(depth_union, breaks = depth_bounds_orig)) # level indices
  ilev0[is.na(ilev0)] <- max(ilev0, na.rm = TRUE) # fill deep levels with lowest data
  ilev1 <- as.numeric(cut(depth_union, breaks = depth_bounds))
  ## Define a transformation matrix (tm) with one row for each target depth level
  ## and one column for each source depth level.  The sum of each row is 1.
  ## An element in tm is the fraction of the target layer contained in the
  ## corresponding source layer.
  tm <- matrix(data = 0, nrow = length(height), ncol = length(depth_bounds_orig) - 1)
  for (n in seq_along(depth)) {
    ilev <- which(ilev1 %in% n)
    tm[n, ilev0[ilev]] <- ave(height_union[ilev] / height[n], ilev0[ilev], FUN = sum)
  }
  nms <- setdiff(names(oppla), c("date", "depth", "height", "dt", "tag"))
  u <- lapply(X = oppla[j = .SD, .SDcols = nms], FUN = units::deparse_unit)
  ts <- as.data.table(do.call(what = rbind,
                              args = by(data = oppla[j = .SD, .SDcols = nms], INDICES = oppla$date,
                                        FUN = function(x) tm %*% as.matrix(x))))
  ts[j = date := rep(unique(oppla$date), each = length(depth))]
  ndates <- length(unique(oppla$date))
  if ("dt" %in% names(oppla)) ts[j = dt := oppla$dt[which(ts$date %in% oppla$date)]]
  ts[j = c("height", "depth") := .(rep(height, times = ndates), rep(depth, times = ndates))]
  for (nm in nms) ts[j = c(nm) := units::set_units(.SD[[1]], u[[nm]]), .SDcols = nm]
    units(ts[[nm]]) <- u[[nm]]
  data.table::setattr(ts, name = "depth_edges",
                      value = units::set_units(rbind(head(depth_bounds, n = -1), depth_bounds[-1]),
                                               units(depth_bounds)))
  return(ts)
}

#' Add SVM tracer-trait products to an initial-conditions file.
#'
#' The days of the year of ascent and descent (\code{doya} and \code{doyd}) are multiplied
#' with the corresponding zooplankton (identified by \code{state}) biomass and added to the
#' initial-conditions file.
#'
#' @param icfile    name of the initial-conditions file
#' @param state     name of the functional group
#' @param doya,doyd initial values of the days of ascent and descent
#' @export
add_svm_traits <- function(icfile, state, doya, doyd) {
  nc <- oppla::onc_info(filename = icfile, write = TRUE)
  sc_name <- paste(state, "C", sep = "_")
  sc_data <- RNetCDF::var.get.nc(ncfile = nc$self, variable = sc_name)
  vars <- setNames(list(sc_data * doya, sc_data * doyd), paste(state, c("CxDa", "CxDd"), sep = "_"))
  for (varname in names(vars)) {
    RNetCDF::var.def.nc(ncfile = nc$self, varname = varname, vartype = "NC_DOUBLE",
                        dimensions = nc$var[[sc_name]]$dimensions)
    RNetCDF::att.put.nc(ncfile = nc$self, variable = varname, name = "units", type = "NC_CHAR",
                        value = paste("d", nc$var[[sc_name]]$units, sep = "*"))
    RNetCDF::var.put.nc(ncfile = nc$self, variable = varname, data = vars[[varname]])
  }
  RNetCDF::close.nc(nc$self)
}
