#' @useDynLib oppla
NULL

#' Cumulative probability density distribution
#'
#' @param x data for which to estimate the density distribution
#' @param run vector of points for which densities are estimated
#' @param quiet set to \code{TRUE} to suppress warning message for
#' zero inter-quartile range
#' @return Named list with elements \code{cdd}, the cumulative density
#' distribution, \code{sigma}, the standard deviation of \code{x}, and
#' \code{"25%", "50%", "75%"}, containing the respective percentiles of \code{x}.
#' If the inter-quartile range is 0, the list contains only the percentiles
#' and \code{sigma}, with a warning.
#' @export
oppla_cdd <- function(x, run, quiet = FALSE) {
  rng <- quantile(x, c(0.25, 0.5, 0.75))
  R   <- diff(rng, lag = 2, names = FALSE)   # inter-quartile range
  sigma <- units::keep_units(sd, x)
  if (R == 0 * x[1]) {
    if (! quiet) warning("Empty inter-quartile range.")
    return(as.list(c(rng, sigma = sigma)))
  }
  nx   <- length(x)
  nrun <- length(run)
  cdd  <- .Fortran("oppla_cdd", nx = nx, nrun = nrun, x = x, run = run, R = R, sigma = sigma,
                   cdd = rep_len(0, nrun))
  return(c(cdd["cdd"], sigma = list(sigma), as.list(rng)))
}

#' Kernel density estimate
#'
#' @param x data for which to estimate the density distribution
#' @param run vector of points for which densities are estimated
#' @param quiet set to \code{TRUE} to suppress warning message for
#' zero inter-quartile range
#' @return Named list with elements \code{kde}, the kernel density estimates, \code{pdd},
#' the probability density distribution, \code{cdd}, the cumulative density distribution,
#' \code{sigma}, the standard deviation of \code{x}, and \code{"25%", "50%", "75%"},
#' containing the respective percentiles of \code{x}.  If the inter-quartile range is 0,
#' the list contains only the percentiles and \code{sigma}, with a warning.
#' @export
oppla_kde <- function(x, run, quiet = FALSE) {
  rng <- quantile(x, c(0.25, 0.5, 0.75))
  R   <- diff(rng, lag = 2, names = FALSE)   # inter-quartile range
  sigma <- units::keep_units(sd, x)
  if (R == 0 * x[1]) {
    if (! quiet) warning("Empty inter-quartile range.")
    return(as.list(c(rng, sigma = sigma)))
  }
  nx   <- length(x)
  nrun <- length(run)
  kde  <- .Fortran("oppla_kde", nx = nx, nrun = nrun, x = x, run = run, R = R, sigma = sigma,
                   kde = matrix(0, nrow = nx, ncol = nrun), pdd = rep_len(0, nrun), cdd = rep_len(0, nrun))
  return(c(kde[c("kde", "pdd", "cdd")], sigma = list(sigma), as.list(rng)))
}
