#' Open NetCDF file and retrieve info about its content.
#'
#' The retrieved info is obtained in a structure similar to \code{ncdf4::nc_open}.
#'
#' @param filename name of a NetCDF file
#' @param close if \code{TRUE} close the file upon return, otherwise the file must be closed by
#' \code{RNetCDF::close.nc(con = nc$self)}
#' @param ... arguments passed on to [`RNetCDF::open.nc`]
#'
#' @return List describing the structure of the NetCDF dataset with the following components:
#' \item{self}{(only for \code{close = FALSE}) object of class [`NetCDF`][RNetCDF::RNetCDF]
#' pointing to the NetCDF dataset}
#' \item{filename}{name of the NetCDF file}
#' \item{att}{list of global attributes in file or group}
#' \item{ngatts}{number of global attributes in file or group}
#' \item{dim}{list of dimensions as described in [`RNetCDF::dim.inq.nc`] plus \code{vals}
#' components where corresponding dimension variables exist}
#' \item{unlimids}{NetCDF ids of unlimited dimensions (if any); note that you have to add 1 to find
#' the corresponding dimension(s) in the \code{dim} component}
#' \item{var}{list of variables as described in [`RNetCDF::var.inq.nc`] plus attributes and \code{dimensions}}
#' \item{group}{list of groups (if any), each containing the components described here except
#' \code{filename} plus \code{name} and \code{fullname}}
#' \item{type}{list of types as described in [`RNetCDF::type.inq.nc`]}
#'
#' @details For \code{close = TRUE} the returned list does not contain \code{self}.
#'
#' You can retrieve all attributes of a variable \code{x} (one of the elements of \code{nc$var})
#' as \code{x[length(x) - (x$natts:1)]}.
#' @export
onc_info <- function(filename, close = FALSE, ...) {
  grpinfo <- function(ncid) {
    ## get group info recursively
    gi <- RNetCDF::grp.inq.nc(ncid = ncid)
    atts <- lapply(X = seq(from = 0, length.out = gi$ngatts), # inquire about group attributes
                   FUN = function(aid) RNetCDF::att.inq.nc(ncfile = ncid, variable = "NC_GLOBAL", attribute = aid))
    gi$att <- sapply(X = atts, FUN = function(att) {
      setNames(list(structure(RNetCDF::att.get.nc(ncfile = ncid, variable = "NC_GLOBAL", attribute = att$name),
                              type = att$type)), att$name)
    })
    ## read group dimensions
    gi$dim <- sapply(X = gi$dimids, FUN = function(dimid) {
      d <- RNetCDF::dim.inq.nc(ncfile = ncid, dimension = dimid)
      return(setNames(list(d), d$name))
    })
    ## read group types
    gi$type <- sapply(X = gi$typeids, FUN = function(type) {
      t <- RNetCDF::type.inq.nc(ncfile = ncid, type = type)
      return(setNames(list(t), t$name))
    })
    gi$var <- list() # read variables in group
    for (varid in gi$varids) {
      vi <- RNetCDF::var.inq.nc(ncfile = ncid, variable = varid)
      gi$var[[vi$name]] <- vi
      ## read variable attributes
      for (na in seq(from = 0, length.out = vi$natts)) {
        ai <- RNetCDF::att.inq.nc(ncfile = ncid, variable = varid, attribute = na)
        gi$var[[vi$name]][[ai$name]] <- structure(RNetCDF::att.get.nc(ncfile = ncid, variable = varid, attribute = na),
                                                  type = ai$type)
      }
      ## add dimension names and read dimension vars for convenience
      if (! is.na(vi$dimids[1])) {
        gi$var[[vi$name]]$dimensions <- sapply(X = gi$dim[match(x = vi$dimids, table = gi$dimids)],
                                               FUN = function(dim) dim$length)
        if (vi$name %in% names(gi$dim))
          gi$dim[[vi$name]]$vals <- RNetCDF::var.get.nc(ncfile = ncid, variable = varid)
      }
    }
    gi$group <- sapply(X = gi$grps, FUN = function(grp) {
      sgi <- grpinfo(ncid = grp)
      return(setNames(list(sgi), sgi$name))
    })
    return(gi[setdiff(names(gi), c("grps", "dimids", "varids", "typeids"))])
  }
  nc <- RNetCDF::open.nc(con = filename, ...)
  fmt <- RNetCDF::file.inq.nc(ncfile = nc)$format
  fi <- grpinfo(ncid = nc)
  if (close) {
    RNetCDF::close.nc(fi$self)
    fi$self <- NULL
  }
  return(c(fi[setdiff(names(fi), c("name", "fullname"))], filename = filename, format = fmt))
}

#' Create NetCDF file from list as obtained with [`onc_info`].
#'
#' @param onc named list with NetCDF information obtained with [`onc_info`]
#' @param filename name of the NetCDF file to be created
#' @param data (optional) named list with data for the NetCDF file as obtained with [`RNetCDF::read.nc`]
#' @details If \code{data} is not provided, only the dimension data in \code{onc} are written to the file.
#' The names of the \code{data} list must match the group and variable names in \code{onc}.
#' @export
onc_create <- function(onc, filename = onc$filename, data) {
  create_group <- function(ncid, group) {
    for (att in names(group$att)) {
      RNetCDF::att.put.nc(ncfile = ncid, variable = "NC_GLOBAL", name = att, # write global attributes
                          type = attr(group$att[[att]], "type"), value = group$att[[att]])
    }
    for (dim in group$dim)
      RNetCDF::dim.def.nc(ncfile = ncid, dimname = dim$name, dimlength = dim$length, unlim = dim$unlim)
    for (type in group$type) {
      switch(EXPR = type$class,
             vlen = RNetCDF::type.def.nc(ncfile = ncid, typename = type$name, class = type$class, size = type$size,
                                         basetype = type$basetype),
             enum = RNetCDF::type.def.nc(ncfile = ncid, typename = type$name, class = type$class, size = type$size,
                                         basetype = type$basetype, names = names(type$value), values = type$value),
             compound = RNetCDF::type.def.nc(ncfile = ncid, typename = type$name, class = type$class, size = type$size,
                                             names = names(type$subtype), subtypes = type$subtype, dimsizes = type$dimsizes),
             RNetCDF::type.def.nc(ncfile = ncid, typename = type$name, class = type$class, size = type$size)
             )
    }
    for (var in group$var) {
      RNetCDF::var.def.nc(ncfile = ncid, varname = var$name, vartype = var$type, dimensions = names(var$dimensions))
      for (att in Filter(f = function(name) ! is.null(attr(var[[name]], "type")), x = names(var)))
        RNetCDF::att.put.nc(ncfile = ncid, variable = var$name, name = att, # write attributes
                            type = attr(var[[att]], "type"), value = var[[att]])
      if (var$name %in% names(group$dim)) # write dimension variables
        RNetCDF::var.put.nc(ncfile = ncid, variable = var$name, data = group$dim[[var$name]]$vals)
    }
    for (grp in group$group)
      create_group(ncid = RNetCDF::grp.def.nc(ncid = ncid), group = grp)
  }
  write_group <- function(ncid, group, data) {
    for (name in names(data)) {
      if (is.list(data[[name]])) {
        write_group(ncid = RNetCDF::grp.inq.nc(ncid = ncid, grpname = name)$self,
                    group = group$group[[name]], data = data[[name]])
      } else {
        if (group$var[[name]]$type != "NC_CHAR") dim(data[[name]]) <- group$var[[name]]$dimensions
        RNetCDF::var.put.nc(ncfile = ncid, variable = name, data = data[[name]])
      }
    }
  }
  nc <- RNetCDF::create.nc(filename = filename, prefill = FALSE, format = onc$format)
  create_group(ncid = nc, group = onc)
  if (! missing("data"))
    write_group(ncid = nc, group = onc, data = data)
  RNetCDF::close.nc(nc)
}

## Create a set of axis dimensions and associated dimension variables.
## argument ncfile: Object of class "NetCDF" which points to the NetCDF dataset
##                  (as returned from RNetCDF::open.nc).
## argument dims:   Named list of lists with dimension specifications.
##                  The names are used as the names of the dimensions.
## Each specification-list must contain vals and unlim.  Further elements are regarded as attributes.
## returns: Character vector of the names of the dimensions
onc_make_axes <- function(ncfile, dims) {
  dimnames <- names(dims)
  for (dimname in dimnames) {
    RNetCDF::dim.def.nc(ncfile = ncfile, dimname = dimname, dimlength = length(dims[[dimname]]$vals),
                        unlim = dims[[dimname]]$unlim)
    RNetCDF::var.def.nc(ncfile = ncfile, varname = dimname, vartype = "NC_DOUBLE", dimensions = dimname)
    for (name in setdiff(names(dims[[dimname]]), c("vals", "unlim")))
      RNetCDF::att.put.nc(ncfile = ncfile, variable = dimname, name = name,
                          type = "NC_CHAR", value = dims[[dimname]][[name]])
    RNetCDF::var.put.nc(ncfile = ncfile, variable = dimname, data = dims[[dimname]]$vals)
  }
  return(dimnames)
}
