!> \brief functions for bacteria and DOM
MODULE bac
  USE et
  IMPLICIT NONE
  PRIVATE

  TYPE, EXTENDS(fungroup), PUBLIC :: bacteria
     REAL(dp) :: QN_param=0.2, QP_param=0.01,&
          ADIM=1.E0_dp,&        !< affinity for DIN and DIP
          ADOP=0._dp,&          !< DOP affinity
          fTem,&
          ggem=0.3D0,&          !< max. bacterial gross growth efficiency
          lambda=10._dp,&       !< straightness of the DOC vs. growth relation
          N2P,&                 !< N:P ratio
          VDIN, VDIP,&
          RC,&
          Vm,&
          Vmax=5D0,&
          zeta=0.6D0            !< cost of N assimilation in molN/molC
     CHARACTER(LEN=100) :: DOM='DOM' !< name of labile DOM compartment
     INTEGER :: idom=0
     INTEGER, POINTER :: idocnp(:)
     REAL(dp), ALLOCATABLE :: DOCNP(:), VDOM(:)
     REAL(dp), POINTER :: DOC, DON, DOP, VDOC, VDON, VDOP, QDON, QDOP
     PROCEDURE(bac_grow), POINTER :: grow
     PROCEDURE(bac_uptake), POINTER :: uptake
   CONTAINS
     PROCEDURE :: flux => bac_flux
     PROCEDURE :: read => bac_read
     PROCEDURE :: set => bac_set
  END TYPE bacteria

  TYPE, EXTENDS(fungroup), PUBLIC :: dom
     REAL(dp) :: trlc=0D0,& !< transformation rate from refractory to labile DOC in \unit{d^{-1}} -> \unit{s^{-1}}
          trln=0D0,& !< transformation rate from refractory to labile DON in \unit{d^{-1}} -> \unit{s^{-1}}
          ldlc=0D0,& !< light-dependent decay rate of labile DOC in \unit{(d.W.m^{-2})^{-1}} -> \unit{(s.W.m^{-2})^{-1}}
          ldln=0D0,& !< light-dependent decay rate of labile DON in \unit{(d.W.m^{-2})^{-1}} -> \unit{(s.W.m^{-2})^{-1}}
     relac, relan
   CONTAINS
     PROCEDURE :: flux => dom_flux
     PROCEDURE :: read => dom_read
     PROCEDURE :: set => dom_set
  END TYPE dom

CONTAINS

  SUBROUTINE bac_read (grp, env, lun)
    IMPLICIT NONE
    CLASS(bacteria), TARGET, INTENT(INOUT) :: grp
    TYPE(local), TARGET, INTENT(IN) :: env
    INTEGER, INTENT(IN) :: lun
    CHARACTER(LEN=512) :: DOM, species, fT='', fn, msg
    LOGICAL :: fTlam
    REAL(dp) :: Q10, ggem, sticky
    TYPE(quantity) :: QN, QP, Tref, Vmax, lambda, ADIM, zeta
    NAMELIST /bac/ DOM, species, fT, fTlam, Q10, ggem, sticky,&
         QN, QP, Tref, Vmax, lambda, ADIM, zeta
    grp%nsv = 1
    species = ''
    DOM = grp%DOM
    REWIND (lun)
    DO
       fTlam = .FALSE.
       ggem = grp%ggem
       lambda = quantity(grp%lambda, 'm3 molC^-1')
       QN = quantity(grp%QN_param, 'molN molC^-1')
       QP = quantity(grp%QP_param, 'molP molC^-1')
       zeta = quantity(grp%zeta, 'molC molN^-1')
       Q10 = grp%Q10
       sticky = grp%sticky
       Tref = quantity(grp%Tref, '°C')
       Vmax = quantity(grp%Vmax, 'd-1')
       ADIM = quantity(grp%ADIM, 'm3 molC-1 d-1')
       READ (lun, NML=bac, END=100, IOMSG=msg, ERR=900)
       IF (TRIM(species).EQ.TRIM(grp%name)) EXIT
    END DO
    grp%ggem = ggem
    grp%lambda = convert(lambda, 'm3 mmolC^-1')
    grp%QN_param = convert(QN, 'molN molC^-1')
    grp%QP_param = convert(QP, 'molP molC^-1')
    grp%zeta = convert(zeta, 'molC molN^-1')
    grp%Q10 = Q10
    grp%sticky = sticky
    grp%Tref = convert(Tref, '°C')
    grp%Vmax = convert(Vmax, 's-1')
    grp%ADIM = convert(ADIM, 'm3 mmolC-1 s-1')
    CALL grp%fT_select (fT, 'bac_read')     ! set temperature function
    ALLOCATE (grp%constituents(grp%nsv))
    grp%constituents = 'C'
    grp%DOM = TRIM(DOM(1:100))
    grp%grow => bac_grow
    IF (fTlam) THEN
       grp%uptake => bac_uptake_fT
    ELSE
       grp%uptake => bac_uptake
    END IF
    RETURN
100 CONTINUE
    INQUIRE (UNIT=lun, NAME=fn)
    WRITE (stderr, '("bac_read: missing namelist bac for species ",A," in file ",A,".")')&
         TRIM(grp%name), TRIM(fn)
    STOP 1
900 CONTINUE
    WRITE (stderr,'("bac_read:",A)') TRIM(msg)
    STOP 1
  END SUBROUTINE bac_read

  SUBROUTINE bac_set (grp, grps, env)
    IMPLICIT NONE
    CLASS(bacteria), TARGET, INTENT(INOUT) :: grp
    TYPE(state), DIMENSION(:), INTENT(IN) :: grps
    TYPE(local), TARGET, INTENT(INOUT) :: env
    INTEGER :: n, ndom
    DO n=1,env%nft
       IF (TRIM(grp%DOM).EQ.TRIM(grps(n)%var%name)) grp%idom = n
    END DO
    IF (grp%idom.EQ.0) THEN
       WRITE (stderr, '("bac_set: Compartment ",A," does not exist.")') TRIM(grp%DOM)
       STOP 1
    END IF
    grp%idocnp => grps(grp%idom)%var%isv
    ndom = SIZE(grp%idocnp)
    ALLOCATE (grp%DOCNP(ndom), grp%VDOM(ndom))
    grp%QN = grp%QN_param
    grp%QP = grp%QP_param
    grp%QDON => grps(grp%idom)%var%QN
    grp%QDOP => grps(grp%idom)%var%QP
    grp%N2P = grp%QN/grp%QP
    grp%DOC => grp%DOCNP(1)
    grp%VDOC => grp%VDOM(1)
    IF (ndom.GT.1) THEN
       grp%DON => grp%DOCNP(2)
       grp%VDON => grp%VDOM(2)
    ELSE
       ALLOCATE (grp%DON, grp%VDON)
    END IF
    IF (ndom.GT.2) THEN
       grp%DOP => grp%DOCNP(3)
       grp%VDOP => grp%VDOM(3)
    ELSE
       ALLOCATE (grp%DOP, grp%VDOP)
    END IF
    env%smsk(grp%isv(1),(/grp%idocnp, env%idic, env%io2, env%idin, env%idip/),:) = .TRUE.
  END SUBROUTINE bac_set

  SUBROUTINE bac_flux (grp, grps, env, box, times)
    IMPLICIT NONE
    CLASS(bacteria), INTENT(INOUT) :: grp
    TYPE(state), DIMENSION(:), INTENT(IN) :: grps
    TYPE(local), INTENT(INOUT) :: env
    TYPE(layer), INTENT(INOUT) :: box(0:)
    TYPE(timing), INTENT(IN) :: times
    grp%DOCNP = box(1)%stv(grp%idocnp)
    CALL grp%grow (box(1))
    box(1)%soma(grp%idocnp(1),grp%isv(1)) = grp%VDOC - grp%RC
    box(1)%soma(grp%isv(1),grp%idocnp) = -grp%VDOM
    box(1)%soma(grp%isv(1),env%idic) = grp%RC
    box(1)%soma(grp%isv(1),env%io2)  = -grp%RC*env%RQ + grp%VDIN*2D0 ! O2 consumption
    box(1)%soma(grp%isv(1),env%idin) = -grp%VDIN
    box(1)%soma(grp%isv(1),env%idip) = -grp%VDIP
  END SUBROUTINE bac_flux

  SUBROUTINE bac_grow (bac, box)
    IMPLICIT NONE
    CLASS(bacteria), INTENT(INOUT) :: bac
    TYPE(layer), INTENT(IN) :: box
    REAL(dp) :: CANi, APi, APo, dmdNi, dmdP, VNimx, VNipt, VPipt, VPopt
    bac%fTem = bac%fT(box)
    bac%Vm = bac%fTem*bac%Vmax
    bac%VDOC = bac%uptake()   ! potential DOC uptake (bac_uptake or bac_uptake_fT)
    bac%VDON = bac%VDOC*bac%QDON
    CANi = bac%OC*bac%ADIM*box%DIN
    APi = bac%ADIM*box%DIP
    APo = bac%ADOP*bac%DOP
    VPipt = bac%OC*APi/(Api/(bac%Vm*bac%QP) + 1D0) ! potential DIP uptake
    VPopt = bac%OC*APo/(Apo/(bac%Vm*bac%QP) + 1D0) ! potential DOP uptake via alkaline phosphatase
    bac%VDOP = bac%VDOC*bac%QDOP                   ! DOP as part of labile DOM
    VNimx = (VPipt + bac%VDOP + VPopt)*bac%N2P     ! mini chain model: P uptake limits N assimilation
    VNipt = MIN(VNimx*CANi/(CANi + VNimx), VNimx - bac%VDON) ! potential DIN uptake
    dmdNi = bac%VDOC*bac%ggem*bac%QN - bac%VDON - MIN(VNipt, 0D0) ! DIN demand (always >= 0)
    IF (dmdNi.GT.(VNipt*(1D0 + bac%zeta*bac%QN))) THEN
       bac%VDIN = VNipt         ! DIN uptake or DON remineralization (if VNipt < 0)
       bac%RC = bac%VDOC - (bac%VDIN + bac%VDON)/bac%QN ! respire excess DOC
    ELSEIF (dmdNi.GT.0D0) THEN  ! DIN uptake according to demand
       bac%VDIN = dmdNi/(1D0 + bac%zeta*bac%QN) ! VNipt is always >= 0 here
       bac%RC = bac%VDOC*(1D0 - bac%ggem) + bac%zeta*bac%VDIN
    ELSE
       bac%VDIN = dmdNi         ! remineralize excess DON (dmdNi <= 0)
       bac%RC = bac%VDOC*(1D0 - bac%ggem)
    END IF
    dmdP = (bac%VDOC - bac%RC)*bac%QP - bac%VDOP
    bac%VDIP = MIN(dmdP, VPipt)
    bac%VDOP = bac%VDOP + dmdP - bac%VDIP
  END SUBROUTINE bac_grow

  !> \brief saturation of DOC uptake depends on temperature
  FUNCTION bac_uptake_fT (bac) RESULT (VC)
    IMPLICIT NONE
    CLASS(bacteria), INTENT(IN) :: bac
    REAL(dp) :: VC, LC
    LC = bac%lambda*bac%DOC
    VC = bac%OC*bac%Vm*LC*(1D0 - EXP(-bac%fTem*LC))
  END FUNCTION bac_uptake_fT
    
  !> \brief saturation of DOC uptake (independent of temperature)
  FUNCTION bac_uptake (bac) RESULT (VC)
    IMPLICIT NONE
    CLASS(bacteria), INTENT(IN) :: bac
    REAL(dp) :: VC, LC
    LC = bac%lambda*bac%DOC
    VC = bac%OC*bac%Vm*LC*(1D0 - EXP(-LC))
  END FUNCTION bac_uptake

  !> read namelist dom and set dom parameters, constituents, number of states
  SUBROUTINE dom_read (grp, env, lun)
    IMPLICIT NONE
    CLASS(dom), TARGET, INTENT(INOUT) :: grp
    TYPE(local), TARGET, INTENT(IN) :: env
    INTEGER, INTENT(IN) :: lun
    CHARACTER(LEN=512) :: fn, fT='', msg
    REAL(dp) :: Q10
    TYPE(quantity) :: ldlc, ldln, trlc, trln, Tref
    NAMELIST /dom/ ldlc, ldln, trlc, trln, fT, Q10, Tref
    grp%nsv= 3
    ALLOCATE (grp%constituents(grp%nsv))
    grp%constituents = (/'C', 'N', 'P'/)
    ldlc = quantity(grp%ldlc, 'd-1 W-1 m2')
    ldln = quantity(grp%ldln, 'd-1 W-1 m2')
    trlc = quantity(grp%trlc, 'd-1')
    trln = quantity(grp%trln, 'd-1')
    Tref = quantity(grp%Tref, '°C')
    Q10 = grp%Q10
    REWIND (lun)
    READ (lun, NML=dom, END=100, IOMSG=msg, ERR=900)
    grp%ldlc = convert(ldlc, 's-1 W-1 m2')
    grp%ldln = convert(ldln, 's-1 W-1 m2')
    grp%trlc = convert(trlc, 's-1')
    grp%trln = convert(trln, 's-1')
    grp%Tref = convert(Tref, '°C')
    grp%Q10 = Q10
    CALL grp%fT_select (fT, 'dom_read')     ! set temperature function
    RETURN
100 CONTINUE
    INQUIRE (UNIT=lun, NAME=fn)
    WRITE (stderr, '("Namelist dom missing in file ",A,".")') TRIM(fn)
    STOP 1
900 CONTINUE
    WRITE (stderr,'("dom_read:",A)') TRIM(msg)
    STOP 1
  END SUBROUTINE dom_read

  !> flag DOM-related elements of soma for output
  SUBROUTINE dom_set (grp, grps, env)
    IMPLICIT NONE
    CLASS(dom), TARGET, INTENT(INOUT) :: grp
    TYPE(state), DIMENSION(:), INTENT(IN) :: grps
    TYPE(local), TARGET, INTENT(INOUT) :: env
    env%smsk(env%irdoc,grp%isv(1),:) = .TRUE.
    env%smsk(env%irdon,grp%isv(2),:) = .TRUE.
  END SUBROUTINE dom_set

  !> set DOM-related fluxes in the source matrix soma
  SUBROUTINE dom_flux (grp, grps, env, box, times)
    IMPLICIT NONE
    CLASS(dom), INTENT(INOUT) :: grp
    TYPE(state), DIMENSION(:), INTENT(IN) :: grps
    TYPE(local), INTENT(INOUT) :: env
    TYPE(layer), INTENT(INOUT) :: box(0:)
    TYPE(timing), INTENT(IN) :: times
    REAL(dp) :: fTem
    fTem = grp%fT(box(1))
    grp%relac = grp%trlc*fTem*box(1)%stv(env%irdoc) - grp%ldlc*box(1)%iPAR*box(1)%stv(grp%isv(1))
    grp%relan = grp%trln*fTem*box(1)%stv(env%irdon) - grp%ldln*box(1)%iPAR*box(1)%stv(grp%isv(2))
    box(1)%soma(env%irdoc,grp%isv(1)) = grp%relac  ! refractory to labile DOC
    box(1)%soma(env%irdon,grp%isv(2)) = grp%relan  ! refractory to labile DON
    box(1)%soma(grp%isv(1),env%irdoc) = -grp%relac
    box(1)%soma(grp%isv(2),env%irdon) = -grp%relan
  END SUBROUTINE dom_flux

END MODULE bac
