!> \brief Functions for solar radiation and the diurnal light cycle
!!
!! The diurnal light cycle is described without twilight after \cite Brock:1981
!! and with twilight after \cite Forsythe:1995.
MODULE brock81
  USE et
  USE julian
  IMPLICIT NONE
  PRIVATE
  PUBLIC :: brock81_init
  REAL(dp), PARAMETER :: d15=rad*15._dp,& !< 15° in rad
       solar=1.353E3_dp,&                 !< solar constant in W/m²
       tilt=rad*23.45_dp,&                !< tilt of the earth against its orbital plane in radians
       yearDays=365.2425_dp,&             !< length of a year in days
       fd=rad*360._dp/yearDays,&          !< conversion factor from doy to rad
       cfhds=0.5_dp*cfds                  !< 12h in s
  REAL(dp) :: sdl                         !< short form for SIN(d1)*SIN(latrad) (+ SIN(ptl))
CONTAINS

  !> \brief Initialise the brock81 module.
  !!
  !! \param[in] envir%daylenfun  how to determine solar declination: 'Brock81' is without, 'Forsythe95' with twilight
  !! \param[in] envir%dicy       diurnal light cycle: '' or 'none' means constant light, 'Brock81' natural light cycle,
  !! 'MesoAqua' light cycle for MesoAqua experiments, 'LD' fixed rectangular light cycle
  SUBROUTINE brock81_init (envir, times)
    IMPLICIT NONE
    TYPE(local), INTENT(INOUT) :: envir
    TYPE(timing), TARGET, INTENT(INOUT) :: times
    times%daylength => daylength
    IF (.NOT.ASSOCIATED(times%latrad)) THEN
       ALLOCATE (times%latrad)
       times%latrad = times%latdeg*rad   ! times%latdeg from physical forcing file
    END IF
    SELECT CASE (TRIM(envir%daylenfun))
    CASE ('Brock81')
       times%declination => declination_b81
    CASE ('Forsythe95')
       times%declination => declination_f95
    CASE DEFAULT
       STOP 'Unknown daylength function in namelist envi.'
    END SELECT
    SELECT CASE (TRIM(envir%dicy)) ! set diurnal light cycle
    CASE ('', 'none')              ! constant light
       envir%PARfun => PAR_const
       times%sunday => sunday_const
       times%sun = NINT(times%end, KIND=INT64)
    CASE ('Brock81', 'brock81') ! natural light cycle after Brock (1981)
       envir%PARfun => PAR_brock81
       times%sunday => sunday_brock81
    CASE ('MesoAqua', 'mesoaqua', 'MESOAQUA') ! light cycle for MesoAqua experiments
       envir%PARfun => PAR_MesoAqua           ! (Lewandowska et al. 2014)
       times%sunday => sunday_MesoAqua
    CASE ('LD')                 ! fixed rectangular light-dark cycle
       envir%PARfun => PAR_LD
       times%sunday => sunday_LD
    CASE DEFAULT
       STOP 'Unknown light cycle (dicy) in namelist envi.'
    END SELECT
    times%sunrise => times%sun(1)
    times%sunset => times%sun(3)
  END SUBROUTINE brock81_init

  !> sunrise, noon, sunset and day length at times\%now after \cite Brock:1981
  !! \param[in] times%now     current time in s
  !! \param[in] times%latrad  latitude N in radians
  !! \param[out] times%daylen current daylength in d
  !! \param[out] times%sun    closest next sunrise, noon, sunset in UTC
  SUBROUTINE sunday_brock81 (times, env)
    IMPLICIT NONE
    CLASS(timing), INTENT(INOUT) :: times
    TYPE(local), INTENT(INOUT) :: env
    INTEGER(KIND=INT64) :: tsec, seconds
    INTEGER(KIND=4) :: dutc, year, yr, month, day
    INTEGER(4), SAVE :: y0=0, m0=0
    tsec = NINT(times%now + times%tbase, KIND=INT64)
    dutc = INT(tsec/cfds, KIND=4)
    seconds = MODULO(tsec, cfds) ! current time of day in seconds
    CALL Jul_YDofDUTC (dutc, yr, times%doy)
    CALL daylength (times)
    times%sun = NINT(times%now + times%noon*(1._dp + (/-times%daylen, 0._dp, times%daylen, 1._dp/)),&
                     KIND=INT64) - seconds
    IF (env%calcPAR) env%dayPAR = csrad(times, 12._dp)/rdrad(times, 12._dp)
    CALL Jul_YMDofDUTC (dutc, year, month, day)
    IF (year.NE.y0) THEN
       y0 = year
       WRITE (stdout,'(/,I4)', ADVANCE='NO') year
    END IF
    IF (month.NE.m0) THEN
       m0 = month
       WRITE (stdout,'(/,I2,":")', ADVANCE='NO') month
    END IF
    WRITE (stdout,'(I2)', ADVANCE='NO') day
  END SUBROUTINE sunday_brock81

  !> instantaneous surface irradiance
   SUBROUTINE PAR_brock81 (ambient, time)
    IMPLICIT NONE
    CLASS(local), INTENT(INOUT) :: ambient
    TYPE(timing), INTENT(INOUT) :: time
    ambient%sPAR = rdrad(times=time, hour=time%hour)*ambient%dayPAR
    ENTRY PAR_const (ambient, time)
  END SUBROUTINE PAR_brock81

  !> \brief sunrise, noon, sunset and daylength at times%now for MesoAqua
  !!
  !! Light cycle of the MesoAqua experiments by \cite Lewandowska:2010
  SUBROUTINE sunday_MesoAqua (times, env)
    IMPLICIT NONE
    CLASS(timing), INTENT(INOUT) :: times
    TYPE(local), INTENT(INOUT) :: env
    env%daylen = (578.4D0 + FLOOR(times%now + times%tbase))/864D2
    times%sun = NINT(times%now + times%noon*(1D0 + (/-env%daylen, 0._dp, env%daylen, 1._dp/)))
  END SUBROUTINE sunday_MesoAqua

  SUBROUTINE PAR_MesoAqua (ambient, time)
    IMPLICIT NONE
    CLASS(local), INTENT(INOUT) :: ambient
    TYPE(timing), INTENT(INOUT) :: time
    ambient%sPAR = MAX(ambient%dayPAR*MIN((1.6D0 - 3.6D0*ABS(0.5D0 - time%hour/24._dp))&
                                          /ambient%daylen - 0.6D0, 1D0), 0D0)
  END SUBROUTINE PAR_MesoAqua

  !> sunrise, noon, sunset for artificial LD cycle with constant daylength
  !! \param[in]  times%now     current time in s
  !! \param[in]  env%daylen    daylength in d
  !! \param[out] times%sun     latitude N in radians
  SUBROUTINE sunday_LD (times, env)
    IMPLICIT NONE
    CLASS(timing), INTENT(INOUT) :: times
    TYPE(local), INTENT(INOUT) :: env
    times%sun = NINT(times%now - MOD(times%now, 86.4E3_dp)&
         + times%noon*(1._dp + (/-env%daylen, 0._dp, env%daylen, 1._dp/)))
  END SUBROUTINE sunday_LD

  SUBROUTINE PAR_LD (ambient, time)
    IMPLICIT NONE
    CLASS(local), INTENT(INOUT) :: ambient
    TYPE(timing), INTENT(INOUT) :: time
    IF (ABS(1._dp - time%hour/12._dp).LE.ambient%daylen) THEN
       ambient%sPAR = ambient%dayPAR
    ELSE
       ambient%sPAR = 0._dp
    END IF
  END SUBROUTINE PAR_LD

  SUBROUTINE sunday_const (times, env)
    IMPLICIT NONE
    CLASS(timing), INTENT(INOUT) :: times
    TYPE(local), INTENT(INOUT) :: env
  END SUBROUTINE sunday_const

  !> \brief clear-sky insolation at sea level
  !!
  !! The difference between the functions for \cite Brock:1981 and
  !! \cite Forsythe:1995 is stored in sdl and times\%d1.
  !! \param[in] times%doy day of year
  !! \param[in] times%latrad latitude N in rad
  !! \param[in] times%d1 solar declination
  !! \param[in] hour time of day in decimal hours local solar time
  !! \param[in] sdl SIN(times\%latrad)*SIN(times\%d1)
  !! \return insolation in \unit{W.m^{-2}}
  FUNCTION csrad (times, hour) RESULT (it)
    IMPLICIT NONE ! ac0, ab0, k0 are base values for ac, ab, and k for 0 altitude
    REAL(dp), PARAMETER :: ac0=0.12814_dp, ab0=0.7568875_dp, k0=0.387225_dp ! (20-22)
    CLASS(timing), INTENT(IN) :: times
    REAL(dp), INTENT(IN) :: hour
    REAL(dp) :: sr, w2, ir1s, cosz, fak, ac, ab, k, tr, td, it
    w2 = d15*(hour - 12._dp)                                         ! hour-angle
    cosz = MAX(sdl + times%cdl*COS(w2), 0._dp)                       ! z = zenith angle
    ir1s = 1._dp + 0.033_dp*COS(fd*times%doy)                        ! 1/r1², r1 is the radius vector
    sr = solar*ir1s*cosz                                             ! rad. at top of atmosphere
    fak = COS(times%latrad)*(1D0 - SIGN(1D0, times%latrad)*times%d1) ! from Hottel (1976)?
    ac = ac0*(0.95_dp + 0.05_dp*fak)
    ab = ab0*(0.98_dp + 0.02_dp*fak)
    k = k0*(1.02_dp - 0.02_dp*fak)
    tr = ac + ab*EXP(-MIN(k/cosz, 700D0)) ! transmission factor  (19)
    td = 0.2710_dp - 0.2939_dp*tr         ! ratio of diffuse to extraterr. rad. (24)
    it = sr*(tr + td)
  END FUNCTION csrad

  !> \brief ratio of actual to average daily irradiance
  !!
  !! Equations 14-16 in Brock (1981) don't seem to work; thus, dr is
  !! calculated here from the ratio of Eq. 26 over Eq. 10 (without the
  !! factor 24).
  !! The difference between the functions for \cite Brock:1981 and
  !! \cite Forsythe:1995 is stored in sdl
  !! \param[in] times%doy day of the year
  !! \param[in] times%latrad latitude N in rad
  !! \param[in] hour time of day in decimal hours
  !! \return irradiance ratio
  FUNCTION rdrad (times, hour) RESULT (dr)
    IMPLICIT NONE
    CLASS(timing), INTENT(IN) :: times
    REAL(dp), INTENT(IN) :: hour
    REAL(dp) :: w2, cosz, dr
    w2 = d15*(hour - 12._dp)                                       ! hour-angle
    cosz = MAX(sdl + times%cdl*COS(w2), 0D0)                       ! z = zenith angle
    IF (times%w1.GT.0D0) THEN
       dr = pi*cosz/(times%w1*(sdl + times%cdl))
    ELSE
       dr = 0D0
    END IF
  END FUNCTION rdrad

  !> \brief solar declination after \cite Brock:1981
  !! \param[in] times%latrad latitude N in rad
  !! \param[in] times%doy day of year
  !! \param[out] times%d1 declination
  !! \param[out] sdl SIN(times\%latrad)*SIN(times\%d1)
  SUBROUTINE declination_b81 (times)
    IMPLICIT NONE
    CLASS(timing), INTENT(INOUT) :: times
    times%d1 = tilt*SIN(fd*(284._dp + times%doy)) ! declination
    sdl = SIN(times%latrad)*SIN(times%d1)
  END SUBROUTINE declination_b81

  !> \brief solar declination after Forsythe:1995
  !! \param[in] times%latrad latitude N in rad
  !! \param[in] times%doy day of year
  !! \param[in] times%ptl twilight parameter in rad
  !! \param[out] times%d1 declination
  !! \param[out] sdl SIN(times\%ptl) + SIN(times\%latrad)*SIN(times\%d1)
  SUBROUTINE declination_f95 (times)
    IMPLICIT NONE
    REAL(dp), PARAMETER :: a=0.39795_dp, b=0.2163108_dp, c=0.9671396_dp, d=pi/yearDays
    CLASS(timing), INTENT(INOUT) :: times
    times%d1 = ASIN(a*COS(b + 2._dp*ATAN(c*TAN(d*(times%doy - 186))))) ! declination
    sdl = SIN(times%ptl) + SIN(times%latrad)*SIN(times%d1)
  END SUBROUTINE declination_f95

  !> \brief daylength in decimal days
  !!
  !! \param[in] times%latrad latitude N in rad
  !! \param[in] times%doy day of year
  !! \param[out] times%daylen daylength decimal fraction
  !! \return daylength (light period) as a fraction of 24 h
  SUBROUTINE daylength (times)
    IMPLICIT NONE
    CLASS(timing), INTENT(INOUT) :: times
    REAL(dp) :: cw1
    CALL times%declination()
    times%cdl = COS(times%latrad)*COS(times%d1)
    cw1 = -sdl/(times%cdl)                           ! COS(sunset hour-angle)
    times%w1 = ACOS(SIGN(MIN(ABS(cw1), 1._dp), cw1)) ! sunset hour-angle
    times%daylen = times%w1/pi                       ! day length
  END SUBROUTINE daylength
END MODULE brock81
