!> \brief Optimal current-feeding model for zooplankton
!!
!! This module implements the model described by \cite Pahlow:2010.
!! Zooplankton groups can feed on different types of prey identified by their species names.
!! Different zooplankton groups can also be distinguished by their vertical migration behaviour.
MODULE cfo
  USE lambert
  USE et
  IMPLICIT NONE
  PRIVATE

  !> \brief Associate prey type with prey capture coefficient
  TYPE :: pcc
     CHARACTER(LEN=100) :: type !< prey type, corresponds to group%name
     TYPE(quantity) :: phi      !< prey capture coefficient
  END TYPE pcc

  !> \brief Zooplankton\n
  !! Components depsvm, v0svm, dtsvma, and (doya and doyd) or svm must be set to enable seasonal vertical migration.\n
  !! Components depdvm and vdvm must be set to enable diel vertical migration.
  TYPE, EXTENDS(fungroup), PUBLIC :: zoocfo
     REAL(dp) :: ngr,&          !< net growth rate
          Imax,&                !< max. specific ingestion rate (1/d)
          Imaf,&                !< Imax for ambush feeding
          Rm,&                  !< specific maintenance respiration (1/d)
          Rm0=0._dp,&           !< default Rm
          fhRM=1._dp,&          !< hibernation factor for Rm
          fhphi=1._dp,&         !< hibernation factor for phi
          beta=0.2_dp,&         !< digestion (assimilation) coefficient
          ca=0.1_dp,&           !< cost of assimilation coefficient
          cf0=0.1_dp, cf,&      !< cost of foraging coefficient
          d0dvm=0._dp,&         !< upper depth of diel vertical migration (m)
          depdvm=0._dp,&        !< lower depth of diel vertical migration (m)
          dtdvm=1.5_dp,&        !< time offset (before sunrise/sunset) for DVM (s)
          d0svm=0._dp,&         !< upper depth of seasonal vertical migration (m)
          depsvm=0._dp,&        !< lower depth of seasonal vertical migration (m)
          dtsvma=0._dp,&        !< temporal spread around doya (d)
          dtsvmd=0._dp,&        !< temporal spread around doyd (d)
          v0svm=0._dp,&         !< potential seasonal vertical migration velocity (m/d)
          vdvm=0._dp,&          !< diel vertical migration velocity
          vsvm0=0._dp,&         !< migration velocity when leaving the surface (m/s)
          vsvm1=0._dp,&         !< migration velocity when leaving hibernation depth (m/s)
          Emax=0.99_dp,&        !< max. assimilation efficiency
          Cmax=1._dp,&          !< maximal clearance rate (\unit{m^3.(mol.C)^{-1}})
          Af=0._dp,&            !< foraging activity
          At,&                  !< total activity
          mort=0._dp,&          !< fish mortality in the surface layer(s)
          I=0._dp,&             !< ingestion
          E=1._dp,&             !< assimilation efficiency
          R=0._dp,&             !< respiration
          g=0._dp,&             !< net growth rate
          gmax,&                !< max. net growth rate
          Pth,&                 !< feeding threshold (\unit{mmolC.m^{-3}})
          thcsvm=1.E-6_dp,&     !< threshold for svm in \unit{mmolC.m^{-3}}
          fsvm,&                !< extension factor of the migration time window
          toy,&                 !< time of year (decimal days)
          QN_param=0.16_dp,&    !< 
          QP_param=0.01_dp
     CHARACTER(LEN=512) :: egestion='det',& !< name of compartment for egestion
          DOM='DOM'                         !< name of labile DOM compartment
     INTEGER, DIMENSION(:), ALLOCATABLE :: idet, idim, imot, inot, icdet, idig
     INTEGER, POINTER :: idom(:), irda, irdd
     INTEGER :: ida=0, idd=0, isda, isdd,&
          id1dvm=0,&             !< index of day-time level for DVM
          id0svm=0,&             !< index of summer-time level for SVM
          id0svm1,&              !< id0svm + 1
          id1svm=0,&             !< index of winter-time level for SVM
          id1svm1                !< id1svm - 1
     LOGICAL :: switch=.FALSE.,& !< switching bewteen motile/non-motile prey
          dodvm=.FALSE.,&        !< diel vertical migration
          dodvm0,&               !< default for dodvm
          dosvm=.FALSE.,&        !< seasonal vertical migration
          dynsvm=.FALSE.         !< dynamic traits for seasonal vertical migration
     TYPE(pcc), DIMENSION(:), ALLOCATABLE :: pcc  !< prey capture coefficient associations
     REAL(dp), DIMENSION(:), ALLOCATABLE :: fde,& !< dissolved fraction of egestion
          fpx,&             !< particulate fraction of excretion
          IP,&              !< individual prey loss rates (prey mortalities)
          phi,&             !< food preferences -> prey capture coefficients (\unit{m^3.(mol.C)^{-1}})
          phi0,&            !< default phi for SVM
          PQ,&              !< prey cell quotas
          RQ,&              !< remineralisation of dissolved inorganic nutrients
          PP,&              !< effective prey (food concentration*phi)
          EQ,&              !< particulate egestion
          XQ,&              !< total excretion including egestion
          PPm,&             !< same as PP but for motile prey only
          PPn,&             !< same as PP but for non-motile prey only
          RFF,&             !< ratio of PP/SUM(PP), corrected for assimilation efficiencies
          phim,&            !< phis for motile prey
          phin              !< phis for non-motile prey
     REAL(dp), DIMENSION(:), POINTER :: RFFm,& !< RFF for motile prey
          RFFn,&                               !< RFF for non-motile prey
          quotas                 !< zooplankton nutrient:C ratios
     REAL(dp), POINTER :: CxDa,& !< product of biomass and doya
          fdPIC,&                !< dissolved fraction of egested PIC
          fdChl,&                !< dissolved fraction of egested Chl
          vsvm,&                 !< seasonal vertical migration velocity (m/d)
          doya,&                 !< day of year of ascent
          doyd,&                 !< day of year of descent
          CxDd                   !< product of biomass and doyd
     PROCEDURE(forage_cfo), POINTER :: forage
     PROCEDURE(excrete_DOM), POINTER :: excrete
     PROCEDURE(egest_pm), POINTER :: egest
     PROCEDURE(cfo_svm_static), POINTER :: svmig !< function for seasonal vertical migration (SVM)
     PROCEDURE(cfo_dvm), POINTER :: dvmig        !< function for diel vertical migration (DVM)
   CONTAINS
     PROCEDURE :: flux => cfo_flux
     PROCEDURE :: read => cfo_read
     PROCEDURE :: set => cfo_set
  END TYPE zoocfo

CONTAINS

  !> \brief Read namelist cfo for matching species
  SUBROUTINE cfo_read (grp, env, lun)
    IMPLICIT NONE
    CLASS(zoocfo), TARGET, INTENT(INOUT) :: grp
    TYPE(local), TARGET, INTENT(IN) :: env
    INTEGER, INTENT(IN) :: lun
    CHARACTER(LEN=512) :: species, DOM, fn, egest, forage, fT, msg, svm
    TYPE(pcc), DIMENSION(:), ALLOCATABLE :: phi
    TYPE(quantity) :: Imax, Rm, Cmax, QN, QP, Tref, mort, d0svm, depsvm, dtsvm,&
         dtsvma, dtsvmd, vsvm, doya, doyd, thcsvm, d0dvm, depdvm, dtdvm, vdvm
    REAL(dp) :: Emax, beta, ca, cf, px, Q10, fde, fdPIC, feChl, fhRM, fhphi
    LOGICAL :: motile, digest_Chl
    NAMELIST /cfo/ species, beta, ca, cf, Emax, phi, Imax, Rm, Cmax, QN, QP,&
         px, motile, forage, egest, fT, DOM, Q10, Tref, fde, fdPIC, feChl, doya, doyd,&
         d0svm, depsvm, dtsvm, dtsvma, dtsvmd, vsvm, svm, thcsvm,&
         d0dvm, depdvm, dtdvm, vdvm, fhRm, fhphi, digest_Chl, mort
    ALLOCATE (phi(env%nft+1), grp%phi(env%nft), grp%phi0(env%nft), grp%pcc(env%nft),&
              grp%IP(env%nft), grp%PP(env%nft), grp%constituents(1),&
              grp%fde(env%ncon), grp%fpx(env%ncon))
    grp%constituents = 'C'
    grp%fdPIC => grp%fde(4)
    grp%fdChl => grp%fde(5)
    grp%IP = 0D0
    DOM = grp%DOM
    species = ''
    REWIND (lun)
    DO
       fT = ''
       egest = grp%egestion
       forage = 'cfo'
       motile = .TRUE.
       beta = grp%beta
       ca = grp%ca
       cf = grp%cf0
       Emax = grp%Emax
       Cmax = quantity(grp%Cmax, 'm3 molC-1')
       phi = pcc('', quantity(0D0, 'm3 molC-1'))
       Imax = quantity(grp%Imax, 'd-1')
       Rm = quantity(grp%Rm0, 'd-1')
       QN = quantity(grp%QN_param, 'molN molC-1')
       QP = quantity(grp%QP_param, 'molP molC-1')
       Tref = quantity(grp%Tref, '°C')
       Q10 = grp%Q10
       fde = 0._dp
       fdPIC = 0._dp
       feChl = 1._dp
       fhRm = grp%fhRm
       fhphi = grp%fhphi
       digest_Chl = .TRUE.
       mort = quantity(grp%mort, 'd-1')
       doyd = quantity(-1._dp, 'd')
       doya = quantity(-1._dp, 'd')
       d0svm = quantity(grp%d0svm, 'm')
       depsvm = quantity(grp%depsvm, 'm')
       dtsvm = quantity(0._dp, 'd')
       dtsvma = quantity(grp%dtsvma, 'd')
       dtsvmd = quantity(grp%dtsvmd, 'd')
       vsvm = quantity(grp%v0svm, 'm d-1')
       thcsvm = quantity(grp%thcsvm, 'mmolC m-3')
       svm = ''
       d0dvm = quantity(grp%d0dvm, 'm')
       depdvm = quantity(grp%depdvm, 'm')
       dtdvm = quantity(grp%dtdvm, 'h')
       vdvm = quantity(grp%vdvm, 'm d-1')
       READ (lun, NML=cfo, END=100, IOMSG=msg, ERR=900)
       IF (LEN_TRIM(species).GT.100) &
            WRITE (*,'("Warning: truncating species to ",A100)') species(1:100)
       IF (TRIM(species(1:100)).EQ.TRIM(grp%name)) EXIT
    END DO
    grp%beta = beta
    grp%ca = ca
    grp%cf0 = cf
    grp%Emax = Emax
    grp%Cmax = convert(Cmax, 'm3 molC-1')
    grp%Imax = convert(Imax, 's-1')
    grp%Rm0 = convert(Rm, 's-1')
    grp%Rm = grp%Rm0
    grp%QN_param = convert(QN, 'molN molC-1')
    grp%QP_param = convert(QP, 'molP molC-1')
    grp%Q10 = Q10
    grp%fde = fde
    grp%fdPIC = fdPIC
    grp%fdChl = 1._dp - feChl
    grp%fhRm = fhRm
    grp%fhphi = fhphi
    grp%Tref = convert(Tref, '°C')
    IF (phi(env%nft+1)%phi%value.GT.0._dp) THEN
       INQUIRE (UNIT=lun, NAME=fn)
       WRITE (stderr,'("cfo_read: too many entries for phi in namelist cfo in file ",A,".")') TRIM(fn)
       STOP 1
    END IF
    phi(1:env%nft)%phi%value = phi(1:env%nft)%phi%value*grp%Cmax
    grp%pcc = phi(1:env%nft)
    grp%fpx = 1._dp
    grp%DOM = DOM
    grp%motile = motile
    grp%d0svm = convert(d0svm, 'm')
    grp%depsvm = convert(depsvm, 'm')
    IF ((dtsvm%value.GT.0._dp).AND.(dtsvma%value.LE.0._dp)) dtsvma = dtsvm
    IF ((dtsvm%value.GT.0._dp).AND.(dtsvmd%value.LE.0._dp)) dtsvmd = dtsvm
    grp%dtsvma = convert(dtsvma, 'd')
    grp%dtsvmd = convert(dtsvmd, 'd')
    grp%mort = convert(mort, 's-1')
    grp%v0svm = convert(vsvm, 'm s-1')
    grp%thcsvm = convert(thcsvm, 'mmolC m-3')
    grp%dosvm = (grp%depsvm.GT.0._dp).AND.(grp%v0svm.GT.0._dp).AND.(grp%dtsvma.GT.0._dp) &
         .AND.(((doya%value.GT.0._dp).AND.(doyd%value.GT.0._dp)).OR.(TRIM(svm).EQ.'dynamic'))
    grp%d0dvm = convert(d0dvm, 'm')
    grp%depdvm = convert(depdvm, 'm')
    grp%dtdvm = convert(dtdvm, 's')
    grp%vdvm = convert(vdvm, 'm s-1')
    grp%dodvm = (grp%depdvm.GT.0._dp).AND.(grp%vdvm.GT.0._dp)
    grp%dodvm0 = grp%dodvm      ! default setting, needed for combining DVM and SVM
    grp%egestion = TRIM(egest(1:100))
    ALLOCATE(grp%idig(3+COUNT((/digest_Chl/))))
    grp%idig(1:3) = (/1, 2, 3/)
    IF (digest_Chl) grp%idig(4) = 5 ! digest Chl in parallel with C
    SELECT CASE (TRIM(forage))
    CASE ('switch')
       grp%switch = .TRUE.
       grp%forage => forage_switch
!       grp%At = grp%Imax/grp%beta*(-1D0 - lwm1(-EXP(-1D0 - grp%beta))) ! (E21)
       grp%At = grp%Imax &                    ! total activity   (15)
         *(-1D0 - lwm1(-(1D0 - grp%cf0/grp%Emax/(1D0 - grp%ca))/EXP(1D0 + grp%beta)))
       grp%Imaf = grp%Imax
!       grp%Imaf = grp%At*grp%beta/(-1D0 - lwm1(-EXP(-1D0 - grp%beta))) ! (E21)
    CASE ('cfo')
       grp%switch = .FALSE.
       grp%forage => forage_cfo
       grp%cf = grp%cf0
       grp%At = grp%Imax &                      ! total activity (15)
         *(-1D0 - lwm1(-(1D0 - grp%cf/grp%Emax/(1D0 - grp%ca))/EXP(1D0 + grp%beta)))
       grp%Pth = -LOG((1D0 - grp%cf/grp%Emax/(1D0 - grp%ca)))*(1._dp + 1.E-15_dp) ! /grp%phi
    CASE DEFAULT
       STOP 'cfo_read: Invalid choice for forage in namelist cfo'
    END SELECT
    grp%nsv = 1
    IF (grp%dosvm) THEN
       SELECT CASE (TRIM(svm))
       CASE ('static', '')
          ALLOCATE (grp%doya, grp%doyd)
          grp%doya = convert(doya, 'd')
          grp%doyd = convert(doyd, 'd')
          grp%svmig => cfo_svm_static
          grp%vsvm => grp%v0svm
          grp%dtsvma = 0.5_dp*grp%dtsvma
          grp%dtsvmd = 0.5_dp*grp%dtsvmd
       CASE ('dynamic')
          grp%nsv = 3
          grp%ida = 2
          grp%idd = 3
          grp%svmig => cfo_svm_dyn
          grp%dynsvm = .TRUE.
          ALLOCATE (grp%vsvm)
          grp%vsvm = grp%v0svm
       CASE DEFAULT
          STOP 'cfo_read: Invalid choice for svm in namelist cfo'
       END SELECT
    ELSE
       grp%svmig => cfo_nosvm
    END IF
    IF (grp%dodvm) grp%dvmig => cfo_dvm
    CALL grp%fT_select (fT, 'cfo_read')
    RETURN
100 CONTINUE
    INQUIRE (UNIT=lun, NAME=fn)
    WRITE(stderr,FMT='("cfo/cfo_read: missing namelist cfo for species ",A," in file ",A,".")')&
         TRIM(grp%name), TRIM(fn)
    STOP 1
900 CONTINUE
    WRITE (stderr,'("cfo_read:",A)') TRIM(msg)
    STOP 1
  END SUBROUTINE cfo_read

  SUBROUTINE cfo_set (grp, grps, env)
    IMPLICIT NONE
    CLASS(zoocfo), TARGET, INTENT(INOUT) :: grp
    TYPE(state), DIMENSION(:), INTENT(IN) :: grps
    TYPE(local), TARGET, INTENT(INOUT) :: env
    INTEGER :: n, itype(1), idep(1), nm, nn
    grp%quotas => env%quotas(:,grp%ifg)
    grp%QN = grp%QN_param
    grp%QP = grp%QP_param
    ALLOCATE (grp%PQ(env%ncon), grp%RQ(env%ncon), grp%EQ(env%ncon), grp%XQ(env%ncon), grp%idim(env%ncon))
    grp%idim = (/env%idic, env%idin, env%idip/)
    NULLIFY (grp%idom, grp%egest)
    grp%phi = 0D0
    IF (grp%switch) THEN ! switch between motile and non-motile prey
       nm = COUNT((/(grps(n)%var%motile, n=1,env%nft)/)) ! No. of motile types
       nn = env%nft - nm                                 ! No. of non-motile types
       ALLOCATE (grp%RFF(env%nft), grp%imot(nm), grp%PPm(nm), grp%phim(nm),&
                 grp%inot(nn), grp%PPn(nn), grp%phin(nn))
       grp%RFFm => grp%RFF(1:nm)
       grp%RFFn => grp%RFF(nm+1:nm+nn)
       nm = 0
       nn = 0
    END IF
    IF (grp%dosvm) THEN
       idep = MINLOC(env%depth_edges(2,:), MASK=env%depth_edges(2,:).GE.grp%depsvm)
       grp%id1svm = MIN(idep(1), env%nbox) ! index of winter-time level
       IF (grp%dynsvm) THEN                ! dynamic traits (days of ascent and descent) for svm
          grp%id1svm1 = grp%id1svm - 1     ! index of winter-time level - 1
          grp%doya => grp%ratios(1)
          grp%doyd => grp%ratios(2)
          grp%irda => grp%ir(1)            ! doya in box n is box(n)%ratios(grp%irda)
          grp%irdd => grp%ir(2)
          grp%isda = grp%isv(grp%ida)
          grp%isdd = grp%isv(grp%idd)
          ! traits are stored as tracer-trait products
          grp%names(2) = TRIM(grp%name)//'_CxDa'  ! product of biomass (C) and doya
          grp%names(3) = TRIM(grp%name)//'_CxDd'  ! product of biomass (C) and doyd
          grp%units(2:3) = 'd '//TRIM(grp%units(1))
          idep = MAXLOC(env%depth_edges(1,:), MASK=env%depth_edges(1,:).LE.grp%d0svm)
          grp%id0svm = idep(1)             ! index of summer-time level
          grp%id0svm1 = grp%id0svm + 1     ! index of summer-time level + 1
          grp%vsvm0 = 10._dp*env%depth_edges(2,grp%id0svm)/grp%dtsvmd/cfds
          grp%vsvm1 = 10._dp*env%heights(1,grp%id1svm)/grp%dtsvma/cfds
          grp%fsvm = grp%depsvm/grp%vsvm/grp%dtsvma/cfds + 1._dp
       END IF
    END IF
    IF (grp%dodvm) THEN
       idep = MINLOC(env%depth_edges(2,:), MASK=env%depth_edges(2,:).GE.grp%depdvm)
       grp%id1dvm = MIN(idep(1), env%nbox) ! index of day-time level
       idep = MAXLOC(env%depth_edges(1,:), MASK=env%depth_edges(1,:).LE.grp%d0dvm)
       grp%d0dvm = MAX(grp%d0dvm, env%depth(1)) ! no upward motion in the surface layer
    END IF
    env%smsk(grp%isv(1),(/grp%idim, env%io2/),:) = .TRUE.
    DO n=1,env%nft
       IF (ANY(grps(n)%var%name.EQ.grp%pcc%type)) THEN ! associate phi with prey type
          itype = PACK([(nn, nn = 1, env%nft)], MASK=grps(n)%var%name.EQ.grp%pcc%type)
          grp%phi(n) = convert(grp%pcc(itype(1))%phi, 'm3 mmolC-1')
       END IF
       IF (TRIM(grp%egestion).EQ.TRIM(grps(n)%var%name)) THEN ! egest particulate matter to detritus
          ALLOCATE(grp%idet(grps(n)%var%ncon), grp%icdet(grps(n)%var%ncon))
          grp%icdet = grps(n)%var%icon
          grp%idet = grps(n)%var%isv
          grp%egest => egest_pm
       END IF
       IF (TRIM(grp%DOM).EQ.TRIM(grps(n)%var%name)) grp%idom => grps(n)%var%isv
       IF (grp%switch) THEN
          IF (grps(n)%var%motile) THEN
             nm = nm + 1
             grp%imot(nm) = n
             grp%phim(nm) = grp%phi(n)
          ELSE
             nn = nn + 1
             grp%inot(nn) = n
             grp%phin(nn) = grp%phi(n)
          END IF
       END IF
       env%smsk(grp%isv(1),grps(n)%var%isv,:) = .TRUE.
    END DO
    IF (.NOT.ASSOCIATED(grp%egest)) grp%egest => egest_dm
    grp%phi0 = grp%phi
    IF (ASSOCIATED(grp%idom)) THEN
       grp%excrete => excrete_dom
       env%smsk(grp%isv(1),grp%idom,:) = .TRUE.
    ELSE
       grp%excrete => excrete_dim
       env%smsk(grp%isv(1),grp%idim,:) = .TRUE.
    END IF
  END SUBROUTINE cfo_set

  !> flux routine for the cfo module
  SUBROUTINE cfo_flux (grp, grps, env, box, times)
    IMPLICIT NONE
    CLASS(zoocfo), INTENT(INOUT) :: grp
    TYPE(state), DIMENSION(:), INTENT(IN) :: grps
    TYPE(local), INTENT(INOUT) :: env
    TYPE(layer), INTENT(INOUT) :: box(0:)
    TYPE(timing), INTENT(IN) :: times
    INTEGER :: n
    IF (box(1)%nb.EQ.grp%id1svm) THEN
       grp%Rm = grp%fhRm*grp%Rm0    ! reduced maintenance loss at hibernation depth
       grp%phi = grp%fhphi*grp%phi0 ! no feeding activity at hibernation depth
    ELSE
       grp%Rm = grp%Rm0
       grp%phi = grp%phi0
    END IF
    CALL grp%forage (env, box(1))
    CALL grp%svmig (box=box, times=times)                   ! seasonal vertical migration (SVM)
    IF (grp%dodvm) CALL grp%dvmig (box=box(1), times=times) ! diel vertical migration (DVM)
    ! store all fluxes in row isv(1) of the flux matrix
    box(1)%soma(grp%isv(1),grp%idim) = grp%RQ(1:3)   ! release of inorganic nutrients (DIC, DIN, DIP)
    DO n = 1,env%nft
       box(1)%soma(grp%isv(1),grps(n)%var%isv) = -grp%IP(n)*box(1)%stv(grps(n)%var%isv)
    END DO
    box(1)%soma(grp%isv(1),grp%isv(1)) = box(1)%soma(grp%isv(1),grp%isv(1)) + grp%ngr
    CALL grp%excrete (env, box(1))
    CALL grp%egest (box(1))
    ! add the dissolved fraction of released PIC to DIC
    box(1)%soma(grp%isv(1),grp%idim(1)) = box(1)%soma(grp%isv(1),grp%idim(1)) + grp%fdPIC*grp%EQ(4)
    box(1)%soma(grp%isv(1),env%io2) = -grp%R*(env%RQ + grp%QN*2D0) ! O2 consumption
  END SUBROUTINE cfo_flux

  !> \brief Vertical velocity for diel vertical migration
  !!
  !! \param[in] zoo%id1dvm    index of day-time level
  !! \param[in] zoo%d0dvm     night-time depth of diel vertical migration
  !! \param[in] zoo%depdvm    day-time target depth of diel vertical migration
  !! \param[in] zoo%dtdvm     time window of diel vertical migration
  !! \param[in] box%depth     depth of current layer
  !! \param[in] times%current current time
  !! \param[in] times%sunrise sunrise of current day in s
  !! \param[in] times%sunset  suset of current day in s
  !! \param[out] zoo%vv       vertical velocity in \unit{m.s^{-1}}
  SUBROUTINE cfo_dvm (zoo, box, times)
    IMPLICIT NONE
    CLASS(zoocfo), INTENT(INOUT) :: zoo
    TYPE(layer), INTENT(IN) :: box
    TYPE(timing), INTENT(IN) :: times
    REAL(dp) :: xdown, xup
    IF (box%nb.LT.zoo%id1dvm) THEN
       xdown = MAX(zoo%dtdvm - ABS(times%current - times%sunrise), 0._dp)**8
       xup = MAX(zoo%dtdvm - ABS(times%current - times%sunset), 0._dp)**8
    ELSE
       xdown = 0._dp
       xup = 0._dp
    END IF
    zoo%vv = zoo%vdvm*(xdown/(xdown + cfds) - &
         xup/(xup + cfds)*MIN(2._dp*MAX(1._dp - zoo%d0dvm/box%depth, 0._dp), 1._dp))
  END SUBROUTINE cfo_dvm

  !> \brief Vertical velocity for seasonal vertical migration and days of
  !! ascent and descent
  !!
  !! Subroutine cfo_svm_dyn implements the days of ascent and descent (doya and
  !! doyd) as dynamic traits.
  !! Animals leaving their target layer (depsvm for the ascent and d0svm for
  !! the descent) take the current day with them as the corresponding trait.
  !! This is implemented here by adding a correction to the corresponding
  !! source term in the adjacent layer in the direction of movement.
  !!
  !! \param[in] zoo%dtsvma width of time window for seasonal vertical ascent
  !! \param[in] zoo%dtsvmd width of time window for seasonal vertical descent
  !! \param[in] zoo%depsvm winter depth for seasonal vertical migration
  !! \param[in] zoo%d0svm  summer depth for seasonal vertical migration
  !! \param[in] box%depth  mid-depth of current layer
  !! \param[in] box%height height of current layer
  SUBROUTINE cfo_svm_dyn (zoo, box, times)
    IMPLICIT NONE
    CLASS(zoocfo), INTENT(INOUT) :: zoo
    TYPE(layer), INTENT(IN) :: box(0:)
    TYPE(timing), INTENT(IN) :: times
    REAL(dp), SAVE :: fda, fdd=0._dp
    ! Zooplankton growth should have no effect on doya and doyd, so the
    ! corresponding trait-tracer products must change at the same relative rate
    ! (ngr) as zooplankton biomass.
    box(1)%soma(zoo%isda,zoo%isda) = zoo%ngr*zoo%doya                 ! (2.14)
    box(1)%soma(zoo%isdd,zoo%isdd) = zoo%ngr*zoo%doyd
    ! When the animals leave their target svm layers (boxes), id0svm in summer
    ! and id1svm in winter, they take the current time with them as their doya
    ! (for the ascent) or doyd (for the descent).  This is implemented here by
    ! modifying the vertical velocity of the tracer-trait products within the
    ! surface and hibernation ranges during migration, so that effectively the
    ! SVM traits transport the current time of year.
    IF (box(1)%nb.LT.zoo%id1svm) THEN            ! surface and migration ranges
       fda = (times%toy - box(2)%ratios(zoo%irda))/zoo%dtsvma + 0.5_dp
       IF ((fda.GT.0._dp).AND.(fda.LE.zoo%fsvm)) THEN       ! ascent
          IF (box(1)%nb.LE.zoo%id0svm) THEN                 ! surface range
             zoo%vv = -0.5_dp*zoo%vsvm0
          ELSEIF (box(1)%nb.LT.zoo%id1svm1) THEN            ! migration range
             zoo%vv = -zoo%vsvm
          ELSEIF ((box(1)%nb.EQ.zoo%id1svm1).AND.(fda.LE.1._dp)) THEN ! above hibernation layer
             zoo%vv = -fda*zoo%vsvm1                        ! velocity out of hibernation layer
             zoo%vv(zoo%ida) = times%toy/box(2)%ratios(zoo%irda)*zoo%vv(1)
          ELSE
             zoo%vv = 0._dp
          END IF
       ELSE
          ! the fdd used here is from the previous call of this subroutine, one
          ! layer above
          IF ((box(1)%nb.EQ.zoo%id0svm1).AND.(fdd.GT.0._dp).AND.(fdd.LE.1.0_dp)) THEN
             box(0)%soma(zoo%isdd,zoo%isdd) = box(0)%soma(zoo%isdd,zoo%isdd) &
                  + box(0)%bbc(zoo%isdd)*(times%toy/box(0)%ratios(zoo%irdd) - 1._dp)&
                  *fdd*zoo%vsvm0/box(0)%height
          END IF
          fdd = (times%toy - zoo%doyd)/zoo%dtsvmd + 0.5_dp
          IF ((fdd.GT.0._dp).AND.(fdd.LE.zoo%fsvm)) THEN    ! descent
             IF (box(1)%nb.GT.zoo%id0svm) THEN              ! migration range
                zoo%vv = zoo%vsvm
             ELSEIF (fdd.LE.1._dp) THEN                     ! surface range
                zoo%vv = fdd*zoo%vsvm0
                IF (box(1)%nb.EQ.zoo%id0svm) zoo%vv(zoo%idd) = times%toy/zoo%doyd*zoo%vv(1)
             ELSE
                zoo%vv = 0._dp
             END IF
          ELSE
             zoo%vv = 0._dp
          END IF
       END IF
    ELSEIF (box(1)%nb.EQ.zoo%id1svm) THEN                ! hibernation layer
       fda = (times%toy - zoo%doya)/zoo%dtsvma + 0.5_dp
       IF ((fda.GT.0._dp).AND.(fda.LE.1._dp)) THEN          ! ascent
          box(1)%soma(zoo%isda,zoo%isda) = box(1)%soma(zoo%isda,zoo%isda) &
               + box(0)%bbc(zoo%isda)*(times%toy/zoo%doya - 1._dp)&
               *fda*zoo%vsvm1/box(1)%height
       END IF
       zoo%vv = 0._dp
    ELSE
       zoo%vv = 0._dp
    END IF
    zoo%dodvm = zoo%dodvm0.AND.& ! no DVM during hibernation
         (times%toy.GT.(zoo%doya + zoo%dtsvma)).AND.(times%toy.LT.(zoo%doyd - zoo%dtsvmd))
  END SUBROUTINE cfo_svm_dyn

  !> \brief Vertical velocities and days of ascent and descent
  !!
  !! Subroutine cfo_svm_static is for static days of ascent and descent, i.e., doya
  !! and doyd are parameters in the corresponding cfo namelist.
  !! \param[in] zoo%doya   day of year of ascent
  !! \param[in] zoo%doyd   day of year of descent
  !! \param[in] zoo%dtsvma width of time window for seasonal vertical ascent
  !! \param[in] zoo%dtsvmd width of time window for seasonal vertical descent
  !! \param[in] zoo%depsvm winter depth for seasonal vertical migration
  !! \param[in] zoo%d0svm  summer depth for seasonal vertical migration
  !! \param[in] box%depth  mid-depth of current layer
  !! \param[in] box%height height of current layer
  SUBROUTINE cfo_svm_static (zoo, box, times)
    IMPLICIT NONE
    CLASS(zoocfo), INTENT(INOUT) :: zoo
    TYPE(layer), INTENT(IN) :: box(0:)
    TYPE(timing), INTENT(IN) :: times
    REAL(dp) :: fdoy
    ! Svm motions start at doya/doyd - dtsvm and end at doya/doyd + dtsvm.
    ! Svm ascent slows down before reaching d0svm and is about 0 above.
    ! The function is based on the symmetric part of (6) in Moisan et al. (2002).
    fdoy = ((times%toy - zoo%doyd)/zoo%dtsvmd)**8 + (box(1)%depth/zoo%depsvm)**16
    zoo%vv = 0._dp
    IF (fdoy.LT.200._dp) THEN
       zoo%vv = zoo%vsvm*EXP(-fdoy)                             ! descent (positive)
    ELSE
       fdoy = ((times%toy - zoo%doya)/zoo%dtsvma)**8 &
            + ((box(1)%depth - zoo%depsvm - zoo%d0svm)/(zoo%depsvm - zoo%d0svm/2._dp))**4
       IF (fdoy.LT.200._dp) zoo%vv = -zoo%vsvm*EXP(-fdoy)       ! ascent (negative)
    END IF
    IF (box(1)%nb.EQ.zoo%id1svm) THEN
       zoo%Rm = zoo%fhRm*zoo%Rm0    ! reduced maintenance loss at hibernation depth
       zoo%phi = zoo%fhphi*zoo%phi0 ! no feeding activity at hibernation depth
    ELSE
       zoo%Rm = zoo%Rm0
       zoo%phi = zoo%phi0
    END IF
    zoo%dodvm = zoo%dodvm0.AND.& ! no DVM during hibernation
         (times%toy.GT.(zoo%doya + zoo%dtsvma)).AND.(times%toy.LT.(zoo%doyd - zoo%dtsvmd))
    ENTRY cfo_nosvm (zoo, box, times)
  END SUBROUTINE cfo_svm_static

  SUBROUTINE excrete_dom (zoo, env, box)
    IMPLICIT NONE
    CLASS(zoocfo), INTENT(INOUT) :: zoo
    TYPE(local), INTENT(INOUT) :: env
    TYPE(layer), INTENT(INOUT) :: box
    REAL(dp) :: add_DOC
    box%soma(zoo%isv(1),zoo%idom) = zoo%EQ(1:3)*zoo%fde(1:3)
    ! mobilize refractory DOC by adsorption of freshly-excreted DON, in proportion to the
    ! refractory DOC:DON ratio:
    add_DOC = zoo%EQ(2)*zoo%fde(2)*box%stv(env%irdoc)/MAX(box%stv(env%irdon), 1.E-30_dp)
    box%soma(env%irdoc,zoo%idom(1)) = box%soma(env%irdoc,zoo%idom(1)) + add_DOC
    box%soma(zoo%idom(1),env%irdoc) = box%soma(zoo%idom(1),env%irdoc) - add_DOC
  END SUBROUTINE excrete_dom

  SUBROUTINE excrete_dim (zoo, env, box)
    IMPLICIT NONE
    CLASS(zoocfo), INTENT(INOUT) :: zoo
    TYPE(local), INTENT(INOUT) :: env
    TYPE(layer), INTENT(INOUT) :: box
    ! PIC release is accounted for in cfo_flux and there is no inorganic Chl
    box%soma(zoo%isv(1),zoo%idim) = box%soma(zoo%isv(1),zoo%idim) + zoo%EQ(1:3)*zoo%fde(1:3)
  END SUBROUTINE excrete_dim

  SUBROUTINE egest_dm (zoo, box)
    IMPLICIT NONE
    CLASS(zoocfo), INTENT(IN) :: zoo
    TYPE(layer), INTENT(INOUT) :: box
    ! add egested matter to associated inorganic nutrients
    box%soma(zoo%isv(1),zoo%idim) = box%soma(zoo%isv(1),zoo%idim) + zoo%EQ(1:3)*(1D0 - zoo%fde(1:3))
  END SUBROUTINE egest_dm

  SUBROUTINE egest_pm (zoo, box)
    IMPLICIT NONE
    CLASS(zoocfo), INTENT(IN) :: zoo
    TYPE(layer), INTENT(INOUT) :: box
    ! add egested particles to associated detritus
    box%soma(zoo%isv(1),zoo%idet) = box%soma(zoo%isv(1),zoo%idet) &
         + zoo%EQ(zoo%icdet)*(1D0 - zoo%fde(zoo%icdet))
  END SUBROUTINE egest_pm

  SUBROUTINE forage_cfo (zoo, env, box)
    IMPLICIT NONE
    CLASS(zoocfo), INTENT(INOUT) :: zoo !< functional type
    TYPE(local), INTENT(IN) :: env   !< global environment
    TYPE(layer), INTENT(IN) :: box   !< local environment
    REAL(dp) :: Rm, fQ, EI, SPP, XC, mort
    zoo%PP = zoo%phi*env%OC     ! effective prey (food concentration*phi)
    SPP = MAX(SUM(zoo%PP), 1.E-30_dp)
    zoo%PQ = MATMUL(env%quotas, zoo%PP/SPP) ! prey cell quotas
    ! fQ accounts for different N:C ratios in predator and prey, after
    ! Kiørboe (1989).  C Ingestion is unaffected by food C:N ratio according
    ! to Kiørboe (1989), so Af should not be affected either.  Thus, food
    ! C:N should influence only R and ngr.  The same concept is applied here
    ! also to C:P.
    fQ = 1._dp
    IF (SPP.GT.1.E-30_dp) fQ = MINVAL(zoo%PQ(1:3)/zoo%quotas(1:3))
    ! If it is assumed that the additional respiration can be used to increase
    ! the prey capture coefficient, this increases the effective strength of the
    ! feeding current.  The choice of the functional relation (fphi) is arbitrary
    ! as no information is available other than that N growth efficiency and
    ! maximum ingestion should not be affected.
    !    fphi = EXP(1D0 - fQ)
    CALL foract (zoo, box, PC=SPP, EI=EI, Rm=Rm)
    mort = box%fish*zoo%mort*zoo%OC     ! fish mortality in the surface layer(s)
    zoo%ngr = (EI*(1D0 - zoo%ca) - zoo%cf*zoo%Af - Rm)*fQ - mort ! net growth  (9)
    zoo%R  = EI - zoo%ngr - mort                                 ! respiration (2)
    XC = zoo%I - EI + mort              ! particulate egestion (faecal pellets) + fish mortality
    zoo%fpx(zoo%idig) = XC/MAX(XC + zoo%R, 1D-50) ! ratio of egestion : metabolic losses
    zoo%XQ = zoo%I*zoo%PQ - zoo%ngr*zoo%quotas
    zoo%RQ = zoo%XQ*(1._dp - zoo%fpx)   ! remineralisation
    zoo%EQ = zoo%XQ*zoo%fpx             ! particulate egestion
    zoo%IP = zoo%I*zoo%phi/SPP          ! prey loss rates in 1/s
  END SUBROUTINE forage_cfo

  SUBROUTINE forage_switch (zoo, env, box)
    IMPLICIT NONE
    CLASS(zoocfo), INTENT(INOUT) :: zoo
    TYPE(local), INTENT(IN) :: env
    TYPE(layer), INTENT(IN) :: box
    REAL(dp) :: I0m, Rm, fQ, Em, EI, EIm, EIn, SPPm, SPPn, Im, In, XC
    zoo%PP = zoo%phi*env%OC     ! effective prey (food concentration*phi)
    zoo%PPm = zoo%PP(zoo%imot)  ! effective motile prey
    zoo%PPn = zoo%PP(zoo%inot)  ! effective non-motile prey
    SPPm = SUM(zoo%PPm)      
    SPPn = SUM(zoo%PPn)        
    I0m = 1D0 - EXP(-SPPm)      ! ingestion saturation for motile prey
    Em = zoo%Emax*(1D0 - EXP(-zoo%beta*(1D0/MAX(I0m, 1D-4) - 1D0)))
    zoo%cf = zoo%cf0 + Em*I0m*(1D0 - zoo%ca)
    zoo%Pth = -LOG((1D0 - zoo%cf/zoo%Emax/(1D0 - zoo%ca)))*(1._dp + 1.E-15_dp) ! /zoo%phi
    CALL foract (zoo, box, SPPn, EIn, Rm) ! current-feeding activity
    In = zoo%I
    Im = (zoo%Imaf - zoo%Af)*I0m          ! motile-prey ingestion (E15)
    zoo%I = In + Im
    EIm = Em*Im
    EI = EIn + EIm
    zoo%E = EI/zoo%I
    ! fQ accounts for different N:C ratios in predator and ingested prey, after
    ! Kiørboe (1989).  C Ingestion is unaffected by food C:N ratio according
    ! to Kiørboe (1989), so Af should not be affected either.  Thus, food
    ! C:N should influence only R and ngr.  The same concept is applied here
    ! also to C:P.
    zoo%RFFm = EIm*zoo%PPm/SPPm/MAX(EI, 1D-30)
    zoo%RFFn = EIn*zoo%PPn/SPPn/MAX(EI, 1D-30)
    zoo%PQ = MATMUL(env%quotas(:,(/zoo%imot, zoo%inot/)), zoo%RFF)
    fQ = MINVAL(zoo%PQ(1:3)/zoo%quotas(1:3))
    zoo%IP(zoo%imot) = Im*zoo%phim/SPPm ! motile prey loss rates in 1/s
    zoo%IP(zoo%inot) = In*zoo%phin/SPPn ! non-motile prey loss rates in 1/s
    zoo%ngr = (EI*(1D0 - zoo%ca) - zoo%cf0*zoo%Af - Rm)*fQ ! net growth  (2.2)
    zoo%R  = EI - zoo%ngr                                  ! respiration (2.8)
    XC = zoo%I - EI                               ! particulate egestion (faecal pellets)
    zoo%fpx(zoo%idig) = XC/MAX(XC + zoo%R, 1D-50) ! ratio of egestion : metabolic losses
    zoo%XQ = zoo%I*zoo%PQ - zoo%ngr*zoo%quotas
    zoo%RQ = zoo%XQ*(1._dp - zoo%fpx)   ! remineralisation
    zoo%EQ = zoo%XQ*zoo%fpx             ! particulate egestion
  END SUBROUTINE forage_switch
  
  SUBROUTINE foract (zoo, box, PC, EI, Rm)
    IMPLICIT NONE
    CLASS(zoocfo), INTENT(INOUT) :: zoo
    TYPE(layer), INTENT(IN) :: box
    REAL(dp), INTENT(IN) :: PC
    REAL(dp), INTENT(OUT) :: EI, Rm
    REAL(dp) :: At, fTC, I0, rAtf
    fTC = zoo%OC*zoo%fT(box)
    Rm = zoo%Rm*fTC
    IF (PC.LE.zoo%Pth) THEN     ! food below feeding threshold
       zoo%Af = 0D0
       zoo%E = zoo%Emax
       zoo%I = 0D0
    ELSE
       I0 = 1D0 - EXP(-PC)      ! ingestion saturation for non-motile prey (2.3)
       At = fTC*zoo%At          ! temperature dependence
       ! rAtf is the ratio At/Af; this is needed to prevent division by 0 when
       ! zoo%OC = 0 in the calculation of zoo%E; \beta was removed from At above
       rAtf = (-1D0 - lwm1(-(1D0 - zoo%cf/(zoo%Emax*(1 - zoo%ca)*I0)) &  ! (2.7)
                            /EXP(1D0 + zoo%beta)))
       ! Af: foraging activity
       zoo%Af = At/rAtf
       ! E: assimilation efficiency; note that \beta was removed from At above
       zoo%E = zoo%Emax*(1D0 - EXP(zoo%beta - rAtf))                     ! (2.2)
       zoo%I = I0*zoo%Af        ! non-motile ingestion (E14)
    END IF
    EI = zoo%E*zoo%I
  END SUBROUTINE foract

END MODULE cfo
