!> \brief process command-line options
MODULE clops
  IMPLICIT NONE
CONTAINS
  !> \brief sort arguments according to an option string
  !!
  !! option | meaning
  !! ------ | -------
  !! \-<x>  | option <x> with argument or switch <x>
  !! \-?    | prints a usage message
  !! \--    | signals the end of options; arguments without option letters are left unsorted
  !!
  !! \param[in]    opts      string with option letters; a : must be included
  !!   to signal the start of switches, if left out, all letters are switches
  !! \param[inout] args      string array to hold arguments; the first n
  !!   elements will be the arguments to the options in opts, with n = length of opts
  !! \param[out]   flags     logical array to indicate which switches are present
  !! \param[in]    usage     message to display for invalid input
  !! \param[in]    help      help message displayed for -?
  SUBROUTINE getopt (opts, args, flags, usage, help)
    IMPLICIT NONE
    CHARACTER(LEN=*), INTENT(IN) :: opts
    CHARACTER(LEN=*), INTENT(IN) :: usage
    CHARACTER(LEN=*), INTENT(IN), OPTIONAL :: help
    CHARACTER(LEN=*), DIMENSION(:), INTENT(INOUT) :: args
    LOGICAL, DIMENSION(:), INTENT(OUT) :: flags
    INTEGER :: i, l, nopts, no, nargc, nparg, npargs, mxlen
    CHARACTER (LEN=LEN(args)) :: argv, opt
    LOGICAL :: cov=.TRUE.
    mxlen = LEN(args)
    flags = .FALSE.
    nparg = 0
    no = 0
    nargc = COMMAND_ARGUMENT_COUNT()
    nopts = MAX(SCAN(opts, ':') - 1, 0) ! number of arguments with values
    npargs = SIZE(args) - MAX(nopts,0)  ! max. number of positional arguments
    DO WHILE (no.LT.nargc)
       no = no + 1
       CALL GET_COMMAND_ARGUMENT (number=no, value=argv)
       IF (cov.AND.(argv(1:1).EQ.'-')) THEN
          i = SCAN(opts, argv(2:2))
          IF (i.GT.nopts) THEN         ! flag argument (no value)
             flags(i-nopts-1) = .TRUE.
          ELSEIF (i.GT.0) THEN
             l = LEN_TRIM(argv) - 2
             IF (l.EQ.0) THEN
                no = no + 1            ! next argument is the value
                CALL GET_COMMAND_ARGUMENT (number=no, value=opt)
                l = MIN(LEN_TRIM(opt), mxlen)
             ELSE                      ! value attached to option letter
                opt = argv(3:l+2)
             END IF
             args(i) = opt(1:l)
          ELSE
             IF (SCAN(argv(2:2),'hH?').GT.0) THEN
                IF (PRESENT(help)) THEN
                   WRITE (*,help) usage//' | -?'; STOP
                ELSE
                   WRITE (*,'("Usage: ",A," | -?")') usage
                END IF
             END IF
             cov = argv(2:2).NE.'-'     ! -- stops option processing; remaining
             IF (cov) THEN              ! arguments will be appended to args
                WRITE (*,'("Unrecognized option ",A2,"."/"Usage: ",A," | -?")')&
                     argv(1:2), usage
                STOP 1
             END IF
          END IF
       ELSEIF (nparg.LT.npargs) THEN
          l = MIN(LEN_TRIM(argv), mxlen)
          nparg = nparg + 1
          args(nparg+nopts) = argv(1:l)
       ELSE
          WRITE (*,'("Usage: ",A)') usage//' | -?'; STOP 1
       END IF
    END DO
    IF (nparg.GT.npargs) THEN
       WRITE  (*,'("Usage: ",A)') usage//' | -?'; STOP 1
    END IF
  END SUBROUTINE getopt
END MODULE clops
