!> \brief Chain-model of optimal phytoplankton growth and diazotrophy
!!
!! This module implements the model described by \cite Pahlow:2013a.
MODULE cmo
  USE logarithmic_integral, ONLY : inloin, dei
  USE lambert
  USE et
  IMPLICIT NONE
  PRIVATE
  !> Ordinary phytoplankton and diazotrophs
  !!
  !! if fF = 0, it is ordinary phytoplankton, if fF = 1, it is a diazotroph
  TYPE, EXTENDS(fungroup), PUBLIC :: phycmo
     REAL(dp) :: A0,&           !< nutrient affinity in \unit{m^3.mol^{-1}.d^{-1}}
          alpha,&               !< light affinity in \unit{m^2.d^{-1}.mol.W^{-1}.gChl^{-1}}
          alphaT,&              !< temperature-dependent alpha
          QsN=0.02,&            !< structure-bound N quota in \unit{mol.mol^{-1}}
          rdl=1.0,&             !< day-length parameter
          RC=0.1_dp,&           !< Chl maintenance respiration in \unit{d^{-1}}
          RCT,&                 !< temperature-dependent RC
          V0=5._dp,&            !< maximum rate parameter in \unit{d^{-1}}
          F0N=0._dp,&           !< maximum \ce{N2} fixation in \unit{molN.molC^{-1}.d^{-1}}
          zC,&                  !< cost of photosynthesis in \unit{mol.gChl^{-1}}
          zN,&                  !< cost of biosynthesis in \unit{mol.mol^{-1}}
          zNu=0.6_dp,&          !< cost of biosynthesis for DIN use
          zNf=2._dp,&           !< cost of biosynthesis for \ce{N2} fixation
          A,&                   !< short form of light-dependent terms
          mu0,&                 !< potential rate of C fixation in \unit{d^{-1}}
          VPh0,&                !< local rate of P uptake
          sVPh,&                !< \ff{\sqrt{\widehat{V}^{\text{P}}_{0}}}
          fV,&                  !< N allocation to nutrient acquisition
          fN,&                  !< local N allocation to N acquisition
          fC,&                  !< N allocation to C fixation
          fTemp,&               !< temperature factor
          fPIC=0._dp,&          !< PIC factor: VPIC (calcification) = fPIC * VDIC (\ce{CO2} fixation)
          fF=0._dp,&            !< N allocation to \ce{N2} fixation (0 for ordinary phytoplankton, 1 for diazotrophs)
          SI,&                  !< degree of light saturation
          tch,&                 !< local (chloroplast) Chl:C ratio in \unit{g.mol^{-1}}
          rctht,&               !< rate of change of the Chl:C ratio
          chl0=-1.0_dp,&        !< initial Chl (for Alex' and Steffi's data)
          VN, VP,&              !< rates of N, P uptake
          N2F=0._dp,&           !< rate of \ce{N2} fixation
          VPIC,&                !< PIC production (calcification)
          vsink=0._dp,&         !< sinking velocity in \unit{m.d^{-1}}
          fcdtdq=0._dp          !< \ff{\displaystyle\frac{f^{\text{C}}}{\theta}\dv{\theta}{Q^\text{N}}}
     REAL(dp), POINTER :: daylen,& !< daylength as a fraction of 24h
          dlfA,&                   !< day length factor for A
          PAR,&                    !< irradiance in \unit{mol.m^{-2}.d^{-1}}
          Q0N, Q0P,&               !< subsistence N, P quotas
          QPIC,&                   !< PIC:POC ratio
          Chl,&                    !< Chl concentration in \unit{gChl.m^{-3}}
          fTalpha,&                !< temperature factor for alpha
          fTRC,&                   !< temperature factor for RC
          theta,&                  !< Chl:C ratio in \unit{gChl.mol^{-1}}
          VDOC,&                   !< DOC loss (\unit{VDOC \le 0}) under unbalanced growth
          ngr,&                    !< rate of N, P, C acquisition
          VDIC                     !< \ce{CO2} fixation
     REAL(dp), ALLOCATABLE :: VC(:)
     CHARACTER(LEN=100) :: DOM='DOM' !< name of labile DOM compartment
     INTEGER, ALLOCATABLE :: ic(:)
     INTEGER :: ichl, idoc, ipic
     LOGICAL :: dynamicChl=.TRUE., PIC=.FALSE.
     PROCEDURE(ftnf_Eppley), POINTER :: fTNF => ftnf_Eppley !< temperature function for \ce{N2} fixation
     PROCEDURE(grow), POINTER :: growth    !< growth function
     PROCEDURE(uptake), POINTER :: nuts    !< nutrient uptake function
     PROCEDURE(chl_dyn), POINTER :: chldyn !< light acclimation
     PROCEDURE(tchdyn), POINTER :: pachl   !< chloroplast Chl:C ratio
     PROCEDURE(cmo_pic), POINTER :: calc   !< calcification
   CONTAINS
     PROCEDURE :: flux => cmo_flux
     PROCEDURE :: read => cmo_read
     PROCEDURE :: set => cmo_set
  END TYPE phycmo

CONTAINS

  !> Read parameters for optimality-based chain model
  SUBROUTINE cmo_read (grp, env, lun)
    IMPLICIT NONE
    CLASS(phycmo), TARGET, INTENT(INOUT) :: grp !< funcional type
    TYPE(local), TARGET, INTENT(IN) :: env   !< local environment
    INTEGER, INTENT(IN) :: lun  !< logical unit number for namelist cmo
    CHARACTER(LEN=512) :: DOM, species, msg
    REAL(dp) :: rdl, fF, fPIC, Q10
    TYPE(quantity) :: A0, alpha, Q0N, Q0P, RC, V0, F0N, zC, zN, zF,&
         Tref, Topt, sticky, chl0, vsink
    CHARACTER(LEN=15) :: pa, fT, FTNF
    INTEGER :: ncon
    LOGICAL :: fTRC, fTalpha, PIC
    NAMELIST /cmo/ A0, alpha, Q0N, Q0P, rdl, RC, V0, F0N, zC, zN, zF, fF, sticky,&
         Tref, Topt, pa, fPIC, fT, fTNF, Q10, DOM, fTRC, fTalpha, PIC, species, vsink
    CALL inloin   ! initialise dei function from logarithmic integral module
    species = ''
    DOM = grp%DOM
    REWIND(lun)
    DO
       A0 = quantity(grp%A0, 'm3 molC-1 d-1')
       alpha = quantity(grp%alpha, 'm2 molC W-1 d-1 gChl-1')
       Q0N = quantity(2._dp*grp%QsN, 'molN molC-1')
       Q0P = quantity(0.001_dp, 'molP molC-1')
       rdl = grp%rdl
       fF = grp%fF
       sticky = quantity(grp%sticky, '')
       RC = quantity(grp%RC, 'd-1')
       V0 = quantity(grp%V0, 'd-1')
       F0N = quantity(grp%F0N, 'molN molC-1 d-1')
       zC = quantity(grp%zC, 'molC gChl-1')
       zN = quantity(grp%zNu, 'molC molN-1')
       zF = quantity(grp%zNf, 'molC molN-1')
       Tref = quantity(grp%Tref, '°C')
       Topt = quantity(grp%Topt, '°C')
       chl0 = quantity(grp%Chl0, 'mg m-3')
       vsink = quantity(grp%vsink, 'm d-1')
       fPIC = grp%fPIC
       Q10 = grp%Q10
       pa = 'dynamic'
       fT = ''
       fTNF = ''
       fTRC = .TRUE.
       fTalpha = .FALSE.
       PIC = .FALSE.
       READ (lun, NML=cmo, END=100, IOMSG=msg, ERR=900)
       IF (TRIM(species(1:100)).EQ.TRIM(grp%name)) EXIT
    END DO
    grp%A0 = convert(A0, 'm3 mmolC-1 s-1')
    grp%alpha = convert(alpha, 'm2 molC W-1 s-1 gChl-1')
    grp%QsN = 0.5_dp*convert(Q0N, 'molN molC-1')
    grp%rdl = rdl
    grp%fF = fF
    grp%sticky = sticky%value
    grp%RC = convert(RC, 's-1')
    grp%V0 = convert(V0, 's-1')
    grp%F0N = convert(F0N, 'molN molC-1 s-1')
    grp%zC = convert(zC, 'molC gChl-1')
    grp%zNu = convert(zN, 'molC molN-1')
    grp%zNf = convert(zF, 'molC molN-1')
    grp%Tref = convert(Tref, '°C')
    grp%Topt = convert(Topt, '°C')
    grp%Chl0 = convert(chl0, 'mg m-3')
    grp%vsink = convert(vsink, 'm s-1')
    grp%Q10 = Q10
    grp%fPIC = fPIC
    grp%daylen => env%daylen
    grp%DOM = TRIM(DOM(1:100))
    SELECT CASE (TRIM(pa))
    CASE ('dynamic')
       grp%pachl => tchdyn
       grp%chldyn => chl_dyn
    CASE ('slow')
       grp%pachl => tchdyn
       grp%chldyn => chl_dyn_slow
    CASE DEFAULT
       grp%pachl => tchia
       grp%chldyn => chl_dummy
       grp%dynamicChl = .FALSE.
    END SELECT
    grp%PIC = PIC
    ncon = 3 + COUNT((/grp%PIC, grp%dynamicChl/))
    grp%nsv = ncon
    grp%growth => grow
    IF (grp%fF.GT.0) THEN
       grp%nuts => facn2f
       IF (grp%F0N.LE.0._dp) &
            WRITE (*,'("cmo_read: N2 fixation is on but F0N = ",ES10.3,".")') grp%F0N
    ELSE
       grp%nuts => uptake
       grp%zN = grp%zNu
    END IF
    grp%nq0 = 2
    ALLOCATE (grp%Q0(grp%nq0), grp%iq0(grp%nq0), grp%constituents(grp%nsv))
    grp%iq0 = (/2, 3/)
    grp%constituents(1:3) = (/'C', 'N', 'P'/)
    IF (grp%PIC) grp%constituents(4) = 'PIC'
    IF (grp%dynamicChl) grp%constituents(ncon) = 'Chl'
    grp%Q0N => grp%Q0(1)
    grp%Q0P => grp%Q0(2)
    grp%Q0N = convert(Q0N, 'molN molC-1')
    grp%Q0P = convert(Q0P, 'molP molC-1')
    CALL grp%fT_select (fT, 'cmo_read')     ! set temperature function
    SELECT CASE (fTNF)
    CASE ('Eppley')
       grp%fTNF => ftnf_Eppley
    CASE ('Houlton')
       grp%fTNF => ftnf_Houlton
    CASE ('oppla')
       grp%fTNF => ftnf_oppla
    END SELECT
    IF (fTalpha) THEN
       grp%fTalpha => grp%fTemp
    ELSE
       ALLOCATE (grp%fTalpha)
       grp%fTalpha = 1._dp
    END IF
    IF (fTRC) THEN
       grp%fTRC => grp%fTemp
    ELSE
       ALLOCATE (grp%fTRC)
       grp%fTRC = 1._dp
    END IF
    RETURN
100 CONTINUE
    WRITE (stderr,FMT='("cmo/cmo_read: missing namelist cmo for species ",A,".")') TRIM(grp%name)
    STOP 1
900 CONTINUE
    WRITE (stderr,'("cmo_read:",A)') TRIM(msg)
    STOP 1
  END SUBROUTINE cmo_read

  SUBROUTINE cmo_set (grp, grps, env)
    IMPLICIT NONE
    CLASS(phycmo), TARGET, INTENT(INOUT) :: grp
    TYPE(state), DIMENSION(:), INTENT(IN) :: grps
    TYPE(local), TARGET, INTENT(INOUT) :: env
    INTEGER :: n
    IF (grp%dynamicChl) THEN
       grp%theta => grp%ratios(grp%nsv-1)
    ELSE
       ALLOCATE (grp%theta)
    END IF
    SELECT CASE (TRIM(env%dicy)) ! diurnal light cycle
    CASE ('', 'none')            ! no light cycle (constant light)
       grp%dlfA => env%daylen    ! calculate daily average growth
    CASE DEFAULT                 ! in this case the light cycle is resolved
       ALLOCATE(grp%dlfA)
       grp%dlfA = 1._dp          ! calculate instantaneous growth rate
    END SELECT
    grp%idoc = 0
    DO n=1,env%nft
       IF (TRIM(grp%DOM).EQ.TRIM(grps(n)%var%name)) grp%idoc = grps(n)%var%isv(1)
    END DO
    IF (grp%idoc.EQ.0) THEN
       ALLOCATE (grp%VDIC, grp%VDOC, grp%ic(1), grp%VC(1))
       grp%ngr => grp%VC(1)     ! growth = net C assimilation
    ELSE       
       ALLOCATE (grp%ngr, grp%ic(2), grp%VC(2))
       grp%ic(2) = grp%idoc     ! DOC release
       grp%VDIC => grp%VC(1)    ! CO2 fixation
       grp%VDOC => grp%VC(2)    ! DOC uptake (always .LE. 0)
    END IF
    grp%ic(1) = env%idic
    IF (grp%PIC) THEN
       grp%QPIC => grp%ratios(3)
       grp%calc => cmo_pic
       grp%ipic = grp%isv(4)
       grp%ichl = grp%isv(5)
       env%smsk(grp%ic(1),grp%ipic,:) = .TRUE.
    ELSE
       grp%calc => cmo_pic_dummy
       grp%ichl = grp%isv(4)
    END IF
    env%smsk(grp%ic,grp%isv(1),:) = .TRUE.
    env%smsk(grp%isv(1),env%io2,:) = .TRUE.
    env%smsk(grp%isv(2),env%idin,:) = .TRUE.
    env%smsk(grp%isv(3),env%idip,:) = .TRUE.
    IF (.NOT.ASSOCIATED(grp%stick)) ALLOCATE (grp%stick)
  END SUBROUTINE cmo_set

  SUBROUTINE cmo_flux (grp, grps, env, box, times)
    IMPLICIT NONE
    CLASS(phycmo), INTENT(INOUT) :: grp
    TYPE(state), DIMENSION(:), INTENT(IN) :: grps
    TYPE(local), INTENT(INOUT) :: env
    TYPE(layer), INTENT(INOUT) :: box(0:)
    TYPE(timing), INTENT(IN) :: times
    CALL grp%growth (ambient=env, box=box(1))
    CALL grp%calc (box=box(1))
    grp%stick = grp%sticky*MAX(2._dp - grp%QN/grp%Q0N, 0._dp)       ! points to env%sticky
    grp%vv = grp%vsink*(2._dp - MIN(grp%QN/grp%Q0N, 3._dp))
    box(1)%soma(grp%isv(1),grp%ic) = -grp%VC                        ! net DIC, DOC uptake
    box(1)%soma(grp%ic,grp%isv(1)) = grp%VC
    box(1)%soma(grp%isv(2),env%idin) = -grp%VN                      ! DIN uptake
    box(1)%soma(grp%isv(3),env%idip) = -grp%VP                      ! DIP uptake
    box(1)%soma(grp%isv(1),env%io2)  = grp%VDIC*env%RQ + grp%VN*2._dp ! O2 production
    box(1)%soma(env%idin,grp%isv(2)) = grp%VN - grp%ngr*grp%Q0N     ! subtract subsistence N
    box(1)%soma(env%idip,grp%isv(3)) = grp%VP - grp%ngr*grp%Q0P     ! subtract subsistence P
  END SUBROUTINE cmo_flux

  !> \brief Rates of growth, nutrient uptake, and chlorophyll synthesis
  !!  or steady-state Chl:C ratio
  !! \param[in] ambient%daylen daylength as a fraction of \qty{24,h}
  !! \param[in] box%DIN       DIN concentration / \unit{mmol.m^{-3}}
  !! \param[in] box%DIP       DIP concentration / \unit{mmol.m^{-3}}
  !! \param[in] box%dPAR      average daytime PAR
  !! \param[in] box%bPAR      PAR at bottom of box (layer)
  !! \param[in] box%tPAR      PAR at top of box (layer)
  !! \param[in] phy%QN        N cell quota (N:C ratio)
  !! \param[in] phy%QP        P cell quota (P:C ratio)
  !! \param[in] phy%A0        potential nutrient affinity
  !! \param[in] phy%alpha     potential light affinity (initial slope)
  !! \param[in] phy%QsN       non-allocatable N quota (0.5*phy%Q0N)
  !! \param[in] phy%Q0N       subsistence N quota (minimum N:C ratio)
  !! \param[in] phy%Q0P       subsistence P quota (minimum P:C ratio)
  !! \param[in] phy%RC        cost of Chl maintenance
  !! \param[in] phy%rdl       daylength parameter
  !! \param[in] phy%V0        maximum-rate parameter
  !! \param[in] phy%zC        cost of Chl synthesis
  !! \param[in] phy%zN        cost of biosynthesis
  !! \param[in] phy%fT        temperature function for growth
  !! \param[out] phy%A        light-dependence of growth
  !! \param[out] phy%fC       allocation factor for \ce{CO2} fixation
  !! \param[out] phy%fN       allocation factor for N uptake
  !! \param[out] phy%fV       allocation factor for nutrient uptake
  !! \param[out] phy%mu0      potential gross \ce{CO2} fixation
  !! \param[out] phy%SI       light saturation
  !! \param[out] phy%rctht    rate of change of the Chl:C ratio
  !! \param[out] phy%tch      chloroplast Chl:C ratio
  !! \param[out] phy%ngr      net growth rate (rate of C assimilation)
  !! \param[out] phy%VN       rate of N uptake
  !! \param[out] phy%VP       rate of P uptake
  !! \param[out] phy%VNh0     local potential N uptake
  !! \param[out] phy%VPh0     local potential P uptake
  SUBROUTINE grow (phy, ambient, box)
    IMPLICIT NONE
    CLASS(phycmo), INTENT(INOUT) :: phy
    TYPE(local), INTENT(IN) :: ambient
    TYPE(layer), INTENT(IN) :: box
    REAL(dp) :: V0
    ! numbers in parentheses refer to equations in ../doc/equations.pdf
    phy%fTemp = phy%fT(box)
    V0 = phy%V0*phy%fTemp
    phy%alphaT = phy%alpha*phy%fTalpha
    phy%RCT = phy%RC*phy%fTRC
    phy%fV = MAX(phy%QsN/phy%QN - phy%zN*(phy%QN - phy%Q0N), 0._dp)     ! (1.23)
    phy%fC = 1._dp - phy%fV - phy%QsN/phy%QN
    phy%mu0 = V0/(0.5_dp + ambient%daylen)**phy%rdl
    CALL phy%pachl (ambient, box)                    ! photo-acclimation (tchdyn or tchia)
    phy%A = (1._dp - ambient%ice)*phy%dlfA*phy%mu0*phy%SI*(1._dp - phy%zC*phy%tch) & ! (1.4)
          - phy%RCT*phy%zC*phy%tch
    CALL phy%nuts (box, V0)                          ! uptake of facn2f
    phy%VP = phy%OC*phy%fV*(1._dp - phy%fN)*phy%VPh0                    ! (1.28)
    phy%VDIC = phy%OC*phy%fC*phy%A - phy%zN*(phy%VN + phy%N2F)          ! (1.3)
    ! release (DO)C to avoid outgrowing Q0P
    phy%VDOC = -MAX(phy%VDIC*phy%Q0P/phy%QP - phy%VP/phy%Q0P, 0._dp)&
            *MAX(2._dp - phy%QP/phy%Q0P, 0._dp)
    phy%ngr = phy%VDIC + phy%VDOC
    CALL phy%chldyn (box)       ! chl_dyn, chl_dyn_slow, or chl_dummy
  END SUBROUTINE grow

  SUBROUTINE uptake (phy, box, V0)
    IMPLICIT NONE
    CLASS(phycmo), INTENT(INOUT) :: phy
    TYPE(layer), INTENT(IN) :: box
    REAL(dp), INTENT(IN) :: V0
    REAL(dp) :: an, ap, VNmax, VNh0
    VNmax = V0*MAX(1._dp - phy%Q0P/phy%QP, 0._dp)                       ! (1.30)
    an = phy%A0*box%DIN
    ap = phy%A0*box%DIP
    VNh0 = an/(SQRT(MAX(an, 1.E-30_dp)) + SQRT(VNmax))**2               ! VNh0/VNmax (1.29)
    phy%VPh0 = ap/(SQRT(ap/V0) + 1._dp)**2                              ! (1.29)
    phy%sVPh = SQRT(phy%VPh0)
    phy%fN = phy%sVPh/(MAX(phy%sVPh, 1.e-30_dp) + SQRT(V0*VNh0**1.5*phy%Q0P/phy%QN)) ! (1.39)
    phy%VN = phy%OC*phy%fV*phy%fN*VNh0*VNmax                            ! (1.28)
  END SUBROUTINE uptake

  SUBROUTINE facn2f (phy, box, V0)
    IMPLICIT NONE
    CLASS(phycmo), INTENT(INOUT) :: phy
    TYPE(layer), INTENT(IN) :: box
    REAL(dp), INTENT(IN) :: V0
    REAL(dp) :: FNmax, fnF, fvF
    CALL uptake (phy, box, V0)
    FNmax = phy%F0N*phy%fTNF(box=box)
    fnF = phy%sVPh/(SQRT(FNmax*phy%Q0P/phy%QN) + phy%sVPh)              ! (1.56)
    fvF = MAX(phy%QsN/phy%QN - phy%zNf*(phy%QN - phy%Q0N), 0._dp)
    phy%N2F = phy%OC*fvF*fnF*(1 - phy%Q0P/phy%QP)*FNmax
    IF (phy%VN.LT.phy%N2F) THEN ! fix N2
       phy%zN = phy%zNf
       phy%fN = fnF
       phy%fV = fvF
       phy%VN = 0._dp
    ELSE                        ! use DIN
       phy%zN = phy%zNu
       phy%N2F = 0._dp
    END IF
    box%soma(phy%isv(2),phy%isv(2)) = phy%N2F
  END SUBROUTINE facn2f

  SUBROUTINE chl_dyn (phy, box)
    IMPLICIT NONE
    CLASS(phycmo), INTENT(INOUT) :: phy
    TYPE(layer), INTENT(IN) :: box
    ! ngr .GE. 0 prevents positive feedback between high theta and Chl synthesis
    ! fcdtdq - fC*dot(QN)/theta*dtheta/dQN                                (1.15)
    phy%fcdtdq = (phy%VN + phy%N2F - MAX(phy%ngr, 0._dp)*phy%QN)*(phy%Q0N/phy%QN**2 + phy%zN)
    ENTRY chl_dyn_slow (phy, box)
    ! Chl synthesis                                                       (1.13)
    box%soma(phy%ichl,phy%ichl) = phy%theta*(phy%ngr + phy%rctht*phy%OC + phy%fcdtdq)
    ENTRY chl_dummy (phy, box)
  END SUBROUTINE chl_dyn

  !> \brief Dynamic photo-acclimation via Chl synthesis
  SUBROUTINE tchdyn (phy, ambient, box)
    IMPLICIT NONE
    CLASS(phycmo), INTENT(INOUT) :: phy !< current phytoplankton group
    TYPE(local), INTENT(IN) :: ambient  !< ambient environment
    TYPE(layer), INTENT(IN) :: box      !< current layer environment
    REAL(dp) :: atm, atmb, atmd, SId
    phy%tch = phy%theta/MAX(phy%fC, 1.E-30_dp)                          ! (1.4)
    atm = phy%alphaT*phy%tch/MAX(phy%mu0, 1.E-30_dp)
    atmb = atm*box%bPAR
    IF (atmb.GT.0._dp) THEN
       phy%SI = 1._dp - (dei(-atm*box%tPAR) - dei(-atmb))/box%lch       ! (1.62)
    ELSE
       phy%SI = 0._dp
    END IF
    ! use approx. daytime average SI (SId) for photo-acclimation, assuming a
    ! triangular light cycle
    atmd = atm*2._dp*box%dPAR   ! atmd is 0 if Chl is
    SId = (EXP(-atmd) - 1._dp + atmd)/MAX(atmd, 1.E-30_dp)
    phy%rctht = (ambient%daylen*(phy%alphaT*box%dPAR*(1._dp - SId)&
                                 *(1._dp/phy%zC - phy%tch) - SId*phy%mu0)&
                 - phy%RCT)*phy%fC  ! (fC/zC)*d(rgr)/d(theta)
  END SUBROUTINE tchdyn

  !> \brief Instantaneous photo-acclimation to daytime-average light intensity
  SUBROUTINE tchia (phy, ambient, box)
    IMPLICIT NONE
    CLASS(phycmo), INTENT(INOUT) :: phy
    TYPE(local), INTENT(IN) :: ambient
    TYPE(layer), INTENT(IN) :: box
    REAL(dp) :: aim, atm, I0
    aim = phy%alphaT*box%dPAR/phy%mu0
    I0 = phy%RCT*phy%zC/(phy%alphaT*ambient%daylen) ! threshold light intensity (1.12)
    IF (box%dPAR.GT.I0) THEN          ! steady-state Chl:C ratio                (1.11)
       phy%tch = 1._dp/phy%zC + (1._dp - lambertw((1._dp + phy%RCT/(ambient%daylen*phy%mu0))&
                                                  *EXP(1._dp + aim/phy%zC), 0, 0))/aim
       atm = phy%tch*phy%alphaT/phy%mu0
       phy%SI = 1._dp - (dei(-atm*box%tPAR) - dei(-atm*box%bPAR))/box%lch
    ELSE                              ! below light threshold: no Chl, no growth
       phy%tch = 0._dp
       phy%SI = 0._dp
    END IF
    phy%theta = phy%tch*phy%fC
  END SUBROUTINE tchia

  SUBROUTINE cmo_pic (phy, box)
    IMPLICIT NONE
    CLASS(phycmo), INTENT(INOUT) :: phy
    TYPE(layer), INTENT(INOUT) :: box
    phy%VPIC = phy%fPIC*phy%VDIC
    box%soma(phy%ic(1),phy%ipic) = phy%VPIC    ! calcification
    box%soma(phy%ipic,phy%ic(1)) = -phy%VPIC
    ENTRY cmo_pic_dummy (phy, box)
  END SUBROUTINE cmo_pic

  FUNCTION ftnf_oppla (phy, box) RESULT (fT)
    IMPLICIT NONE
    CLASS(phycmo), INTENT(IN) :: phy
    TYPE(layer), INTENT(IN) :: box
    REAL(dp) :: fT
    fT = ft_oppla(phy, box)
  END FUNCTION ftnf_oppla

  FUNCTION ftnf_Eppley (phy, box) RESULT (fT)
    IMPLICIT NONE
    CLASS(phycmo), INTENT(IN) :: phy
    TYPE(layer), INTENT(IN) :: box
    REAL(dp) :: fT
    fT = ft_Eppley(phy, box)
  END FUNCTION ftnf_Eppley

  FUNCTION ftnf_Houlton (phy, box) RESULT (fT)
    IMPLICIT NONE
    CLASS(phycmo), INTENT(IN) :: phy
    TYPE(layer), INTENT(IN) :: box
    REAL(dp) :: fT
    fT = ft_Houlton(phy, box)
  END FUNCTION ftnf_Houlton
END MODULE cmo
