!> \brief detritus functions
MODULE det
  USE et
  IMPLICIT NONE
  PRIVATE

  TYPE, EXTENDS(fungroup), PUBLIC :: detritus
     REAL(dp) :: sink=0_dp,&    !< sinking velocity in \unit{m.s^{-1}}
          omax=15._dp,&         !< maximal Omega above which aragonite will precipitate
          remiC=0._dp, remiN=0._dp, remiP=0._dp,&
          remiPIC=-1._dp,&      !< PIC dissolution rate in \unit{s^{-1}}
          formPIC=0._dp,&       !< PIC formation rate in \unit{s^{-1}}
          remiChl=-1._dp,&      !< Chl decay rate in \unit{s^{-1}}
          kag=-1._dp            !< aggregation kernel (units should be \unit{m^3.mmolC^{-1}.d^{-1}})
     REAL(dp), ALLOCATABLE :: remi(:), loss(:)
     CHARACTER(LEN=512) :: decay='' !< name of compartment for detritus decay
     LOGICAL :: detChl,&            !< flag for Chl in detritus
          detPIC,&                  !< flag for PIC in detritus
          aggregates=.FALSE.        !< flag whether this compartment represents aggregates
     INTEGER :: nrec
     INTEGER, ALLOCATABLE :: iloss(:),& !< indices of extra loss terms for Chl and O2
          isvl(:),&                     !< state-variable indices corresponding to iloss
          irec(:)                       !< indices of recipients of decayed matter
     PROCEDURE(det_fluxes), POINTER :: fluxes
     PROCEDURE(det_loss), POINTER :: losses
   CONTAINS
     PROCEDURE :: flux => det_flux
     PROCEDURE :: read => det_read
     PROCEDURE :: set => det_set
  END TYPE detritus

CONTAINS

  !> read detritus-related parameters, set detritus parameters, constituents, number of states
  SUBROUTINE det_read (grp, env, lun)
    IMPLICIT NONE
    CLASS(detritus), TARGET, INTENT(INOUT) :: grp !< funcional type
    TYPE(local), TARGET, INTENT(IN) :: env        !< local environment
    INTEGER, INTENT(IN) :: lun                    !< unit number of open file with namelist det
    CHARACTER(LEN=512) :: fn, decay, fT, msg, species='det'
    REAL(dp) :: omax
    TYPE(quantity) :: remiC, remiN, remiP, remiPIC, formPIC, remiChl, sink,&
                      sticky, kag
    LOGICAL :: aggregates
    INTEGER :: ncon
    NAMELIST /det/ remiC, remiN, remiP, remiPIC, formPIC, omax, remiChl, sink, decay,&
                   fT, sticky, kag, species, aggregates
    REWIND (lun)
    DO
       remiC = quantity(grp%remiC, 'd-1')
       remiN = quantity(grp%remiN, 'd-1')
       remiP = quantity(grp%remiP, 'd-1')
       remiPIC = quantity(grp%remiPIC, 'd-1')
       formPIC = quantity(grp%formPIC, 'd-1')
       omax = grp%omax
       remiChl = quantity(grp%remiChl, 'd-1')
       sink  = quantity(grp%sink, 'm d-1')
       sticky  = quantity(grp%sticky, '')
       kag = quantity(-1._dp, 'd-1')
       aggregates = grp%aggregates
       decay = ''
       fT = 'Eppley'
       READ (lun, NML=det, END=100, IOMSG=msg, ERR=900)
       IF (TRIM(species(1:100)).EQ.TRIM(grp%name)) EXIT
    END DO
    grp%detPIC = remiPIC%value.GE.0
    grp%detChl = remiChl%value.GE.0
    grp%decay = decay
    ncon = 3 + COUNT((/grp%detPIC, grp%detChl/))
    grp%nsv = ncon
    grp%nrec = 3 + COUNT((/grp%detPIC/))
    ALLOCATE (grp%constituents(ncon))
    grp%constituents(1:3) = (/'C', 'N', 'P'/)
    grp%sticky = sticky%value
    IF (aggregates.AND.(kag%value.LT.0._dp)) THEN
       WRITE (stderr,'("Missing kag in namelist det for ",A,".")') TRIM(grp%name)
       STOP
    END IF
    grp%aggregates = aggregates
    grp%kag = convert(kag, 's-1')
    grp%remiC = convert(remiC, 's-1')
    grp%remiN = convert(remiN, 's-1')
    grp%remiP = convert(remiP, 's-1')
    IF (grp%detPIC) THEN
       grp%constituents(4) = 'PIC'
       grp%remiPIC = MAX(convert(remiPIC, 's-1'), 0._dp)
       grp%formPIC = convert(formPIC, 's-1')
       grp%omax = omax
    END IF
    IF (grp%detChl) THEN
       grp%constituents(ncon) = 'Chl'
       grp%remiChl = MAX(convert(remiChl, 's-1'), 0._dp)
    END IF
    grp%sink = convert(sink, 'm s-1')
    CALL grp%fT_select (fT, 'det_read')
    RETURN
100 CONTINUE
    INQUIRE (UNIT=lun, NAME=fn)
    WRITE (stderr, '("Namelist det missing in file ",A,".")') TRIM(fn)
    STOP 1
900 CONTINUE
    WRITE (stderr,'("det_read:",A)') TRIM(msg)
    WRITE (stderr,'("searching/reading namelist det for species ",A)') TRIM(grp%name)
    STOP 1
  END SUBROUTINE det_read

  !> flag detritus fluxes in soma for output, set-up remineralisation scheme
  SUBROUTINE det_set (grp, grps, env)
    IMPLICIT NONE
    CLASS(detritus), TARGET, INTENT(INOUT) :: grp
    TYPE(state), DIMENSION(:), INTENT(IN) :: grps
    TYPE(local), TARGET, INTENT(INOUT) :: env
    INTEGER :: n, igrec, nloss, nxl
    LOGICAL rDIC
    igrec = 0
    ! find index of recipient group (DOM or inorganic matter) of decaying detritus
    DO n=1,env%nft
       IF (TRIM(grp%decay).EQ.TRIM(grps(n)%var%name)) igrec = n
    END DO
    rDIC = igrec.EQ.0
    nloss = grp%ncon + COUNT((/rDIC/))
    nxl = nloss - grp%nrec  ! number of extra losses
    ALLOCATE (grp%irec(grp%nrec), grp%remi(nloss), grp%loss(nloss), grp%isvl(nloss), grp%iloss(nxl))
    grp%remi(1:3) = (/grp%remiC, grp%remiN, grp%remiP/)
    grp%isvl(1:grp%nrec) = grp%isv(1:grp%nrec)
    IF (rDIC) THEN
       IF (env%ndom.GT.0) &
            WRITE (stderr,'("det_set: DOM present but not identified as decay in namelist det (decay = ",A,")!")') &
            TRIM(grp%decay)
       grp%irec(1:3) = (/env%idic, env%idin, env%idip/)  ! if no DOM: remineralize directly
       grp%iloss(nxl) = nloss
       grp%isvl(grp%iloss(nxl)) = env%io2
       grp%remi(nloss) = grp%remiC*env%RQ ! remineralisation of POC to DIC consumes O2
       env%smsk(grp%isv(1),grp%isvl(grp%iloss(nxl)),:) = .TRUE.
    ELSE
       grp%irec(1:3) = grps(igrec)%var%isv(1:3)
    END IF
    IF (grp%detPIC) THEN        ! PIC becomes DIC (alkalinity changes are accounted for in plankton:alkalinity)
       grp%irec(4) = env%idic
       grp%losses => det_loss_PIC
    ELSE
       grp%losses => det_loss
    END IF
    IF (grp%detChl) THEN        ! Chl disappears
       grp%iloss(1) = grp%nrec + 1
       grp%isvl(grp%iloss(1)) = grp%isv(grp%ncon)
       grp%remi(grp%iloss(1)) = grp%remiChl
       env%smsk(grp%isv(1),grp%isvl(grp%iloss(1)),:) = .TRUE.
    END IF
    IF (nxl.GT.0) THEN
       grp%fluxes => det_fluxes
    ELSE
       grp%fluxes => det_fluxes_decay
    END IF
    DO n=1,grp%nrec
       env%smsk(grp%isv(n),grp%irec(n),:) = .TRUE.
    END DO
  END SUBROUTINE det_set

  SUBROUTINE det_flux (grp, grps, env, box, times)
    IMPLICIT NONE
    CLASS(detritus), INTENT(INOUT) :: grp
    TYPE(state), DIMENSION(:), INTENT(IN) :: grps
    TYPE(local), INTENT(INOUT) :: env
    TYPE(layer), INTENT(INOUT) :: box(0:)
    TYPE(timing), INTENT(IN) :: times
    grp%vv = grp%sink
    CALL grp%losses (box=box(1))
    CALL grp%fluxes (box=box(1))
  END SUBROUTINE det_flux

  SUBROUTINE det_loss_PIC (grp, box)
    IMPLICIT NONE
    CLASS(detritus), INTENT(INOUT) :: grp
    TYPE(layer), INTENT(IN) :: box
    grp%remi(4) = grp%remiPIC - MAX(box%omega - grp%omax, 0._dp)*grp%formPIC
    ENTRY det_loss (grp, box)
    grp%loss = grp%fT(box)*grp%remi*box%stv(grp%isvl)
  END SUBROUTINE det_loss_PIC

  SUBROUTINE det_fluxes (grp, box)
    IMPLICIT NONE
    CLASS(detritus), INTENT(IN) :: grp
    TYPE(layer), INTENT(INOUT) :: box
    INTEGER :: n
    box%soma(grp%isv(1),grp%isvl(grp%iloss)) = -grp%loss(grp%iloss)
    ENTRY det_fluxes_decay (grp, box)
    DO n=1,grp%nrec
       box%soma(grp%irec(n),grp%isvl(n)) = -grp%loss(n)
       box%soma(grp%isvl(n),grp%irec(n)) = grp%loss(n)
    END DO
  END SUBROUTINE det_fluxes

END MODULE det
