!> \brief Functions for the carbonate system and alkalinity
!!
!! Expressions for the equilibrium constants are from SolveSAPHE (\cite Munhoven:2013)
!! except for ammonia.  Concentrations are converted from \unit{mmol.m^{-3}} to
!! \unit{mol.kg^{–1}} in subroutine \ref dicalp.
!!
!! variable | description
!! ------- | -----------
!! at      | total alkalinity in \unit{eq.kg^{-1}}
!! ^       | at = \ce{[HCO3^-] + 2[CO3^2-] + [B(OH)4^-] + [OH^-] + [SiO(OH)3^-] + [HS^-]}
!! ^       |      -- (\ce{[H^+]F + [HSO4^-] + [HF]})
!! ^       | \ce{H2S} and \ce{H3PO4} systems not included
!! bt      | total boron concentration in \unit{mol.kg^{-1}}
!! co2     | \ce{CO2} concentration in \unit{mol.kg^{-1}}
!! co3     | \ce{CO3^2-} concentration in \unit{mol.kg^{-1}}
!! ct      | DIC concentration in \unit{mol.kg^{-1}}
!! f       | \ce{CO2} fugacity in Pa
!! fa      | \ce{CO2} fugacity in air in Pa
!! hc      | proton concentration in \unit{mol.kg^{-1}}
!! hco3    | \ce{HCO3^-} concentration in \unit{mol.kg^{-1}}
!! i0      | ionic strength
!! k0      | solubility of \ce{CO2} in sea water in \unit{mol.kg^{-1}.Pa^{-1}}
!! ^       | k0 = \ce{[CO2]}/f\ce{CO2}
!! k1      | first dissociation constant of the carbonate system in \unit{mol.kg^{-1}}
!! ^       | k1 = \ce{[H+][HCO3^-]/[CO2]}
!! k1p     | first dissociation constant of the phosphate system in \unit{mol.kg^{-1}}
!! ^       | k1p = \ce{[H+][H2PO4^-]/[H3PO4]}
!! k2      | second dissociation constant of the carbonate system in \unit{mol.kg^{-1}}
!! ^       | k2 = \ce{[H+][CO3^2-]/[HCO3^-]}
!! k2p     | second dissociation constant of the phosphate system in \unit{mol.kg^{-1}}
!! ^       | k2p = \ce{[H+][HPO4^2-]/[H2PO4^-]}
!! k3p     | third dissociation constant of the phosphate system in \unit{mol.kg^{-1}}
!! ^       | k3p = \ce{[H+][PO4^3-]/[HPO4^2-]}
!! kb      | equilibrium constant of the borate system in \unit{mol.kg^{-1}}
!! ^       | kb = \ce{[H+][B(OH)4^-]/[B(OH)3]}
!! kf      | dissociation constant of hydrogen flouride in \unit{mol.kg^{-1}}
!! ^       | kf = \ce{[H+][F^-]/[HF]}
!! knh4    | dissociation constant of ammonium in \unit{mol.kg^{-1}}
!! ^       | knh4 = \ce{[H+][NH3]/[NH4+]}
!! kso4    | equilibrium constant of the sulfate system in \unit{mol.kg^{-1}}
!! ^       | kso4 = \ce{[H+]F[SO4^2-]/[HSO4^-]}
!! ksi     | equilibrium constant of the silicate system in \unit{mol.kg^{-1}}
!! ^       | ksi = \ce{[H+][SiO(OH)3^-]/[Si(OH)4]}
!! kw      | equilibrium constant for water in \unit{mol^2.kg^{-2}}
!! ^       | kw = \ce{[H+][OH^-]}
!! kspa    | equilibrium constant for aragonite in \unit{mol.kg^{-1}}
!! ^       | kspc = \ce{[Ca^{2+}][CO3^{2-}]}
!! kspc    | equilibrium constant for calcite in \unit{mol.kg^{-1}}
!! ^       | kspc = \ce{[Ca^{2+}][CO3^{2-}]}
!! lnx     | ln(x)
!! prat    | atmospheric pressure in Pa
!! pt      | total inorganic phosphorus concentration in \unit{mol.kg^{-1}}
!! Rgas    | ideal gas constant in \unit{dJ.K^{-1}.mol^{-1}} = \unit{dPa.m^3.K^{-1}.mol^{-1}}
!! sal     | salinity in psu
!! schnco2 | Schmidt number for \ce{CO2}
!! schno2  | Schmidt number for \ce{O2}
!! SO4     | total sulfate concentration in \unit{mol.kg^{-1}}
!! tk      | temperature in K
!! tv      | \ce{CO2} air-sea gas transfer velocity (piston velocity) in \unit{m.s^{-1}}
!! xco2    | mole fraction of \ce{CO2} in air

MODULE dic
  USE et
  IMPLICIT NONE
  PRIVATE
  REAL(dp), PARAMETER :: t0ck=273.15_dp    !< 0 °C in K
  REAL(dp), PARAMETER :: Rgas = 83.1446261815324_dp !< ideal gas constant in
  !! \unit{dPa.m^3.K^{-1}.mol^{-1}} (this value is from wikipedia, citing 2018 CODAT, NIST)
  TYPE, PUBLIC :: dicsys
     REAL(dp) :: at, pt, tk, pH0=8.1_dp, Ca=10.28E-3_dp,&
          kw, k1, k12, k2, kb, kf, k12p, ksi, kSO4, k1p, k2p, k3p, knh4,&
          kspc, kspa, SO4, bt, ct, ft, sit, ac, gmk, acd,&
          rrnsi,&               !< N:Si Redfield ratio
          rrnp,&                !< N:P Redfield ratio
          potalk=0._dp,&        !< potential alkalinity (Key et al. 2004) in \unit{mmol.m^{-3}}
          totalk=0._dp,&
          r_tot_free,&          !< \ce{[H+]_T}/\ce{[H+]_f}
          r_tot_sws,&           !< \ce{[H+]_T}/\ce{[H+]_{SWS}}
          faco2, k0co2,&
          schnco2,&
          schno2,&
          tvco2,&               !< transfer (piston) velocity for \ce{CO2} in \unit{m.s^{-1}}
          tvo2                  !< transfer (piston) velocity for \ce{O2} in \unit{m.s^{-1}}
     REAL(dp), POINTER :: asfco2, asfo2
     INTEGER :: niter=2
     PROCEDURE(dic_alk), POINTER :: alkalinity !< set alkalinity
     PROCEDURE(pivel_lm86), POINTER :: pivel   !< calculate transfer velocities for \ce{CO2} and \ce{O2}
     PROCEDURE(Hsolve_F06), POINTER :: Hsolve
     PROCEDURE(Hini_F06), POINTER :: init
   CONTAINS
     PROCEDURE :: read => indic
     PROCEDURE :: asf => asfco2
     PROCEDURE :: co2 => co2
  END TYPE dicsys
CONTAINS

  !> \brief Initialise DIC module
  !!
  !! pvfun | piston velocity function
  !! ----- | ------------------------
  !! LM86  | \cite Liss:1986 (default)
  !! W92   | \cite Wanninkhof:1992
  !! WG99  | \cite Wanninkhof:1999
  !! W14   | \cite Wanninkhof:2014
  !! \param[in]  dics%totalk  total alkalinity in \unit{mmol.m^{-3}}
  !! \param[out] dics%rrnsi   N:Si Redfield ratio
  !! \param[out] dics%rrnp    N:P Redfield ratio
  !! \param[out] dics%hc      initial guess for \ce{H+} concentration
  ! Values from namelist envi (in plankton_init) values for potalk, co2, and atmprs are used only if
  ! no corresponding datasets are found in the forcing file
  SUBROUTINE indic (dics, lun, env, fluxes)
    IMPLICIT NONE
    CLASS(dicsys), INTENT(INOUT) :: dics
    INTEGER, INTENT(IN) :: lun  !< of file with namelist dic
    TYPE(local), INTENT(INOUT) :: env
    REAL(dp), DIMENSION(:,0:), TARGET, INTENT(IN), OPTIONAL :: fluxes
    REAL(dp) :: pH
    CHARACTER(LEN=15) :: pvfun='LM86', pHfun='Munhoven 2013'
    CHARACTER(LEN=512) :: iomsg
    INTEGER :: niter
    TYPE(quantity) :: rfrnsi, rfrnp
    NAMELIST /dic/ pH, rfrnsi, rfrnp, pvfun, pHfun, niter
    IF (PRESENT(fluxes)) THEN
       dics%asfco2 => fluxes(env%idic,0)
       dics%asfo2 => fluxes(env%io2,0)
    ELSE
       ALLOCATE (dics%asfco2, dics%asfo2)
       dics%asfco2 = 0._dp
       dics%asfo2 = 0._dp
    END IF
    pH = dics%pH0
    niter = dics%niter
    rfrnsi = quantity(1D0, 'molN/molSi') ! N:Si Redfield ratio
    rfrnp = quantity(16D0, 'molN/molP')  ! N:P  Redfield ratio
    REWIND (lun)
    READ (lun,NML=dic,END=10,IOMSG=iomsg,ERR=900)
10  CONTINUE
    dics%niter = niter
    SELECT CASE (TRIM(pvfun))
    CASE ('LM86')                               ! Liss & Merlivat (1986)
       dics%pivel => pivel_lm86
    CASE ('W92')                                ! Wanninkhof (1992)
       dics%pivel => pivel_w92
    CASE ('WG99')                               ! Wanninkhof and McGillis (1999)
       dics%pivel => pivel_w99
    CASE ('W14')                                ! Wanninkhof (2014)
       dics%pivel => pivel_w14
    CASE DEFAULT
       STOP 'indic: Unknown piston velocity function.'
    END SELECT
    SELECT CASE (TRIM(pHfun))
    CASE ('Follows 2006', 'F06', 'Follows06')
       dics%init => Hini_F06
       dics%Hsolve => Hsolve_F06
       dics%pH0 = pH
    CASE ('Munhoven 2013', 'M13', 'Munhoven13')
       dics%init => Hini_M13
       dics%Hsolve => Hsolve_M13
    CASE DEFAULT
       STOP 'Invalid pHfun.'
    END SELECT
    dics%rrnsi = convert(rfrnsi, 'molN/molSi')
    dics%rrnp = convert(rfrnp, 'molN/molP')
    IF ((env%ialk.EQ.0).AND.(dics%potalk.GT.0D0)) THEN
       IF (dics%totalk.GT.0D0) STOP &
            'indic: Specify only one of potalk or totalk in namelist envi'
       dics%alkalinity => dic_potalk
    ELSE
       dics%alkalinity => dic_alk
    END IF
    RETURN
900 CONTINUE
    WRITE (stderr,'("indic:",A)') TRIM(iomsg)
    STOP
  END SUBROUTINE indic

  !> \brief Air-to-sea flux of \ce{CO2} and \ce{O2}
  !!
  !! Oxygen saturation concentration is calculated after \cite Garcia:1992a,
  !! following their recommendation to use the first column of their Table 1.
  !! - See also OCMIP-2
  !! \param[in]  box%salinity    salinity in psu
  !! \param[in]  box%temperature temperature in °C
  !! \param[in]  env%pressure    atmospheric pressure in Pa
  !! \param[in]  env%xco2        atmospheric \ce{CO2} mole fraction
  !! \param[in]  env%wind        wind speed at \qty{10,m} above surface in \unit{m.s^{-1}}
  !! \param[in]  dic%co2         concentration of \ce{CO2} in \unit{mol.kg^{-1}}
  !! \param[out] dic%faco2       fugacity of \ce{CO2} in air in Pa
  !! \param[out] dic%k0co2       solubility of \ce{CO2} in \unit{mol.kg^{-1}.Pa^{-1}}
  !! \param[out] dic%asfco2      air-sea flux of \ce{CO2} in \unit{mmol.m^{-2}.s^{-1}}
  !! \param[out] dic%asfo2       air-sea flux of \ce{O2} in \unit{mmol.m^{-2}.s^{-1}}
  SUBROUTINE asfco2 (dic, env, box)
    IMPLICIT NONE
    CLASS(dicsys), INTENT(INOUT) :: dic
    TYPE(local), INTENT(IN) :: env
    TYPE(layer), INTENT(INOUT) :: box
    CALL dic%pivel (env=env, box=box) ! piston velocity
    dic%faco2 = faco2(box%tk, env%pressure, env%xco2)
    dic%k0co2 = k0co2(box)
    dic%asfco2 = dic%tvco2*(dic%k0co2*dic%faco2 - box%CO2)*1D6 & ! convert m*mol/kg/s to mmol/m²/s
         + env%sfDIC
    dic%asfo2 = asf_o2(box) + env%sfO2
  CONTAINS
    ELEMENTAL FUNCTION asf_o2 (box) RESULT (asfo2)
      IMPLICIT NONE
      REAL(dp), PARAMETER :: A0=5.80871_dp, A1=3.20291_dp, A2=4.17887_dp,&
           A3=5.10006_dp, A4=-9.86643E-2_dp, A5=3.80369_dp, B0=-7.01577E-3_dp,&
           B1=-7.70028E-3_dp, B2=-1.13864E-2_dp, B3=-9.51519E-3_dp, C0=-2.75915E-7_dp
      TYPE(layer), INTENT(IN) :: box
      REAL(dp) :: asfo2, Ts, o2sat
      Ts = LOG((298.15D0 - box%temperature)/(t0ck + box%temperature))
      ! O2 saturation concentration [mol/kg=mmol/m3]
      o2sat = EXP(A0 + Ts*(A1 + Ts*(A2 + Ts*(A3 + Ts*(A4 + Ts*A5)))) &
           + box%salinity*(B0 + Ts*(B1 + Ts*(B2 + Ts*B3)) + box%salinity*C0))
      asfo2 = dic%tvo2*(o2sat - box%o2)
    END FUNCTION asf_o2
  END SUBROUTINE asfco2

  !> \brief Piston velocity after \cite Liss:1986
  !!
  !! \param[in]  env%wind            wind speed at \qty{10,m} in \unit{m.s^{-1}}
  !! \param[out] dic%tvco2           transfer (piston) velocity for \ce{CO2} in \unit{m.s^{-1}}
  !! \param[in]  dic%tvo2            transfer (piston) velocity for \ce{O2} in \unit{m.s^{-1}}
  SUBROUTINE pivel_lm86 (dic, env, box)
    IMPLICIT NONE
    CLASS(dicsys), INTENT(INOUT) :: dic
    TYPE(local), INTENT(IN) :: env
    TYPE(layer), INTENT(IN) :: box
    IF (env%wind.LE.3.6D0) THEN
       dic%tvco2 = 0.17D0*env%wind/3.6D5
    ELSEIF (env%wind.LE.13D0) THEN
       dic%tvco2 = (2.85D0*env%wind - 9.65D0)/3.6D5
    ELSE
       dic%tvco2 = (5.9D0*env%wind - 49.3D0)/3.6D5
    END IF
    dic%tvo2 = dic%tvco2
  END SUBROUTINE pivel_lm86

  !> \brief Piston velocity after \cite Wanninkhof:1992
  !!
  !! \param[in]  box%temperature sea-surface temperature in °C
  !! \param[in]  env%wind        wind speed at \qty{10,m} in \unit{m.s^{-1}}
  !! \param[out] dic%schnco2     Schmidt No. for \ce{CO2}
  !! \param[out] dic%schno2      Schmidt No. for \ce{O2}
  !! \param[out] dic%tvco2       transfer (piston) velocity for \ce{CO2} in \unit{m.s^{-1}}
  !! \param[out] dic%tvo2        transfer (piston) velocity for \ce{O2} in \unit{m.s^{-1}}
  SUBROUTINE pivel_w92 (dic, env, box)
    IMPLICIT NONE
    CLASS(dicsys), INTENT(INOUT) :: dic
    TYPE(local), INTENT(IN) :: env
    TYPE(layer), INTENT(IN) :: box
    REAL(dp) :: fwind
    ! Schmidt number for CO2 after Wanninkhof (1992)
    ! Schmidt number for O2 from UVic 2.9
    dic%schnco2 = 2073.1_dp - box%temperature*(125.62_dp &
         - box%temperature*(3.6276_dp - box%temperature*0.043219_dp))
    dic%schno2 = 1638._dp - box%temperature*(81.83_dp &
           - box%temperature*(1.483_dp - box%temperature*8.004E-3_dp))
    fwind = 0.31_dp*env%wind**2/3.6E5_dp ! convert cm/h to m/s
    dic%tvco2 = fwind/SQRT(dic%schnco2/660._dp)
    dic%tvo2 = fwind/SQRT(dic%schno2/660._dp)
  END SUBROUTINE pivel_w92

  !> \brief Piston velocity after \cite Wanninkhof:1999
  !!
  !! \param[in]  box%temperature sea-surface temperature in °C
  !! \param[in]  env%wind        wind speed at \qty{10,m} in \unit{m.s^{-1}}
  !! \param[out] dic%schnco2     Schmidt No. for \ce{CO2}
  !! \param[out] dic%schno2      Schmidt No. for \ce{O2}
  !! \param[out] dic%tvco2       transfer (piston) velocity for \ce{CO2} in \unit{m.s^{-1}}
  !! \param[out] dic%tvo2        transfer (piston) velocity for \ce{O2} in \unit{m.s^{-1}}
  SUBROUTINE pivel_w99 (dic, env, box)
    IMPLICIT NONE
    CLASS(dicsys), INTENT(INOUT) :: dic
    TYPE(local), INTENT(IN) :: env
    TYPE(layer), INTENT(IN) :: box
    REAL(dp) :: fwind
    ! Schmidt number for CO2 after Wanninkhof (1992)
    ! Schmidt number for O2 from UVic 2.9
    dic%schnco2 = 2073.1_dp - box%temperature*(125.62_dp &
         - box%temperature*(3.6276_dp - box%temperature*0.043219_dp))
    dic%schno2 = 1638._dp - box%temperature*(81.83_dp &
           - box%temperature*(1.483_dp - box%temperature*8.004E-3_dp))
    fwind = 0.0283_dp*env%wind**3/3.6E5_dp ! convert cm/h to m/s
    dic%tvco2 = fwind/SQRT(dic%schnco2/660._dp)
    dic%tvo2 = fwind/SQRT(dic%schno2/660._dp)
  END SUBROUTINE pivel_w99

  !> \brief Piston velocity after \cite Wanninkhof:2014
  !!
  !! \param[in]  box%temperature sea-surface temperature in °C
  !! \param[in]  env%wind        wind speed at \qty{10,m} in \unit{m.s^{-1}}
  !! \param[out] dic%schnco2     Schmidt No. for \ce{CO2}
  !! \param[out] dic%schno2      Schmidt No. for \ce{O2}
  !! \param[out] dic%tvco2       transfer (piston) velocity for \ce{CO2} in \unit{m.s^{-1}}
  !! \param[out] dic%tvo2        transfer (piston) velocity for \ce{O2} in \unit{m.s^{-1}}
  SUBROUTINE pivel_w14 (dic, env, box)
    IMPLICIT NONE
    CLASS(dicsys), INTENT(INOUT) :: dic
    TYPE(local), INTENT(IN) :: env
    TYPE(layer), INTENT(IN) :: box
    REAL(dp) :: fwind
    ! Schmidt numbers for CO2 and O2 after Wanninkhof (2014)
    dic%schnco2 = 2116.8_dp - box%temperature*(136.25_dp - box%temperature*(4.7353_dp &
         - box%temperature*(9.2307E-2_dp - box%temperature*7.555E-4_dp)))
    dic%schno2 = 1920.4_dp - box%temperature*(135.6_dp - box%temperature*(5.2122_dp &
         - box%temperature*(0.10939_dp - box%temperature*9.3777E-4_dp)))
    fwind = 0.251_dp*env%wind**2/3.6E5_dp ! convert cm/h to m/s
    dic%tvco2 = fwind/SQRT(dic%schnco2/660._dp)
    dic%tvo2 = fwind/SQRT(dic%schno2/660._dp)
  END SUBROUTINE pivel_w14

  !> \brief \ce{CO2} solubility
  !!
  !! Equation (7.1.3) from Chapter 5 in \cite Dickson:1997a.
  !! \param[in] box%salinity  salinity in psu
  !! \param[in] box%tk        temperature in K
  !! \return \ce{CO2} solubility in \unit{mol.kg^{-1}.Pa^{-1}}
  ELEMENTAL FUNCTION k0co2 (box) RESULT (k0)
    IMPLICIT NONE
    REAL(dp), PARAMETER :: atm=101325._dp ! 1 atm in Pa
    TYPE(layer), INTENT(IN) :: box
    REAL(dp) :: lnk0, k0
    lnk0 = 9345.17_dp/box%tk - 167.81077_dp + 23.3585_dp*LOG(box%tk) & ! (7.1.3)
         + box%salinity*(2.3517E-2_dp - box%tk*(2.3656E-4_dp - box%tk*4.7036E-7_dp))
    k0 = EXP(lnk0)/atm
  END FUNCTION k0co2

  !> \brief Fugacity of \ce{CO2} in air
  !!
  !! Equations from SOP 24 in \cite Dickson:1997a.
  !! See also OCMIP-2
  !! \param[in] tk      temperature in K
  !! \param[in] prat    atmospheric pressure in Pa
  !! \param[in] xco2    atmospheric \ce{CO2} mole fraction
  !! \return fugacity of \ce{CO2} in air in Pa
  ELEMENTAL FUNCTION faco2 (tk, prat, xco2) RESULT (fugacity)
    IMPLICIT NONE
    REAL(dp), INTENT(IN) :: tk   !< temperature in K
    REAL(dp), INTENT(IN) :: prat !< atmospheric pressure in Pa
    REAL(dp), INTENT(IN) :: xco2 !< atmospheric \ce{CO2} mole fraction
    REAL(dp) :: vcCO2,&          ! virial coefficient of pure CO2 in m3/mol
         vcAirCO2,&              ! virial coefficient of CO2 in air in m3/mol
         fugacity
    vcCO2 = -1636.75E-6_dp &
           + tk*(12.0408E-6_dp - tk*(3.27957E-8_dp - tk*3.16528E-11_dp)) ! SOP 24 (19)
    vcAirCO2 = 57.7E-6_dp - 0.118E-6_dp*tk                               ! SOP 24 (20)
    fugacity = xco2*prat*EXP((vcCO2 + 2._dp*(1._dp - xco2)**2*vcAirCO2)&
                             *prat*10._dp/Rgas/tk)                       ! SOP 24 (15)
  END FUNCTION faco2

  !> Carbonate SYSTEM
  !!
  !! kw = \ce{[H+][OH^-]}: equilibrium constant for water in \unit{mol^2.kg^{-2}}
  !! \param[in] prs pressure in bar
  ELEMENTAL FUNCTION lnkw_sws (sal, tk, lntk, tc, prs, prt, sqs) RESULT (lnkw)
    IMPLICIT NONE
    REAL(dp), INTENT(IN) :: sal, tk, lntk, tc, prs, prt, sqs
    REAL(dp) :: lnkw, dvi, dki
    dvi = -20.02_dp + (0.1119_dp - 0.1409E-02_dp*tc)*tc
    dki = -5.13E-3_dp + 0.0794E-3_dp*tc
    lnkw = -13847.26_dp/tk + 148.9802_dp - 23.6521_dp*lntk&
         + (118.67_dp/tk - 5.977_dp + 1.0495_dp*lntk)*sqs - 0.01615_dp*sal &
         + (-dvi + 0.5_dp*dki*prs)*prt ! pressure correction
  END FUNCTION lnkw_sws

  !>  k1 = \ce{[H+][HCO3^-]/[CO2]}: equilibrium constant for \ce{CO2} and \ce{HCO3-}
  !!  in \unit{mol.kg^{-1}} for SWS
  !! \param[in] prs pressure in bar
  ELEMENTAL FUNCTION lnk1_sws (sal, tk, lntk, tc, prs, prt, ds, sqs) RESULT (lnk1)
    IMPLICIT NONE
    REAL(dp), INTENT(IN) :: sal, tk, lntk, tc, prs, prt, ds, sqs
    REAL(dp) :: lnk1, dvi, dki
    dvi = -25.5_dp - 0.151_dp*ds + 0.1271_dp*tc
    dki = -3.08E-3_dp - 0.578E-3_dp*ds + 0.0877E-3_dp*tc
    lnk1 = -2275.036_dp/tk + 2.18867_dp - 1.468591_dp*lntk &
         + (-9.33291_dp/tk - 0.138681_dp - 0.00574938_dp*sal)*sqs + 0.0726483_dp*sal &
         + (-dvi + 0.5_dp*dki*prs)*prt
  END FUNCTION lnk1_sws

  !>  k2 = \ce{[H+][CO3^2-]/[HCO3^-]}: equilibrium constant for \ce{HCO3-}
  !!  and \ce{CO3^2-} in \unit{mol.kg^{-1}} for SWS
  !! \param[in] prs pressure in bar
  ELEMENTAL FUNCTION lnk2_sws (sal, tk, lntk, tc, prs, prt, ds, sqs) RESULT (lnk2)
    IMPLICIT NONE
    REAL(dp), INTENT(IN) :: sal, tk, lntk, tc, prs, prt, ds, sqs
    REAL(dp) :: lnk2, dvi, dki
    dvi = -15.82_dp + 0.321_dp*ds - 0.0219_dp*tc
    dki = 1.13E-3_dp - 0.314E-3_dp*ds - 0.1475E-3_dp*tc
    lnk2 = -3741.1288_dp/tk - 0.84226_dp - 1.437139_dp*lntk &
         + (-24.41239_dp/tk - 0.128417_dp - 0.0091284_dp*sal)*sqs &
         + 0.1195308_dp*sal + (-dvi + 0.5_dp*dki*prs)*prt
  END FUNCTION lnk2_sws

  !>  k1 = \ce{[H+][HCO3^-]/[CO2]}: equilibrium constant for \ce{CO2} and \ce{HCO3-}
  !!  in \unit{mol.kg^{-1}} for the total scale
  !! \param[in] prs pressure in bar
  ELEMENTAL FUNCTION lnk1_total (sal, tk, lntk, tc, prs, prt, ds) RESULT (lnk1)
    IMPLICIT NONE
    REAL(dp), INTENT(IN) :: sal, tk, lntk, tc, prs, prt, ds
    REAL(dp) :: lnk1, dvi, dki
    dvi = -25.5_dp - 0.151_dp*ds + 0.1271_dp*tc
    dki = -3.08E-3_dp - 0.578E-3_dp*ds + 0.0877E-3_dp*tc
    lnk1 = (-3633.86_dp/tk + 61.2172_dp - 9.6777_dp*lntk &
         + (0.011555_dp - 1.152E-4_dp*sal)*sal)*LOG(10._dp) &
         + (-dvi + 0.5_dp*dki*prs)*prt
  END FUNCTION lnk1_total

  !>  k2 = \ce{[H+][CO3^2-]/[HCO3^-]}: equilibrium constant for \ce{HCO3-}
  !!  and \ce{CO3^2-} in \unit{mol.kg^{-1}} for the total scale
  !! \param[in] prs pressure in bar
  ELEMENTAL FUNCTION lnk2_total (sal, tk, lntk, tc, prs, prt, ds) RESULT (lnk2)
    IMPLICIT NONE
    REAL(dp), INTENT(IN) :: sal, tk, lntk, tc, prs, prt, ds
    REAL(dp) :: lnk2, dvi, dki
    dvi = -15.82_dp + 0.321_dp*ds - 0.0219_dp*tc
    dki = 1.13E-3_dp - 0.314E-3_dp*ds - 0.1475E-3_dp*tc
    lnk2 = (-471.78_dp/tk - 25.929_dp + 3.16967_dp*lntk &
         + (0.01781_dp - 1.122E-4_dp*sal)*sal)*LOG(10._dp) &
         + (-dvi + 0.5_dp*dki*prs)*prt
  END FUNCTION lnk2_total

  !> \brief Sulfate system
  !!
  !! kSO4 = \ce{[H+][SO4^2-]/[HSO4-]} : equilibrium constant of sulfate system in \unit{mol.kg^{-1}}
  !! \param[in] prs   pressure in bar
  !! \param[in] i0    ionic strength
  !! \return natural logarithm of kSO4
  ELEMENTAL FUNCTION lnks_free (tk, lntk, tc, prs, prt, i0) RESULT (lnks)
    IMPLICIT NONE
    REAL(dp), INTENT(IN) :: tk, lntk, tc, prs, prt, i0
    REAL(dp) :: lnks, dvi, dki
    dvi = -18.03_dp + tc*(0.0466_dp + tc*0.316E-03_dp)
    dki = -4.53E-3_dp + tc*0.09E-3_dp
    lnks = -4276.1_dp/tk + 141.328_dp - 23.093_dp*lntk &
         + (-13856._dp/tk + 324.57_dp - 47.986_dp*lntk)*SQRT(i0) &
         + (35474._dp/tk - 771.54_dp + 114.723_dp*lntk)*i0 &
         + (-2698._dp*SQRT(i0) + 1776._dp*i0)*i0/tk &
         + (-dvi + 0.5_dp*dki*prs)*prt ! pressure correction
  END FUNCTION lnks_free

  !> \brief Boron system
  !!
  !!     kb = [H+][B(OH)4-]/[B(OH)3] : equilibrium constant of boric acid system
  !!                                   in mol/kg
  !! \param[in] prs pressure in bar
  ELEMENTAL FUNCTION lnkb_total (sal, tk, lntk, tc, prs, prt, ds, sqs) RESULT (lnkb)
    IMPLICIT NONE
    REAL(dp), INTENT(IN) :: sal, tk, lntk, tc, prs, prt, ds, sqs
    REAL(dp) :: lnkb, dvi, dki
    dvi = -29.48_dp + 0.295_dp*ds + (0.1622_dp - 0.002608_dp*tc)*tc
    dki = -2.84E-3_dp + 0.354E-3_dp*ds
    lnkb = (-8966.9_dp - 2890.53_dp*sqs - (77.942_dp - 1.728_dp*sqs + 0.0996*sal)*sal)/tk &
         + 148.0248_dp + (137.1942_dp + 0.053105_dp*tk)*sqs + 1.62142_dp*sal &
         + (-24.4344_dp - 25.085_dp*sqs - 0.2474_dp*sal)*lntk &
         + (-dvi + 0.5_dp*dki*prs)*prt ! pressure correction
  END FUNCTION lnkb_total

  !> \brief Ammonium system (SWS)
  !!
  !!    khf(f) = [H+]f[F-]/[HF] : equilibrium constant of fluoric acid system
  !!                              in mol/kg
  !! \param[in] prs pressure in bar
  ELEMENTAL FUNCTION lnknh4_sws (sal, tk, tc, prs, prt, sqs) RESULT (lnknh4)
    IMPLICIT NONE
    REAL(dp), INTENT(IN) :: sal, tk, tc, prs, prt, sqs
    REAL(dp) :: lnknh4, dvi, dki
    dvi = -26.43_dp + tc*(0.0889_dp - tc*0.905E-03_dp) ! volume change for ionization
    dki = -5.03E-3_dp + tc*0.0814E-3_dp                ! dki compressibility change for ionization
    lnknh4 = -0.25444_dp - 6285.33_dp/tk + 0.0001635_dp*tk &
         + (0.46532_dp - 123.7184_dp/tk)*sqs + (-0.01992_dp + 3.17556_dp/tk)*sal &
         + (-dvi + 0.5_dp*dki*prs)*prt ! pressure correction
  END FUNCTION lnknh4_sws

  !> \brief Flouride system (free scale)
  !!
  !!    khf(f) = [H+]f[F-]/[HF] : equilibrium constant of fluoric acid system
  !!                              in mol/kg
  !! \param[in] prs pressure in bar
  ELEMENTAL FUNCTION lnkf_free (tk, tc, prs, prt, i0) RESULT (lnkf)
    IMPLICIT NONE
    REAL(dp), INTENT(IN) :: tk, tc, prs, prt, i0
    REAL(dp) :: lnkf, dvi, dki
    dvi = -9.78_dp + tc*(-0.009_dp - tc*0.942E-03_dp) ! volume change for ionization
    dki = -3.91E-3_dp + tc*0.054E-3_dp                ! dki compressibility change for ionization
    lnkf = 1590.2_dp/tk - 12.641_dp + 1.525_dp*SQRT(i0) &
         + (-dvi + 0.5_dp*dki*prs)*prt ! pressure correction
  END FUNCTION lnkf_free

  !> \brief Flouride system (total scale)
  !!
  !!    khf = [H+][F-]/[HF] : equilibrium constant of fluoric acid system
  !!                          in mol/kg
  !! \param[in] prs pressure in bar
  ELEMENTAL FUNCTION lnkf_total (tk, tc, prs, prt, sqs) RESULT (lnkf)
    IMPLICIT NONE
    REAL(dp), INTENT(IN) :: tk, tc, prs, prt, sqs
    REAL(dp) :: lnkf, dvi, dki
    dvi = -9.78_dp + tc*(-0.009_dp - tc*0.942E-03_dp) ! volume change for ionization
    dki = -3.91E-3_dp + tc*0.054E-3_dp                ! dki compressibility change for ionization
    lnkf = 874._dp/tk - 9.68_dp + 0.111_dp*sqs &
         + (-dvi + 0.5_dp*dki*prs)*prt ! pressure correction
  END FUNCTION lnkf_total

  ! equilibrium constants of phosphate system in mol/kg
  !     k1p = [H+][H2PO4-]/[H3PO4]
  !     k2p = [H+][HPO4^2-]/[H2PO4-]
  !     k3p = [H+][PO4^3-]/[HPO4^2-]

  !> \brief First equilibrium constant of the phosphate system
  ELEMENTAL FUNCTION lnk1p_sws (sal, tk, lntk, tc, prs, prt, sqs) RESULT (lnk1p)
    IMPLICIT NONE
    REAL(dp), INTENT(IN) :: sal, tk, lntk, tc, prs, prt, sqs
    REAL(dp) :: lnk1p, dvi, dki
    dvi = -14.51_dp + (0.1211_dp - 0.321E-03_dp*tc)*tc ! volume change for ionization
    dki = -2.67E-3_dp + 0.0427E-3_dp*tc                ! compressibility change for ionization
    lnk1p = -4576.752_dp/tk + 115.54_dp - 18.453_dp*lntk &
         + (-106.736_dp/tk + 0.69171_dp)*sqs + (-0.65643_dp/tk - 1.844E-2_dp)*sal &
         + (-dvi + 0.5_dp*dki*prs)*prt ! pressure correction
  END FUNCTION lnk1p_sws

  !> \brief Second equilibrium constant of the phosphate system
  !!
  !! \param[in] prs pressure in bar
  ELEMENTAL FUNCTION lnk2p_sws (sal, tk, lntk, tc, prs, prt, sqs) RESULT (lnk2p)
    IMPLICIT NONE
    REAL(dp), INTENT(IN) :: sal, tk, lntk, tc, prs, prt, sqs
    REAL(dp) :: lnk2p, dvi, dki
    dvi = -23.12_dp + (0.1758_dp -2.647E-03_dp*tc)*tc ! volume change for ionization
    dki = -5.15E-3_dp +   0.09E-3_dp*tc               ! compressibility change for ionization
    lnk2p = -8814.715_dp/tk + 172.1033_dp - 27.927_dp*lntk&
         + (-160.34_dp/tk + 1.3566_dp)*sqs + (0.37335_dp/tk - 5.778D-2)*sal &
         + (-dvi + 0.5_dp*dki*prs)*prt ! pressure correction
  END FUNCTION lnk2p_sws

  !> \brief Third equilibrium constant of the phosphate system
  !!
  !! \param[in] prs pressure in bar
  !! \param[in] prt (pressure in bar)/(R*T)
  ELEMENTAL FUNCTION lnk3p_sws (sal, tk, tc, prs, prt, sqs) RESULT (lnk3p)
    IMPLICIT NONE
    REAL(dp), INTENT(IN) :: sal, tk, tc, prs, prt, sqs
    REAL(dp) :: lnk3p, dvi, dki
    dvi = -26.57_dp + (0.202_dp -3.042E-03_dp*tc)*tc ! volume change for ionization
    dki = -4.08E-3_dp + 0.0714E-3_dp*tc              ! compressibility change for ionization
    lnk3p = -3070.75_dp/tk - 18.126_dp + (17.27039_dp/tk + 2.81197_dp)*sqs&
         + (-44.99486_dp/tk - 9.984E-2_dp)*sal &
         + (-dvi + 0.5_dp*dki*prs)*prt ! pressure correction
  END FUNCTION lnk3p_sws

  !> \brief Equilibrium constant for aragonite after \cite Millero:1995
  !!
  !! \param[in] sal salinity in psu
  !! \param[in] tk  temperature in K
  !! \param[in] tc  temperature in °C
  !! \param[in] prs pressure in bar
  !! \return common logarithm of the equilibrium constant for aragonite
  ELEMENTAL FUNCTION lgkspa (sal, tk, tc, prs) RESULT (lgk)
    IMPLICIT NONE
    REAL(dp), INTENT(IN) :: sal, tk, tc, prs
    REAL(dp) :: dV, dk, lgk
    dV = -35._dp + 0.5304_dp*tc         ! Eq. (91)
    dk = -11.76E-3_dp + 0.3692E-3_dp*tc ! Eq. (92)
    lgk = -171.945_dp - 0.077993_dp*tk + 2903.293_dp/tk + 71.595_dp*LOG10(tk) & ! Eq. (79)
         + (-0.068393_dp + 1.7276E-3_dp*tk + 88.135_dp/tk + 5.9415E-3_dp*sal)*SQRT(sal) - 0.10018_dp*sal &
         - (dV - 0.5_dp*dk*prs)/(Rgas*tk)*prs ! Eq. (90)
  END FUNCTION lgkspa

  !> \brief Equilibrium constant for calcite after \cite Millero:1995
  !!
  !! \param[in] sal salinity in psu
  !! \param[in] tk  temperature in K
  !! \param[in] tc  temperature in °C
  !! \param[in] prs pressure in bar
  !! \return common logarithm of the equilibrium constant for calcite
  ELEMENTAL FUNCTION lgkspc (sal, tk, tc, prs) RESULT (lgk)
    IMPLICIT NONE
    REAL(dp), INTENT(IN) :: sal, tk, tc, prs
    REAL(dp) :: dV, dk, lgk
    dV = -48.76_dp + 0.5304_dp*tc       ! Eq. (91)
    dk = -11.76E-3_dp + 0.3692E-3_dp*tc ! Eq. (92)
    lgk = -171.9065_dp - 0.077993_dp*tk + 2839.319_dp/tk + 71.595_dp*LOG10(tk) & ! Eq. (78)
         + (-0.77712_dp + 2.8426E-3_dp*tk + 178.34_dp/tk + 4.1249E-3_dp*sal)*SQRT(sal) - 0.07711_dp*sal &
         - (dV - 0.5_dp*dk*prs)/(Rgas*tk)*prs ! Eq. (90)
  END FUNCTION lgkspc

  !> parameters for the DIC system and alkalinity (total scale)
  !!  \param[in]   box%tk         temperature in K
  !!  \param[in]   box%salinity   salinity in psu
  !!  \param[in]   box%alkalinity alkalinity in \unit{mmol.m^{-3}}
  !!  \param[in]   box%depth      depth in m
  !!  \param[in]   box%DIC        DIC in \unit{mmol.m^{-3}}
  !!  \param[out]  box%rho        density in \unit{g.m^{-3}}
  !!  \param[out]  box%at         alkalinity in \unit{mol.kg^{-1}}
  !!  \param[out]  box%ct         DIC in \unit{mol.kg^{-1}}
  SUBROUTINE dicalp (dic, box)
    IMPLICIT NONE
    CLASS(dicsys), INTENT(INOUT) :: dic
    TYPE(layer), INTENT(INOUT) :: box
    REAL(dp), PARAMETER :: rcf=68E-6_dp/35._dp, rcb=0.232D-3, rcSO4=0.14_dp/96.062_dp/1.80655_dp
    REAL(dp) :: i0, lntk, lnksi, sqs, mH2O, kSO4_free, prs, prt, delsal
    prs = 0.1_dp*box%depth                  ! pressure in bar
    delsal = box%salinity - 34.8_dp
    sqs = SQRT(box%salinity)
    mH2O = 1._dp - 1.005E-3_dp*box%salinity ! mass of H2O in 1 kg seawater
    i0 = 19.924E-3_dp*box%salinity/mH2O     ! ionic strength
    box%tk = box%temperature + t0ck         ! convert °C to K
    prt = prs/(Rgas*box%tk)                 ! (pressure in bar)/(R*T)
    lntk = LOG(box%tk)
    box%rho = eos80_rho(tmp=box%temperature, sal=box%salinity)
    dic%at = box%alkalinity/box%rho ! convert mmol/m3 to mol/kg
    dic%ct = box%DIC/box%rho
    dic%pt = box%DIP/box%rho
    dic%sit = box%silc/box%rho
    dic%ft = rcf*box%salinity
    dic%bt = rcb/10.811_dp*box%salinity/1.80655_dp
    kSO4_free = EXP(lnks_free(tk=box%tk, lntk=lntk, tc=box%temperature, prs=prs, prt=prt, i0=i0))*mH2O
    dic%SO4 = rcSO4*box%salinity
    dic%r_tot_free = 1._dp + dic%SO4/kSO4_free
    dic%r_tot_sws = dic%r_tot_free/(dic%ft/EXP(lnkf_free(tk=box%tk, tc=box%temperature, prs=prs, &
                                               prt =prt, i0=i0)) + dic%r_tot_free)
    ! water
    dic%kw = EXP(lnkw_sws(sal=box%salinity, tk=box%tk, lntk=lntk, tc=box%temperature, &
                          prs=prs, prt=prt, sqs=sqs))*dic%r_tot_sws
    ! carbon system
    dic%k1 = EXP(lnk1_total(sal=box%salinity, tk=box%tk, lntk=lntk, tc=box%temperature, &
                            prs=prs, prt=prt, ds=delsal))
    dic%k2 = EXP(lnk2_total(sal=box%salinity, tk=box%tk, lntk=lntk, tc=box%temperature, &
                            prs=prs, prt=prt, ds=delsal))
    dic%k12 = dic%k1*dic%k2
    ! boron buffer
    dic%kb = EXP(lnkb_total(sal=box%salinity, tk=box%tk, lntk=lntk, tc=box%temperature, &
                            prs=prs, prt=prt, ds=delsal, sqs=sqs))
    ! sulphate buffer
    dic%kSO4 = kSO4_free*dic%r_tot_free
    ! flouride
    dic%kf = EXP(lnkf_total(tk=box%tk, tc=box%temperature, prs=prs, prt=prt, sqs=sqs))
    ! silicic acid
    lnksi = -8904.2_dp/box%tk + 117.4_dp - 19.334_dp*lntk&
         + (-458.79_dp/box%tk + 3.5913_dp)*SQRT(i0) + ((188.74_dp/box%tk - 1.5998_dp) &
         + (-12.1652_dp/box%tk + 7.871E-2_dp)*i0)*i0
    dic%ksi = EXP(lnksi)*mH2O*dic%r_tot_sws
    ! phosphate buffer
    dic%k1p = EXP(lnk1p_sws(sal=box%salinity, tk=box%tk, lntk=lntk, tc=box%temperature, &
                            prs=prs, prt=prt, sqs=sqs))*dic%r_tot_sws
    dic%k2p = EXP(lnk2p_sws(sal=box%salinity, tk=box%tk, lntk=lntk, tc=box%temperature, &
                            prs=prs, prt=prt, sqs=sqs))*dic%r_tot_sws
    dic%k3p = EXP(lnk3p_sws(sal=box%salinity, tk=box%tk, tc=box%temperature, &
                            prs=prs, prt=prt, sqs=sqs))*dic%r_tot_sws
    dic%kspa = 10**(lgkspa(sal=box%salinity, tk=box%tk, tc=box%temperature, prs=prs))
    dic%kspc = 10**(lgkspc(sal=box%salinity, tk=box%tk, tc=box%temperature, prs=prs))
  END SUBROUTINE dicalp

  !> \brief Concentrations of protons and DIC species and aragonite and calcite saturation
  !!
  !! \param[in]  box%alkalinity   total alkalinity in \unit{\micro mol.L^{-1}} (or box%DIN and dic%potalk)
  !! \param[in]  box%salinity     salinity in psu
  !! \param[in]  box%temperature  temperature in °C
  !! \param[in]  box%DIC          DIC concentration in \unit{mmol.m^{-3}}
  !! \param[in]  box%DIP          phosphate concentration in \unit{mmol.m^{-3}}
  !! \param[in]  box%silc         silicate concentration in \unit{mmol.m^{-3}}
  !! \param[inout] box%hc         proton concentration in \unit{mol.kg^{-1}}
  !! \param[out]   box%tk            temperature in K
  !! \param[out]   box%CO2           \ce{CO2} concentration in \unit{mol.kg^{-1}}
  !! \param[out]   box%HCO3          \ce{HCO3^-} concentration in \unit{mol.kg^{-1}}
  !! \param[out]   box%CO3           \ce{CO3^2-} concentration in \unit{mol.kg^{-1}}
  !! \param[out]   box%rho           density in \unit{g.m^{-3}}
  !! \param[out]   box%omega         aragonite saturation, \cite Millero:1995, Eq. (81)
  !! \param[out]   box%omega_calcite calcite saturation, \cite Millero:1995, Eq. (80)
  SUBROUTINE co2 (dic, box, niter)
    IMPLICIT NONE
    CLASS(dicsys), INTENT(INOUT) :: dic
    TYPE(layer), INTENT(INOUT) :: box
    INTEGER, INTENT(IN) :: niter
    CALL dicalp (dic, box)     ! equilibrium constants etc.
    CALL dic%Hsolve (box=box, niter=niter)
    dic%acd = box%hc*(box%hc + dic%k1) + dic%k12
    box%CO2 = box%hc*box%hc*dic%ct/dic%acd
    box%HCO3 = box%hc*dic%ct*dic%k1/dic%acd
    box%CO3 = dic%ct*dic%k12/dic%acd
    box%omega = dic%Ca*box%CO3/dic%kspa         ! Millero (1995), Eq. (81)
    box%omega_calcite = dic%Ca*box%CO3/dic%kspc ! Millero (1995), Eq. (80)
  END SUBROUTINE co2

  !> \brief Initialise DIC system after \cite Follows:2006
  !!
  !! \param[out] box%hc    initial estimate for \ce{[H+]} in \unit{mol.kg^{-1}}
  SUBROUTINE Hini_F06 (dic, box)
    IMPLICIT NONE
    CLASS(dicsys), INTENT(INOUT) :: dic
    TYPE(layer), INTENT(INOUT) :: box(0:)
    INTEGER :: nb
    box%hc = 10._dp**(-dic%pH0)
    DO nb=1,(SIZE(box) - 2)
       CALL dic%co2 (box=box(nb), niter=6)
    END DO
  END SUBROUTINE Hini_F06

  !> Concentrations of \ce{H+} from alkalinity and total DIC concentration
  !! \ce{[H+]} is calculated after \cite Follows:2006 but including sulphate and fluoride systems.
  SUBROUTINE Hsolve_F06 (dic, box, niter)
    IMPLICIT NONE
    CLASS(dicsys), INTENT(INOUT) :: dic
    TYPE(layer), INTENT(INOUT) :: box
    INTEGER, INTENT(IN) :: niter
    REAL(dp) :: k123
    INTEGER :: i
    dic%k12p = dic%k1p*dic%k2p
    k123 = dic%k12p*dic%k3p
    DO i=1,niter
       dic%ac = MIN(dic%at - dic%bt*dic%kb/(box%hc + dic%kb) - dic%kw/box%hc&
            - dic%pt*(box%hc*(dic%k12p - box%hc*box%hc) + 2._dp*k123)&
             /(box%hc*(dic%k12p + box%hc*(dic%k1p + box%hc)) + k123) &
            - dic%ksi*dic%sit/(box%hc + dic%ksi)&
            + box%hc*(dic%kSO4/(dic%kSO4 + dic%SO4) + dic%SO4/(box%hc + dic%kSO4 + dic%SO4)&
                      + dic%ft/(box%hc + dic%kf)), 1.9_dp*dic%ct)
       dic%gmk = (dic%ct/dic%ac - 1._dp)*dic%k1
       box%hc = 0.5D0*dic%gmk + SQRT(dic%gmk*(0.25D0*dic%gmk + dic%k2) + dic%k1*dic%k2*dic%ct/dic%ac)
    END DO
  END SUBROUTINE Hsolve_F06

  !> alkalinity as a function of potential alkalinity and DIN
  SUBROUTINE dic_potalk (dic, box)
    IMPLICIT NONE
    CLASS(dicsys), INTENT(IN) :: dic
    TYPE(layer), INTENT(INOUT) :: box
    box%alkalinity = (dic%potalk*box%salinity - box%DIN)
    ENTRY dic_alk (dic, box)
  END SUBROUTINE dic_potalk

  !> seawater density in \unit{g.m^{-3}} after \cite Millero:1980
  !! \param[in] tmp  temperature in °C
  !! \param[in] sal  salinity in psu
  REAL(dp) ELEMENTAL FUNCTION eos80_rho (tmp, sal) RESULT (rho)
    IMPLICIT NONE
    REAL(dp), INTENT(IN) :: tmp, sal
    rho = 999.842594E3_dp &
         + tmp*(67.93952_dp + tmp*(-9.09529_dp + tmp*(0.1001685_dp + tmp*(-1.120083E-3_dp + tmp*6.536332E-6_dp)))) &
         + sal*(824.493_dp + tmp*(-4.0899_dp + tmp*(7.6438E-2_dp + tmp*(-8.2467E-4_dp + tmp*5.3875E-6_dp))) &
                + SQRT(sal)*(-5.72466_dp + tmp*(0.10227_dp - tmp*1.6546E-3_dp)) + sal*4.8314E-1_dp)
  END FUNCTION eos80_rho

  !> Initial guess for [H+] after \cite Munhoven:2013
  !!
  !! This is an adaptation of subroutine AHINI_FOR_AT from \cite Munhoven:2013.
  !! Total alkalinity (box%alkalinity) is used as an approximation for
  !! carbonate alkalinity.
  SUBROUTINE Hini_M13 (dic, box)
    IMPLICIT NONE
    CLASS(dicsys), INTENT(INOUT) :: dic
    TYPE(layer), INTENT(INOUT) :: box(0:)
    INTEGER :: nb
    DO nb=1,(SIZE(box) - 2)
       CALL Hset_M13 (dic, box(nb))
    END DO
  CONTAINS
    SUBROUTINE Hset_M13 (dic, box)
      CLASS(dicsys), INTENT(INOUT) :: dic
      TYPE(layer), INTENT(INOUT) :: box
      REAL(dp) :: ca, ba, a0, a1, a2, dis, sdis, hmin
      CALL dic%alkalinity (box=box)          ! set alkalinity
      CALL dicalp (dic, box)                 ! equilibrium constants etc.
      IF (box%alkalinity.LE.0._dp) THEN
         box%hc = 1E-3_dp
      ELSE IF (box%alkalinity.GE.(2._dp*dic%ct + dic%bt)) THEN
         box%hc = 1E-10_dp
      ELSE
         ca = dic%ct/box%alkalinity
         ba = dic%bt/box%alkalinity
         ! coefficients of the cubic polynomial
         a0 = dic%k12*dic%kb*(1._dp - ba - 2._dp*ca)
         a1 = dic%k1*dic%kb*(1._dp - ba - ca) + dic%k12*(1._dp - 2._dp*ca)
         a2 = dic%kb*(1._dp - ba) + dic%k1*(1._dp - ca)
         dis = a2*a2 - 3._dp*a1              ! discriminant of the quadratic equation
         IF (dis > 0._dp) THEN               ! if the discriminant is positive
            sdis = SQRT(dis)
            IF (a2 < 0._dp) THEN
               hmin = (-a2 + sdis)/3._dp
            ELSE
               hmin = -a1/(a2 + sdis)
            END IF
            box%hc = hmin + SQRT(-(a0 + hmin*(a1 + hmin*(a2 + hmin)))/sdis)
         ELSE
            box%hc = 1E-7_dp
         END IF
      END IF
      CALL dic%co2 (box=box, niter=4)
    END SUBROUTINE Hset_M13
  END SUBROUTINE Hini_M13

  !> Solve for [H+] given alkalinity and total DIC concentration.
  !!
  !! This is an adaptation of function SOLVE_AT_GENERAL_SEC (without ammonium and H2S)
  !! from \cite Munhoven:2013 for the total pH scale.  Munhoven's routine has more iterations but here the
  !! result from the previous call is used as the initial guess.  Also,
  !! iterations are done only if the error in the alkalinity equation
  !! (DeltaAlk) is greater than 1e-9 (relative error > 1e-6).
  SUBROUTINE Hsolve_M13 (dic, box, niter)
    IMPLICIT NONE
    CLASS(dicsys), INTENT(INOUT) :: dic
    TYPE(layer), INTENT(INOUT) :: box
    INTEGER, INTENT(IN) :: niter
    REAL(dp) :: delta, Hmin, Hmax, H0, H1, H2, dA0, dA1, dA2, absmin_dA,&
         alkInf,&              !< infimum (lower bound) of non-water contributions to alkalinity
         alkSup                !< supremum (upper bound) of non-water contributions to alkalinity
    INTEGER :: n
    alkInf = -dic%pt - dic%SO4 - dic%ft                 ! infinimum
    alkSup = 2._dp*(dic%ct + dic%pt) + dic%bt + dic%sit ! supremum
    ! lower limit for [H+]
    delta = SQRT((dic%at - alkInf)**2 + 4._dp*dic%kw/dic%r_tot_free)
    IF (dic%at.GE.alkInf) THEN
       Hmin = 2._dp*dic%kw/(dic%at - alkInf + delta)
    ELSE
       Hmin = dic%r_tot_free*(alkInf - dic%at + delta)/2._dp
    END IF
    ! upper limit for [H+]
    delta = SQRT((dic%at - alkSup)**2 + 4._dp*dic%kw/dic%r_tot_free)
    IF (dic%at.LE.alkSup) THEN
       Hmax = dic%r_tot_free*(alkSup - dic%at + delta)/2._dp
    ELSE
       Hmax = 2._dp*dic%kw/(dic%at - alkSup + delta)
    END IF
    H0 = MAX(MIN(box%hc, Hmax), Hmin)
    H2 = H0
    dA2 = deltaAlk(dic, hplus=H2)
    absmin_dA = ABS(dA2)
    IF (dA2.LT.-1E-9_dp) THEN
       Hmax = H2
       H1 = MAX(Hmax/1.25_dp, SQRT(Hmin*Hmax))
    ELSE IF (dA2.GT.1E-9_dp) THEN
       Hmin = H2
       H1 = MIN(Hmin*1.25_dp, SQRT(Hmin*Hmax))
    ELSE                        ! good enough
       box%hc = H2
       RETURN
    END IF
    dA1 = deltaAlk(dic, hplus=H1)
    IF (dA1.GT.1E-9_dp) THEN
       Hmin = H1
    ELSE IF (dA1.LT.-1E-9_dp) THEN
       Hmax = H1
    ELSE                        ! good enough
       box%hc = H1
       RETURN
    END IF
    IF (ABS(dA1).GT.absmin_dA) THEN
       H0 = H1
       dA0 = dA1
       H1 = H2
       dA1 = dA2
       H2 = H0
       dA2 = dA0
    ELSE
       absmin_dA = abs(dA1)
    END IF
    H0 = H1 - dA1*(H2 - H1)/(dA2 - dA1)
    IF (H0.GT.Hmax) H0 = SQRT(H1*Hmax)
    IF (H0.LT.Hmin) H0 = SQRT(H1*Hmin)
    DO n=1,niter
       dA0 = deltaAlk(dic, hplus=H0)
       IF (dA0.GT.0._dp) Hmin = H0
       IF (dA0.LT.0._dp) Hmax = H0
       H2 = H1
       dA2 = dA1
       H1 = H0
       dA1 = dA0
       IF (ABS(dA0).GE.0.5_dp*absmin_dA) THEN
          H0 = SQRT(Hmax*Hmin)
       ELSE
          H0 = H1 - dA1*(H2 - H1)/(dA2 - dA1)
          IF (H0.LT.Hmin) H0 = SQRT(H1*Hmin)
          IF (H0.GT.Hmax) H0 = SQRT(H1*Hmax)
       END IF
       absmin_dA = MIN(ABS(dA0), absmin_dA)
    END DO
    box%hc = H0
  CONTAINS
    FUNCTION deltaAlk (dic, hplus) RESULT (dAlk)
      IMPLICIT NONE
      CLASS(dicsys), INTENT(IN) :: dic
      REAL(dp), INTENT(IN) :: hplus
      REAL(dp) :: dAlk, alkDic, alkBor, alkPO4, alkSil, alkSO4, alkF, alkH2O
      alkDic = dic%ct*(2._dp*dic%k12 + hplus*dic%k1)/(dic%k12 + hplus*(dic%k1 + hplus))
      alkBor = dic%bt*dic%kb/(dic%kb + hplus)
      alkPO4 = dic%pt*(3._dp*dic%k3p + hplus*(2._dp*dic%k2p + hplus*dic%k1p))&
                     /(dic%k3p + hplus*(dic%k2p + hplus*(dic%k1p + hplus)))
      alkSil = dic%sit*dic%ksi/(dic%ksi + hplus)
      alkSO4 = dic%SO4*(dic%kSO4/(dic%kSO4 + hplus) - 1._dp)
      alkF = dic%ft*(dic%kf/(dic%kf + hplus) - 1._dp)
      alkH2O = dic%kw/hplus - hplus/dic%r_tot_free
      dAlk = alkDic + alkBor + alkPO4 + alkSil + alkSO4 + alkF + alkH2O - dic%at
    END FUNCTION deltaAlk
  END SUBROUTINE Hsolve_M13
END MODULE dic
