!> \brief VODE\_F90 interface for oppla
MODULE dvode
  USE DVODE_F90_M
  USE NETCDF, ONLY : NF90_MAX_NAME
  USE stdunt
  IMPLICIT NONE
  PRIVATE
  PUBLIC :: VODE_OPTS, SET_OPTS, GET_STATS, VODE_F90, XSETF
  TYPE, PUBLIC :: ode
     REAL(dp), DIMENSION(:), ALLOCATABLE :: atol,& !< absolute tolerances
          rtol,&                                   !< relative tolerances
          mxatol,&                                 !< maxima of atol
          mxrtol,&                                 !< maxima of rtol
          clower,&                                 !< lower constraints
          cupper                                   !< upper constraints
     REAL(dp) :: h0=0D0, hmin=0D0, hmax=0D0
     INTEGER :: itask=4, mf, mxstep=500, mxhnil=0, maxord=12, meth=1, miter=3,&
                jsv=1, neq=0, nconstr=0, moss=0
     INTEGER, DIMENSION(:), ALLOCATABLE :: constrained
     CHARACTER(LEN=NF90_MAX_NAME), DIMENSION(:), ALLOCATABLE :: names
     LOGICAL :: quiet=.FALSE., constr=.TRUE., sparse_jacobian
   CONTAINS
     PROCEDURE :: read => read_ode
     PROCEDURE :: set_atol => set_atol
  END TYPE ode

CONTAINS

  !> \brief Read DVODE parameters
  SUBROUTINE read_ode (parode, lun, nconstr)
    IMPLICIT NONE
    CLASS(ode), INTENT(INOUT) :: parode
    INTEGER, INTENT(IN) :: lun     !< logical unit number for namelist vode
    INTEGER, INTENT(IN) :: nconstr !< number of constrained states
    CHARACTER(LEN=512) :: msg='', filename
    REAL(dp) :: atol, rtol, mxatol, mxrtol, h0, hmin, hmax
    INTEGER :: itask, mxstep, mxhnil, maxord, meth, miter, jsv, n
    LOGICAL :: quiet, constr
    NAMELIST /vode/ atol, rtol, h0, hmin, hmax, quiet, constr, itask,&
         mxrtol, mxatol, mxstep, mxhnil, maxord, meth, miter, jsv
    parode%nconstr = nconstr
    atol = 1.E-9_dp
    mxatol = 0._dp
    rtol = 1.E-6_dp
    mxrtol = 1.E-5_dp
    h0 = parode%h0
    hmin = parode%hmin
    hmax = parode%hmax
    quiet = parode%quiet
    constr = parode%constr
    itask = parode%itask
    jsv = parode%jsv
    mxstep = parode%mxstep
    mxhnil = parode%mxhnil
    maxord = parode%maxord
    meth = parode%meth
    miter = parode%miter
    REWIND (UNIT=lun)
    READ (UNIT=lun, NML=vode, END=10, IOMSG=msg, ERR=900)
    IF (meth.EQ.1) maxord = MIN(maxord, 12)
    IF (meth.EQ.2) maxord = MIN(maxord, 5)
    parode%h0 = h0
    parode%hmin = hmin
    parode%hmax = hmax
    parode%quiet = quiet
    parode%constr = constr
    parode%itask = itask
    parode%jsv = jsv
    parode%mxstep = mxstep
    parode%mxhnil = mxhnil
    parode%maxord = maxord
    parode%meth = meth
    parode%miter = miter
10  CONTINUE
    IF (LEN_TRIM(msg).GT.0) THEN
       INQUIRE (UNIT=lun, NAME=filename)
       WRITE (stderr,'("read_ode: namelist vode not in ",A,".")') TRIM(filename)
    END IF
    parode%sparse_jacobian = miter.EQ.7
    IF (parode%sparse_jacobian) parode%moss = 2
    parode%mf = parode%moss*100 + jsv*(10*meth + miter)
    IF (ALLOCATED(parode%constrained)) &
         DEALLOCATE (parode%constrained, parode%clower, parode%cupper)
    ALLOCATE(parode%atol(parode%neq), parode%mxatol(parode%neq),&
         parode%rtol(parode%neq), parode%mxrtol(parode%neq),&
         parode%constrained(parode%nconstr),&
         parode%clower(parode%nconstr), parode%cupper(parode%nconstr))
    parode%constrained = [(n, n=1,parode%nconstr)] ! all plankton states are constrained
    parode%atol = atol
    IF (mxatol.LE.0._dp) THEN
       parode%mxatol = atol * 10._dp
    ELSE
       parode%mxatol = mxatol
    END IF
    parode%rtol = rtol
    IF (mxrtol.LE.0._dp) THEN
       parode%mxrtol = rtol * 10._dp
    ELSE
       parode%mxrtol = mxrtol
    END IF
    parode%clower = 0D0
    parode%cupper = 1D6
    RETURN
900 CONTINUE
    WRITE (stderr,'("read_ode:",A)') TRIM(msg)
    STOP 1
  END SUBROUTINE read_ode

  !> \brief Set absolute tolerances for each state individually
  !!
  !! Absolute tolerances atol and mxatol are assigned individually if provided
  !! as negative values in namelist vode.
  SUBROUTINE set_atol (parode, stv)
    IMPLICIT NONE
    CLASS(ode), INTENT(INOUT) :: parode
    REAL(dp), INTENT(IN) :: stv(*)
    IF (ALL(parode%atol.LT.0.0_dp)) THEN
       parode%atol(1:parode%nconstr) = -parode%atol(1:parode%nconstr)*stv(1:parode%nconstr)
       parode%atol(parode%nconstr+1:parode%neq) = ABS(parode%atol(parode%nconstr+1:parode%neq))
    END IF
    IF (ALL(parode%mxatol.LT.0.0_dp)) THEN
       parode%mxatol(1:parode%nconstr) = -parode%mxatol(1:parode%nconstr)*stv(1:parode%nconstr)
       parode%mxatol(parode%nconstr+1:parode%neq) = ABS(parode%mxatol(parode%nconstr+1:parode%neq))
    END IF
  END SUBROUTINE set_atol

END MODULE dvode
