!> \brief Ecological types
!!
!! This module defines four types: group, state, local, layer, and timing.  The
!! types group and state form the core of representation of the plankton
!! community.  The state type has only one component, var, which is an
!! allocatable array of pointers of class group.  The group type is extended in
!! the modules for the functional types, cfo, cmo, bac, etc.  Each element of
!! the var array represents one functional group of the model plankton
!! ecosystem.
!!
!! This module also provides a suite of temperature functions and ths
!! subroutine fT_select, which are used by the functional-group modules to
!! select temperature functions for individual processes.
MODULE et
  USE stdunt
  USE fudu
  USE ISO_FORTRAN_ENV, ONLY: INT64
  IMPLICIT NONE
  !> \brief Common components for all functional groups
  !!
  !! - Properties common to all groups, e.g., all have a name and track one or
  !!   more constituents (C, N, P, ...).
  !! - TYPE(group) is extended in the functional-group modules bac, cmo,
  !!   cfo, etc., to declare the actual types of the functional groups which are
  !!   then associated with pointer var in variable fungroup (module plankton) of TYPE(state)
  !! - The central part of this type are the deferred procedure pointers flux,
  !!   read, and set, which must be associated (over-ridden) in the functional-group modules.
  REAL(dp), PARAMETER :: pi=3.14159265358979323_dp, rad=pi/180._dp
  CHARACTER(LEN=*), PARAMETER :: constituents(*)=(/'C  ', 'N  ', 'P  ', 'PIC', 'Chl'/),&
       constituent_units(*)=(/'mmol m-3 ', 'mmol m-3 ', 'mmol m-3 ', 'mmol m-3 ', 'mgChl m-3'/)
  TYPE, ABSTRACT :: fungroup
     CHARACTER(LEN=512) :: name                    !< name of functional group
     CHARACTER(LEN=100), ALLOCATABLE :: names(:),& !< names of state variables
          units(:),&                               !< units of state variables
          constituents(:)                          !< constituents of this functional group
     INTEGER :: ifg,&                              !< index of functional group
          ncon,&                                   !< number of constituents
          nsv=1,&                                  !< number of state variables
          nq0=0                                    !< number of subsistence quotas
     INTEGER, ALLOCATABLE :: isv(:),& !< vector of state-variable indices
          icon(:),&                   !< local constituent indices
          ir(:),&                     !< global ratio indices
          iq0(:)                      !< group-sepcific indices of subsistence quotas
     LOGICAL :: motile=.FALSE.
     REAL(dp), ALLOCATABLE :: Q0(:) !< vector of subsistence quotas
     REAL(dp), POINTER :: OC,&      !< organic C
          QN,&                      !< N:C ratio (cell quota)
          QP,&                      !< P:C ratio (cell quota)
          stick,&                   !< stickiness (\unit{m^3.(mmol.C)^{-1}})
          vv(:),&                   !< vertical velocities
          ratios(:)                 !< variable ratios of this group
     REAL(dp) :: Q10=1.89_dp, Tref=27._dp, Topt=15._dp, Tspr=12._dp, sticky=0._dp
     PROCEDURE(fT_Eppley), POINTER :: fT => fT_Eppley ! temperature function
   CONTAINS
     PROCEDURE :: fT_select => fT_select
     PROCEDURE(fluxes), DEFERRED :: flux !< calculate fluxes in the source matrix soma
     PROCEDURE(init), DEFERRED :: read   !< read parameters of each functional type
     PROCEDURE(preset), DEFERRED :: set  !< define ratios, set indices, pointers
  END TYPE fungroup
  TYPE :: state
     CLASS(fungroup), POINTER :: var !< pointer to functional group of type bacteria, phycmo, zoocfo, etc.
  END TYPE state
  TYPE :: local
     INTEGER :: idic=0, io2=0, idin=0, idip=0, ialk=0, isi=0, irdoc=0, irdon=0,&
          idagg=0,&                !< index of dretritus group representing aggregates
          nft,&                    !< No. of functional types
          nbox,&                   !< No. of boxes (layers)
          ncon,&                   !< No. of different material constituents
          nq0,&                    !< No. of subsistence quotas
          nrs,&                    !< No. of quotas
          ndm,&                    !< No. of dissolved-matter variables except labile DOM
          ndom,&                   !< No. of DOM groups
          nagg,&                   !< No. of types involved in aggregation
          nbac, ndet, nphy, nzoo   !< No. of bacteria, detritus, phytoplankton, zooplankton groups
     INTEGER, DIMENSION(:), ALLOCATABLE :: &
          iq0r,&                !< ratio-indices of subsistence quotas
          irc,&                 !< indices of C tracers (reference tracers for ratios)
          irt,&                 !< tracer-indices of quota tracers
          ichl,&                !< indices of Chl tracers
          ipic,&                !< indices of PIC tracers
          ipoc,&                !< indices of POC tracers
          ifalk,&               !< indices of tracers influencing alkalinity
          isfalk,&              !< indices of surface fluxes influencing alkalinity
          iphy, iagg
     CHARACTER(LEN=256) :: dicy='',&                  !< diurnal light cycle
          daylenfun=''
     LOGICAL :: calcPAR=.TRUE.,& !< should PAR be calculated from clear-sky radiation in sunday_brock81?
          xsfalk                 !< do any surface fluxes affect alkalinity?
     !> quota mask: .TRUE. if the quota is variable (a dynamic ratio)
     LOGICAL, ALLOCATABLE :: qmsk(:,:),&
          fmsk(:,:),&                                 !< fluxes-matrix mask
          smsk(:,:,:)                                 !< source-matrix mask
     REAL(dp), DIMENSION(:,:,:), ALLOCATABLE :: soma  !< source matrices
     !> biomass (C)-normalised quotas for all groups, dimensions are (nq,nft):\n
     !! quotas(1,:) = 1, quotas(2,:) = QN, quotas(3,:) = QP, etc.
     REAL(dp), DIMENSION(:,:), ALLOCATABLE :: quotas,&
          fluxes,&                               !< flux vectors
          heights,&                              !< layer heights
          all_ratios,&                           !< all ratios in all boxes
          kij                                    !< coagulation kernel matrix
     REAL(dp), DIMENSION(:), ALLOCATABLE :: Q0,& !< subsistence quotas
          ratios,&                               !< ratios of all groups
          PAR,&                                  !< surface PAR for each layer
          fPAR,&           !< fraction of sPAR arriving at the top of each layer
          fALK,&           !< alkalinity factors for DIN, PIC, etc.
          fsfALK,&         !< alkalinity factors for surface fluxes
          OC,&             !< DOC or POC for all groups
          sticky, agg,&
          depth,&
          vvs,&                 !< relative vertical velocities (sinking, vertical migration, etc.)
          vvstv,&               !< absolute vertical velocities, including up- or downwelling
          dydz,&
          bbc                      !< bottom-boundary conditions for layers
     REAL(dp), POINTER :: dayPAR,& !< average day-time surface PAR
          sPAR,&                   !< instantaneous surface PAR
          lacw,&                   !< light-attenuation coefficient of water in \unit{m^{-1}}
          pressure,&               !< atmospheric surface pressure in Pa
          xCO2,&                   !< mole fraction of \ce{CO2} in air
          sfDIC,&                  !< suface flux of DIC in \unit{mmol.m^{-2}.s^{-1}}
          sfO2,&                   !< suface flux of \ce{O2} in \unit{mmol.m^{-2}.s^{-1}}
          wind,&                   !< wind speed in \unit{m.s^{-1}}
          daylen,&                 !< day length fraction
          ice,&                    !< ice cover fraction
          depth_edges(:,:)         !< depths of layer edges
     REAL(dp) :: lachl=16D0, lapon=16D0,&    !< light attenuation coefficient of chl, pon
          RQ=1.2D0,&               !< respiratory quotient in \unit{mol.O2.(mol.C)^{-1}}
          temperature,&            !< constant temperature if no profile data available
          salinity,&               !< constant salinity if no profile data available
          vdc,&                    !< constant vertical mixing if no profile data available
          velocity                 !< constant vertical velocity if no profile data available
     PROCEDURE(light), POINTER :: PARfun !< calculate or read surface irradiance
     PROCEDURE(attenuate), POINTER :: attenuation !< light attenuation
     PROCEDURE(fpar), POINTER :: PARfrac          !< fractions of surface PAR arriving at the box bottoms
     PROCEDURE(boxbc), POINTER :: boxbc           !< box-boundary concentrations/conditions
     PROCEDURE(boxalk), POINTER :: alkalinity     !< alkalinity
     PROCEDURE(aggregation), POINTER :: aggregate !< aggregation fluxes
  END TYPE local
  !> \brief Conditions and fluxes local to each box (layer)
  !! - stv is a pointer to the state variables in the current box, i.e., stv(nb,:)
  !!   in module plankton and stvdat(nb,:) in subroutine invar (module onf)
  !! - soma is a pointer to the source matrix of fluxes among state variables within the
  !!   current box (layer); the first index denotes the source, the second index
  !!   the target of the flux; diagonal elements store rates of change not
  !!   affecting other state variables; the first row (first index = 0) is for
  !!   aggregation
  TYPE :: layer
     INTEGER :: nb,&            !< number of current box (layer); nb = 1 is the surface box
          fish=0                !< factor for fish mortality (0 or 1)
     REAL(dp) :: dz,&           !< distance between current and above boxes mid-depths
          depth,&               !< mid-depth of current box in m
          height=0D0,&          !< height of current box in m
          lch,&                 !< (light-attenuation coefficient)*(box height)
          iPAR,&                !< instantaneous depth-averaged irradiance
          dPAR,&                !< depth average of daytime irradiance
          delvel,&              !< velocity difference between top and bottom of box
          rdl,&                 !< ratio of lower box height to total height of two boxes
          hc,&                  !< proton concentration in \unit{mol.kg^{-1}}
          CO2=0._dp,&           !< CO2 concentration in \unit{mol.kg^{-1}}
          HCO3,&                !< HCO3 concentration in \unit{mol.kg^{-1}}
          CO3,&                 !< CO3 concentration in \unit{mol.kg^{-1}}
          rho,&                 !< density in \unit{g.m^{-3}}
          omega_calcite,&       !< calcite saturation
          omega,&               !< aragonite saturation
          tk,&                  !< temperature in K
          botperm=1._dp         !< bottom permeability (for sinking etc.)
     REAL(dp), ALLOCATABLE :: bbc(:) !< bottom boundary conditions
     REAL(dp), POINTER :: tPAR,&     !< irradiance (PAR) at top of box
          bPAR,&                     !< irradiance (PAR) at bottom of box
          temperature, salinity, alkalinity,&
          stv(:),&              !< state variables in the current box
          DIN, DIP, DIC, O2, rDOC, rDON,&
          silc,&         !< total Si concentration in \unit{mol.kg^{-1}}
          vdc,&          !< vertical diffusivity in \unit{m^2.s^{-1}}
          velocity,&     !< vertical velocity in \unit{m.s^{-1}} (positive downward)
          ratios(:),&    !< ratios in this layer
          flux(:),&      !< vector of vertical fluxes toward the layer beneath
          soma(:,:)      !< fluxes among states within the current box
  END TYPE layer
  TYPE, PUBLIC :: timing
     REAL(dp) :: londeg, latdeg,&
          now,&                 !< current model time in s in the main program
          previous,&            !< model time at previous output
          start=0._dp,&         !< time of start of simulation
          end,&                 !< time of end of simulation
          hour,&                !< current hour of the day
          sample=0D0,&
          daylen=0.5_dp,&       !< day length fraction
          ptl=0.8333_dp,&       !< twighlight parameter in degrees
          w1,&                  !< sunset hour angle
          d1,&                  !< declination
          cdl,&                 !< COS(d1)*COS(latrad)
          dto=1D0,&             !< output time step (not related to time step for integration)
          tin=0D0,&             !< time of initial conditions in initial-conditions file
          toy,&                 !< time of year in decimal days
          noon=12._dp*cfhs      !< noon-time of day in s
     REAL(dp), POINTER :: latrad
     REAL(dp), DIMENSION(:), ALLOCATABLE :: dil_time, dil_fact, dil_DIN, dil_DIP
     INTEGER :: doy,&           !< day of year
          dtutc,&               !< difference between local time and UTC in s
          ndil=0,&              !< number of dilution events
          timeout=0
     INTEGER(INT64) :: current,& !< current model time in s during integration in subroutine plankton:plankton_ode
          sun(4),&               !< current-day sunrise, noon, sunset, midnight in s
          tbase                  !< time 0 in s with respect to date units in forcing file
     INTEGER(KIND=INT64), POINTER :: sunrise, sunset
     LOGICAL :: dilute=.FALSE.
     PROCEDURE(sunday), POINTER :: sunday
     PROCEDURE(decl), POINTER :: declination,& !< declination function
          daylength                            !< day length function
  END TYPE timing

  ABSTRACT INTERFACE
     !> Calculate fluxes in the source matrix soma
     SUBROUTINE fluxes (grp, grps, env, box, times)
       IMPORT dp, fungroup, state, local, layer, timing
       IMPLICIT NONE
       CLASS(fungroup), INTENT(INOUT) :: grp
       TYPE(state), DIMENSION(:), INTENT(IN) :: grps
       TYPE(local), INTENT(INOUT) :: env
       TYPE(layer), INTENT(INOUT) :: box(:)
       TYPE(timing), INTENT(IN) :: times
     END SUBROUTINE fluxes
     !> Read namelists
     SUBROUTINE init (grp, env, lun)
       IMPORT fungroup, local
       IMPLICIT NONE
       CLASS(fungroup), TARGET, INTENT(INOUT) :: grp !< functional type
       TYPE(local), TARGET, INTENT(IN) :: env     !< local environment
       INTEGER, INTENT(IN) :: lun                 !< logical unit number for namelist input
     END SUBROUTINE init
     !> Define ratios, set indices, pointers
     SUBROUTINE preset (grp, grps, env)
       IMPORT dp, fungroup, state, local
       IMPLICIT NONE
       CLASS(fungroup), TARGET, INTENT(INOUT) :: grp
       TYPE(state), DIMENSION(:), INTENT(IN) :: grps
       TYPE(local), TARGET, INTENT(INOUT) :: env
     END SUBROUTINE preset
     !> Calculate aggregation fluxes
     SUBROUTINE aggregation (env, grps, box)
       IMPORT dp, local, state, layer
       IMPLICIT NONE
       CLASS(local), INTENT(INOUT) :: env
       TYPE(state), INTENT(IN) :: grps(:)
       TYPE(layer), INTENT(INOUT) :: box
     END SUBROUTINE aggregation
     !> Calculate or read surface irradiance
     SUBROUTINE light (ambient, time)
       IMPORT local, timing
       IMPLICIT NONE
       CLASS(local), INTENT(INOUT) :: ambient
       TYPE(timing), INTENT(INOUT) :: time
     END SUBROUTINE light
     !> Light attenuation
     SUBROUTINE attenuate (ambient, boxes, stv)
       IMPORT dp, local, layer
       IMPLICIT NONE
       CLASS(local), INTENT(IN) :: ambient
       TYPE(layer), INTENT(INOUT) :: boxes(:)
       REAL(dp), INTENT(IN) :: stv(:,:)
     END SUBROUTINE attenuate
     !> Fractions of surface PAR arriving at the box bottoms
     SUBROUTINE fpar (env, boxes, nbox)
       IMPORT local, layer
       CLASS(local), INTENT(INOUT) :: env
       TYPE(layer), INTENT(INOUT) :: boxes(0:)
       INTEGER, INTENT(IN) :: nbox
     END SUBROUTINE fpar
     SUBROUTINE boxbc (env, boxes)
       IMPORT dp, local, layer
       IMPLICIT NONE
       CLASS(local), INTENT(IN) :: env
       TYPE(layer), INTENT(INOUT) :: boxes(:)
     END SUBROUTINE boxbc
     SUBROUTINE boxalk (env, thisbox)
       IMPORT dp, local, layer
       IMPLICIT NONE
       CLASS(local), INTENT(IN) :: env
       TYPE(layer), INTENT(INOUT) :: thisbox
     END SUBROUTINE boxalk
     SUBROUTINE sunday (times, env)
       IMPORT local, timing
       CLASS(timing), INTENT(INOUT) :: times
       TYPE(local), INTENT(INOUT) :: env
     END SUBROUTINE sunday
     !> daylength function
     SUBROUTINE decl (times)
       IMPORT dp, timing
       IMPLICIT NONE
       CLASS(timing), INTENT(INOUT) :: times
     END SUBROUTINE decl
  END INTERFACE
CONTAINS
  !> Temperature function after Eppley (1972)
  FUNCTION fT_Eppley (grp, box) RESULT (fT)
    IMPLICIT NONE
    CLASS(fungroup), INTENT(IN) :: grp
    TYPE(layer), INTENT(IN) :: box
    REAL(dp) :: fT
    fT = grp%Q10**(0.1_dp*(box%temperature - grp%Tref))
  END FUNCTION fT_Eppley
  !> Temperature function after Houlton et al. (2008)
  FUNCTION ft_Houlton (grp, box) RESULT (ftp)
    IMPLICIT NONE
    CLASS(fungroup), INTENT(IN) :: grp
    TYPE(layer), INTENT(IN) :: box
    REAL(dp) :: ftp
    ftp = EXP(-3.62D0 + 0.27D0*box%temperature*(1D0 - box%temperature/(2D0*grp%Tref)))
  END FUNCTION ft_Houlton
  !> Temperature function based on Moisan et al. (2002)
  FUNCTION fT_oppla (grp, box) RESULT (fT)
    IMPLICIT NONE
    CLASS(fungroup), INTENT(IN) :: grp
    TYPE(layer), INTENT(IN) :: box
    REAL(dp) :: fT
    fT = grp%Q10**(0.1_dp*(box%temperature - grp%Tref)) &
         *EXP(-(ABS(box%temperature - grp%Topt)/grp%Tspr)**3)
  END FUNCTION fT_oppla
  !> Temperature function for Mytilus edulis
  FUNCTION fT_ME (grp, box) RESULT (fT)
    IMPLICIT NONE
    CLASS(fungroup), INTENT(IN) :: grp
    TYPE(layer), INTENT(IN) :: box
    REAL(dp) :: fT
    fT = (1D0 + (box%temperature - 15D0)*9.5D0)/260D0
  END FUNCTION fT_ME
  !> \brief Select and associate temperature-dependence function
  SUBROUTINE fT_select (grp, fT, caller)
    IMPLICIT NONE
    CLASS(fungroup), INTENT(INOUT) :: grp   !< functional group
    CHARACTER(LEN=*), INTENT(IN) :: fT,& !< name of temperature function ('oppla', 'Eppley', 'Houlton', or 'ME')
         caller                          !< name of caller (used for error messages)
    SELECT CASE (TRIM(fT))
    CASE ('Eppley', '')
       grp%fT => fT_Eppley
    CASE ('Houlton')
       grp%fT => fT_Houlton
    CASE ('oppla')
       grp%fT => fT_oppla
    CASE ('ME')
       grp%fT => fT_ME
    CASE DEFAULT
       WRITE (stderr, '("fT_select, called by ",A,": invalid temperature function ",A,".")') TRIM(caller), TRIM(fT)
       STOP 1
    END SELECT
  END SUBROUTINE fT_select
END MODULE et
