!> \brief FORTRAN interface for the UDUNITS2 library
!!
!! see the
!! <a href="https://www.unidata.ucar.edu/software/udunits/#documentation">UDUNITS2 documentation</a>
!! for details of the interfaced C functions
MODULE fudu
  USE, INTRINSIC :: ISO_C_BINDING
  USE ISO_FORTRAN_ENV, ONLY: INT64
  IMPLICIT NONE
  PRIVATE
  PUBLIC :: inudu, exudu, convert, slot
  !> \brief Type for keeping values and units together; enables easy input of units
  !! in namelists
  TYPE, PUBLIC :: quantity
     REAL(C_DOUBLE) :: value                !< numeric value
     CHARACTER(KIND=C_CHAR,LEN=250) :: unit !< units string
  END TYPE quantity
  INTEGER, PARAMETER, PUBLIC :: cfhs=3600
  INTEGER(KIND=INT64), PARAMETER, PUBLIC :: cfds=24*cfhs
  INTEGER(C_INT), PARAMETER :: ASCII=0, LATIN1=1, UTF8=2, UTE_SUCCESS=0, UTE_BAD_ARG=1,&
       UTE_EXISTS=2, UTE_NO_UNIT=3, UTE_OS=4, UTE_NOT_SAME_SYSTEM=5, UTE_MEANINGLESS=6,&
       UTE_NO_SECOND=7, UTE_VISIT_ERROR=8, UTE_CANT_FORMAT=9, UTE_SYNTAX=10, UTE_UNKNOWN=11,&
       UTE_OPEN_ARG=12, UTE_OPEN_ENV=13, UTE_OPEN_DEFAULT=14, UTE_PARSE=15
  INTEGER(C_INTPTR_T), PRIVATE :: utsystem
  INTEGER(C_INT), PRIVATE :: utencoding
  INTERFACE
     !> \brief Set-up and return units system from xml database
     INTEGER(C_INTPTR_T) FUNCTION ut_read_xml (FILE) BIND(C)
       USE, INTRINSIC :: ISO_C_BINDING
       IMPLICIT NONE
       CHARACTER(KIND=C_CHAR) :: FILE !< xml database
     END FUNCTION ut_read_xml
     !> \brief Parse unit string
     INTEGER(C_INTPTR_T) FUNCTION ut_parse (UTSYS, US, UTENC) BIND(C)
       USE, INTRINSIC :: ISO_C_BINDING
       IMPLICIT NONE
       INTEGER(C_INTPTR_T), VALUE :: UTSYS !< units system
       CHARACTER(KIND=C_CHAR) :: US        !< units string
       INTEGER(C_INT), VALUE :: UTENC      !< encoding of units string
     END FUNCTION ut_parse
     REAL(C_DOUBLE) FUNCTION ut_encode_time (year, month, day, hour, minute,&
          second) BIND(C)
       USE, INTRINSIC :: ISO_C_BINDING
       IMPLICIT NONE
       INTEGER(C_INT), VALUE :: year, month, day, hour, minute
       REAL(C_DOUBLE), VALUE :: second
     END FUNCTION ut_encode_time
     INTEGER(C_INTPTR_T) FUNCTION ut_get_converter (UP1, UP2) BIND(C)
       USE, INTRINSIC :: ISO_C_BINDING
       IMPLICIT NONE
       INTEGER(C_INTPTR_T), VALUE :: UP1, UP2
     END FUNCTION ut_get_converter
     INTEGER(C_INT) FUNCTION ut_get_status () BIND(C)
       USE, INTRINSIC :: ISO_C_BINDING
       IMPLICIT NONE
     END FUNCTION ut_get_status
     SUBROUTINE ut_decode_time (time, year, month, day, hour, minute, second, res) BIND(C)
       USE, INTRINSIC :: ISO_C_BINDING
       IMPLICIT NONE
       REAL(C_DOUBLE), VALUE :: time
       INTEGER(C_INT) :: year, month, day, hour, minute
       REAL(C_DOUBLE) :: second, res
     END SUBROUTINE ut_decode_time
     SUBROUTINE cv_free (CVCNV) BIND(C)
       USE, INTRINSIC :: ISO_C_BINDING
       IMPLICIT NONE
       INTEGER(C_INTPTR_T), VALUE :: CVCNV
     END SUBROUTINE cv_free
     SUBROUTINE ut_free (UP) BIND(C)
       USE, INTRINSIC :: ISO_C_BINDING
       IMPLICIT NONE
       INTEGER(C_INTPTR_T), VALUE :: UP
     END SUBROUTINE ut_free
     SUBROUTINE ut_free_system (UTSYS) BIND(C)
       USE, INTRINSIC :: ISO_C_BINDING
       IMPLICIT NONE
       INTEGER(C_INTPTR_T), VALUE :: UTSYS
     END SUBROUTINE ut_free_system
  END INTERFACE
  INTERFACE cv_convert
     REAL(C_FLOAT) FUNCTION cv_convert_float (CVCNV, VAL) BIND(C)
       USE, INTRINSIC :: ISO_C_BINDING
       IMPLICIT NONE
       INTEGER(C_INTPTR_T), VALUE :: CVCNV
       REAL(C_FLOAT), VALUE :: VAL
     END FUNCTION cv_convert_float
     REAL(C_DOUBLE) FUNCTION cv_convert_double (CVCNV, VAL) BIND(C)
       USE, INTRINSIC :: ISO_C_BINDING
       IMPLICIT NONE
       INTEGER(C_INTPTR_T), VALUE :: CVCNV
       REAL(C_DOUBLE), VALUE :: VAL
     END FUNCTION cv_convert_double
  END INTERFACE cv_convert
CONTAINS
  !> \brief Initialise the UDUNITS2 library and read the units database
  SUBROUTINE inudu
    IMPLICIT NONE
    CHARACTER(LEN=*), PARAMETER :: ududb=UDUXML
    utencoding = UTF8
#ifdef GFORTRAN
    utsystem = ut_read_xml(ududb//char(0))
#else
    utsystem = ut_read_xml(ududb)
#endif
    IF (utsystem.EQ.0) CALL uduerr ('UT_READ_XML', ududb)
  END SUBROUTINE inudu
  !> \brief Unregister the units database associated with the global variable utsystem
  SUBROUTINE exudu
    IMPLICIT NONE
    CALL ut_free_system (utsystem)
  END SUBROUTINE exudu
  !> \brief Convert a quantity to another unit
  FUNCTION convert (qty, unit) RESULT (val)
    IMPLICIT NONE
    TYPE(quantity), INTENT(IN) :: qty                !< quantity to be converted
    CHARACTER(KIND=C_CHAR,LEN=*), INTENT(IN) :: unit !< target units
    INTEGER(C_INTPTR_T) :: up1, up2, cvcnv
    REAL(C_DOUBLE) :: val       !< value of quantity in target units
    up1 = ut_parse(utsystem, TRIM(qty%unit)//C_NULL_CHAR, utencoding)
    IF (up1.EQ.0) CALL uduerr ('UT_PARSE(convert)', qty%unit)
    up2 = ut_parse(utsystem, TRIM(unit)//C_NULL_CHAR, utencoding)
    IF (up2.EQ.0) CALL uduerr ('UT_PARSE(convert)', unit)
    cvcnv = ut_get_converter(up1, up2)
    IF (ut_get_status().NE.0) CALL uduerr ('UT_GET_CONVERTER(convert)',&
         qty%unit, unit)
    val = cv_convert(cvcnv, qty%value)
    CALL cv_free (cvcnv)
    CALL ut_free (up1)
    CALL ut_free (up2)
  END FUNCTION convert
  !> \brief Slope and offset for unit conversion
  !!
  !! A quantity \ff{q_{\text{us1}}} in units given by units string us1 can be converted
  !! to units given by us2 by \eq{q_{\text{us2}} = \text{offset} + \text{slope}
  !! \cdot q_{\text{us1}}}
  !! \param[in] us1 original units string
  !! \param[in] us2 target units string
  SUBROUTINE slot (us1, us2, slope, offset)
    IMPLICIT NONE
    CHARACTER(KIND=C_CHAR,LEN=*), INTENT(IN) :: us1, us2
    REAL(C_DOUBLE), INTENT(OUT) :: slope              !< slope of units conversion
    REAL(C_DOUBLE), INTENT(OUT), OPTIONAL :: offset   !< offset of units conversion
    INTEGER(C_INTPTR_T) :: up1, up2, cvcnv
    up1 = ut_parse(utsystem, TRIM(us1)//C_NULL_CHAR, utencoding)
    IF (up1.EQ.0) CALL uduerr ('UT_PARSE(slot)', us1)
    up2 = ut_parse(utsystem, TRIM(us2)//C_NULL_CHAR, utencoding)
    IF (up2.EQ.0) CALL uduerr ('UT_PARSE(slot)', us2)
    cvcnv = ut_get_converter(up1, up2)
    IF (ut_get_status().NE.0) CALL uduerr ('UT_GET_CONVERTER(slot)',&
         us1, us2)
    slope = cv_convert(cvcnv, REAL(1, C_DOUBLE))
    IF (PRESENT(offset)) offset = cv_convert(cvcnv, REAL(0, C_DOUBLE))
    CALL cv_free (cvcnv)
    CALL ut_free (up1)
    CALL ut_free (up2)
  END SUBROUTINE slot
  SUBROUTINE uduerr (udfun, unit, unit2)
    IMPLICIT NONE
    CHARACTER(LEN=*), DIMENSION(15), PARAMETER :: ut_errmsg=(/ CHARACTER(LEN=18) :: 'UT_BAD_ARG',&
         'UT_EXISTS', 'UT_NO_UNIT', 'UT_OS', 'UT_NOT_SAME_SYSTEM', 'UT_MEANINGLESS',&
         'UT_NO_SECOND', 'UT_VISIT_ERROR', 'UT_CANT_FORMAT', 'UT_SYNTAX', 'UT_UNKNOWN',&
         'UT_OPEN_ARG', 'UT_OPEN_ENV', 'UT_OPEN_DEFAULT', 'UT_PARSE'/)
    CHARACTER(LEN=*), INTENT(IN) :: udfun
    CHARACTER(LEN=*), INTENT(IN), OPTIONAL :: unit, unit2
    IF (PRESENT(unit2)) THEN
       WRITE (0,'("UDUNITS:",A,": ",A," for ",A," and ",A,".")') TRIM(udfun),&
            TRIM(ut_errmsg(ut_get_status())), TRIM(unit), TRIM(unit2)
    ELSEIF (PRESENT(unit)) THEN
       WRITE (0,'("UDUNITS:",A,": ",A," for ",A,".")') TRIM(udfun),&
            TRIM(ut_errmsg(ut_get_status())), TRIM(unit)
    ELSE
       WRITE (0,'("UDUNITS:",A,": ",A,".")') TRIM(udfun),&
            TRIM(ut_errmsg(ut_get_status()))
    END IF
    STOP
  END SUBROUTINE uduerr
END MODULE fudu
