!> \brief Fortran implementation of the
!! [julian library](https://pds-rings.seti.org/toolkits/julian_133_html/aareadme.html).
!!
!! The main differences are that the conversion routines from/to character
!! srings (Jul_FormatDate and Jul_ParseDT) are replaced by wrappers around
!! strftime and strptime and seconds are mostly represented as integers.
MODULE julian
  USE, INTRINSIC :: ISO_C_BINDING, ONLY: C_CHAR, C_INT, C_NULL_CHAR
  IMPLICIT NONE
  PRIVATE
  PUBLIC :: Jul_DHMSofSec, Jul_FormatDate, Jul_ParseDT, Jul_YDofDUTC, Jul_YMDofDUTC
  INTEGER, PARAMETER :: dp=SELECTED_REAL_KIND(KIND(0D0)),&
       JUL_UTC_TYPE=0, JUL_TAI_TYPE=1, JUL_ET_TYPE=2,&
       JDN_OF_J2000=2451545,&   !< Julian day numer of 1 January 2000, i.e., since 4713-01-01 BCE = –4712-01-01
       gregorian_year=1582,&
       gregorian_month=10, gregorian_day=15, gregorian_dutc=-152384
  REAL(dp), PARAMETER :: MJD_OF_J2000_NOON=51544.5_dp, JD_OF_J2000_NOON=2451545.0_dp
  TYPE :: yms
     INTEGER :: year,&          !< year for leap second
          month,&               !< month for leap second
          seconds               !< cumulative leap seconds at month, year
  END TYPE yms
  TYPE :: leap
     INTEGER :: nmin,&          !< number of leap second before ymin = 1972
          nmax,&                !< number of leap seconds as of ymax
          ymin,&                !< first year of leap-seconds table (1972)
          ymax,&                !< last year of leap-seconds table
          size                  !< size of leap-seconds table
  END TYPE leap

  !> \brief Derived type for compatibility with C and C++ struct tm.
  !!
  !! Enables calling strftime and strptime using iso_c_binding.
  !! See https://www.cplusplus.com/reference/ctime/tm for reference.
  TYPE, BIND(C) :: tm_struct
     INTEGER(C_INT) :: tm_sec = 0   !< Seconds [0--60] (1 leap second)
     INTEGER(C_INT) :: tm_min = 0   !< Minutes [0--59]
     INTEGER(C_INT) :: tm_hour = 0  !< Hours [0--23]
     INTEGER(C_INT) :: tm_mday = 0  !< Day [1--31]
     INTEGER(C_INT) :: tm_mon = 0   !< Month [0--11]
     INTEGER(C_INT) :: tm_year = 0  !< Year – 1900
     INTEGER(C_INT) :: tm_wday = 0  !< Day of week [0--6]
     INTEGER(C_INT) :: tm_yday = 0  !< Days in year [0--365]
     INTEGER(C_INT) :: tm_isdst = 0 !< DST [–1/0/1]
  END TYPE tm_struct
  TYPE(yms), DIMENSION(*), PARAMETER :: leap_defaults=(/&
       yms(1972, 1, 10),&
       yms(1972, 7, 11),&
       yms(1973, 1, 12),&
       yms(1974, 1, 13),&
       yms(1975, 1, 14),&
       yms(1976, 1, 15),&
       yms(1977, 1, 16),&
       yms(1978, 1, 17),&
       yms(1979, 1, 18),&
       yms(1980, 1, 19),&
       yms(1981, 7, 20),&
       yms(1982, 7, 21),&
       yms(1983, 7, 22),&
       yms(1985, 7, 23),&
       yms(1988, 1, 24),&
       yms(1990, 1, 25),&
       yms(1991, 1, 26),&
       yms(1992, 7, 27),&
       yms(1993, 7, 28),&
       yms(1994, 7, 29),&
       yms(1996, 1, 30),&
       yms(1997, 7, 31),&
       yms(1999, 1, 32),&
       yms(2006, 1, 33),&
       yms(2009, 1, 34),&
       yms(2012, 7, 35),&
       yms(2015, 7, 36),&
       yms(2017, 1, 37)/)
  TYPE(leap) :: leaps
  INTEGER, ALLOCATABLE :: leap_table(:) !< leap-seconds table
  INTERFACE
     !> Interface to POSIX strftime.
     !! Returns a formatted time string, given input time struct and format.
     !! See https://www.cplusplus.com/reference/ctime/strftime for reference.
     FUNCTION c_strftime (str, slen, form, tm) BIND (C, name='strftime') RESULT (rc)
       IMPORT :: C_CHAR, C_INT, tm_struct
       CHARACTER(KIND=C_CHAR), INTENT(OUT) :: str(*)   !> result string
       INTEGER(C_INT), VALUE, INTENT(IN) :: slen       !> string length
       CHARACTER(KIND=C_CHAR), INTENT(IN) :: form(*)   !> time format
       TYPE(tm_struct), INTENT(IN) :: tm               !> tm_struct instance
       INTEGER(C_INT) :: rc                            !> return code
     END FUNCTION c_strftime
     !> Interface to POSIX strptime.
     !! Returns a time struct object based on the input time string str and format.
     !! See https://man7.org/linux/man-pages/man3/strptime.3.html for reference.
     FUNCTION c_strptime (str, format, tm) BIND (C, name='strptime') RESULT (rc)
       IMPORT :: C_CHAR, C_INT, tm_struct
       CHARACTER(KIND=C_CHAR), INTENT(IN) :: str(*)    !> input string
       CHARACTER(KIND=C_CHAR), INTENT(IN) :: format(*) !> time format
       TYPE(TM_STRUCT), INTENT(INOUT) :: tm            !> result tm_struct
       CHARACTER(KIND=C_CHAR, LEN=1) :: rc             !> return code
     END FUNCTION c_strptime
  END INTERFACE
CONTAINS
  !> \brief Convert Julian date to a number of seconds relative to J2000 TAI.
  !!
  !! This function converts a Julian date to a number of seconds relative to
  !! J2000 TAI.  The Julian date may be given in any of three possible time
  !! systems: Universal time (UTC), Atomic time (TAI), or Ephemeris time (ET).
  !!
  !! If the date type is UTC, fractional days are scaled by the number of
  !! seconds between one UTC noon and the next; since some days have leap
  !! seconds, this makes the spacing of UTC Julian dates slightly
  !! non-uniform.
  !!
  !! \return number of seconds since noon J2000 TAI
  FUNCTION Jul_TAIofJD (jd, type) RESULT (tai)
    IMPLICIT NONE
    REAL(dp), INTENT(IN) :: jd  !< Julian date, in one of three possible time systems
    INTEGER, INTENT(IN) :: type !< type of the Julian date, one of the
    !! constants JUL_UTC_TYPE (0), JUL_TAI_TYPE (1), or JUL_ET_TYPE (2)
    REAL(dp) :: et, tai
    SELECT CASE (type)
    CASE (JUL_TAI_TYPE)
       tai = (jd - JD_OF_J2000_NOON)*86400._dp
    CASE (JUL_ET_TYPE)
       et = (jd - JD_OF_J2000_NOON)*86400._dp
       tai = Jul_TAIofET(et)
    CASE DEFAULT
       tai = Jul_TAIofDUTC(jd - JD_OF_J2000_NOON)
    END SELECT
  END FUNCTION Jul_TAIofJD
  !> \brief Convert from ET (ephemeris time) to TAI (atomic time).
  !!
  !! This function converts from ET (ephemeris time) to TAI (Atomic time).
  !! All times are given as seconds relative to J2000.  The algorithm is
  !! based on that used in the SPICE toolkit.
  !!
  !! \return corresponding atomic time (seconds relative to J2000 TAI)
  FUNCTION Jul_TAIofET (et) RESULT (tai)
    IMPLICIT NONE
    REAL(dp), PARAMETER :: M0=6.239996_dp, M1=1.99096871E-7_dp, EB=1.671E-2_dp,&
         K=1.657E-3_dp, DELTA_T_A=32.184_dp
    REAL(dp), INTENT(IN) :: et  !< ephemeris time
    REAL(dp) :: tai, mean_anom, ecc_anom
    mean_anom = M0 + M1*et
    ecc_anom = mean_anom + EB*SIN(mean_anom)
    tai = et - DELTA_T_A - K*SIN(ecc_anom)
  END FUNCTION Jul_TAIofET
  !> \brief This internal function calculates the number of seconds TAI from the
  !! floating-point number of days UTC relative to noon J2000.
  !!
  !! \return seconds from J2000 noon TAI
  FUNCTION Jul_TAIofDUTC (dutc) RESULT (tai)
    IMPLICIT NONE
    REAL(dp), INTENT(IN) :: dutc !< days relative to J2000 noon UTC
    REAL(dp) :: tai, frac, leaps
    INTEGER :: dutc_floor, dutc_noon
    dutc_floor = FLOOR(dutc)
    dutc_noon = dutc_floor
    leaps = Jul_LeapSecs(dutc_noon)
    frac = dutc - dutc_floor
    IF (frac.GT.0._dp) frac = frac*(86400._dp + Jul_LeapSecs(dutc_noon + 1) - leaps)
    tai = 86400._dp*dutc_noon + leaps + frac
  END FUNCTION Jul_TAIofDUTC
  !> \brief number of leap seconds elapsed before a given day
  FUNCTION Jul_LeapSecs (dutc) RESULT (seconds)
    IMPLICIT NONE
    INTEGER, INTENT(IN) :: dutc
    INTEGER :: seconds, year, month, day
    CALL Jul_YMDofDUTC (dutc, year, month, day)
    seconds = Jul_LeapSecsYM(year, month)
  END FUNCTION Jul_LeapSecs
  FUNCTION leap_index (year, month) RESULT (index)
    IMPLICIT NONE
    INTEGER, INTENT(IN) :: year, month
    INTEGER :: index
    index = 2*(year - leaps%ymin) + (month - 1)/6 + 1
  END FUNCTION leap_index
  FUNCTION Jul_LeapSecsYM (year, month) RESULT (seconds)
    IMPLICIT NONE
    INTEGER, INTENT(IN) :: year, month
    INTEGER :: seconds, index
    IF (.NOT.ALLOCATED(leap_table)) CALL Jul_InitLeaps
    index = leap_index(year, month)
    IF (index < 0) THEN
       seconds = leaps%nmin
    ELSE IF (index >= leaps%size) THEN
       seconds = leaps%nmax
    ELSE
       seconds = leap_table(index)
    END IF
  END FUNCTION Jul_LeapSecsYM
  !> Determines whether a given day contains a leap second.
  FUNCTION Jul_IsLeapDay (dutc) RESULT (isLeapDay)
    IMPLICIT NONE
    INTEGER, INTENT(IN) :: dutc
    LOGICAL :: isLeapDay
    INTEGER :: year, month, day
    IF (.NOT.ALLOCATED(leap_table)) CALL Jul_InitLeaps
    CALL Jul_YMDofDUTC (dutc, year, month, day)
    isLeapDay = .FALSE.
    IF (day.LE.29) RETURN
    IF (day.EQ.30) THEN
       IF (month.NE.6) RETURN
    ELSEIF (day.EQ.31) THEN
       IF (month.NE.12) RETURN
    ELSE
       RETURN      ! invalid date
    END IF
    isLeapDay = Jul_LeapSecsYM(year, month + 1).NE.Jul_LeapSecsYM(year, month)
  END FUNCTION Jul_IsLeapDay
  FUNCTION Jul_DUTCofYMD (year, month, day) RESULT (dutc)
    IMPLICIT NONE
    INTEGER, INTENT(INOUT) :: year, month
    INTEGER, INTENT(IN) :: day
    INTEGER :: dutc
    CALL Jul_FixYM (year, month)
    IF ((year.GT.gregorian_year).OR.((year.EQ.gregorian_year)&
         .AND.((month.GT.gregorian_month)&
         .OR.((month.EQ.gregorian_month).AND.(day.GE.gregorian_day))))) THEN
       dutc = Jul_JDofGregYMD(year, month, day) - JDN_OF_J2000
    ELSE
       dutc = Jul_JDofJulYMD(year, month, day) - JDN_OF_J2000
    END IF
  END FUNCTION Jul_DUTCofYMD
  FUNCTION Jul_JDofGregYMD (year, month, day) RESULT (jd)
    IMPLICIT NONE
    INTEGER, INTENT(IN) :: year, month, day
    INTEGER :: jd, tmp
    tmp = 0
    IF (month.LE.2) tmp = -1
    jd = (1461*(year + 4800 + tmp))/4 + (367*(month - 2 - 12*tmp))/12 &
         - (3*((year + 4900 + tmp)/100))/4 + day - 32075
  END FUNCTION Jul_JDofGregYMD
  FUNCTION Jul_JDofJulYMD(year, month, day) RESULT (jd)
    IMPLICIT NONE
    INTEGER, INTENT(IN) :: year, month, day
    INTEGER :: jd
    jd = 367*year - (7*(year + 5001 + (month-9)/7))/4 + (275*month)/9 &
         + day + 1729777
  END FUNCTION Jul_JDofJulYMD
  !> \brief Returns the calendar year and day-of-year number for a day number
  !! relative to January 1, 2000 for a given date.
  !!
  !! This subroutine returns the calendar year and day-of-year for a day number
  !! relative to January 1, 2000.
  SUBROUTINE Jul_YDofDUTC(dutc, year, doy)
    IMPLICIT NONE
    INTEGER, INTENT(IN) :: dutc    !< number of days relative to January 1, 2000
    INTEGER, INTENT(OUT) :: year   !< year value
    INTEGER, INTENT(OUT) :: doy    !< day-of-year (1-366)
    INTEGER :: month, day, jan=1
    CALL Jul_YMDofDUTC(dutc, year, month, day)
    doy = dutc - Jul_DUTCofYMD(year, jan, 1) + 1
  END SUBROUTINE Jul_YDofDUTC
  !> Returns the calendar year, month and day for a day number relative to
  !! January 1, 2000 for a given date.
  !!
  !! This subroutine returns the calendar year, month and day for a day number
  !! relative to January 1, 2000.
  SUBROUTINE Jul_YMDofDUTC (dutc, year, month, day)
    IMPLICIT NONE
    INTEGER, INTENT(IN) :: dutc   !< number of days relative to January 1, 2000
    INTEGER, INTENT(OUT) :: year  !< calendar year
    INTEGER, INTENT(OUT) :: month !< month (1-12)
    INTEGER, INTENT(OUT) :: day   !< day (1-31)
    IF (dutc >= gregorian_dutc) THEN
       CALL Jul_GregYMDofJD (dutc + JDN_OF_J2000, year, month, day)
    ELSE
       CALL Jul_JulYMDofJD (dutc + JDN_OF_J2000, year, month, day)
    END IF
  END SUBROUTINE Jul_YMDofDUTC
  SUBROUTINE Jul_GregYMDofJD(jd, year, month, day)
    IMPLICIT NONE
    INTEGER, INTENT(IN) :: jd
    INTEGER, INTENT(OUT) :: year, month, day
    INTEGER :: l, n, i, j
    l = jd + 68569
    n = (4*l)/146097
    l = l - (146097*n + 3)/4
    i = (4000*(l+1))/1461001
    l = l - (1461*i)/4 + 31
    j = (80*l)/2447
    day = l - (2447*j)/80
    l = j/11
    month = j + 2 - 12*l
    year = 100*(n-49) + i + l
  END SUBROUTINE Jul_GregYMDofJD
  SUBROUTINE Jul_JulYMDofJD (jd, year, month, day)
    IMPLICIT NONE
    INTEGER, INTENT(IN) :: jd
    INTEGER, INTENT(OUT) :: year, month, day
    INTEGER :: l, n, i, j, k
    j = jd + 1402;
    k = (j-1)/1461;
    l = j - 1461*k;
    n = (l-1)/365 - l/1461;
    i = l - 365*n + 30;
    j = (80*i)/2447;
    day = i - (2447*j)/80;
    i = j/11;
    month = j + 2 - 12*i;
    year = 4*k + n + i - 4716;
  END SUBROUTINE Jul_JulYMDofJD
  SUBROUTINE Jul_FixYM (year, month)
    IMPLICIT NONE
    INTEGER, INTENT(INOUT) :: year, month
    INTEGER :: delta_year
    DO WHILE (month < 1)
       month = month + 12
       year = year - 1
    END DO
    delta_year = (month - 1) / 12
    month = month - 12 * delta_year
  END SUBROUTINE Jul_FixYM
  SUBROUTINE Jul_InitLeaps
    IMPLICIT NONE
    INTEGER :: n, nsecs
    IF (ALLOCATED(leap_table)) RETURN
    leaps%ymin = leap_defaults(1)%year
    leaps%nmin = leap_defaults(1)%seconds - 1
    leaps%ymax = leap_defaults(SIZE(leap_defaults))%year
    leaps%nmax = leap_defaults(SIZE(leap_defaults))%seconds
    leaps%size = 2 * (leaps%ymax - leaps%ymin + 1) + 1
    ALLOCATE (leap_table(leaps%size))
    leap_table = leaps%nmin
    DO n=1,SIZE(leap_defaults)
       leap_table(leap_index(leap_defaults(n)%year, leap_defaults(n)%month)) = leap_defaults(n)%seconds
    END DO
    nsecs = leaps%nmin
    DO n=1,SIZE(leap_defaults)
       IF (leap_table(n) == leaps%nmin) THEN
          leap_table(n) = nsecs
       ELSE
          IF (leap_table(n).NE.nsecs + 1) THEN
             WRITE (*,'("Illegal leap second step from ",I0," to ",I0)') nsecs, leap_table(n)
             STOP
          END IF
          nsecs = leap_table(n)
       END IF
    END DO
  END SUBROUTINE Jul_InitLeaps
  !> \brief convert number of seconds to days, hours, minutes and seconds
  !!
  !! This function converts a number of seconds to days, hours, minutes and
  !! seconds.  All days are assumed to contain 86400 seconds; leap seconds
  !! are not supported.
  SUBROUTINE Jul_DHMSofSec (secs, day, hour, minute, seconds)
    IMPLICIT NONE
    INTEGER, INTENT(IN) :: secs   !< number of seconds
    INTEGER, INTENT(OUT) :: day,& !< number of days
         hour,&                   !< number of hours into day (0--23)
         minute,&                 !< number of minutes into hour (0--59)
         seconds                  !< number of seconds into minute (0--60)
    CALL Jul_DSofSec(secs, day, seconds)
    hour = FLOOR(seconds / 3600._dp)
    seconds = seconds - 3600*hour
    minute = FLOOR(seconds / 60._dp)
    seconds = seconds - 60*minute
  END SUBROUTINE Jul_DHMSofSec
  ! convert number of seconds to days, plus seconds into day
  SUBROUTINE Jul_DSofSec (secs, day, seconds)
    IMPLICIT NONE
    INTEGER, INTENT(IN) :: secs
    INTEGER, INTENT(OUT) :: day, seconds
    day = FLOOR(secs / 86400._dp)
    seconds = secs - 86400*day
    IF (seconds.LT.0) THEN
       day = day - 1
       seconds = seconds + 86400
    END IF
  END SUBROUTINE Jul_DSofSec
  !> \brief Interprets a character string as a date and time.
  !!
  !! This is a wrapper for the
  !! [strptime C function](https://man7.org/linux/man-pages/man3/strptime.3.html).
  SUBROUTINE Jul_ParseDT (string, fmt, dutc, secs)
    IMPLICIT NONE
    CHARACTER(LEN=*), INTENT(IN) :: string,& !< the character string to parse
         fmt                                 !< format string
    INTEGER, INTENT(OUT) :: dutc,&           !< inferred days relative to January 1, 2000
         secs                                !< inferred seconds into day
    TYPE(tm_struct) :: tm
    CHARACTER(KIND=C_CHAR, LEN=1) :: status
    INTEGER :: year, month
    tm = tm_struct()
    status = c_strptime(TRIM(string)//C_NULL_CHAR, format=TRIM(fmt)//C_NULL_CHAR, tm=tm)
    year = tm%tm_year + 1900
    month = tm%tm_mon + 1
    dutc = Jul_DUTCofYMD(year, month, tm%tm_mday)
    secs = tm%tm_sec + tm%tm_min*60 + tm%tm_hour*3600
  END SUBROUTINE Jul_ParseDT
  !> Format a date and time according to a format string.
  !!
  !! This is a wrapper function for the
  !! [strftime C function](https://www.cplusplus.com/reference/ctime/strftime).
  !!
  !! \return formatted string
  FUNCTION Jul_FormatDate (dutc, seconds, fmt) RESULT (string)
    IMPLICIT NONE
    INTEGER, INTENT(IN) :: dutc, seconds
    CHARACTER(LEN=*), INTENT(IN) :: fmt
    CHARACTER(LEN=200) :: string
    INTEGER :: len, year, month, day
    TYPE(tm_struct) :: tm
    tm = tm_struct()
    CALL Jul_YMDofDUTC (dutc, year, month, day)
    tm%tm_year = year - 1900
    tm%tm_mon = month - 1
    tm%tm_mday = day
    tm%tm_hour = seconds / 3600
    tm%tm_min = (seconds - 3600 * tm%tm_hour) / 60
    tm%tm_sec = seconds - 3600 * tm%tm_hour - 60 * tm%tm_min
    len = c_strftime(str=string, slen=200, form=TRIM(fmt)//C_NULL_CHAR, tm=tm)
    string = string(1:len)
  END FUNCTION Jul_FormatDate
END MODULE julian
