!> \brief NetCDF interface for oppla

MODULE onf
  USE NETCDF
  USE et
  IMPLICIT NONE
  PRIVATE
  PUBLIC :: NF90_MAX_NAME
  TYPE :: dataset
     CHARACTER(LEN=NF90_MAX_NAME) :: type='',& !< data type, e.g., Temperature, Salinity, etc.
          name=''
  END TYPE dataset
  TYPE :: nf90stratt
     INTEGER :: id
     CHARACTER(LEN=NF90_MAX_NAME) :: name, value
  END TYPE nf90stratt
  TYPE :: nf90var
     CHARACTER(LEN=NF90_MAX_NAME) :: name='',&
          type='',&             !< data type, e.g., Temperature, Salinity, etc.
          units=''
     INTEGER :: varid,&         !< variable ID
          ndims                 !< rank (number of dimensions)
   CONTAINS
     PROCEDURE :: read => read_ds
  END TYPE nf90var
  TYPE, EXTENDS(nf90var) :: nf90grd
     REAL(KIND(0D0)), DIMENSION(:,:), POINTER :: values
  END TYPE nf90grd
  TYPE, EXTENDS(nf90var) :: nf90vec
     REAL(KIND(0D0)), DIMENSION(:), POINTER :: values
  END TYPE nf90vec
  TYPE :: dimension
     INTEGER :: id=0,&          !< dimension ID
          varid                 !< dimension-variable ID
     CHARACTER(LEN=NF90_MAX_NAME) :: name='', units=''
     REAL(dp), ALLOCATABLE :: values(:)
     TYPE(nf90grd) :: bounds
   CONTAINS
     PROCEDURE :: read => read_dim
  END TYPE dimension
  TYPE :: nf90group
     CHARACTER(LEN=NF90_MAX_NAME) :: filename
     INTEGER :: id=0, it0, ilat, ilon, ntim, nbox=0,&
          timediff=0 !< difference between local time and UTC in s
     TYPE(dimension) :: depth, depth_flux, time, lat, lon, state, bounds
     TYPE(nf90stratt), ALLOCATABLE :: attributes(:)
   CONTAINS
     PROCEDURE :: info => nf90info
  END TYPE nf90group
  TYPE, EXTENDS(nf90group), PUBLIC :: phydat        !< physical data
     INTEGER, DIMENSION(:), ALLOCATABLE :: isflx, ibot
     TYPE(nf90vec), DIMENSION(:), ALLOCATABLE :: surface, bottom, surflx, profile
     TYPE(nf90grd), DIMENSION(:), ALLOCATABLE :: data
     CHARACTER(LEN=NF90_MAX_NAME) :: advect='upwind'
     REAL(dp), DIMENSION(:), POINTER :: surflux, botbc, surPAR
     REAL(dp), ALLOCATABLE :: sds(:), pds(:,:), cds(:,:)
     REAL(dp) :: fPAR=0.43_dp   !< PAR fraction in total-irradiance data
     LOGICAL :: clsbot=.FALSE.
     PROCEDURE(rpd), POINTER :: get => nopd
   CONTAINS
     PROCEDURE :: init => inpd
  END TYPE phydat
  TYPE, EXTENDS(nf90group), PUBLIC :: statevars
     INTEGER :: nsv_one_box,&   !< No. of plankton state variables in each layer
          nsv_all_boxes,&       !< total number of plankton state variables (across all layers)
          nsv_total,&           !< total number of all (including diagnostic) state variables
          nfl,&                 !< No. of flux levels in output
          i0flux,&              !< starting index of fluxes in state-variable vector
          i0sms,&               !< starting index of soma in state-variable vector
          nsms                  !< No. of elements in soma to write to output file
     INTEGER, POINTER :: it(:)
     INTEGER, DIMENSION(:), ALLOCATABLE :: start, count
     LOGICAL :: write_soma=.FALSE.
     LOGICAL, POINTER :: smsk(:,:)
     CHARACTER(LEN=NF90_MAX_NAME) :: svgn=''
     REAL(dp), DIMENSION(:,:), POINTER :: data,& !< array of state variables (pointer to all_states)
          roc                                    !< array of rates of change (pointer to ydot in plankton_ode)
     REAL(dp), DIMENSION(:), ALLOCATABLE :: all_states, flux_depth
     TYPE(nf90vec), DIMENSION(:), ALLOCATABLE :: var
     TYPE(nf90vec) :: flux, soma
     PROCEDURE(write_states), POINTER :: write_output => write_states
   CONTAINS
     PROCEDURE :: open => innf, init => invar, write => outsv, close => exof!, read => readsv
  END TYPE statevars
  INTEGER :: nbvds, nsvds, nsfds, ngdds, npfds
  CHARACTER(LEN=14), PARAMETER, PUBLIC :: pfan='Parameter File',&
       plfan='Plankton File', dfan='Data File', infan='Start File',&
       bfan='Bottom File', sufan='Spin-up File', timan='Time',&
       latan='latitude', lonan='longitude'
  CHARACTER(LEN=NF90_MAX_NAME), PUBLIC :: timunt='s'
  INTEGER, PUBLIC :: iprsds, ico2ds
  REAL(dp), ALLOCATABLE :: offsets(:,:)
  REAL(dp), ALLOCATABLE, TARGET :: pdsds(:,:), pdgds(:,:,:)
  ABSTRACT INTERFACE
     SUBROUTINE rpd (pdf, time)
       IMPORT phydat, dp
       IMPLICIT NONE
       CLASS(phydat), INTENT(INOUT) :: pdf
       REAL(dp), INTENT(IN) :: time
     END SUBROUTINE rpd
  END INTERFACE
CONTAINS

  !> Initialise NETCDF interface and read associated parameters
  !!
  !! Subroutine innf is called as states%open in plankton
  SUBROUTINE innf (states, phys, lun, atts)
    IMPLICIT NONE
    CLASS(statevars), INTENT(OUT) :: states !< state-variable info
    TYPE(phydat), INTENT(OUT) :: phys       !< info about physical forcing
    INTEGER, INTENT(IN) :: lun              !< unit number of the control file
    TYPE(dataset), DIMENSION(100) :: forcing, profile, bottom, surface, surflx
    LOGICAL :: clsbot, soma
    CHARACTER(LEN=NF90_MAX_NAME) :: time, depth, lat, lon, bounds, iomsg, svgroup, advect
    REAL(dp) :: fPAR            !< PAR fraction in total-irradiance data
    TYPE(nf90stratt), DIMENSION(:), ALLOCATABLE, INTENT(OUT), OPTIONAL :: atts
    NAMELIST /start/ time, depth, lat, lon, bounds, svgroup, soma
    NAMELIST /physics/ time, depth, lat, lon, bounds, forcing, profile, bottom, surface, surflx,&
         clsbot, advect, fPAR
    time = 'Time'               ! NetCDF names
    depth = 'Depth'
    lat = 'latitude'
    lon = 'longitude'
    bounds = 'bounds'
    soma = .FALSE.
    svgroup = ''
    REWIND (lun)
    READ (lun, NML=start, END=900, ERR=900, IOMSG=iomsg)
    IF (soma) WRITE (stderr,'("innf: Ignoring soma in namelist start. Please use namelist fluxes!")')
    states%time%name = TRIM(time)
    states%depth%name = TRIM(depth)
    states%lat%name = TRIM(lat)
    states%lon%name = TRIM(lon)
    states%bounds%name = TRIM(bounds)
    states%svgn = TRIM(svgroup)
    time = 'Time'
    depth = 'Depth'
    lat = 'latitude'
    lon = 'longitude'
    bounds = 'bounds'
    advect = phys%advect
    clsbot = phys%clsbot
    fPAR = phys%fPAR
    REWIND (lun)
    READ (lun, NML=physics, END=900, ERR=900, IOMSG=iomsg)
    phys%time%name = TRIM(time)
    phys%depth%name = TRIM(depth)
    phys%lat%name = TRIM(lat)
    phys%lon%name = TRIM(lon)
    phys%bounds%name = TRIM(bounds)
    phys%fPAR = fPAR
    phys%advect = advect
    phys%clsbot = clsbot
    ngdds = COUNT(forcing%type.NE.'') ! No. of gridded physical forcings
    npfds = COUNT(profile%type.NE.'') ! No. of temporally constant vertical profiles
    nbvds = COUNT(bottom%type.NE.'')  ! number of bottom value datasets
    nsvds = COUNT(surface%type.NE.'') ! number of surface value datasets
    nsfds = COUNT(surflx%type.NE.'')  ! number of surface flux datasets
    ALLOCATE (phys%data(ngdds), phys%profile(npfds), phys%bottom(nbvds), phys%surface(nsvds), &
              phys%surflx(nsfds))
    phys%data%name = forcing(1:ngdds)%name
    phys%data%type = forcing(1:ngdds)%type
    phys%bottom%name  = bottom(1:nbvds)%name
    phys%bottom%type  = bottom(1:nbvds)%type
    phys%surface%name = surface(1:nsvds)%name
    phys%surface%type = surface(1:nsvds)%type
    phys%surflx%name = surflx(1:nsfds)%name
    phys%surflx%type = surflx(1:nsfds)%type
    phys%profile%name = profile(1:npfds)%name
    phys%profile%type = profile(1:npfds)%type
    IF (PRESENT(atts)) THEN
       ALLOCATE (atts(6))
       atts%name = (/dfan, infan, pfan, plfan, bfan, sufan/) ! attribute names
    END IF
    RETURN
900 CONTINUE
    WRITE (stderr,'("innf: ",A)') TRIM(iomsg)
    STOP
  END SUBROUTINE innf

  !> open group%filename, read global attributes and coordinates (depth, lon, lat, time)
  !! \param[in] group%depth%name name of depth dataset
  !! \param[in] group%filename name of file to read
  !! \param[in] group%lat%name name of latitude dataset
  !! \param[in] group%lon%name name of longitude dataset
  !! \param[in] group%time%name name of time dataset
  !! \param[in] group%timediff difference between local solar time and UTC in s
  !! \param[out] group%depth%values depth (z) axis
  !! \param[out] group%lat%values latitude (y) axis
  !! \param[out] group%depth%values longitude (x) axis
  !! \param[out] group%time%values time axis
  SUBROUTINE nf90info (group, latdeg, londeg, t0, t1, writeable)
    IMPLICIT NONE
    CLASS(nf90group), INTENT(INOUT) :: group
    REAL(dp), INTENT(INOUT), OPTIONAL :: latdeg, londeg
    REAL(dp), INTENT(INOUT), OPTIONAL :: t0    !< start time in s
    REAL(dp), INTENT(INOUT), OPTIONAL :: t1    !< end time in s
    LOGICAL, INTENT(IN), OPTIONAL :: writeable !< whether to open the file with read-write access
    INTEGER :: n, nt, natts, nferr, i(1), rank, mode
    REAL(dp), ALLOCATABLE :: edges(:)
    IF (group%id.EQ.0) THEN
       mode = NF90_NOWRITE
       IF (PRESENT(writeable)) THEN
          IF (writeable) mode = NF90_WRITE
       END IF
       nferr = NF90_OPEN(path=TRIM(group%filename), mode=mode, ncid=group%id)
       IF (nferr.NE.NF90_NOERR) CALL onf_error ('nf90info', 'NF90_OPEN', TRIM(group%filename), nferr)
    END IF
    nferr = NF90_INQUIRE(ncid=group%id, nAttributes=natts)
    nferr = NF90_INQ_DIMID(ncid=group%id, name=group%time%name, dimid=group%time%id)
    IF (nferr.NE.NF90_NOERR) CALL onf_error ('nf90info', 'NF90_INQ_DIMID', group%time%name, nferr)
    CALL group%time%read (ncid=group%id, caller='nf90info')
    nferr = NF90_INQ_DIMID(ncid=group%id, name=group%lat%name, dimid=group%lat%id)
    IF (nferr.EQ.NF90_NOERR) THEN                              ! latitude stored as a dataset
       CALL group%lat%read (ncid=group%id, caller='nf90info')
    ELSEIF (LEN_TRIM(group%lat%name).GT.0) THEN                ! latitude stored as an attribute
       ALLOCATE (group%lat%values(1))
       nferr = NF90_GET_ATT(ncid=group%id, varid=NF90_GLOBAL, name=group%lat%name, &
                            values=group%lat%values)
       IF (nferr.NE.NF90_NOERR) CALL onf_error ('nf90info', 'NF90_GET_ATT', group%lat%name, nferr)
    ELSE
       ALLOCATE (group%lat%values(0))
    END IF
    IF (PRESENT(latdeg).AND.(SIZE(group%lat%values).GT.0)) THEN
       i = MINLOC(ABS(group%lat%values - latdeg))
       group%ilat = i(1)
       IF (ABS(group%lat%values(group%ilat) - latdeg).GT.1D-6) WRITE (stdout,100) &
            TRIM(group%lat%name), latdeg, TRIM(group%filename), group%lat%values(group%ilat)
       ! use actual lat/lon from forcing for light cycle and output file
       latdeg = group%lat%values(group%ilat)
    ELSE
       group%ilat = 0
    END IF
    nferr = NF90_INQ_DIMID(ncid=group%id, name=group%lon%name, dimid=group%lon%id)
    IF (nferr.EQ.NF90_NOERR) THEN
       CALL group%lon%read (ncid=group%id, caller='nf90info')
    ELSEIF (LEN_TRIM(group%lon%name).GT.0) THEN
       ALLOCATE (group%lon%values(1))
       nferr = NF90_GET_ATT(ncid=group%id, varid=NF90_GLOBAL, name=group%lon%name, &
                            values=group%lon%values)
       IF (nferr.NE.NF90_NOERR) CALL onf_error ('nf90info', 'NF90_GET_ATT', group%lon%name, nferr)
    ELSE
       ALLOCATE (group%lon%values(0))
    END IF
    IF (PRESENT(londeg).AND.(SIZE(group%lon%values).GT.0)) THEN
       i = MINLOC(ABS(group%lon%values - londeg))
       group%ilon = i(1)
       IF (ABS(group%lon%values(group%ilon) - londeg).GT.1D-6) WRITE (stdout,100) &
            TRIM(group%lon%name), londeg, TRIM(group%filename), group%lon%values(group%ilon)
       ! use actual lat/lon from forcing for light cycle and output file
       londeg = group%lon%values(group%ilon)
    ELSE
       group%ilon = 0
    END IF
    nferr = NF90_INQ_DIMID(ncid=group%id, name=group%bounds%name, dimid=group%bounds%id)
    IF (nferr.NE.NF90_NOERR) CALL onf_error ('nf90info', 'NF90_INQ_DIMID', group%bounds%name, nferr)
    nferr = NF90_INQ_DIMID(ncid=group%id, name=group%depth%name, dimid=group%depth%id)
    IF (nferr.NE.NF90_NOERR) CALL onf_error ('nf90info', 'NF90_INQ_DIMID', group%depth%name, nferr)
    CALL group%depth%read (ncid=group%id, caller='nf90info')
    group%nbox = SIZE(group%depth%values)
    IF (LEN_TRIM(group%depth%bounds%name).EQ.0) THEN
       ALLOCATE (group%depth%bounds%values(2,group%nbox), edges(0:group%nbox))
       group%depth%bounds%name = TRIM(group%depth%name)//'_edges'
       group%depth%bounds%units = group%depth%units
       edges(0) = 0._dp
       DO n=1,group%nbox
          edges(n) = 2._dp*group%depth%values(n) - edges(n-1)
       END DO
       group%depth%bounds%values(1,:) = edges(0:group%nbox-1)
       group%depth%bounds%values(2,:) = edges(1:group%nbox)
       DEALLOCATE (edges)
    END IF
    nt = SIZE(group%time%values)
    IF (mode.EQ.NF90_WRITE) THEN ! only for appending to previous output
       group%it0 = nt
       group%ntim = 1
    ELSEIF (PRESENT(t0).AND.(nt.GT.1)) THEN
       IF (t0.LE.group%time%values(1)) t0 = t0 + group%time%values(1)
       i = MAXLOC(group%time%values, MASK=group%time%values.LE.t0)
       group%it0 = i(1)
       IF (PRESENT(t1)) THEN
          t1 = MIN(t1, group%time%values(nt))
          i = MINLOC(group%time%values(group%it0:), MASK=group%time%values(group%it0:).GE.t1)
          group%ntim = i(1)
       ELSE
          group%ntim = 1
       END IF
    ELSE
       group%it0 = 1
       group%ntim = nt
    END IF
    group%time%values(1:group%ntim) = group%time%values(group%it0:group%it0+group%ntim-1) &
         + group%timediff ! convert UTC to local time
100 FORMAT ("nf90info: closest ",A," to",G12.5,"in ",A," is",G12.5)
  END SUBROUTINE nf90info

  !> Read initial state-variable values and create and initialize output file
  !!
  !! Subroutine invar is called as states%init in plankton:plankton_init
  !! \param[inout]  stv     state-variable structure (type statevars)
  !! \param[in]     resume  append output to initial-conditions file
  !! \param[in]     force   overwrite output file if it exists
  !! \param[in]     ofn     name of output file
  !! \param[out]    ystv    allocated real(dp) vector of initial state-variable values
  SUBROUTINE invar (stv, ofn, control, dfn, pfn, force, resume)
    IMPLICIT NONE
    CLASS(statevars), TARGET, INTENT(INOUT) :: stv
    CHARACTER(LEN=*), INTENT(IN) :: ofn, control, dfn, pfn
    LOGICAL, INTENT(IN) :: force, resume
    TYPE(dimension) :: sms
    CHARACTER(LEN=23) :: date
    CHARACTER(LEN=500) :: prefix="", prefix_ic="", prefix_dfn="", prefix_pfn=""
    CHARACTER(LEN=:), DIMENSION(:), ALLOCATABLE :: ftnames, from, to, fromto
    LOGICAL :: exists
    INTEGER, DIMENSION(:,:), ALLOCATABLE :: ifrom, ito ! indices in smsk
    INTEGER :: dimids(4), datevals(8)
    INTEGER :: id, n, nn, nferr, namelen=0, stringid, ndims, lunof
    ALLOCATE (offsets(stv%nsv_one_box,stv%nbox), stv%all_states(stv%nsv_total), stv%start(4), stv%count(4))
    stv%data(1:stv%nsv_one_box,1:stv%nbox) => stv%all_states(1:stv%nsv_all_boxes)
    offsets = 0D0
    id = stv%id                 ! save file ID
    IF (LEN_TRIM(stv%svgn).GT.0) THEN
       nferr = NF90_INQ_NCID(ncid=id, name=TRIM(stv%svgn), grp_ncid=stv%id)
       IF (nferr.NE.NF90_NOERR) &
            CALL onf_error (caller='invar', nffun='NF90_INQ_NCID', name=stv%svgn, nferr=nferr)
    END IF
    DO n=1,stv%nsv_one_box              ! read initial conditions for all states
       stv%var(n)%values => stv%data(n,:)
       CALL stv%var(n)%read (group=stv, data=stv%var(n)%values)
    END DO
    stv%it => stv%start(2:2)
    stv%count = (/stv%nbox, 1, 1, 1/)
    stv%start = (/1, 0, 1, 1/)
    IF (resume) THEN
       stv%it = stv%it0
       stv%id = id
       RETURN
    END IF
    nferr = NF90_CLOSE(id)
    INQUIRE (FILE=TRIM(ofn), EXIST=exists)
    IF (exists) THEN
       IF (force.AND.(SCAN(TRIM(ofn), '*?').EQ.0)) THEN
          OPEN (NEWUNIT=lunof, FILE=TRIM(ofn))
          CLOSE (UNIT=lunof, STATUS='DELETE')
       ELSE
          WRITE (stderr,'("Output file ",A," exists. Use -f to overwrite.")') TRIM(ofn)
          STOP
       END IF
    END IF
    stv%depth%bounds%name = TRIM(stv%depth%name)//'_edges'
    stv%depth%bounds%units = stv%depth%units
    CALL DATE_AND_TIME (values = datevals)
    WRITE (date, '(I4.4,"-",I2.2,"-",I2.2," ",I2.2,":",I2.2,":",I2.2,".",I3.3)') &
         datevals((/(n, n = 1,3), (n ,n = 5,8)/))
    nn = 0
    IF (ofn(1:1).NE.'/') THEN   ! if ofn is a relative path
       DO n=1,LEN_TRIM(ofn)     ! make the paths of control, parameter, data files relative to ofn
          SELECT CASE (ofn(n:n))
          CASE ('.')
             nn = nn + 1
          CASE ('/')
             IF (nn.EQ.2) THEN     ! this means ofn contains ../
                prefix = ''        ! in this case the path cannot be reconstructed
                EXIT
             ELSEIF (nn.EQ.0) THEN
                prefix = TRIM(prefix)//'../'
             END IF
          CASE DEFAULT
             nn = 0
          END SELECT
       END DO
       IF (stv%filename(1:1).NE.'/') prefix_ic = prefix ! use prefix only for relative file names
       IF (dfn(1:1).NE.'/') prefix_dfn = prefix
       IF (pfn(1:1).NE.'/') prefix_pfn = prefix
    ELSE
       IF (control(1:1).NE.'/') WRITE (stdout,100) 'control', 'control'
       IF (pfn(1:1).NE.'/') WRITE (stdout,100) 'parameter', 'parameter'
       IF (dfn(1:1).NE.'/') WRITE (stdout,100) 'data', 'data'
       IF (stv%filename(1:1).NE.'/') WRITE (stdout,100) 'initial-conditions', 'initial-conditions'
    END IF
    nferr = NF90_CREATE(TRIM(ofn), cmode=NF90_NETCDF4, ncid=stv%id) ! create output file
    IF (nferr.NE.NF90_NOERR) CALL onf_error ('invar', 'NF90_CREATE', TRIM(ofn), nferr)
    nferr = NF90_PUT_ATT(ncid=stv%id, varid=NF90_GLOBAL, name='date', values=date)
    nferr = NF90_PUT_ATT(ncid=stv%id, varid=NF90_GLOBAL, name='initial_conditions',&
                         values=TRIM(prefix_ic)//TRIM(stv%filename))
    nferr = NF90_PUT_ATT(ncid=stv%id, varid=NF90_GLOBAL, name='control_file', values=TRIM(prefix)//TRIM(control))
    nferr = NF90_PUT_ATT(ncid=stv%id, varid=NF90_GLOBAL, name='data_file', values=TRIM(prefix_dfn)//TRIM(dfn))
    nferr = NF90_PUT_ATT(ncid=stv%id, varid=NF90_GLOBAL, name='parameter_file', values=TRIM(prefix_pfn)//TRIM(pfn))
    ! create dimensions and dimension-variables
    nferr = NF90_DEF_DIM(ncid=stv%id, name=stv%depth%name, len=stv%nbox, dimid=stv%depth%id)
    nferr = NF90_DEF_DIM(ncid=stv%id, name=stv%bounds%name, len=2, dimid=stv%bounds%id)
    nferr = NF90_DEF_VAR(ncid=stv%id, name=stv%depth%name, xtype=NF90_DOUBLE,&
                         dimids=(/stv%depth%id/), varid=stv%depth%varid)
    nferr = NF90_PUT_ATT(ncid=stv%id, varid=stv%depth%varid, name='units', values=TRIM(stv%depth%units))
    nferr = NF90_PUT_ATT(ncid=stv%id, varid=stv%depth%varid, name='axis', values='Z')
    nferr = NF90_PUT_ATT(ncid=stv%id, varid=stv%depth%varid, name='positive', values='down')
    nferr = NF90_PUT_ATT(ncid=stv%id, varid=stv%depth%varid, name='bounds', values=stv%depth%bounds%name)
    nferr = NF90_DEF_VAR(ncid=stv%id, name=stv%depth%bounds%name, xtype=NF90_DOUBLE,&
                         dimids=(/stv%bounds%id, stv%depth%id/), varid=stv%depth%bounds%varid)
    nferr = NF90_PUT_ATT(ncid=stv%id, varid=stv%depth%bounds%varid, name='units', values=TRIM(stv%depth%bounds%units))
    nferr = NF90_DEF_DIM(ncid=stv%id, name=stv%time%name, len=NF90_UNLIMITED, dimid=stv%time%id)
    nferr = NF90_DEF_VAR(ncid=stv%id, name=stv%time%name, xtype=NF90_DOUBLE,&
                         dimids=(/stv%time%id/), varid=stv%time%varid)
    nferr = NF90_PUT_ATT(ncid=stv%id, varid=stv%time%varid, name='units', values=TRIM(stv%time%units))
    nferr = NF90_PUT_ATT(ncid=stv%id, varid=stv%time%varid, name='axis', values='T')
    nferr = NF90_PUT_ATT(ncid=stv%id, varid=stv%time%varid, name='calendar', values='standard')
    IF (ALL(LEN_TRIM((/stv%lon%name, stv%lat%name/)).GT.0)) THEN
       nferr = NF90_DEF_DIM(ncid=stv%id, name=stv%lat%name, len=1, dimid=stv%lat%id)
       nferr = NF90_DEF_VAR(ncid=stv%id, name=stv%lat%name, xtype=NF90_DOUBLE,&
                            dimids=(/stv%lat%id/), varid=stv%lat%varid)
       nferr = NF90_PUT_ATT(ncid=stv%id, varid=stv%lat%varid, name='units', values=TRIM(stv%lat%units))
       nferr = NF90_PUT_ATT(ncid=stv%id, varid=stv%lat%varid, name='axis', values='Y')
       nferr = NF90_DEF_DIM(ncid=stv%id, name=stv%lon%name, len=1, dimid=stv%lon%id)
       nferr = NF90_DEF_VAR(ncid=stv%id, name=stv%lon%name, xtype=NF90_DOUBLE,&
                            dimids=(/stv%lon%id/), varid=stv%lon%varid)
       nferr = NF90_PUT_ATT(ncid=stv%id, varid=stv%lon%varid, name='units', values=TRIM(stv%lon%units))
       nferr = NF90_PUT_ATT(ncid=stv%id, varid=stv%lon%varid, name='axis', values='X')
       ndims = 4
    ELSE
       ndims = 2
    END IF
    dimids = (/stv%depth%id, stv%time%id, stv%lon%id, stv%lat%id/)
    DO n=1,stv%nsv_one_box
       nferr = NF90_DEF_VAR(ncid=stv%id, name=stv%var(n)%name, xtype=NF90_DOUBLE,&
                            dimids=dimids(1:ndims), varid=stv%var(n)%varid)
       IF (nferr.NE.NF90_NOERR) &
            CALL onf_error ('invar', 'NF90_DEF_VAR', TRIM(stv%var(n)%name), nferr)
       nferr = NF90_PUT_ATT(ncid=stv%id, varid=stv%var(n)%varid, name='units',&
                            values=TRIM(stv%var(n)%units))
    END DO
    IF (stv%write_soma) THEN
       stv%write_output => write_output
       stv%flux%values => stv%all_states(stv%i0flux:stv%i0sms-1)
       stv%soma%values => stv%all_states(stv%i0sms:stv%nsv_total)
       nferr = NF90_PUT_ATT(ncid=stv%id, varid=NF90_GLOBAL, name='soma', values=TRIM(stv%soma%name))
       nferr = NF90_PUT_ATT(ncid=stv%id, varid=NF90_GLOBAL, name='flux', values=TRIM(stv%flux%name))
       stv%depth_flux%name = 'depth_flux'
       nferr = NF90_DEF_DIM(ncid=stv%id, name=stv%depth_flux%name, len=stv%nfl, dimid=stv%depth_flux%id)
       nferr = NF90_DEF_VAR(ncid=stv%id, name=stv%depth_flux%name, xtype=NF90_DOUBLE,&
                            dimids=(/stv%depth_flux%id/), varid=stv%depth_flux%varid)
       nferr = NF90_PUT_ATT(ncid=stv%id, varid=stv%depth_flux%varid, name='units', values=TRIM(stv%depth%units))
       nferr = NF90_PUT_ATT(ncid=stv%id, varid=stv%depth_flux%varid, name='axis', values='Z')
       nferr = NF90_PUT_ATT(ncid=stv%id, varid=stv%depth_flux%varid, name='positive', values='down')
       namelen = MAX(MAXVAL(LEN_TRIM(stv%var%name)), 11)*2 + 2
       ALLOCATE (ifrom(0:stv%nsv_one_box,stv%nsv_one_box), ito(0:stv%nsv_one_box,stv%nsv_one_box))
       ALLOCATE (CHARACTER(namelen)::ftnames(0:stv%nsv_one_box),&
                 from(stv%nsms), to(stv%nsms), fromto(stv%nsms))
       ito(0,:) = (/(nn, nn=1,stv%nsv_one_box)/) ! set up index matrices for state names
       ftnames(0) = 'aggregation'
       DO n=1,stv%nsv_one_box
          ftnames(n) = stv%var(n)%name
          ifrom(:,n) = (/(nn, nn=0,stv%nsv_one_box)/)
          ito(n,:) = (/(nn, nn=1,stv%nsv_one_box)/)
       END DO
       sms%name = 'sms'
       from = ftnames(PACK(ifrom, MASK=stv%smsk))
       to = ftnames(PACK(ito, MASK=stv%smsk))
       DO n=1,stv%nsms
          fromto(n) = TRIM(from(n))//'->'//TRIM(to(n))
       END DO
       nferr = NF90_DEF_DIM(ncid=stv%id, name=stv%state%name, len=stv%nsv_one_box, dimid=stv%state%id)
       nferr = NF90_DEF_DIM(ncid=stv%id, name=sms%name, len=stv%nsms, dimid=sms%id)
       nferr = NF90_DEF_DIM(ncid=stv%id, name='string', len=namelen, dimid=stringid)
       nferr = NF90_DEF_VAR(ncid=stv%id, name=stv%state%name, xtype=NF90_CHAR,&
                            dimids=(/stringid, stv%state%id/), varid=stv%state%varid)
       nferr = NF90_DEF_VAR(ncid=stv%id, name=sms%name, xtype=NF90_CHAR,&
                            dimids=(/stringid, sms%id/), varid=sms%varid)
       nferr = NF90_DEF_VAR(ncid=stv%id, name=stv%flux%name, xtype=NF90_DOUBLE,&
                            dimids=(/stv%state%id, stv%depth_flux%id, dimids(2:ndims)/), varid=stv%flux%varid)
       IF (nferr.NE.NF90_NOERR) &
            CALL onf_error (caller='onf:invar', nffun='NF90_DEF_VAR', name=stv%flux%name, nferr=nferr)
       nferr = NF90_DEF_VAR(ncid=stv%id, name=stv%soma%name, xtype=NF90_DOUBLE,&
                            dimids=(/sms%id, dimids(1:ndims)/), varid=stv%soma%varid)
       IF (nferr.NE.NF90_NOERR) &
            CALL onf_error (caller='onf:invar', nffun='NF90_DEF_VAR', name=stv%soma%name, nferr=nferr)
    END IF
    nferr = NF90_ENDDEF(ncid=stv%id)
    nferr = NF90_PUT_VAR(ncid=stv%id, varid=stv%depth%varid, values=stv%depth%values)
    nferr = NF90_PUT_VAR(ncid=stv%id, varid=stv%depth%bounds%varid, values=stv%depth%bounds%values)
    IF (ndims.EQ.4) THEN
       nferr = NF90_PUT_VAR(ncid=stv%id, varid=stv%lat%varid, values=stv%lat%values)
       nferr = NF90_PUT_VAR(ncid=stv%id, varid=stv%lon%varid, values=stv%lon%values)
    END IF
    IF (stv%write_soma) THEN
       nferr = NF90_PUT_VAR(ncid=stv%id, varid=sms%varid, values=fromto)
       DO n=1,stv%nsv_one_box
          nferr = NF90_PUT_VAR(ncid=stv%id, varid=stv%state%varid, values=TRIM(stv%var(n)%name),&
                               start=(/1, n/))
       END DO
       DEALLOCATE (ifrom, ito, ftnames, from, to, fromto)
       nferr = NF90_PUT_VAR(ncid=stv%id, varid=stv%depth_flux%varid, values=stv%flux_depth)
    END IF
100 FORMAT ("Warning: cannot determine the location of the ",A," file relative to the output file. "/&
         "Either both or neither of the ",A," and output files should be absolute paths.")
  END SUBROUTINE invar

  !> Open physical data file and read forcing datasets
  !!
  !! Subroutine inpd is called as phys%init in plankton:plankton_init
  !! \param [inout] pdf%data      gridded (depth-time) forcing data
  !! \param [inout] pdf%profile   temporally constant vertical profiles
  !! \param [inout] pdf%surface   surface data
  SUBROUTINE inpd (pdf, times, env, box, svd)
    USE, INTRINSIC :: IEEE_ARITHMETIC
    IMPLICIT NONE
    CLASS(phydat), TARGET, INTENT(INOUT) :: pdf
    TYPE(timing), INTENT(INOUT) :: times
    TYPE(local), INTENT(INOUT) :: env
    TYPE(layer), INTENT(INOUT) :: box(0:)
    TYPE(nf90vec), DIMENSION(:), INTENT(IN) :: svd
    CHARACTER(LEN=NF90_MAX_NAME) :: descr
    INTEGER :: ind(1), iPAR=0, n, nn, nb, nsv, nferr
    INTEGER, DIMENSION(:), ALLOCATABLE :: ibot
    LOGICAL :: bottom
    IF ((ngdds.GT.0).AND.(pdf%ntim.EQ.1)) &
         STOP 'inpd:specify temporally constant profiles as profile in namelist physics.'
    nsv = SIZE(svd)
    ALLOCATE (pdsds(pdf%ntim,nsfds+nsvds+nbvds), pdgds(pdf%nbox,pdf%ntim,ngdds),&
              pdf%attributes(nsfds+nsvds+nbvds+ngdds+npfds), pdf%sds(nbvds+nsvds+nsfds),&
              pdf%pds(pdf%nbox,ngdds), pdf%cds(pdf%nbox,npfds), pdf%isflx(nsfds), ibot(nbvds))
    DO nb=0,pdf%nbox            ! nullify so that association status can be checked
       NULLIFY (box(nb)%temperature, box(nb)%alkalinity, box(nb)%salinity, box(nb)%vdc,&
                box(nb)%velocity, box(nb)%silc)
    END DO
    pdf%surflux => pdf%sds(1:nsfds)
    pdf%botbc => pdf%sds(nsfds+1:nsfds+nbvds)
    ibot = 0D0
    DO n=1,ngdds                ! gridded (depth-time) forcing data
       bottom = .FALSE.         ! whether data apply to the bottom of the box
       SELECT CASE (TRIM(pdf%data(n)%type))
       CASE ('Temperature')
          pdf%data(n)%units = '°C'
          DO nb=1,pdf%nbox; box(nb)%temperature => pdf%pds(nb,n); END DO
       CASE ('Salinity')
          pdf%data(n)%units = 'psu'
          DO nb=1,pdf%nbox; box(nb)%salinity => pdf%pds(nb,n); END DO
       CASE ('Alkalinity')
          IF (env%ialk.GT.0) STOP &
               'Alkalinity is specified both in forcing in namelist physics and &
               &in dima in namelist coco as a state variable.  Please remove from&
               & physics or coco.'
          pdf%data(n)%units = 'mmol m-3'
          DO nb=1,pdf%nbox; box(nb)%alkalinity => pdf%pds(nb,n); END DO
       CASE ('VDC')             ! vertical diffusivity at bottom of box
          bottom = .TRUE.       ! set to 0 at bottom of model domain if closed
          pdf%data(n)%units = 'm2 s-1'
          ALLOCATE (box(0)%vdc)
          box(0)%vdc = 0D0      ! 0 at surface of model domain
          DO nb=1,pdf%nbox; box(nb)%vdc => pdf%pds(nb,n); END DO
       CASE ('Velocity')        ! vertical velocity at bottom of box
          bottom = .TRUE.       ! set to 0 at bottom of model domain if closed
          pdf%data(n)%units = 'm s-1'
          ALLOCATE (box(0)%velocity)
          box(0)%velocity = 0D0 ! 0 at surface of model domain
          DO nb=1,pdf%nbox; box(nb)%velocity => pdf%pds(nb,n); END DO
       CASE DEFAULT
          WRITE (stdout,110) 'data', TRIM(pdf%data(n)%type)
          STOP
       END SELECT
       WRITE (stdout,100) TRIM(pdf%data(n)%type), TRIM(pdf%data(n)%name)
       CALL pdf%data(n)%read (group=pdf, data=pdgds(:,:,n))
       CALL set_att (group=pdf, ds=pdf%data(n), natt=n)
       IF (pdf%clsbot.AND.bottom) pdgds(pdf%nbox,:,n) = 0D0
    END DO
    NULLIFY (env%sfDIC, env%sfO2)
    DO n=1,nsfds                ! surface-flux data
       ind = PACK([(nn, nn = 1,nsv)], MASK=svd%name.EQ.TRIM(pdf%surflx(n)%type), VECTOR=(/0/))
       IF (ind(1).GT.0) THEN
          descr = 'Surface flux of '//TRIM(pdf%surflx(n)%type)
          pdf%surflx(n)%units = TRIM(svd(ind(1))%units)//' m s-1'
          pdf%isflx(n) = ind(1) ! connect state-variable index (ind(1)) with surface data index (n)
          IF (ind(1).EQ.env%idic) env%sfDIC => pdf%surflux(n)
          IF (ind(1).EQ.env%io2) env%sfO2 => pdf%surflux(n)
       ELSE
          WRITE (stdout,110) 'surface', TRIM(pdf%surflx(n)%type)
          STOP
       END IF
       WRITE (stdout,100) TRIM(descr), TRIM(pdf%surflx(n)%name)
       CALL pdf%surflx(n)%read (group=pdf, data=pdsds(:,n)) ! subroutine read_ds
       CALL set_att (group=pdf, ds=pdf%surflx(n), natt=n+ngdds)
    END DO
    IF (.NOT.ASSOCIATED(env%sfDIC)) ALLOCATE (env%sfDIC)
    IF (.NOT.ASSOCIATED(env%sfO2)) ALLOCATE (env%sfO2)
    env%sfDIC = 0._dp
    env%sfO2 = 0._dp
    DO n=1,nbvds                ! bottom-boundary data
       ind = PACK([(nn, nn = 1,nsv)], MASK=svd%name.EQ.TRIM(pdf%bottom(n)%type), VECTOR=(/0/))
       IF (ind(1).GT.0) THEN
          ibot(n) = ind(1)
          WRITE (stdout,100) 'bottom '//TRIM(pdf%bottom(n)%type), TRIM(pdf%bottom(n)%name)
          pdf%bottom(n)%units = svd(ind(1))%units
          CALL pdf%bottom(n)%read (group=pdf, data=pdsds(:,n+nsfds))
          CALL set_att (group=pdf, ds=pdf%bottom(n), natt=n+ngdds+nsfds)
       ELSE
          WRITE (stdout,110) 'bottom', TRIM(pdf%bottom(n)%type)
          STOP
       END IF
    END DO
    ! names of surface datasets specified as surface types and names in
    ! namelist physics in subroutine innf (states%open in plankton:plankton_init)
    NULLIFY (pdf%surPAR)
    DO n=1,nsvds                ! surface-value data
       nn = n + nbvds + nsfds
       SELECT CASE (TRIM(pdf%surface(n)%type))
       CASE ('avgPAR', 'avgSWR')
          descr = 'Average daily surface '//pdf%surface(n)%type(4:6)
          pdf%surface(n)%units = 'W m-2'
          DEALLOCATE (env%dayPAR)
          ALLOCATE (env%sPAR)
          env%dayPAR => pdf%sds(nn)
          env%calcPAR = .FALSE.
          IF (iPAR.NE.0) STOP 'Please specify only one surface-irradiance dataset.'
          IF (TRIM(pdf%surface(n)%type).EQ.'avgSWR') iPAR = nn
       CASE ('PAR', 'SWR')
          descr = 'Instantaneous surface '//pdf%surface(n)%type(1:3)
          pdf%surface(n)%units = 'W m-2'
          DEALLOCATE (env%dayPAR)
          env%sPAR => pdf%sds(nn)
          pdf%surPAR => pdsds(:,nn)
          env%calcPAR = .FALSE.
          IF (iPAR.NE.0) STOP 'Please specify only one surface-irradiance dataset.'
          IF (TRIM(pdf%surface(n)%type).EQ.'SWR') iPAR = nn
       CASE ('Wind')
          descr = 'Surface wind'
          pdf%surface(n)%units = 'm s-1'
          DEALLOCATE (env%wind)
          env%wind => pdf%sds(nn)
       CASE ('Ice')
          descr = 'Surface ice cover'
          pdf%surface(n)%units = '1'
          DEALLOCATE (env%ice)
          env%ice => pdf%sds(nn)
       CASE ('Pressure')
          descr = 'Surface pressure'
          pdf%surface(n)%units = 'Pa'
          DEALLOCATE (env%pressure)
          env%pressure => pdf%sds(nn)
       CASE ('CO2')
          descr = 'Atmospheric CO2'
          pdf%surface(n)%units = '1'
          DEALLOCATE (env%xco2)
          env%xco2 => pdf%sds(nn)
       CASE ('Alkalinity')
          IF (env%ialk.GT.0) STOP &
               'Alkalinity is specified both in surface in namelist physics and in dima in &
               & namelist coco as a state variable.  Please remove from physics or coco.'
          IF (ASSOCIATED(box(1)%alkalinity)) STOP &
               'Alkalinity is specified both as forcing and surface in namelist &
               &physics.  Please remove one of them.'
          descr = 'Surface alkalinity'
          pdf%surface(n)%units = 'mmol m-3'
          box(1)%alkalinity => pdf%sds(nn)
       CASE ('law')
          descr = 'Attenuance'
          pdf%surface(n)%units = 'm-1'
          env%lacw => pdf%sds(nn)
       CASE ('Latitude')
          descr = 'Geographic latitude'
          pdf%surface(n)%units = 'rad'
          times%latrad => pdf%sds(nn)
       CASE DEFAULT
          WRITE (stdout,110) 'surface', TRIM(pdf%surface(n)%type)
          STOP
       END SELECT
       WRITE (stdout,100) TRIM(descr), TRIM(pdf%surface(n)%name)
       CALL pdf%surface(n)%read (group=pdf, data=pdsds(:,nn))
       CALL set_att (group=pdf, ds=pdf%surface(n), natt=nn+ngdds)
    END DO
    IF (iPAR.GT.0) pdsds(:,iPAR) = pdsds(:,iPAR)*pdf%fPAR ! convert SWR to PAR
    DO n=1,npfds                ! temporally constant vertical profiles
       bottom = .FALSE.
       SELECT CASE (TRIM(pdf%profile(n)%type))
       CASE ('Temperature')
          pdf%profile(n)%units = '°C'
          DO nb=1,pdf%nbox; box(nb)%temperature => pdf%cds(nb,n); END DO
       CASE ('Salinity')
          pdf%profile(n)%units = 'psu'
          DO nb=1,pdf%nbox; box(nb)%salinity => pdf%cds(nb,n); END DO
       CASE ('Alkalinity')
          pdf%profile(n)%units = 'mmol m-3'
          DO nb=1,pdf%nbox; box(nb)%alkalinity => pdf%cds(nb,n); END DO
       CASE ('VDC')
          bottom = .TRUE.
          pdf%profile(n)%units = 'm2 s-1'
          DO nb=1,pdf%nbox; box(nb)%vdc => pdf%cds(nb,n); END DO
       CASE ('Velocity')
          bottom = .TRUE.
          pdf%profile(n)%units = 'm s-1'
          ALLOCATE (box(0)%velocity)
          box(0)%velocity = 0D0
          DO nb=1,pdf%nbox; box(nb)%velocity => pdf%cds(nb,n); END DO
       CASE DEFAULT
          WRITE (stdout,110) 'data', TRIM(pdf%profile(n)%type)
          STOP
       END SELECT
       WRITE (stdout,100) 'constant '//TRIM(pdf%profile(n)%type), TRIM(pdf%profile(n)%name)
       CALL pdf%profile(n)%read (group=pdf, data=pdf%cds(:,n))
       CALL set_att (group=pdf, ds=pdf%profile(n), natt=n+ngdds+nsfds+nbvds)
       IF (pdf%clsbot.AND.bottom) pdf%cds(pdf%nbox,n) = 0D0
    END DO
    IF (pdf%id.NE.0) THEN
       nferr = NF90_CLOSE(ncid=pdf%id)
       IF (pdf%ntim.GT.1) pdf%get => readpd
    END IF
    ALLOCATE (pdf%ibot(COUNT(ibot.GT.0)))
    pdf%ibot = PACK(ibot, MASK=ibot.GT.0)
    IF (ANY(ISNAN(pdgds)).OR.ANY(ISNAN(pdsds))) THEN
       WRITE (stderr,'("inpd: NaN detected in forcing data.")')
       STOP
    END IF
100 FORMAT ("inpd: ",A," dataset: ",A,".")
110 FORMAT ("inpd: unknown ",A," type: ",A,".")
  CONTAINS
    SUBROUTINE set_att (group, ds, natt)
      IMPLICIT NONE
      CLASS(nf90group), INTENT(INOUT) :: group
      CLASS(nf90var), INTENT(IN) :: ds
      INTEGER, INTENT(IN) :: natt
      group%attributes(natt)%name = TRIM(ds%type)//' dataset'
      group%attributes(natt)%value = TRIM(ds%name)
    END SUBROUTINE set_att
  END SUBROUTINE inpd

  !> read dataset in group described by ds into array data
  !! - if ds%units is set, data will be converted to these units
  SUBROUTINE read_ds (ds, group, data)
    IMPLICIT NONE
    CLASS(nf90group), INTENT(IN) :: group
    CLASS(nf90var), INTENT(INOUT) :: ds
    REAL(dp), DIMENSION(*), INTENT(OUT) :: data
    REAL(dp) :: cf, os
    CHARACTER(LEN=NF90_MAX_NAME) :: nf_units
    INTEGER :: nferr, nd, rank_ds, len, dids(NF90_MAX_VAR_DIMS)
    INTEGER, DIMENSION(:), ALLOCATABLE :: starts, counts
    nferr = NF90_INQ_VARID(ncid=group%id, name=ds%name, varid=ds%varid)
    IF (nferr.NE.NF90_NOERR) &
         CALL onf_error (caller='read_ds ('//TRIM(group%filename)//')',&
         nffun='NF90_INQ_VARID', name=ds%name, nferr=nferr)
    nferr = NF90_INQUIRE_VARIABLE(ncid=group%id, varid=ds%varid, ndims=rank_ds, dimids=dids)
    ALLOCATE (starts(rank_ds), counts(rank_ds))
    starts = 1
    counts = 0
    DO nd=1,rank_ds
       IF (dids(nd).EQ.group%lat%id) THEN
          starts(nd) = group%ilat
          counts(nd) = 1
       ELSEIF (dids(nd).EQ.group%lon%id) THEN
          starts(nd) = group%ilon
          counts(nd) = 1
       ELSEIF (dids(nd).EQ.group%depth%id) THEN
          counts(nd) = group%nbox
       ELSEIF (dids(nd).EQ.group%bounds%id) THEN
          counts(nd) = 2
       ELSEIF (dids(nd).EQ.group%time%id) THEN
          starts(nd) = group%it0
          counts(nd) = group%ntim
       END IF
    END DO
    IF (COUNT(counts.GT.0).NE.rank_ds) THEN
       WRITE (stderr,'("read_ds: rank of ",A," is ",I1,", should be ",I1,".")') &
            TRIM(ds%name), rank_ds, COUNT(counts.GT.0)
       STOP
    END IF
    len = PRODUCT(counts)
    nferr = NF90_GET_VAR(ncid=group%id, varid=ds%varid, values=data(1:len),&
                         start=starts, count=counts)
    IF (nferr.NE.NF90_NOERR) &
         CALL onf_error (caller='read_ds', nffun='NF90_GET_VAR', name=ds%name, nferr=nferr)
    nferr = NF90_GET_ATT(ncid=group%id, varid=ds%varid, name='units', values=nf_units)
    IF ((nferr.NE.NF90_NOERR).AND.(nferr.NE.NF90_ENOTATT)) &
         CALL onf_error ("read_ds", "NF90_GET_ATT", TRIM(ds%name)//'/units', nferr)
    IF ((LEN_TRIM(ds%units).GT.0).AND.(nferr.NE.NF90_ENOTATT)) THEN
       CALL slot (nf_units, ds%units, slope=cf, offset=os)
       data(1:len) = data(1:len)*cf + os
    ELSEIF (LEN_TRIM(ds%units).GT.0) THEN
       WRITE (stdout, '("read_ds: Dataset ",A," in file ",A," has no units, assuming ",A,".")') &
            TRIM(ds%name), TRIM(group%filename), TRIM(ds%units)
    ELSEIF (nferr.EQ.NF90_ENOTATT) THEN
       WRITE (stdout, '("read_ds: Dataset ",A," in file ",A," has no units.")') &
            TRIM(ds%name), TRIM(group%filename)
       STOP
    END IF
  END SUBROUTINE read_ds

!> Read physical data for time interval encompassing phytim
  SUBROUTINE readpd (pda, time)
    IMPLICIT NONE
    CLASS(phydat), INTENT(INOUT) :: pda !< physical data array
    REAL(dp), INTENT(IN) :: time        !< current time
    INTEGER :: j
    INTEGER, SAVE :: ipdt=1
    REAL(dp) :: a, b
    DO WHILE ((time.LT.pda%time%values(ipdt)).AND.(ipdt.GT.1))
       ipdt = ipdt - 1
    END DO
    j = ipdt
    DO WHILE ((time.GT.pda%time%values(j)).AND.(j.LT.pda%ntim))
       j = j + 1
    END DO
    IF (ipdt.EQ.j) THEN
       pda%pds = pdgds(:,ipdt,:)
       pda%sds = pdsds(ipdt,:)
    ELSE
       b = (time - pda%time%values(ipdt))/(pda%time%values(j) - pda%time%values(ipdt))
       a = 1D0 - b
       pda%pds = a*pdgds(:,ipdt,:) + b*pdgds(:,j,:)
       pda%sds = a*pdsds(ipdt,:) + b*pdsds(j,:)
    END IF
    ENTRY nopd (pda, time)
  END SUBROUTINE readpd

  !> Write time in UTC and current state-variable values to file
  SUBROUTINE outsv (stv, times, env)
    IMPLICIT NONE
    CLASS(statevars), INTENT(INOUT) :: stv
    TYPE(timing), INTENT(IN) :: times
    TYPE(local), INTENT(IN) :: env
    INTEGER :: nferr
    stv%it = stv%it + 1
    nferr = NF90_PUT_VAR(ncid=stv%id, varid=stv%time%varid,&
                         values=(/times%now - stv%timediff/), start=stv%it, count=(/1/))
    CALL stv%write_output (env, times)
  END SUBROUTINE outsv

  SUBROUTINE write_output (stv, env, times)
    IMPLICIT NONE
    CLASS(statevars), INTENT(INOUT) :: stv
    TYPE(local), INTENT(IN) :: env
    TYPE(timing), INTENT(IN) :: times
    REAL(dp) :: dt
    INTEGER :: n, nferr
    dt = times%now - times%previous
    stv%all_states(stv%i0flux:stv%nsv_total) = stv%all_states(stv%i0flux:stv%nsv_total)/dt
    nferr = NF90_PUT_VAR(ncid=stv%id, varid=stv%flux%varid, values=stv%flux%values,&
                         start=(/1, stv%start/), count=(/stv%nsv_one_box, stv%nfl, stv%count(2:)/))
    nferr = NF90_PUT_VAR(ncid=stv%id, varid=stv%soma%varid, values=stv%soma%values,&
                         start=(/1, stv%start/), count=(/stv%nsms, stv%count/))
    stv%all_states(stv%nsv_all_boxes+1:stv%nsv_total) = 0._dp
    ENTRY write_states (stv, env, times)
    DO n=1,env%nq0
       offsets(env%irt(env%iq0r(n)),:) = stv%var(env%irc(env%iq0r(n)))%values(1:stv%nbox)*env%Q0(n) ! subsistence N, P, ...
    END DO
    DO n=1,stv%nsv_one_box
       nferr = NF90_PUT_VAR(ncid=stv%id, varid=stv%var(n)%varid,&
                            values=stv%var(n)%values(1:stv%nbox) + offsets(n,:),&
                            start=stv%start, count=stv%count)
    END DO
  END SUBROUTINE write_output

  !> Close output file
  SUBROUTINE exof (stv)
    IMPLICIT NONE
    CLASS(statevars), INTENT(INOUT) :: stv !< state-variable structure of the output file
    INTEGER :: nferr
    DEALLOCATE (stv%var, stv%start, stv%count)
    nferr = NF90_CLOSE(ncid=stv%id)
  END SUBROUTINE exof

  !> Load dimension variable
  !! \param [in]  dim         TYPE(dimension) variable to read
  !! \param [in]  ncid        id of the file or group of the dataset
  !! \param [in]  caller      name of the calling subroutine or function
  !! \param [in]  dim%name    name of the dimension
  !! \param [out] dim%values  allocatable array to hold the data
  !! \param [out] dim%varid   variable id of the dimension variable
  !! \param [out] nferr   NetCDF error code
  SUBROUTINE read_dim (dim, ncid, caller, nferr)
    IMPLICIT NONE
    CLASS(dimension), INTENT(INOUT) :: dim
    INTEGER, INTENT(IN) :: ncid
    CHARACTER(LEN=*), INTENT(IN) :: caller
    INTEGER, INTENT(OUT), OPTIONAL :: nferr
    INTEGER :: nferror, len, dimids(2)
    nferror = NF90_INQ_VARID(ncid, name=TRIM(dim%name), varid=dim%varid)
    IF (PRESENT(nferr)) THEN
       nferr = nferror
       IF (nferr.EQ.NF90_ENOTVAR) RETURN ! variable not found
    END IF
    IF (nferror.NE.NF90_NOERR) CALL onf_error ("read_dim from "//TRIM(caller), "NF90_INQ_VARID", TRIM(dim%name), nferror)
    nferror = NF90_INQUIRE_VARIABLE(ncid, varid=dim%varid, dimids=dimids(1:1))
    IF (nferror.NE.NF90_NOERR) CALL onf_error ("read_dim from "//TRIM(caller), "NF90_INQUIRE_VARIABLE", TRIM(dim%name), nferror)
    dim%id = dimids(1)
    nferror = NF90_INQUIRE_DIMENSION(ncid=ncid, dimid=dim%id, len=len)
    IF (nferror.NE.NF90_NOERR) CALL onf_error ("read_dim from "//TRIM(caller), "NF90_INQUIRE_DIMENSION", TRIM(dim%name), nferror)
    ALLOCATE (dim%values(len))
    nferror = NF90_GET_VAR(ncid, varid=dim%varid, values=dim%values)
    IF (nferror.NE.NF90_NOERR) CALL onf_error ("read_dim from "//TRIM(caller), "NF90_GET_VAR", TRIM(dim%name), nferror)
    nferror = NF90_GET_ATT(ncid, varid=dim%varid, name='units', values=dim%units)
    IF (nferror.EQ.NF90_ENOTATT) THEN
       WRITE (stdout,'("read_dim from ",A,": no units attribute for ",A)') TRIM(caller), TRIM(dim%name)
       dim%units = ''
    ELSEIF (nferror.NE.NF90_NOERR) THEN
       CALL onf_error ("read_dim from "//TRIM(caller), "NF90_GET_ATT", TRIM(dim%name)//'.'//TRIM(dim%units), nferror)
    ENDIF
    nferror = NF90_GET_ATT(ncid, varid=dim%varid, name='bounds', values=dim%bounds%name)
    IF (nferror.EQ.NF90_ENOTATT) THEN
       dim%bounds%name = ''
    ELSE
       nferror = NF90_INQ_VARID(ncid, name=TRIM(dim%bounds%name), varid=dim%bounds%varid)
       IF (nferror.NE.NF90_NOERR) &
            CALL onf_error ("read_dim from "//TRIM(caller), "NF90_INQ_VARID", TRIM(dim%bounds%name), nferror)
       ALLOCATE (dim%bounds%values(2,len))
       nferror = NF90_GET_VAR(ncid, varid=dim%bounds%varid, values=dim%bounds%values)
       IF (nferror.NE.NF90_NOERR) &
            CALL onf_error ("read_dim from "//TRIM(caller), "NF90_GET_VAR", TRIM(dim%bounds%name), nferror)
    END IF
  END SUBROUTINE read_dim

  SUBROUTINE onf_error (caller, nffun, name, nferr)
    IMPLICIT NONE
    CHARACTER(LEN=*), INTENT(IN) :: caller, nffun, name
    INTEGER, INTENT(IN) :: nferr
    WRITE (stderr,'(A,": ",A,"(",A,"): ",A,", nferr = ",I4)') caller, nffun, TRIM(name),&
         TRIM(NF90_STRERROR(nferr)), nferr
    IF (nferr.EQ.-101) WRITE (stderr,'("File ",A," is blocked by another program.")') TRIM(name)
    STOP
  END SUBROUTINE onf_error
END MODULE onf
