!> \brief Set up and drive marine ecosystem dynamics
MODULE plankton
  USE et
  USE onf
  USE bac
  USE cmo
  USE det
  USE cfo
  USE dic
  USE julian
  USE brock81
  USE dvode
  IMPLICIT NONE
  PRIVATE
  PUBLIC :: plankton_init, plankton_ode, stdout, dp, NF90_MAX_NAME, Jul_DHMSofSec
  !> \brief Type for boundary conditions
  TYPE :: bc
     CHARACTER(LEN=100) :: name !< variable name
     TYPE(quantity) :: value    !< value of boundary condition
  END TYPE bc
  TYPE(ode), TARGET, SAVE, PUBLIC :: parode
  TYPE(timing), TARGET, SAVE, PUBLIC :: times
  TYPE(local), TARGET, SAVE, PUBLIC :: envir
  TYPE(layer), ALLOCATABLE :: box(:)
  TYPE(state), ALLOCATABLE :: fungrp(:)            !< vector of functional groups
  TYPE(dom), ALLOCATABLE, TARGET :: doma(:)        !< vector of DOM types
  TYPE(bacteria), ALLOCATABLE, TARGET :: bacpla(:) !< vector of bacteria groups
  TYPE(detritus), ALLOCATABLE, TARGET :: detrit(:) !< vector of detritus groups
  TYPE(phycmo), ALLOCATABLE, TARGET :: phypla(:)   !< vector of phytoplankton groups
  TYPE(zoocfo), ALLOCATABLE, TARGET :: zoopla(:)   !< vector of zooplankton groups
  TYPE(dicsys), SAVE :: dics
  TYPE(phydat), SAVE :: phys
  TYPE(statevars), TARGET, SAVE, PUBLIC :: states
CONTAINS

  !> Read all parameters and boundary-condition information and initialise all modules
  !! \param[in] control main input (namelist) file
  !! \param[in] dfn0    NetCDF file with physical forcing (boundary conditions) data
  !! \param[in] ifn0    NetCDF file with initial conditions
  !! \param[in] ofn0    name of output file
  !! \param[in] pfn0    file with parameter namelists
  SUBROUTINE plankton_init (control, dfn0, ifn0, ofn0, pfn0, sms0, force, resume)
    IMPLICIT NONE
    CHARACTER(LEN=*), INTENT(IN) :: control, dfn0, ifn0, ofn0, pfn0
    LOGICAL, INTENT(IN) :: force,& !< flag whether to overwrite output file
         sms0,&                    !< flag whether to write fluxes to output file
         resume                    !< flag whether this is a continuation simulation
    TYPE(quantity), DIMENSION(15) :: time, DIN, DIP
    TYPE(bc), ALLOCATABLE :: bottom(:),& ! bottom boundary conditions for open configurations
         fsfalk(:)                       ! alkalinity surface-flux factors
    CHARACTER(LEN=NF90_MAX_NAME) :: dfn='', ifn='', ofn='', pfn='',&
         utcdatstr, locdatstr, fmt, flux_name='flux', soma_name='soma', state_name='States'
    CHARACTER(LEN=512) :: iomsg
    LOGICAL :: soma=.FALSE.
    LOGICAL, ALLOCATABLE :: ichl(:), iPIC(:), mask(:)
    REAL(dp) :: factor(15)=1._dp,& !< dilution factors
         flux_depth(100)=-1._dp    !< approximate depths for which to write vertical fluxes to output
    REAL(dp), ALLOCATABLE :: faces(:), ydot(:) ! dummy
    INTEGER, ALLOCATABLE :: flux_level(:), position(:)
    INTEGER :: k, m, n, nagg, nPIC, nfalk,&
         nsfalk,&               ! number of surface fluxes affecting alkalinity
         nbfish=0,&             ! number of boxes (layers) with fish
         np(1), lunctl, lunp, pos, dutc, seconds,&
         nq0, nrs, nrft
    INTEGER(KIND=INT64) :: tsec
    NAMELIST /files/ dfn, ifn, ofn, pfn
    NAMELIST /dilute/ time, DIN, DIP, factor
    NAMELIST /fluxes/ soma, flux_depth, flux_name, soma_name, state_name
    CALL inudu                  ! initialise UDUNITS2 library (module fudu)
    OPEN (NEWUNIT=lunctl, FILE=expand_path(control), ACTION='READ', STATUS='OLD', IOMSG=iomsg, ERR=900)
    pfn = control               ! read everything from main input file by default
    READ (UNIT=lunctl, NML=files, END=10, IOMSG=iomsg, ERR=900)
10  CONTINUE
    IF (LEN_TRIM(pfn0).GT.0) pfn = pfn0 ! command line overrides namelist files
    IF (LEN_TRIM(dfn0).GT.0) dfn = dfn0
    IF (LEN_TRIM(ifn0).GT.0) ifn = ifn0
    IF (LEN_TRIM(ofn0).GT.0) ofn = ofn0
    IF (LEN_TRIM(ifn).EQ.0) STOP 'Please specify initial-condition file (option -i or ifn in namelist files).'
    IF (LEN_TRIM(ofn).EQ.0) STOP 'Please specify output file (option -o or ofn in namelist files).'
    IF (resume) ifn = ofn
    CALL states%open (phys=phys, lun=lunctl) ! initialise NetCDF library (innf in module onf)
    ! read community composition from namelist coco
    IF (TRIM(pfn).EQ.TRIM(control)) THEN
       lunp = lunctl
    ELSE
       OPEN (NEWUNIT=lunp, FILE=expand_path(pfn), ACTION='READ', STATUS='OLD', IOMSG=iomsg, ERR=900)
    END IF
    REWIND (UNIT=lunp)
    envir%ncon = SIZE(constituents)   ! number of constituents (C, N, P, Chl)
    CALL set_community (lun=lunctl, lunpar=lunp)
    CALL set_environment (lun=lunctl)
    CALL set_timing (lun=lunctl)
    states%filename = expand_path(ifn)     ! file with initial conditions
    states%timediff = times%dtutc
    ! states%info => nf90info in module onf (TYPE(nf90group) -> TYPE(statevars))
    CALL states%info (latdeg=times%latdeg, londeg=times%londeg, t0=times%tin, writeable=resume)
    IF (resume) times%start = states%time%values(1)
    IF (LEN_TRIM(dfn).EQ.0) THEN  ! no physical forcing data
       ALLOCATE (phys%depth%values(states%nbox))
       phys%depth = states%depth ! apply depth structure of
       phys%nbox = states%nbox   ! initial conditions
    ELSE
       phys%filename = expand_path(dfn)  ! name of forcing file
       phys%timediff = times%dtutc
       ! phys%info => nf90info (TYPE(nf90group) -> TYPE(phydat) in module onf) 
       CALL phys%info (latdeg=times%latdeg, londeg=times%londeg, t0=times%start, t1=times%end)
       states%depth%bounds%values = phys%depth%bounds%values
    ENDIF
    envir%nbox = phys%nbox
    envir%depth_edges => phys%depth%bounds%values
    REWIND (UNIT=lunctl)
    READ (UNIT=lunctl, NML=fluxes, END=20, IOMSG=iomsg, ERR=900)
20  CONTINUE
    states%write_soma = soma.OR.sms0
    IF (states%write_soma) THEN
       WRITE (*,'("plankton_init: writing fluxes to output file.")')
       ALLOCATE (faces(phys%nbox+1), envir%fmsk(1:states%nsv_one_box,1:phys%nbox+1))
       states%state%name = TRIM(state_name)
       states%flux%name = TRIM(flux_name)
       states%soma%name = TRIM(soma_name)
       states%nfl = COUNT(flux_depth.GE.0._dp)
       faces = (/envir%depth_edges(1,:), envir%depth_edges(2,phys%nbox)/)
       IF (states%nfl.EQ.0) THEN
          states%nfl = phys%nbox + 1
          flux_depth(1:states%nfl) = faces
          envir%fmsk = .TRUE.         ! flux-matrix mask
       ELSE
          envir%fmsk = .FALSE.        ! flux-matrix mask
       END IF
       ALLOCATE (flux_level(states%nfl), states%flux_depth(states%nfl))
       DO n=1,states%nfl
          flux_level(n) = MINLOC(ABS(faces - flux_depth(n)), DIM=1, BACK=.TRUE.)
          envir%fmsk(:,flux_level(n)) = .TRUE.
       END DO
       states%flux_depth = faces(flux_level)
       DEALLOCATE (faces, flux_level)
    END IF
    ! rows 0 of soma are phantom rows for aggregation fluxes
    ! traits are stored and transported as tracer-ratio products
    ALLOCATE (envir%quotas(envir%ncon,envir%nft), envir%qmsk(envir%ncon,envir%nft),&
         envir%Q0(envir%nq0), envir%iq0r(envir%nq0),&
         envir%ratios(envir%nrs), envir%all_ratios(envir%nrs,phys%nbox), &
         envir%OC(envir%nft), envir%soma(0:states%nsv_one_box,states%nsv_one_box,phys%nbox),&
         envir%smsk(0:states%nsv_one_box,states%nsv_one_box,phys%nbox),&
         envir%irc(envir%nrs), envir%irt(envir%nrs), envir%vvs(states%nsv_one_box),&
         envir%vvstv(states%nsv_one_box), envir%dydz(states%nsv_one_box), envir%iagg(envir%nagg),&
         envir%sticky(envir%nagg), envir%kij(envir%nagg,envir%nagg), envir%agg(envir%nagg),&
         envir%heights(states%nsv_one_box, phys%nbox), envir%depth(0:phys%nbox+1),&
         box(0:phys%nbox+1), ichl(states%nsv_one_box), iPIC(states%nsv_one_box),&
         envir%PAR(0:phys%nbox), envir%fPAR(0:phys%nbox))
    envir%depth = (/0._dp, phys%depth%values, 0._dp/)
    box%depth = envir%depth
    box(0)%height = 0._dp
    box%nb = (/(n, n=0,phys%nbox+1)/) ! layer index
    envir%PARfrac => fPAR_vertical
    DO n=1,phys%nbox             ! convert mean depths to box heights
       box(n)%dz = box(n)%depth - box(n-1)%depth
       box(n)%height = (box(n)%dz - box(n-1)%height)
    END DO
    box(1:phys%nbox)%height = 2D0*box(1:phys%nbox)%height
    envir%heights = SPREAD(box(1:phys%nbox)%height, DIM=1, NCOPIES=states%nsv_one_box) ! layer heights
    ! rdl: ratio of lower box height to total height of two boxes
    box(0:phys%nbox-1)%rdl = 1D0/(box(0:phys%nbox-1)%height/box(1:phys%nbox)%height + 1D0)
    box(phys%nbox)%rdl = 0.5D0
    box(phys%nbox+1)%nb = phys%nbox + 1
    box(1:nbfish)%fish = 1
    envir%vvs = 0D0             ! vertical velocities (sinking, DVM, SVM)
    envir%quotas = 0D0          ! default
    envir%quotas(1,:) = 1D0     ! the C quota (C:C) is always 1
    envir%qmsk = .FALSE.        ! quota mask
    envir%smsk = .FALSE.        ! source-matrix mask
    envir%soma = 0D0            ! source matrices (fluxes between states)
    nq0 = 0
    nrs = 0
    nagg = 0
    DO n = 1,envir%nft          ! number of functional types
       fungrp(n)%var%vv => envir%vvs(fungrp(n)%var%isv(1):fungrp(n)%var%isv(fungrp(n)%var%nsv))
       fungrp(n)%var%vv = 0D0   ! vertical velocities for each state in each type
       fungrp(n)%var%OC => envir%OC(n)       ! organic carbon (POC)
       fungrp(n)%var%QN => envir%quotas(2,n) ! N:C quota
       fungrp(n)%var%QP => envir%quotas(3,n) ! P:C quota
       NULLIFY (fungrp(n)%var%stick)
       IF (fungrp(n)%var%sticky.GT.0D0) THEN ! does group n undergo aggregation?
          nagg = nagg + 1
          envir%iagg(nagg) = n                      ! indices of aggregating types
          fungrp(n)%var%stick => envir%sticky(nagg) ! stickiness
          fungrp(n)%var%stick = fungrp(n)%var%sticky
       END IF
       IF (fungrp(n)%var%nsv.GT.1) THEN ! dynamic quotas or tracer-trait products
          nrft = fungrp(n)%var%nsv - 1
          envir%irc(nrs+1:nrs+nrft) = fungrp(n)%var%isv(1)  ! ratio reference (POC) tracer-indices
          envir%irt(nrs+1:nrs+nrft) = fungrp(n)%var%isv(2:) ! ratio tracer-indices
          fungrp(n)%var%ratios => envir%ratios(nrs+1:nrs+nrft)
          fungrp(n)%var%ir = (/(m + nrs, m=1,nrft)/)
          IF (fungrp(n)%var%nq0.GT.0) THEN                  ! subsistence quotas
             envir%iq0r(nq0+1:nq0+fungrp(n)%var%nq0) = nrs + fungrp(n)%var%iq0 - 1
             envir%Q0(nq0+1:nq0+fungrp(n)%var%nq0) = fungrp(n)%var%Q0  ! assign subsistence quota
             nq0 = nq0 + fungrp(n)%var%nq0
          END IF
          DO m=2,fungrp(n)%var%ncon      ! associate dynamic quotas with constituents
             np = PACK([(k, k=2,envir%ncon)],&
                  MASK=TRIM(fungrp(n)%var%constituents(m)).EQ.constituents(2:), VECTOR=(/0/))
             IF (np(1).EQ.0) THEN
                WRITE (stderr,'("Constituent ",A," not implemented!")') &
                     TRIM(fungrp(n)%var%constituents(m))
                STOP
             END IF
             envir%qmsk(np(1),n) = .TRUE. ! quota mask (np(1) = constituent index, n = group index)
          END DO
          nrs = nrs + nrft
       END IF
       ! fungrp(n)%var%set points to the set procedure pointer in each functional group
       ! for example, if fungrp(n)%var is of TYPE(phycmo) (defined in module cmo), then
       ! fungrp(n)%var%set => phypla%set => cmo_set, defined in module cmo
       CALL fungrp(n)%var%set (fungrp, envir) ! initialise group n
       ! names and units of all states in functional group n
       states%var(fungrp(n)%var%isv)%name = fungrp(n)%var%names
       states%var(fungrp(n)%var%isv)%units = fungrp(n)%var%units
    END DO
    ! set bottom boundary conditions from bottom in namelist envi
    ALLOCATE(mask(states%nsv_one_box))
    DO n=1,COUNT(bottom%name.NE.'')
       mask = states%var%name.EQ.bottom(n)%name
       IF (ANY(mask)) THEN
          np = PACK(position, MASK=mask)
          envir%bbc(np(1)) = convert(bottom(n)%value, states%var(np(1))%units)
       ELSE
          WRITE (stderr,'("Ignoring bottom value for non-existent variable ",A,".")') &
               TRIM(bottom(n)%name)
       END IF
    END DO
    IF ((detrit(envir%idagg)%kag.GT.0D0).AND.(envir%nagg.GT.0)) THEN ! does aggregation occur?
       envir%aggregate => aggregate
       envir%kij = detrit(envir%idagg)%kag
       DO n = 1,envir%nagg
          envir%kij(n,n) = 2._dp*envir%kij(n,n)
          envir%smsk(0,fungrp(envir%iagg(n))%var%isv,:) = .TRUE.
       END DO
       envir%smsk(0,detrit(1)%isv,:) = .TRUE.
    ELSE
       envir%aggregate => no_aggregate
    END IF
    states%smsk => envir%smsk(:,:,1)
    states%nsms = COUNT(states%smsk)
    states%nsv_all_boxes = states%nsv_one_box*phys%nbox
    IF (states%write_soma) THEN
       states%nsv_total = states%nsv_all_boxes + states%nsv_one_box*states%nfl + states%nsms*phys%nbox
       states%i0flux = states%nsv_all_boxes + 1
       states%i0sms = states%i0flux + states%nsv_one_box*states%nfl
    ELSE
       states%nsv_total = states%nsv_all_boxes
    END IF
    parode%neq = states%nsv_total
    CALL parode%read (lun=lunctl, nconstr=states%nsv_all_boxes) ! => read_ode in module dvode
    IF (lunp.NE.lunctl) CLOSE (UNIT=lunctl)
    IF (TRIM(states%time%units).NE.TRIM(phys%time%units)) THEN
       WRITE (stdout,'("Time units in ",A," are ",A," and ",A," in ",A,"!")') &
            TRIM(phys%filename), TRIM(phys%time%units), TRIM(states%time%units), TRIM(states%filename)
       IF (resume) STOP 'This should not happen for a continuation!'
    END IF
    IF (times%end.LE.times%start) STOP 'plankton_init: end must be greater than start in namelist timing.'
    states%time%units = phys%time%units
    IF (states%ilat.GT.0) THEN
       states%lat%values(states%ilat) = times%latdeg ! set lat/lon for output file
       states%lon%values(states%ilon) = times%londeg
    END IF
    ! states%init => invar (module onf): create or open output file; read initial conditions
    ! states%nsms and states%smsk must be completely defined at this point
    CALL states%init (ofn=expand_path(ofn), control=expand_path(control),&
         dfn=expand_path(dfn), pfn=expand_path(pfn), force=force, resume=resume)
    CALL parode%set_atol(stv=states%data) ! => set_atol (module dvode)
    NULLIFY (times%latrad)
    CALL phys%init (env=envir, times=times, box=box, svd=states%var) ! => inpd (module onf)
    IF (.NOT.ASSOCIATED(phys%surPAR)) CALL brock81_init (envir=envir, times=times)
    IF (phys%clsbot) box(phys%nbox)%botperm = 0D0
    ALLOCATE (envir%fluxes(1:states%nsv_one_box,0:phys%nbox))
    envir%fluxes = 0D0          ! fluxes between boxes (layers)
    REWIND (UNIT=lunp)
    ! dics%read => indic (module dic): initialise DIC system
    CALL dics%read (lun=lunp, env=envir, fluxes=envir%fluxes)
    ! check for presence of datasets for alkalinity, temperature, etc.
    ! otherwise use constants supplied via namelist envi
    IF (.NOT.ASSOCIATED(box(1)%alkalinity)) THEN
       IF (envir%ialk.GT.0) THEN
          DO n=1,phys%nbox; box(n)%alkalinity => states%data(envir%ialk,n)
          END DO
       ELSEIF (dics%totalk.GT.0D0) THEN
          DO n=1,phys%nbox; ALLOCATE (box(n)%alkalinity); box(n)%alkalinity = dics%totalk
          END DO
       ELSEIF (dics%potalk.LE.0D0) THEN
          STOP 'plankton_init: no alkalinity data found.'
       ELSE
          DO n=1,phys%nbox; ALLOCATE (box(n)%alkalinity)
          END DO
       END IF
    END IF
    IF (.NOT.ASSOCIATED(box(1)%temperature)) THEN
       DO n=1,phys%nbox; box(n)%temperature => envir%temperature
       END DO
    END IF
    IF (.NOT.ASSOCIATED(box(1)%salinity)) THEN
       DO n=1,phys%nbox; box(n)%salinity => envir%salinity
       END DO
    END IF
    IF (.NOT.ASSOCIATED(box(1)%vdc)) THEN
       ALLOCATE (box(0)%vdc); box(0)%vdc = 0D0
       DO n=1,phys%nbox; box(n)%vdc => envir%vdc
       END DO
    END IF
    IF (.NOT.ASSOCIATED(box(1)%velocity)) THEN
       ALLOCATE (box(0)%velocity); box(0)%velocity = 0D0
       DO n=1,phys%nbox; box(n)%velocity => envir%velocity
       END DO
    END IF
    envir%fPAR(0) = 1D0 ! fraction of surface irradiance (PAR) arriving at each layer
    SELECT CASE (TRIM(phys%advect)) ! select advection scheme
    CASE ('CD', 'cd')   ! central differences (modified to be positive-definite)
       envir%boxbc => boxbc_cd
    CASE ('upwind')
       envir%boxbc => boxbc_upwind
    CASE ('dummy')              ! no advection (only mixing)
       envir%boxbc => boxbc_dummy
    CASE DEFAULT
       STOP 'plankton_init:invalid advection scheme (advect in namelist physics).'
    END SELECT
    IF (LEN_TRIM(phys%time%units).EQ.0) THEN  ! no forcing
       dutc = 0
       seconds = 0
    ELSE
       pos = SCAN(TRIM(phys%time%units), '123456789')
       ! time%units should be something like "seconds since 1990-01-01"
       CALL Jul_ParseDT(TRIM(phys%time%units(pos:)), '%Y-%m-%d', dutc, seconds)
       ! dutc is the UTC day of the date in time%units
    END IF
    times%tbase = dutc*cfds + seconds ! date (in s) of the date in phys%time%units
    ! envir%PARfun is one of the PAR functions selected according to envir%dicy above
    CALL envir%PARfun (times)         ! initialise light
    ! time and date in UTC
    tsec = NINT(times%start + times%tbase - times%dtutc, KIND=INT64)
    dutc = FLOOR(REAL(tsec, KIND=dp)/cfds)
    seconds = INT(MODULO(tsec, cfds), KIND=4)
    utcdatstr = Jul_FormatDate(dutc, seconds, fmt='%d %m %Y at %H:%M:%S')
    ! local time and date
    tsec = NINT(times%start + times%tbase, KIND=INT64)
    dutc = FLOOR(REAL(tsec, KIND=dp)/cfds)
    seconds = INT(MODULO(tsec, cfds), KIND=4)
    locdatstr = Jul_FormatDate(dutc, seconds, fmt='%d %m %Y at %H:%M:%S')
    WRITE (stdout,'("plankton_init: Starting simulation on ",A," (",A," UTC).")')&
         TRIM(locdatstr), TRIM(utcdatstr)
    n = MIN(states%nbox, phys%nbox)
    IF ((states%nbox.NE.phys%nbox).OR.(ANY(ABS(states%depth%values(1:n) - envir%depth(1:n)).GT.1.E-12_dp))) &
         STOP 'plankton_init: Depth structures in (re)start and data files must agree.'
    box(0)%flux => envir%fluxes(:,0) ! local fluxes (between boxes)
    DO n=1,phys%nbox
       box(n)%stv => states%data(:,n)        ! all states in layer
       box(n)%tPAR => envir%PAR(n-1) ! PAR at top of layer
       box(n)%bPAR => envir%PAR(n)   ! PAR at bottom of layer
       box(n)%DIN => states%data(envir%idin,n)
       box(n)%DIP => states%data(envir%idip,n)
       box(n)%DIC => states%data(envir%idic,n)
       box(n)%O2 => states%data(envir%io2,n)
       IF (envir%irdoc.GT.0) box(n)%rDOC => states%data(envir%irdoc,n) ! refractory DOC
       IF (envir%irdon.GT.0) box(n)%rDON => states%data(envir%irdon,n)
       box(n)%flux => envir%fluxes(:,n) ! local fluxes (between boxes)
       box(n)%soma(0:states%nsv_one_box,1:states%nsv_one_box) => envir%soma(:,:,n) ! local source matrix (in 1 box)
       box(n)%ratios => envir%all_ratios(:,n)
       ALLOCATE (box(n)%bbc(states%nsv_one_box))
       box(n)%bbc = 0._dp
    END DO
    IF (envir%isi.NE.0) THEN
       DO n=1,phys%nbox; box(n)%silc => states%data(envir%isi,n)
       END DO
    ELSEIF (.NOT.ASSOCIATED(box(1)%silc)) THEN
       DO n=1,phys%nbox; ALLOCATE (box(n)%silc); box(n)%silc => box(n)%DIN
       END DO
    END IF
    ALLOCATE (box(0)%bbc(states%nsv_one_box), box(phys%nbox+1)%bbc(states%nsv_one_box),&
         box(0)%stv(states%nsv_one_box), box(phys%nbox+1)%stv(states%nsv_one_box),&
         box(0)%ratios(nrs), box(phys%nbox+1)%ratios(nrs))
    box(0)%stv = 0._dp               ! ensure that surface concentrations are defined
    box(0)%bbc = 0._dp
    box(phys%nbox)%bbc = envir%bbc   ! set bottom concentrations
    box(phys%nbox+1)%stv = envir%bbc
    box(0)%ratios = 0._dp
    box(phys%nbox+1)%ratios = 0._dp
    ichl = INDEX(states%var%name, '_Chl').GT.0
    IF (any(ichl)) THEN
       ALLOCATE (envir%ichl(COUNT(ichl)))
       envir%ichl = PACK([(k, k=1,states%nsv_one_box)], MASK=ichl) ! indices of all Chl states
       envir%attenuation => lch_chl
       IF (phys%nbox.EQ.1) THEN
          DO n=1,envir%nphy     ! assign initial Chl if specified
             IF (phypla(n)%Chl0.GT.0._dp) states%data(envir%ichl(n),1) = phypla(n)%Chl0
          END DO
       END IF
    ELSE
       DO n=1,envir%nphy
          envir%iphy(n) = fungrp(envir%iphy(n))%var%isv(2)
       END DO
       envir%attenuation => lch_pon
    END IF
    IF (envir%ialk.GT.0) THEN   ! alkalinity is a state variable
       iPIC = INDEX(states%var%name, '_PIC').GT.0
       nPIC = COUNT(iPIC)
       nfalk = 3 + nPIC
       ALLOCATE (envir%ipic(nPIC), envir%ifalk(nfalk), envir%fALK(nfalk))
       envir%ifalk(1:3) = (/envir%idic, envir%idin, envir%idip/)
       ! account for S uptake assuming an S:C ratio of 0.0226 (Wolf-Gladrow et al. 2007,
       ! citing Matrai and Keller 1994), which may be less variable than the S:P ratio
       envir%fALK(1:3) = (/-0.045_dp, -1._dp, -1._dp/)
       IF (any(iPIC)) THEN
          envir%ipic = PACK([(k, k=1,states%nsv_one_box)], MASK=iPIC) ! indices of all PIC states
          envir%ifALK(4:3+nPIC) = envir%ipic  ! the 0.045 compensates for the spurious S uptake
          envir%fALK(4:3+nPIC) = -2.045_dp    ! incurred by fALK(1) for the associated DIC change
       END IF
       ! surface fluxes
       nsfalk = COUNT(fsfalk%name.NE.'')
       IF ((.NOT.ANY(fsfalk%name.EQ.'alkalinity')).AND.&
            ANY(states%var(phys%isflx)%name.EQ.'alkalinity')) THEN
          nsfalk = nsfalk + 1
          fsfalk(nsfalk) = bc('alkalinity', quantity(1._dp, '1'))
       END IF
       n = 1
       DO WHILE (n.LE.nsfalk)
          IF (.NOT.ANY(fsfalk(n)%name.EQ.states%var(phys%isflx)%name)) THEN
             WRITE (stderr,'("Ignoring alkalinity surface-flux factor for non-existent flux ",A,".")') &
                  TRIM(fsfalk(n)%name)
             fsfalk(n) = fsfalk(nsfalk)
             nsfalk = nsfalk - 1
          ELSE
             n = n + 1
          END IF
       END DO
       envir%xsfalk = ((fsfalk(1)%name.NE.'alkalinity').AND.(nsfalk.EQ.1)) &
            .OR.(nsfalk.GT.1)
       IF (envir%xsfalk) THEN
          ALLOCATE (envir%isfalk(nsfalk), envir%fsfALK(nsfalk))
          DO n=1,nsfalk
             envir%isfALK(n:n) = PACK(phys%isflx, MASK=states%var(phys%isflx)%name.EQ.fsfalk(n)%name)
             envir%fsfALK(n) = convert(fsfalk(n)%value, '1')
          END DO
       END IF
       envir%alkalinity => alkalinity
    ELSE
       envir%alkalinity => no_alkalinity
    END IF
    time = quantity(0D0, 'd')
    DIN = quantity(0D0, 'mmolN m-3')
    DIP = quantity(0D0, 'mmolP m-3')
    REWIND (UNIT=lunp)
    ! this is for the PU1, 2 experiments, where the mesocosms were diluted
    READ (UNIT=lunp, NML=dilute, END=30, IOMSG=iomsg, ERR=900)
    times%ndil = COUNT(time%value.GT.0D0) ! number of dilution events
    times%dilute = times%ndil.GT.0        ! is dilution done?
    ALLOCATE (times%dil_time(times%ndil), times%dil_fact(times%ndil),&
              times%dil_DIN(times%ndil), times%dil_DIP(times%ndil))
    DO n=1,times%ndil
       times%dil_time(n) = convert(time(n), 's') + times%start ! times of dilutions
       times%dil_fact(n) = factor(n)                           ! dilution factors
       times%dil_DIN(n) = convert(DIN(n), 'mmolN m-3') ! concentration in added water
       times%dil_DIP(n) = convert(DIP(n), 'mmolP m-3')
    END DO
30  CONTINUE
    CLOSE (UNIT=lunp)
    DO n=1,phys%nbox            ! check consistency of initial conditions
       IF (ANY(states%data(envir%irt(envir%iq0r),n).LT.states%data(envir%irc(envir%iq0r),n)*envir%Q0)) THEN
          WRITE (fmt,'("(""Initial quota(s) too low in layer "",I0,"" for "",",I0,"(A:"", ""))")') &
               SIZE(envir%iq0r)
          WRITE (stderr,TRIM(fmt)) n, PACK((/(TRIM(states%var(envir%irt(envir%iq0r(m)))%name), m=1,&
               SIZE(envir%iq0r))/), MASK=states%data(envir%irt(envir%iq0r),n).LT.states%data(envir%irc(envir%iq0r),n)*envir%Q0)
          STOP
       END IF
       ! delta-tracers for PON, POP, ... with subsistence quotas to have a lower
       ! limit of 0 -> subtract subsistence N, P, ...
       states%data(envir%irt(envir%iq0r),n) = states%data(envir%irt(envir%iq0r),n) &
            - states%data(envir%irc(envir%iq0r),n)*envir%Q0
    END DO
    ! write initial conditions (and fluxes) to output file
    times%now = times%start
    CALL times%sunday (env=envir)
    CALL phys%get (time=times%start)
    CALL dics%init (box=box)    ! set initial [H+]
    IF (states%write_soma) states%all_states(states%i0flux:states%nsv_total) = 0._dp
    ALLOCATE (ydot(states%nsv_total))
    CALL plankton_ode (neq=states%nsv_total, time=times%start, y=states%all_states, ydot=ydot)
    ! write initial conditions to output file unless it is a continuation
    times%now = times%start
    times%previous = times%now - times%dto  ! dummy time step for initial fluxes and sms (these are all 0 here)
    IF (.NOT.resume) CALL states%write (times=times, env=envir) ! => outsv (module onf)
    DEALLOCATE (ydot)
    RETURN
900 CONTINUE
    WRITE (stderr,'("plankton_init: ",A)') TRIM(iomsg)
    STOP 1
  CONTAINS
    !> \brief length of filename without extension
    FUNCTION lfn (fn) RESULT (ix)
      IMPLICIT NONE
      CHARACTER(LEN=*), INTENT(IN) :: fn
      INTEGER :: ix
      ix = SCAN(fn, '.', BACK=.TRUE.) - 1 ! (position of last . in fn) - 1
      IF (ix.LE.0) ix = LEN_TRIM(fn)
    END FUNCTION lfn
    SUBROUTINE set_community (lun, lunpar)
      IMPLICIT NONE
      INTEGER, INTENT(IN) :: lun, lunpar
      CHARACTER(LEN=100), DIMENSION(100) :: dima='', dom(1)='', bact='', detr='', cmo='', cfo=''
      INTEGER :: mm, nn, n0
      LOGICAL :: coco_in_cf=.FALSE.
      NAMELIST /coco/ dima, dom, bact, detr, cmo, cfo  !< community composition
      REWIND (UNIT=lun)
      READ (UNIT=lun, NML=coco, IOMSG=iomsg, END=10, ERR=900)
      coco_in_cf = .TRUE.
10    CONTINUE
      IF (.NOT.coco_in_cf) THEN
         REWIND (UNIT=lunpar)
         READ (UNIT=lunpar, NML=coco, IOMSG=iomsg, ERR=900)
      END IF
      ! dima, detr, etc. are vectors of component names for each functional type
      envir%ndm = COUNT(dima.NE.'')  ! dissolved matter types, e.g., NO3, PO4, O2, ...
      envir%ndom = COUNT(dom.NE.'')  ! DOM types
      envir%ndet = COUNT(detr.NE.'') ! detritus types, e.g., small, large, ...
      envir%nbac = COUNT(bact.NE.'') ! bacteria types
      envir%nphy = COUNT(cmo.NE.'')  ! phytoplankton types
      envir%nzoo = COUNT(cfo.NE.'')  ! zooplankton types
      envir%nft = envir%ndom + envir%ndet + envir%nbac + envir%nphy + envir%nzoo
      ALLOCATE (envir%iphy(envir%nphy), envir%ipoc(envir%nft), fungrp(envir%nft), doma(envir%ndom),&
           detrit(envir%ndet), bacpla(envir%nbac), phypla(envir%nphy), zoopla(envir%nzoo))
      ! vector fungrp represents the community composition; it is a derived-type vector
      ! with just one component, var, which is a pointer to a specific functional group
      ! associate pointers fungrp(n)%var with individual functional groups
      DO nn = 1,envir%ndom         ! DOM groups
         fungrp(nn)%var => doma(nn)
         fungrp(nn)%var%name = dom(nn)
      END DO
      n0 = envir%ndom
      DO nn = 1,envir%ndet         ! detritus groups (small, large, ...)
         fungrp(n0+nn)%var => detrit(nn)
         fungrp(n0+nn)%var%name = detr(nn)
         IF (detrit(nn)%aggregates) envir%idagg = nn
      END DO
      IF (COUNT(detrit%aggregates).GT.1) &
           STOP 'There can be only one aggregates group.'
      IF ((envir%ndet.EQ.1).AND.(envir%idagg.EQ.0)) envir%idagg = 1
      n0 = n0 + envir%ndet
      DO nn=1,envir%nbac           ! bacteria groups
         fungrp(n0+nn)%var => bacpla(nn)
         fungrp(n0+nn)%var%name = bact(nn)
      END DO
      n0 = n0 + envir%nbac
      DO nn=1,envir%nphy           ! phytoplankton (cmo) groups
         envir%iphy(nn) = n0 + nn
         fungrp(n0+nn)%var => phypla(nn)
         fungrp(n0+nn)%var%name = cmo(nn)
      END DO
      n0 = n0 + envir%nphy
      DO nn=1,envir%nzoo           ! zooplankton (cfo) groups
         fungrp(n0+nn)%var => zoopla(nn)
         fungrp(n0+nn)%var%name = cfo(nn)
      END DO
      n0 = n0 + envir%nzoo
      states%nsv_one_box = envir%ndm            ! number of states in each box
      envir%nrs = 0                     ! number of ratios across all functional groups
      envir%nagg = 0
      ! read parameters from namelists for all functional types (groups)
      DO nn = 1,envir%nft                ! number of functional types
         fungrp(nn)%var%ifg = nn          ! index of functional group
         REWIND (UNIT=lunp)
         ! fungrp(nn)%var%read points to the read procedure pointer in each functional group
         ! for example, if fungrp(nn)%var is of TYPE(phycmo) (defined in module cmo), then
         ! fungrp(nn)%var%read => phypla%read => cmo_read, defined in module cmo
         CALL fungrp(nn)%var%read(env=envir, lun=lunp)  ! read group parameters
         fungrp(nn)%var%ncon = SIZE(fungrp(nn)%var%constituents)
         ALLOCATE (fungrp(nn)%var%isv(fungrp(nn)%var%nsv), fungrp(nn)%var%names(fungrp(nn)%var%nsv),&
              fungrp(nn)%var%units(fungrp(nn)%var%nsv), fungrp(nn)%var%icon(fungrp(nn)%var%ncon),&
              fungrp(nn)%var%ir(fungrp(nn)%var%nsv - 1))
         DO mm=1,fungrp(nn)%var%ncon ! set names, units, indices for constituents (C, N, ...) in group
            fungrp(nn)%var%icon(mm:mm) = PACK([(k, k=1,envir%ncon)], &
                                           MASK=fungrp(nn)%var%constituents(mm).EQ.constituents)
            fungrp(nn)%var%names(mm) &
                 = TRIM(fungrp(nn)%var%name)//'_'//TRIM(fungrp(nn)%var%constituents(mm))
            fungrp(nn)%var%units(mm) = TRIM(constituent_units(fungrp(nn)%var%icon(mm)))
         END DO
         fungrp(nn)%var%isv = states%nsv_one_box + [(mm, mm=1,fungrp(nn)%var%nsv)] ! state-variable indices
         envir%ipoc(nn) = fungrp(nn)%var%isv(1)          ! index of biomass (C) of group
         envir%nq0 = envir%nq0 + fungrp(nn)%var%nq0      ! number of subsistence quotas across groups
         envir%nrs = envir%nrs + fungrp(nn)%var%nsv - 1  ! number of ratios (quotas & tracer-trait products)
         states%nsv_one_box = states%nsv_one_box + fungrp(nn)%var%nsv    ! number of state variables
         ! nagg is the number of types (groups) involved in aggregation
         IF (fungrp(nn)%var%sticky.GT.0D0) envir%nagg = envir%nagg + 1
      END DO
      ALLOCATE (states%var(states%nsv_one_box))
      states%var(1:envir%ndm)%name = dima(1:envir%ndm)
      DO nn = 1,envir%ndm                ! envir%ndm is number of dissolved-matter states
         SELECT CASE (TRIM(dima(nn)))
         CASE ('CO2', 'DIC')
            envir%idic = nn
            states%var(nn)%units = 'mmolC m-3'
         CASE ('alkalinity')
            envir%ialk = nn
            states%var(nn)%units = 'mmol m-3'
         CASE ('O2')
            envir%io2 = nn
            states%var(nn)%units = 'mmol m-3'
         CASE ('SiO3')
            envir%isi = nn
            states%var(nn)%units = 'mmol m-3'
         CASE ('DIN', 'NO3')
            envir%idin = nn
            states%var(nn)%units = 'mmolN m-3'
         CASE ('DIP', 'PO4')
            envir%idip = nn
            states%var(nn)%units = 'mmolP m-3'
         CASE ('rDOC')
            envir%irdoc = nn
            states%var(nn)%units = 'mmolC m-3'
         CASE ('rDON')
            envir%irdon = nn
            states%var(nn)%units = 'mmolN m-3'
         CASE DEFAULT
            WRITE (stderr, '("Variable name ",A," is not implemented.")') TRIM(dima(nn))
            STOP
         END SELECT
      END DO
      IF (envir%idic.EQ.0) STOP 'DIC missing in namelist coco' ! DIC and O2 are required
      IF (envir%io2.EQ.0) STOP 'O2 missing in namelist coco'
      IF ((envir%ndom.GT.0).AND.(envir%irdoc*envir%irdon.EQ.0)) &
           STOP 'DOM requires "rDOC" and "rDON" in dima'
      RETURN
900   CONTINUE
      WRITE (stderr,'("set_community: ",A)') TRIM(iomsg)
      STOP 1
    END SUBROUTINE set_community
    SUBROUTINE set_environment (lun)
      IMPLICIT NONE
      INTEGER, INTENT(IN) :: lun
      TYPE(quantity) :: PAR, wind, atmprs, CO2, temperature, salinity, ice,&
           potalk, totalk,& !< potential, total surface alkalinity
           vdc, velocity, lacw, lachl, lapon, kag,&
           RQ,&                   !< respiratory quotient
           ptl                    !< twilight parameter (degrees)
      CHARACTER(LEN=LEN(envir%dicy)) :: dicy
      CHARACTER(LEN=NF90_MAX_NAME) :: daylength='Brock81'
      REAL(dp) :: daylen=0.5_dp, latdeg=0._dp, londeg=0._dp
      NAMELIST /envi/ PAR, daylen, dicy, latdeg, londeg, wind, atmprs, CO2, bottom, fsfalk,&
           temperature, salinity, ice, potalk, totalk, vdc, velocity, lacw, lachl,&
           lapon, RQ, nbfish, daylength, ptl, kag
      ! set default values and units for namelist envi
      ! allocation and assignments for envir%daypar, envir%wind, etc. are overridden in
      ! subroutine inpd (module onf) for those driven by surface datasets
      ALLOCATE (envir%daypar, envir%wind, envir%pressure, envir%xCO2, envir%ice, envir%lacw,&
           envir%bbc(states%nsv_one_box), bottom(states%nsv_one_box), fsfalk(states%nsv_one_box),&
           position(states%nsv_one_box))
      PAR = quantity(0D0, 'W m-2')  ! surface irradiance
      wind = quantity(0D0, 'm s-1') ! wind speed at 10 m
      ice = quantity(0D0, '1')      ! ice cover fraction
      atmprs = quantity(1013.25D0,  'hPa') ! surface atmospheric pressure
      CO2 = quantity(390D0, 'ppm')         ! CO2 concentration in air
      temperature = quantity(10D0,  '°C')  ! surface air temperature
      vdc = quantity(0D0,  'm2 s-1')       ! vertical eddy-diffusion coefficient
      velocity = quantity(0D0,  'm s-1')   ! vertical velocity
      lacw = quantity(4D-2,  'm-1')        ! light-attenuation coefficient of seawater
      salinity = quantity(33,  'psu')
      potalk = quantity(0D0, 'mmol m-3') ! potential alkalinity
      totalk = quantity(0D0, 'mmol m-3') ! total alkalinity
      lachl = quantity(envir%lachl,  'm2 gChl-1') ! light-attenuation coefficient of chl
      lapon = quantity(envir%lapon,  'm2 molN-1') ! light-attenuation coefficient of PON
      RQ = quantity(envir%RQ, 'mol mol-1')        ! respiratory quotient
      dicy = envir%dicy           ! diurnal light cycle
      ptl = quantity(times%ptl, 'degrees')
      bottom = bc('', quantity(0._dp, ''))
      fsfalk = bc('', quantity(0._dp, ''))
      position = (/(n, n=1,states%nsv_one_box)/)
      envir%bbc = 0D0             ! bottom-boundary conditions
      REWIND (UNIT=lun)
      READ (UNIT=lun, NML=envi, IOMSG=iomsg, ERR=900)
      envir%dayPAR = convert(PAR, 'W m-2')
      envir%wind = convert(wind, 'm s-1')
      envir%ice = convert(ice, '1')
      envir%pressure = convert(atmprs, 'Pa')
      IF (detrit(envir%idagg)%kag.LT.0._dp) detrit(envir%idagg)%kag = convert(kag, 's-1')
      dics%potalk = convert(potalk, 'mmol m-3')/35D0
      dics%totalk = convert(totalk, 'mmol m-3')
      envir%xCO2 = convert(CO2, '')   ! convert ppmv to mole fraction
      envir%RQ = convert(RQ, 'mol mol-1')
      envir%lacw = convert(lacw, 'm-1')
      envir%lachl = convert(lachl, 'm2 mgChl-1')
      envir%lapon = convert(lapon, 'm2 mmolN-1')
      envir%daylen => times%daylen
      envir%daylen = daylen
      envir%temperature = convert(temperature, '°C')
      envir%salinity = convert(salinity, 'psu')
      envir%vdc = convert(vdc, 'm2 s-1')
      envir%velocity = convert(velocity, 'm s-1')
      times%latdeg = latdeg       ! latitude in degrees N
      times%londeg = londeg       ! longitude in degrees E
      times%ptl = convert(ptl, 'rad')
      envir%dicy = dicy
      envir%daylenfun = daylength
      RETURN
900   CONTINUE
      WRITE (stderr,'("set_environment: ",A)') TRIM(iomsg)
      STOP 1
    END SUBROUTINE set_environment
    SUBROUTINE set_timing (lun)
      IMPLICIT NONE
      INTEGER, INTENT(IN) :: lun
      TYPE(quantity) :: dto,&     !< output time step
           start, end, sampling, timeout,&
           tin                    !< time for reading initial conditions
      NAMELIST /timing/ dto, start, end, sampling, timeout, tin
      dto = quantity(times%dto, 'h')     ! output time step
      start = quantity(times%start, 'd') ! start time relative to start of forcing file
      end = quantity(times%end, 'd')     ! end time relative to start of forcing file
      tin = quantity(times%tin, 'd')     ! time of initial conditions
      sampling = quantity(times%sample, 'h') ! sampling frequency
      timeout = quantity(REAL(times%timeout, dp), 'h') ! timeout (in case the integration hangs)
      REWIND (UNIT=lun)
      READ (UNIT=lun, NML=timing, IOMSG=iomsg, ERR=900)
      times%dto = convert(dto, 's')
      times%start = convert(start, 's')
      times%end = convert(end, 's')
      times%tin = convert(tin, 's')
      times%sample = convert(sampling, 's')
      times%timeout = INT(convert(timeout, 's'))
      IF (times%londeg.GT.180) times%londeg = times%londeg - 360D0
      times%dtutc = NINT(cfds*times%londeg)/360  ! difference between UTC and local time in s
      RETURN
900   CONTINUE
      WRITE (stderr,'("set_timing: ",A)') TRIM(iomsg)
      STOP 1
    END SUBROUTINE set_timing
    FUNCTION expand_path (file) RESULT (path)
      IMPLICIT NONE
      CHARACTER(LEN=NF90_MAX_NAME), INTENT(IN) :: file
      CHARACTER(LEN=NF90_MAX_NAME) :: path, ev, evn
      INTEGER :: status=0
      SELECT CASE (file(1:1))
      CASE ("~")
         evn = "HOME"
         CALL GET_ENVIRONMENT_VARIABLE (NAME=evn, VALUE=ev, STATUS=status)
         path = TRIM(ev)//file(2:LEN_TRIM(file))
      CASE ("$")
         pos = SCAN(file, "/")
         IF (file(2:2).EQ."{") THEN
            evn = file(3:pos-2)
         ELSE
            evn = file(2:pos-1)
         END IF
         CALL GET_ENVIRONMENT_VARIABLE (NAME=evn, VALUE=ev, STATUS=status)
         path = TRIM(ev)//file(pos:LEN_TRIM(file))
      CASE DEFAULT
         path = file
      END SELECT
      IF (status.EQ.0) RETURN
      SELECT CASE (status)
      CASE (-1)
         WRITE (stderr,'(A," is too long.")') TRIM(evn)
         STOP
      CASE (1)
         WRITE (stderr,'(A," is not set.")') TRIM(evn)
         STOP
      CASE (2)
         STOP 'Environment variables are not supported.'
      END SELECT
    END FUNCTION expand_path
  END SUBROUTINE plankton_init

  !> light attenuation as a function of Chl
  SUBROUTINE lch_chl (ambient, boxes, stv)
    IMPLICIT NONE
    CLASS(local), INTENT(IN) :: ambient
    TYPE(layer), INTENT(INOUT) :: boxes(:)
    REAL(dp), INTENT(IN) :: stv(:,:)
    boxes%lch = (ambient%lacw + ambient%lachl*SUM(stv(ambient%ichl,:), DIM=1))*boxes%height
  END SUBROUTINE lch_chl
  !> light attenuation as a function of PON
  SUBROUTINE lch_pon (ambient, boxes, stv)
    IMPLICIT NONE
    CLASS(local), INTENT(IN) :: ambient
    TYPE(layer), INTENT(INOUT) :: boxes(:)
    REAL(dp), INTENT(IN) :: stv(:,:)
    boxes%lch = (ambient%lacw + ambient%lapon*SUM(stv(ambient%iphy,:), DIM=1))*boxes%height
  END SUBROUTINE lch_pon

  ! fraction of par arriving at box bottom
  SUBROUTINE fPAR_vertical (env, boxes, nbox)
    CLASS(local), INTENT(INOUT) :: env
    TYPE(layer), INTENT(INOUT) :: boxes(:)
    INTEGER, INTENT(IN) :: nbox
    INTEGER :: nb
    DO nb=1,nbox
       env%fPAR(nb) = env%fPAR(nb-1)*EXP(-boxes(nb)%lch)
       boxes(nb)%dPAR = env%dayPAR*(env%fPAR(nb-1) - env%fPAR(nb))/boxes(nb)%lch
    END DO
  END SUBROUTINE fPAR_vertical

  !> aggregation fluxes, forming detritus
  !!
  !! subroutine aggregate cannot be called from the detritus flux routine
  !! because the stickiness of functional types might be variable
  SUBROUTINE aggregate (env, grps, thisbox)
    IMPLICIT NONE
    CLASS(local), INTENT(INOUT) :: env
    TYPE(state), INTENT(IN) :: grps(:)
    TYPE(layer), INTENT(INOUT) :: thisbox
    INTEGER :: n
    thisbox%soma(0,detrit(env%idagg)%isv) = 0D0
    env%agg = env%sticky*MATMUL(env%OC(env%iagg)*env%sticky, env%kij)
    DO n=1,env%nagg
       thisbox%soma(0,grps(env%iagg(n))%var%isv) = -env%agg(n)*thisbox%stv(grps(env%iagg(n))%var%isv)
    END DO
    thisbox%soma(0,detrit(env%idagg)%isv) = thisbox%soma(0,detrit(env%idagg)%isv)&
         + MATMUL(env%quotas(detrit(env%idagg)%icon,env%iagg), env%OC(env%iagg)*env%agg)
    ENTRY no_aggregate (env, grps, thisbox)
  END SUBROUTINE aggregate

  SUBROUTINE alkalinity (env, thisbox)
    IMPLICIT NONE
    CLASS(local), INTENT(IN) :: env
    TYPE(layer), INTENT(INOUT) :: thisbox
    ! air-sea exchange is stored in fluxes(:,0), so does not affect alkalinity in soma
    thisbox%soma(:,env%ialk) = MATMUL(thisbox%soma(:,env%ifalk), env%fALK)
    ENTRY no_alkalinity (env, thisbox)
  END SUBROUTINE alkalinity

  !> Boundary concentrations for modified central differences
  !!
  !! Where vertical motion follows the gradient, the boundary concentrations
  !! are the means of the concentrations in the adjacent layers.  Elsewhere
  !! they are calculated as the geometric means to prevent negative
  !! concentrations.
  SUBROUTINE boxbc_cd (env, boxes)
    IMPLICIT NONE
    CLASS(local), INTENT(IN) :: env
    TYPE(layer), INTENT(INOUT) :: boxes(:)
    WHERE (NINT(SIGN(1._dp, env%dydz)).EQ.NINT(SIGN(1._dp, env%vvstv))) ! arithmetic means
       boxes(1)%bbc = (boxes(1)%stv - boxes(2)%stv)*boxes(1)%rdl + boxes(2)%stv
    ELSEWHERE                   ! geometric means: $x^a*y^{1-a} = EXP(a*LOG(x) + (1-a)*LOG(y))
       boxes(1)%bbc = EXP(boxes(1)%rdl*LOG(boxes(1)%stv) + (1._dp - boxes(1)%rdl)*LOG(boxes(2)%stv))
    END WHERE
    ENTRY boxbc_dummy (env, boxes)
  END SUBROUTINE boxbc_cd

  !> boundary concentrations for upwind scheme
  SUBROUTINE boxbc_upwind (env, boxes)
    IMPLICIT NONE
    CLASS(local), INTENT(IN) :: env
    TYPE(layer), INTENT(INOUT) :: boxes(:)
    WHERE (env%vvstv.GT.0D0)
       boxes(1)%bbc = boxes(1)%stv
    ELSEWHERE
       boxes(1)%bbc = boxes(2)%stv
    END WHERE
  END SUBROUTINE boxbc_upwind

  !> Rates of change for all ODEs
  !!
  !! This subroutine is passed to VODE_F90 in the main program oppla (oppla.F90).
  !! The state-variable and rates-of-change arrays (y and ydot) are declared as
  !! assumed-size (2D) arrays, so that they can be passed as vectors from/to VODE_F90.
  SUBROUTINE plankton_ode (neq, time, y, ydot)
    IMPLICIT NONE
    INTEGER, INTENT(IN) :: neq                  !< number of state variables and ODEs
    REAL(dp), INTENT(IN) :: time                !< current time in s
    REAL(dp), INTENT(INOUT), TARGET :: y(*)     !< state variables (stv in oppla and module plankton)
    REAL(dp), INTENT(OUT), TARGET :: ydot(*)    !< rates of change
    REAL(dp) :: delvel
    INTEGER :: n, nb
    states%roc(1:states%nsv_one_box,1:phys%nbox) => ydot(1:states%nsv_all_boxes)
    IF (ANY(states%data.LT.0._dp)) THEN
       ydot(1:neq) = 0._dp
       RETURN
    END IF
    times%current = NINT(time, KIND=INT64)
    times%hour = MODULO(times%current + times%tbase, cfds)/3.6e3_dp
    times%toy = times%doy + times%hour/24._dp ! time of year in decimal days
    CALL phys%get (time)        ! onf:readpd or onf:nopd -> get mixing, temperature, light, etc.
    CALL envir%PARfun (times)   ! PAR_const, PAR_brock81, PAR_MesoAqua, or PAR_LD
    CALL envir%attenuation (boxes=box(1:phys%nbox), stv=states%data) ! lch_chl or lch_pon
    CALL envir%PARfrac (boxes=box(1:phys%nbox), nbox=phys%nbox)
    envir%PAR = envir%sPAR*envir%fPAR      ! light intensity at box boundaries
    box(0)%flux(phys%isflx) = phys%surflux ! same as envir%fluxes(phys%isflx,0)
    IF (envir%xsfalk) box(0)%flux(envir%ialk) = SUM(box(0)%flux(envir%isfALK)*envir%fsfALK)
    box(phys%nbox+1)%stv(phys%ibot) = phys%botbc
    envir%all_ratios = states%data(envir%irt,:)/MAX(states%data(envir%irc,:), 1.E-30_dp) ! ratios (including quotas)
    DO nb=1,phys%nbox
       ! depth-averages of instantaneous (iPAR) and average daytime (dPAR) irradiance in each box
       box(nb)%iPAR = (box(nb)%tPAR - box(nb)%bPAR)/box(nb)%lch
       CALL dics%alkalinity (box=box(nb))            ! set alkalinity
       CALL dics%co2 (box=box(nb), niter=dics%niter) ! calculate CO2, HCO3, CO3
       ! load ratios, quotas, POC (envir%OC) for current box (layer)
       envir%ratios = box(nb)%ratios
       envir%ratios(envir%iq0r) = envir%ratios(envir%iq0r) + envir%Q0
       ! subsistence N, P are removed to improve model stability (see Christian 2007)
       ! the CONSTRAINED option (see oppla.F90) does not prevent VODE_F90 from
       ! passing negative values in y
       envir%quotas = UNPACK(envir%ratios, MASK=envir%qmsk, FIELD=envir%quotas)
       envir%OC = states%data(envir%ipoc,nb)          ! POC of all groups
       envir%dydz = (box(nb)%stv - box(nb+1)%stv)/box(nb)%dz ! vertical gradients
       ! fungrp(n)%var%flux points to the flux procedure pointer in each functional group
       ! for example, if fungrp(n)%var is of TYPE(phycmo) (defined in module cmo), then
       ! fungrp(n)%var%flux => phypla%flux => cmo_flux, defined in module cmo
       DO n = 1,envir%nft                   ! fluxes within and among functional groups
          CALL fungrp(n)%var%flux (grps=fungrp, env=envir, box=box(nb-1:nb+1), times=times)
       END DO
       CALL envir%aggregate (grps=fungrp, box=box(nb)) ! aggregate or no_aggregate
       ! physical transport: vertical advection and diffusion (mixing)
       delvel = box(nb-1)%velocity - box(nb)%velocity ! needed to correct for convergence
       envir%vvstv = box(nb)%velocity + envir%vvs*box(nb)%botperm ! advection + sinking + SVM + DVM + ...
       CALL envir%boxbc (boxes=box(nb:nb+1)) ! tracer concentrations at box boundaries
       ! flux: vertical flux out of (bottom of) box (advection + convergence + mixing)
       box(nb)%flux = box(nb)%bbc*envir%vvstv + delvel*states%data(:,nb) + envir%dydz*box(nb)%vdc
       CALL envir%alkalinity (thisbox=box(nb))
    END DO
    CALL dics%asf (env=envir, box=box(1)) ! air-sea fluxes of CO2 and O2
    states%roc = SUM(envir%soma, DIM=1) &
         + (envir%fluxes(:,0:phys%nbox-1) - envir%fluxes(:,1:phys%nbox))/envir%heights
    IF (states%write_soma) &
         ydot(states%i0flux:neq) = (/PACK(envir%fluxes, MASK=envir%fmsk), PACK(envir%soma, MASK=envir%smsk)/)
  END SUBROUTINE plankton_ode

END MODULE plankton
